/*
** itest.c - Linux-specific example demonstrating interrupt wait
**           functionality in the ADMXRC3 API.
**
** (C) Copyright 2010 Alpha Data
**
** Demonstrates how to wait for FPGA interrupt notifications using the non-
** blocking direct call technique in an interrupt thread. Also demonstrates
** how to make the interrupt thread terminate cleanly when the application
** shuts down.
*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <pthread.h>
#include <semaphore.h>

#include <admxrc3.h>

#include <sdk_common.h>
#include <platform_endian.h>

/*
** FPGA register addresses on local bus. These are also the offsets from the
** base of space 0, since space 0 begins at absolute local bus address 0x0.
*/
#define REGADDR_ISET 	(0xC0U)
#define REGADDR_ISTAT (0xC4U)
#define REGADDR_IMASK (0xC8U)
#define REGADDR_IARM	(0xCCU)

typedef struct _ThreadContext {
  ADMXRC3_HANDLE hCard;
  unsigned long signalledCount;
  unsigned int targetIndex;
  unsigned int window;
  sem_t waitingSemaphore;
} ThreadContext;

static void*
interruptThreadFunction(
  void* pArgument)
{
  ThreadContext* pContext = (ThreadContext*)pArgument;
  ADMXRC3_STATUS status;
  ADMXRC3_TICKET ticket;
  uint32_t istat;
  uint32_t dummy32 = 0; /* value doesn't matter, but avoid compiler whine */
  
  printf("Interrupt thread started\n");
	
  ADMXRC3_InitializeTicket(&ticket);

  while (1) {
    /* Start a non-blocking wait for an FPGA interrupt */
    status = ADMXRC3_StartNotificationWait(pContext->hCard, &ticket, ADMXRC3_EVENT_FPGAINTERRUPT(pContext->targetIndex));
    if (status != ADMXRC3_PENDING) {
      printf("*** ADMXRC3_StartNotificationWait returned non-pending status: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      break;
    }

    /* We are now waiting */
    sem_post(&pContext->waitingSemaphore);

    /* Rearm the 'FINTI_L' line driven by the FPGA by forcing it high for 1 cycle */
    dummy32 = 0;
    status = ADMXRC3_Write(pContext->hCard, pContext->window, 0, REGADDR_IARM, sizeof(dummy32), &dummy32);

    /* Finish the non-blocking wait for an FPGA interrupt */
    status = ADMXRC3_FinishNotificationWait(pContext->hCard, &ticket, TRUE);
    if (status != ADMXRC3_SUCCESS) {
      if (status == ADMXRC3_CANCELLED) {
        /* The main thread wants us to terminate */
        printf("Interrupt thread wait cancelled.\n");
        break;
      } else {
        printf("*** ADMXRC3_FinishNotificationWait failed: %s\n", ADMXRC3_GetStatusString(status, TRUE));
        break;
      }
    } else {
      pContext->signalledCount++;
    }

    /* We are no longer waiting */
    if (sem_trywait(&pContext->waitingSemaphore)) {
      /* The main thread wants us to terminate */
      printf("Interrupt thread sem_trywait failed.\n");
      break;
    }

    /* Get the pending FPGA interrupts */
    ADMXRC3_Read(pContext->hCard,	pContext->window, 0, REGADDR_ISTAT, sizeof(istat), &istat);

    /* Clear the pending FPGA interrupts */
    ADMXRC3_Write(pContext->hCard, pContext->window, 0, REGADDR_ISTAT, sizeof(istat), &istat);
  }
  
  printf("Interrupt thread exiting.\n");

  return 0;
}

int
main(
  int argc,
  char *argv[])
{
  const double clockFreq = 33.0e6;
  int ret = 0;
  unsigned int index, serial;
  ADMXRC3_STATUS status;
  ADMXRC3_CARD_INFO cardInfo;
  unsigned long numInterruptGenerated;
  const char* pFilename;
  Arguments arguments = g_initArguments;
  Option options[2];
  ThreadContext threadContext = {
    ADMXRC3_HANDLE_INVALID_VALUE, /* hCard */
    0, /* signalledCount */
    0 /* targetIndex */,
    0 /* window */
  };
  pthread_t interruptThread;
  bool_t bInterruptThreadRunning = FALSE;
  bool_t bSemValid = FALSE;
	uint32_t val32;
  int ch;

  options[0].key = "index";
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = "index of card to use";
  options[1].key = "sn";
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = "serial number of card to use";
  arguments.nOption = 2;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;

  if (sampleParseCommandLine("itest", argc, argv, &arguments) < 0) {
    ret = -1;
    goto done;
  }
  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;

  status = sampleOpenCard(index, options[0].specified, serial, options[1].specified, FALSE, &threadContext.hCard);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to open card with index %lu: %s\n", (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  status = ADMXRC3_GetCardInfo(threadContext.hCard, &cardInfo);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to get card info: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }
	
  if (cardInfo.Model < 0x100) {
    /* Set local bus clock to nominal frequency */
    status = ADMXRC3_SetClockFrequency(threadContext.hCard, ADMXRC3_CLOCK_LCLK, 0, clockFreq, NULL);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to set LCLK to %.1fMHz: %s\n", clockFreq / 1.0e6f, ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
  }

  pFilename = sampleGetFpgaFilename(threadContext.hCard, 0, "uber", "uber", &arguments);
  if (NULL == pFilename) {
    printf("*** Failed to determine bitstream filename.\n");
    ret = -1;
    goto done;
  }

  status = ADMXRC3_ConfigureFromFile(threadContext.hCard, 0, 0, pFilename);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to load the bitstream '%s': %s\n", pFilename, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  /* Start up the thread that waits for interrupts. */
  threadContext.signalledCount = 0;
  threadContext.targetIndex = 0;
  threadContext.window = 0;
  sem_init(&threadContext.waitingSemaphore, TRUE, 0);
  bSemValid = TRUE;
  if (pthread_create(&interruptThread, NULL, interruptThreadFunction, (void*)&threadContext)) {
    printf("*** Failed to start interrupt thread\n");
    ret = -1;
    goto done;
  }
  bInterruptThreadRunning = TRUE;
  
  /* Enable interrupts in the FPGA */
  val32 = cpu_to_le32(0x0U);
  status = ADMXRC3_Write(threadContext.hCard, threadContext.window, 0, REGADDR_IMASK, sizeof(val32), &val32);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Write to FPGA_IMASK register failed: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }
  
  /*
  ** Allow the user to generate interrupts by pressing the enter key.
  ** Ctrl-Z or entering 'q' / 'Q' exits this loop.
  */
  numInterruptGenerated = 0;
  do {
    printf("Enter 'q' to quit, or anything else to generate an interrupt:\n");
    ch = getchar();
    if (EOF == ch || 'q' == tolower(ch)) {
      ch = EOF;
    } else {
      val32 = cpu_to_le32(0x1U);
      status = ADMXRC3_Write(threadContext.hCard, threadContext.window, 0, REGADDR_ISET, sizeof(val32), &val32);
      if (status != ADMXRC3_SUCCESS) {
        printf("*** Write to FPGA_TEST register failed: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      }
      
      numInterruptGenerated++;
    }
  } while (EOF != ch);

  printf("Generated %lu interrupts\n", numInterruptGenerated);
  printf("Interrupt thread saw %lu interrupt(s)\n", threadContext.signalledCount);

done:
  /* Make the interrupt thread to terminate cleanly, if it is running */
  if (bInterruptThreadRunning) {
    /* Wait until the interrupt thread is waiting */
    sem_wait(&threadContext.waitingSemaphore);
    /* Cancel any ongoing operations (including the interrupt thread's notification wait) */
    ADMXRC3_Cancel(threadContext.hCard);
    /* Join, so that the interrupt thread is cleaned up */
    if (pthread_join(interruptThread, NULL)) {
      printf("+++ Failed to join with interrupt thread\n");
    }
  }

  if (bSemValid) {
    sem_destroy(&threadContext.waitingSemaphore);
    bSemValid = FALSE;
  }

  if (threadContext.hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(threadContext.hCard);
    threadContext.hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}
