/*
** memtesth.c - Example that performs a host-driven (as opposed to chip-
**              driven) test of the memory devices connected to a target
**              FPGA.
**
** (C) Copyright 2010-2011 Alpha Data
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tmain main
# define _tprintf printf
# define FALSE (0)
# define TRUE (1)

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <admxrc3.h>

#include <sdk_common.h>
#include <platform_endian.h>
#include <platform_time.h>

/* Default maximum number of individual verification errors to report */
#define MAX_ERROR_REPORTED 20

/* Maximum number of independent memory banks that this example supports */
#define MAX_BANKS 7

/* Size of the direct slave window onto the FPGA registers */
#define DS_MAPPING_SIZE 0x400000U

/* Chunk size used for transferring data during reads and writes of memory */
#define TRANSFER_CHUNK 0x100000U

/* Arbirary nonzero starting value for 64-bit XOR LFSR that is used to generate 32-bit pseudorandom values */
#ifdef _MSC_VER
/* MSVC6 compiler doesn't understand unsigned long long */
const uint64_t g_lfsr64StartValue = ((uint64_t)0xDEADBEEFU << 32) | (uint64_t)0xCAFEFACEU;
#else
const uint64_t g_lfsr64StartValue = (uint64_t)0xDEADBEEFCAFEFACEULL;
#endif

/* Attributes of direct slave window onto memory */
#define RAM_DS_WINDOW_BASE  0x200000U                  /* Base OCP address */
#define RAM_DS_WINDOW_ORDER 21U
#define RAM_DS_WINDOW_SIZE (1U << RAM_DS_WINDOW_ORDER) /* Size, in bytes*/
#define RAM_DS_WINDOW_MASK (RAM_DS_WINDOW_SIZE - 1U)

/* OCP addresses of FPGA registers */
#define REGADDR_RAM_DS_BANK          (0x300U)
#define REGADDR_RAM_DS_PAGE          (0x304U)
/* Per-bank registers (7 sets of 8 x 32-bit registers) */
#define REGADDR_BANK_CTRL(bank)      (0x320U + ((bank) << 5U))
#define REGADDR_BANK_OFFSET(bank)    (0x324U + ((bank) << 5U))
#define REGADDR_BANK_LENGTH(bank)    (0x328U + ((bank) << 5U))
#define REGADDR_BANK_INFO(bank)      (0x32CU + ((bank) << 5U))
#define REGADDR_BANK_STAT(bank)      (0x330U + ((bank) << 5U))
#define REGADDR_BANK_ERR_ADDR(bank)  (0x334U + ((bank) << 5U))
#define REGADDR_BANK_MUX_ERR(bank)   (0x338U + ((bank) << 5U))
#define REGADDR_BANK_DDR3_ERR(bank)  (0x33CU + ((bank) << 5U))

/* BANK_CTRL(x) fields */
#define REGBIT_BANK_CTRL_TEST        (0x1U << 8)  /* Write 1 to initiate FPGA-driven memory test */

/* BANK_INFO(x) fields */
#define REGMSK_BANK_INFO_WIDTH       (0xFFU << 16) /* Width of logical memory word in OCP */
#define REGPOS_BANK_INFO_WIDTH       (16)          /* Width of logical memory word in OCP */

/* BANK_STAT(x) fields */
#define REGBIT_BANK_STAT_WLEVELDONE  (0x1U << 0)  /* 1 => Write levelling done */
#define REGBIT_BANK_STAT_RLEVEL1DONE (0x1U << 1)  /* 1 => Read levelling done 1 */
#define REGBIT_BANK_STAT_RLEVEL2DONE (0x1U << 2)  /* 1 => Read levelling done 2 */
#define REGBIT_BANK_STAT_TRAINDONE   (0x1U << 3)  /* 1 => Everything finished training */
#define REGBIT_BANK_STAT_WLEVELERR   (0x1U << 8)  /* 1 => Error during write levelling */
#define REGBIT_BANK_STAT_RLEVEL1ERR  (0x1U << 9)  /* 1 => Error during read levelling 1 */
#define REGBIT_BANK_STAT_RLEVEL2ERR  (0x1U << 10) /* 1 => Error during read levelling 2 */
#define REGBIT_BANK_STAT_TRAINERR    (0x1U << 11) /* 1 => Error during training */
#define REGBIT_BANK_STAT_TESTDONE    (0x1U << 16) /* 1 => FPGA-driven memory test idle/completed */
#define REGBIT_BANK_STAT_TESTERR     (0x1U << 23) /* 1 => FPGA-driven memory test error */
#define REGMSK_BANK_STAT_ERRPHASE    (0xFU << 28) /* 1 => FPGA-driven memory test error phase */
#define REGPOS_BANK_STAT_ERRPHASE    (28)         /* 1 => FPGA-driven memory test error phase */

/* Returns the OCP address of a particular byte in a particular bank (applies only to DMA transfers) */
#define RAM_DMA_ADDRESS(bank, address) ((((uint64_t)(bank) + 1U) << 36U) + (address))

/* Writes a 32-bit value to an FPGA register; beware of write-buffering! */
#define WRITE_FPGA_REG(p, offset, val32) \
  *(volatile uint32_t*)((uint8_t*)(p) + (offset)) = (val32)

/* Reads a 32-bit value from an FPGA register */
#define READ_FPGA_REG(p, offset) \
  *(volatile uint32_t*)((uint8_t*)(p) + (offset))

/* Information and state required by the {read|write}Bank functions */
typedef struct _TestState {
  ADMXRC3_HANDLE hDevice;
  unsigned int dmaChannel; /* DMA channel to use, when testing memory using DMA transfers */
  void* pFpgaSpace;
  unsigned int selectedBank; /* does not apply to DMA */
  uint32_t selectedPage; /* does not apply to DMA */
  admxrc3_time_t spinnerTime;
  unsigned int spinnerIndex;
} TestState;

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

static void
spinner(
  unsigned int bank,
  bool_t bWrite,
  uint64_t address,
  TestState* pTestState)
{
  const TCHAR pattern[] = _T("|/-\\");
  admxrc3_time_t now;

  now = timeGet();
  if (timeAfter(now, pTestState->spinnerTime)) {
    pTestState->spinnerIndex++;
    if (pTestState->spinnerIndex >= ARRAY_LENGTH(pattern)) {
      pTestState->spinnerIndex = 0;
    }
    _tprintf(_T("%s bank %u @ 0x%llx %c                \r"), bWrite ? _T("Writing") : _T("Reading"), bank, (unsigned long long)address, pattern[pTestState->spinnerIndex]);
    fflush(stdout);
    pTestState->spinnerTime = timeAdd(pTestState->spinnerTime, timeMake(0, 100000, FALSE));
  }
}

/*
** Verify that a memory bank has been initialized and trained OK. Return
** value is:
**
**   -ve => error
**     0 => still training
**   +ve => training successful
*/
static int
checkBank(
   void* pFpgaSpace,
   unsigned int bankIndex)
{
  uint32_t val32;

  val32 = READ_FPGA_REG(pFpgaSpace, REGADDR_BANK_STAT(bankIndex));
  if (val32 & REGBIT_BANK_STAT_TRAINDONE) {
    return 1;
  }
  if (val32 & REGBIT_BANK_STAT_TRAINERR) {
    return -1;
  }

  return 0;
}

/* Verify that all banks selected for test have been trained successfully */
static bool_t
checkBanks(
  void* pFpgaSpace,
  unsigned int numBank,
  uint32_t bankMask)
{
  admxrc3_time_t finish;
  uint32_t completedMask = 0, errorMask = 0;
  bool_t bOk = TRUE, bAllTrained;
  unsigned int bank;
  int res;

  finish = timeAdd(timeGet(), timeMake(0, 500000, TRUE));
  while (bOk) {
    bAllTrained = TRUE;
    completedMask = 0;
    errorMask = 0;
    for (bank = 0; bank < numBank; bank++) {
      if (!(bankMask & (0x1U << bank))) {
        continue;
      }
      res = checkBank(pFpgaSpace, bank);
      if (res != 0) {
        completedMask |= 0x1U << bank;
        if (res < 0) {
          errorMask = 0x1U << bank;
        }
      } else {
        bAllTrained = FALSE;
      }
    }
    if (bAllTrained) {
      break;
    }
    if (errorMask) {
      _tprintf(_T("*** Error during memory bank training\n"));
      bOk = FALSE;
      continue;
    }
    if (timeAfter(timeGet(), finish)) {
      /* Timeout */
      _tprintf(_T("*** Timeout waiting for all banks to complete training\n"));
      bOk = FALSE;
      continue;
    }
    sleepMs(50);
  }
  if (!bOk) {
    _tprintf(_T("    Completed bank mask is 0x%lx, error mask is 0x%lx\n"),
      (unsigned long)completedMask, (unsigned long)errorMask);
  }
  return bOk;
}

/* Select which bank of memory is being accessed by direct slave; does not apply to DMA transfers */
static void
selectBank(
  void* pFpgaSpace,
  unsigned int bankIndex)
{
  WRITE_FPGA_REG(pFpgaSpace, REGADDR_RAM_DS_BANK, bankIndex);
}

/* Select which page of memory (in the selected bank) is being accessed by direct slave; does not apply to DMA transfers */
static void
selectPage(
  void* pFpgaSpace,
  unsigned int pageIndex)
{
  WRITE_FPGA_REG(pFpgaSpace, REGADDR_RAM_DS_PAGE, pageIndex);
}

/* Fill a buffer with a constant byte value */
static void
generateConstantByte(
  uint8_t* pBuffer,
  uint8_t val8,
  size_t n)
{
  while (n) {
    *pBuffer++ = val8;
    n--;
  }
}

/* Fill a buffer with own 32-bit address values */
static void
generateOwnAddress32(
  uint8_t* pBuffer,
  uint64_t address,
  size_t n)
{
  uint32_t val32 = (uint32_t)(address >> 2U);

  while (n) {
    *pBuffer++ = (uint8_t)(val32 >>  0U);
    *pBuffer++ = (uint8_t)(val32 >>  8U);
    *pBuffer++ = (uint8_t)(val32 >> 16U);
    *pBuffer++ = (uint8_t)(val32 >> 24U);
    n -= 4;
    val32++;
  }
}

/* Fill a buffer with own 32-bit values generated from a 64-bit LFSR */
static void
generateLfsr64(
  uint8_t* pBuffer,
  uint64_t* pLfsr64Value,
  size_t n)
{
  uint64_t lfsr64Val = *pLfsr64Value;
  uint8_t bit;

  while (n) {
    *pBuffer++ = (uint8_t)(lfsr64Val >>  0U);
    *pBuffer++ = (uint8_t)(lfsr64Val >>  8U);
    *pBuffer++ = (uint8_t)(lfsr64Val >> 16U);
    *pBuffer++ = (uint8_t)(lfsr64Val >> 24U);
    bit = ((lfsr64Val >> 0U) ^ (lfsr64Val >> 1U) ^ (lfsr64Val >> 3U) ^ (lfsr64Val >> 4U)) & 0x1U;
    lfsr64Val = (lfsr64Val >> 1U) | ((uint64_t)bit << 63U);
    n -= 4;
  }
  *pLfsr64Value = lfsr64Val;
}

/* Reads a block of data from a memory bank, using direct slave or DMA */
static ADMXRC3_STATUS
readBank(
  TestState* pTestState,
  bool_t bUseDma,
  unsigned int bankIndex,
  uint64_t address,
  size_t n,
  void* pBuffer,
  ADMXRC3_BUFFER_HANDLE hBuffer)
{
  ADMXRC3_STATUS status;
  uint32_t page, offset;
  size_t chunk;
  uint8_t* p;

  if (bUseDma) {
    status = ADMXRC3_ReadDMALockedEx(pTestState->hDevice, pTestState->dmaChannel, 0, hBuffer, 0, n, RAM_DMA_ADDRESS(bankIndex, address));
    if (ADMXRC3_SUCCESS != status) {
      return status;
    }
  } else {
    if (pTestState->selectedBank != bankIndex) {
      selectBank(pTestState->pFpgaSpace, bankIndex);
      pTestState->selectedBank = bankIndex;
    }
    page = (uint32_t)(address >> RAM_DS_WINDOW_ORDER);
    if (pTestState->selectedPage != page) {
      selectPage(pTestState->pFpgaSpace, page);
      pTestState->selectedPage = page;
    }
    offset = (uint32_t)(address & RAM_DS_WINDOW_MASK);
    p = (uint8_t*)pBuffer;
    while (n) {
      chunk = RAM_DS_WINDOW_SIZE - offset;
      if (chunk > n) {
        chunk = n;
      }
      memcpy(p, (uint8_t*)pTestState->pFpgaSpace + offset + RAM_DS_WINDOW_BASE, chunk);
      address += chunk;
      n -= chunk;
      p += chunk;
      page++;
      selectPage(pTestState->pFpgaSpace, page);
      pTestState->selectedPage = page;
      offset = 0;
    }
  }

  return ADMXRC3_SUCCESS;
}

/* Reads a block of data from a memory bank, using direct slave or DMA */
static ADMXRC3_STATUS
writeBank(
  TestState* pTestState,
  bool_t bUseDma,
  unsigned int bankIndex,
  uint64_t address,
  size_t n,
  const void* pBuffer,
  ADMXRC3_BUFFER_HANDLE hBuffer)
{
  ADMXRC3_STATUS status;
  uint32_t page, offset;
  size_t chunk;
  uint8_t* p;

  if (bUseDma) {
    status = ADMXRC3_WriteDMALockedEx(pTestState->hDevice, pTestState->dmaChannel, 0, hBuffer, 0, n, RAM_DMA_ADDRESS(bankIndex, address));
    if (ADMXRC3_SUCCESS != status) {
      return status;
    }
  } else {
    if (pTestState->selectedBank != bankIndex) {
      selectBank(pTestState->pFpgaSpace, bankIndex);
      pTestState->selectedBank = bankIndex;
    }
    page = (uint32_t)(address >> RAM_DS_WINDOW_ORDER);
    if (pTestState->selectedPage != page) {
      selectPage(pTestState->pFpgaSpace, page);
      pTestState->selectedPage = page;
    }
    offset = (uint32_t)(address & RAM_DS_WINDOW_MASK);
    p = (uint8_t*)pBuffer;
    while (n) {
      chunk = RAM_DS_WINDOW_SIZE - offset;
      if (chunk > n) {
        chunk = n;
      }
      memcpy((uint8_t*)pTestState->pFpgaSpace + offset + RAM_DS_WINDOW_BASE, p, n);
      address += chunk;
      n -= chunk;
      p += chunk;
      page++;
      selectPage(pTestState->pFpgaSpace, page);
      pTestState->selectedPage = page;
      offset = 0;
    }
    READ_FPGA_REG(pTestState->pFpgaSpace, 0); /* Do a read of any FPGA register, to force all writes to be fully propagated */
  }

  return ADMXRC3_SUCCESS;
}

static void
verifyData(
  unsigned int bank,
  uint64_t address,
  const uint8_t* pReferenceBuffer,
  const uint8_t* pDataBuffer,
  size_t n,
  uint64_t maxError,
  uint64_t* pNumError)
{
  uint64_t numError = *pNumError;
  uint8_t expected, actual;

  while (n) {
    expected = *pReferenceBuffer++;
    actual = *pDataBuffer++;
    if (expected != actual) {
      if (numError < maxError) {
        _tprintf(_T("*** Verification error bank=%u address=0x%llx expected=0x%02x actual=0x%02x\n"),
          bank, (unsigned long long)address, (unsigned int)expected, (unsigned int)actual);
      } else if (numError == maxError) {
        _tprintf(_T("*** (More errors)\n"));
      }
      numError++;
    }
    address++;
    n--;
  }
  *pNumError = numError;
}

int
_tmain(
  int argc,
  TCHAR *argv[])
{
  int ret = 0;
  ADMXRC3_STATUS status;
  unsigned int index, serial, numRep, rep;
  ADMXRC3_HANDLE hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_CARD_INFO cardInfo;
  ADMXRC3_BANK_INFO bankInfo[MAX_BANKS];
  uint64_t bankSize[MAX_BANKS];
  uint32_t* pFpgaSpace = NULL;
  uint8_t* pCheckBuffer = NULL;
  ADMXRC3_BUFFER_HANDLE hCheckBuffer = 0;
  uint8_t* pGenerateBuffer = NULL;
  ADMXRC3_BUFFER_HANDLE hGenerateBuffer = 0;
  const TCHAR* pFilename;
  Option options[6];
  Arguments arguments = g_initArguments;
  TestState testState;
  unsigned int bank, phase, direction;
  uint32_t bankMask;
  uint64_t maxError, numError = 0;
  bool_t bUseDma = TRUE;

  /* Set up expected options */
  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
  options[2].key = _T("banks");
  options[2].type = OptionUInt;
  options[2].def.uintVal = 0xFFFFU;
  options[2].meaning = _T("bitmask of banks to be tested");
  options[3].key = _T("dma");
  options[3].type = OptionBoolean;
  options[3].def.booleanVal = TRUE;
  options[3].meaning = _T("TRUE => use DMA to read and write memory");
  options[4].key = _T("maxerror");
  options[4].type = OptionUInt64;
  options[4].def.uint64Val = MAX_ERROR_REPORTED;
  options[4].meaning = _T("Maximum number of errors to display");
  options[5].key = _T("repeat");
  options[5].type = OptionUInt;
  options[5].def.uintVal = 1;
  options[5].meaning = _T("Number of repetitions to perform");
  arguments.nOption = 6;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;

  if (sampleParseCommandLine(_T("memtesth"), argc, argv, &arguments) < 0) {
    ret = -1;
    goto done;
  }
  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;
  bUseDma = options[3].value.uintVal;
  maxError = options[4].value.uint64Val;
  numRep = options[5].value.uintVal;

  status = sampleOpenCard(index, options[0].specified, serial, options[1].specified, FALSE, &hCard);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"), (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  status = ADMXRC3_GetCardInfo(hCard, &cardInfo);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to get card info: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  /* Get the address of FPGA space */
  status = ADMXRC3_MapWindow(hCard, 0, 0x0U, DS_MAPPING_SIZE, (void*)&pFpgaSpace);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to map FPGA space: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  if (cardInfo.Model < 0x100 && cardInfo.NumClockGen > 0) {
    /* Set local bus clock to nominal frequency */
    status = ADMXRC3_SetClockFrequency(hCard, ADMXRC3_CLOCK_LCLK, 0, 33.0e6, NULL);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to set LCLK to %.1lf MHz: %s\n"), 33.0, ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
  }

  /* Assume initially that we are going to test all banks */
  bankMask = (uint32_t)(1U << cardInfo.NumMemoryBank) - 1U;
  /* AND with the bank presence bits from card information */
  bankMask &= cardInfo.MemoryBankPresent;
  /* AND with the bank mask from the '-banks' command line option */
  bankMask &= (uint32_t)options[2].value.uintVal;
  for (bank = 0; bank < cardInfo.NumMemoryBank && bank < MAX_BANKS; bank++) {
    const TCHAR* pTypeStr = NULL;
    uint32_t type;

    status = ADMXRC3_GetBankInfo(hCard, bank, &bankInfo[bank]);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("+++ Warning: failed to get memory bank %u information: %s\n"), bank, ADMXRC3_GetStatusString(status, TRUE));
      /* Don't test this bank, because we couldn't get any information about it */
      bankMask &= ~(1U << bank);
      continue;
    }
    /* Calculate bank size, in bytes */
    bankSize[bank] = bankInfo[bank].PhysicalSize * bankInfo[bank].PhysicalDataWidth / 8;
    type = bankInfo[bank].TypeMask;
    if (type & ADMXRC3_BANK_SDRAM_DDR3) {
      pTypeStr = _T("DDR-3 SDRAM");
    } else if (type & ADMXRC3_BANK_SSRAM_QDR) {
      pTypeStr = _T("QDR SSRAM");
    } else if (type & ADMXRC3_BANK_SDRAM_DDR2) {
      pTypeStr = _T("DDR-2 SDRAM");
    } else if (type & ADMXRC3_BANK_SDRAM_SDR) {
      pTypeStr = _T("SDR SDRAM");
    } else if (type & ADMXRC3_BANK_SSRAM_DDR2) {
      pTypeStr = _T("DDR-2 SSRAM");
    } else if (type & ADMXRC3_BANK_SDRAM_DDR) {
      pTypeStr = _T("DDR SDRAM");
    } else if (type & ADMXRC3_BANK_ZBT_P) {
      pTypeStr = _T("Pipelined ZBT");
    } else if (type & ADMXRC3_BANK_ZBT_FT) {
      pTypeStr = _T("Flowthrough ZBT");
    } else {
      pTypeStr = _T("Unknown type");
    }
    _tprintf(_T("Bank %02u: %s, %llu (0x%llx) kiB\n"), bank, pTypeStr, (unsigned long long)(bankSize[bank] >> 10U), (unsigned long long)(bankSize[bank] >> 10U));
  }
  _tprintf(_T("Bank test mask is 0x%04x\n"), (unsigned int)bankMask);

  pCheckBuffer = (uint8_t*)malloc(RAM_DS_WINDOW_SIZE);
  if (NULL == pCheckBuffer) {
    _tprintf(_T("*** Failed to allocate check buffer of 0x%x bytes.\n"), (unsigned int)RAM_DS_WINDOW_SIZE);
    ret = -1;
    goto done;
  }

  pGenerateBuffer = (uint8_t*)malloc(RAM_DS_WINDOW_SIZE);
  if (NULL == pGenerateBuffer) {
    _tprintf(_T("*** Failed to allocate generate buffer of 0x%x bytes.\n"), (unsigned int)RAM_DS_WINDOW_SIZE);
    ret = -1;
    goto done;
  }

  status = ADMXRC3_Lock(hCard, pCheckBuffer, RAM_DS_WINDOW_SIZE, &hCheckBuffer);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to lock check buffer for DMA transfers: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  status = ADMXRC3_Lock(hCard, pGenerateBuffer, RAM_DS_WINDOW_SIZE, &hGenerateBuffer);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to lock generate buffer for DMA transfers: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  pFilename = sampleGetFpgaFilename(hCard, 0, _T("uber"), _T("uber"), &arguments);
  if (NULL == pFilename) {
    _tprintf(_T("*** Failed to determine bitstream filename.\n"));
    ret = -1;
    goto done;
  }

  status = ADMXRC3_ConfigureFromFile(hCard, 0, 0, pFilename);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to configure target FPGA with '%s': %s\n"), pFilename, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  testState.hDevice = hCard;
  testState.pFpgaSpace = pFpgaSpace;
  testState.dmaChannel = 0;
  testState.selectedBank = 0;
  selectBank(pFpgaSpace, testState.selectedBank);
  testState.selectedPage = 0;
  selectPage(pFpgaSpace, testState.selectedPage);
  testState.spinnerTime = timeGet();
  testState.spinnerIndex = 0;

  /* Check the memory banks have been initialized and trained OK */
  if (!checkBanks(pFpgaSpace, cardInfo.NumMemoryBank, bankMask)) {
    _tprintf(_T("*** FAILED - one or more banks not trained successufully\n"));
    ret = -2;
    goto done;
  }

  rep = 0;
  while (!numRep || rep < numRep) {
    _tprintf(_T("Performing host-driven memory test (repetition %u)...\n"), rep);
    for (phase = 1; phase <= 2; phase++) {
      _tprintf(_T("Phase %u - 0x%02X pattern               \n"), phase, (phase == 1) ? 0x55U : 0xAAU);
      generateConstantByte(pGenerateBuffer, (phase == 1) ? 0x55U : 0xAAU, RAM_DS_WINDOW_SIZE);
      for (bank = 0; bank < cardInfo.NumMemoryBank; bank++) {
        uint64_t address, chunk;

        if (!(bankMask & (0x1U << bank))) {
          continue;
        }
        for (address = 0; address < bankSize[bank];) {
          spinner(bank, TRUE, address, &testState);
          chunk = bankSize[bank] - address;
          if (chunk > RAM_DS_WINDOW_SIZE) {
            chunk = RAM_DS_WINDOW_SIZE;
          }
          if (chunk > TRANSFER_CHUNK) {
            chunk = TRANSFER_CHUNK;
          }
          status = writeBank(&testState, bUseDma, bank, address, (size_t)chunk, pGenerateBuffer, hGenerateBuffer);
          if (ADMXRC3_SUCCESS != status) {
            _tprintf(_T("*** Failed to write memory: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
            ret = -1;
            goto done;
          }
          address += chunk;
        }
      }
      for (bank = 0; bank < cardInfo.NumMemoryBank; bank++) {
        uint64_t address, chunk;

        if (!(bankMask & (0x1U << bank))) {
          continue;
        }
        for (address = 0; address < bankSize[bank];) {
          spinner(bank, FALSE, address, &testState);
          chunk = bankSize[bank] - address;
          if (chunk > RAM_DS_WINDOW_SIZE) {
            chunk = RAM_DS_WINDOW_SIZE;
          }
          if (chunk > TRANSFER_CHUNK) {
            chunk = TRANSFER_CHUNK;
          }
          status = readBank(&testState, bUseDma, bank, address, (size_t)chunk, pCheckBuffer, hCheckBuffer);
          if (ADMXRC3_SUCCESS != status) {
            _tprintf(_T("*** Failed to read memory: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
            ret = -1;
            goto done;
          }
          verifyData(bank, address, pGenerateBuffer, pCheckBuffer, (size_t)chunk, maxError, &numError);
          address += chunk;
        }
      }
      if (numError) {
        goto done;
      }
    }

    _tprintf(_T("Phase %u - own address pattern          \n"), phase);
    for (bank = 0; bank < cardInfo.NumMemoryBank; bank++) {
      uint64_t address, chunk;

      if (!(bankMask & (0x1U << bank))) {
        continue;
      }
      for (address = 0; address < bankSize[bank];) {
        spinner(bank, TRUE, address, &testState);
        chunk = bankSize[bank] - address;
        if (chunk > RAM_DS_WINDOW_SIZE) {
          chunk = RAM_DS_WINDOW_SIZE;
        }
        if (chunk > TRANSFER_CHUNK) {
          chunk = TRANSFER_CHUNK;
        }
        generateOwnAddress32(pGenerateBuffer, address, (size_t)chunk);
        status = writeBank(&testState, bUseDma, bank, address, (size_t)chunk, pGenerateBuffer, hGenerateBuffer);
        if (ADMXRC3_SUCCESS != status) {
          _tprintf(_T("*** Failed to write memory: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
          ret = -1;
          goto done;
        }
        address += chunk;
      }
    }
    for (bank = 0; bank < cardInfo.NumMemoryBank; bank++) {
      uint64_t address, chunk;

      if (!(bankMask & (0x1U << bank))) {
        continue;
      }
      for (address = 0; address < bankSize[bank];) {
        spinner(bank, FALSE, address, &testState);
        chunk = bankSize[bank] - address;
        if (chunk > RAM_DS_WINDOW_SIZE) {
          chunk = RAM_DS_WINDOW_SIZE;
        }
        if (chunk > TRANSFER_CHUNK) {
          chunk = TRANSFER_CHUNK;
        }
        generateOwnAddress32(pGenerateBuffer, address, (size_t)chunk);
        status = readBank(&testState, bUseDma, bank, address, (size_t)chunk, pCheckBuffer, hCheckBuffer);
        if (ADMXRC3_SUCCESS != status) {
          _tprintf(_T("*** Failed to read memory: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
          ret = -1;
          goto done;
        }
        verifyData(bank, address, pGenerateBuffer, pCheckBuffer, (size_t)chunk, maxError, &numError);
        address += chunk;
      }
    }
    phase++;
    if (numError) {
      goto done;
    }

    _tprintf(_T("Phase %u - pseudorandom data            \n"), phase);
    for (bank = 0; bank < cardInfo.NumMemoryBank; bank++) {
      uint64_t address, chunk;
      uint64_t lfsr64Value = g_lfsr64StartValue;

      if (!(bankMask & (0x1U << bank))) {
        continue;
      }
      for (address = 0; address < bankSize[bank];) {
        spinner(bank, TRUE, address, &testState);
        chunk = bankSize[bank] - address;
        if (chunk > RAM_DS_WINDOW_SIZE) {
          chunk = RAM_DS_WINDOW_SIZE;
        }
        if (chunk > TRANSFER_CHUNK) {
          chunk = TRANSFER_CHUNK;
        }
        generateLfsr64(pGenerateBuffer, &lfsr64Value, (size_t)chunk);
        status = writeBank(&testState, bUseDma, bank, address, (size_t)chunk, pGenerateBuffer, hGenerateBuffer);
        if (ADMXRC3_SUCCESS != status) {
          _tprintf(_T("*** Failed to write memory: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
          ret = -1;
          goto done;
        }
        address += chunk;
      }
    }
    for (bank = 0; bank < cardInfo.NumMemoryBank; bank++) {
      uint64_t address, chunk;
      uint64_t lfsr64Value = g_lfsr64StartValue;

      if (!(bankMask & (0x1U << bank))) {
        continue;
      }
      for (address = 0; address < bankSize[bank];) {
        spinner(bank, FALSE, address, &testState);
        chunk = bankSize[bank] - address;
        if (chunk > RAM_DS_WINDOW_SIZE) {
          chunk = RAM_DS_WINDOW_SIZE;
        }
        if (chunk > TRANSFER_CHUNK) {
          chunk = TRANSFER_CHUNK;
        }
        generateLfsr64(pGenerateBuffer, &lfsr64Value, (size_t)chunk);
        status = readBank(&testState, bUseDma, bank, address, (size_t)chunk, pCheckBuffer, hCheckBuffer);
        if (ADMXRC3_SUCCESS != status) {
          _tprintf(_T("*** Failed to read memory: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
          ret = -1;
          goto done;
        }
        verifyData(bank, address, pGenerateBuffer, pCheckBuffer, (size_t)chunk, maxError, &numError);
        address += chunk;
      }
    }
    phase++;
    if (numError) {
      goto done;
    }

    rep++;
  }

  _tprintf(_T("Measuring throughput...                 \n"));
  for (direction = 0; direction < 2; direction++) {
    admxrc3_time_t start, finish, now;
    uint64_t total = 0;
    double elapsed, rate;

    start = timeGet();
    finish = timeAdd(start, timeMake(2, 0, FALSE));
    do {
      if (direction == 0) {
        status = writeBank(&testState, bUseDma, 0, 0, RAM_DS_WINDOW_SIZE, pCheckBuffer, hCheckBuffer);
        if (ADMXRC3_SUCCESS != status) {
          _tprintf(_T("*** Failed to write memory: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
          ret = -1;
          goto done;
        }
      } else {
        status = readBank(&testState, bUseDma, 0, 0, RAM_DS_WINDOW_SIZE, pCheckBuffer, hCheckBuffer);
        if (ADMXRC3_SUCCESS != status) {
          _tprintf(_T("*** Failed to read memory: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
          ret = -1;
          goto done;
        }
      }
      total += RAM_DS_WINDOW_SIZE;
      now = timeGet();
    } while (timeBefore(now, finish));
    elapsed = timeToDouble(timeSubtract(now, start));
    rate = (double)total / elapsed;
    _tprintf(_T("Throughput %s is %.1f MiB/s\n"), (direction == 0) ? _T("from host to memory") : _T("from memory to host"), rate / 1048576.0);
  }

  _tprintf(_T("PASSED                                  \n"));

done:
  if (numError) {
    _tprintf(_T("*** FAILED with %llu (0x%llx) error(s)\n"), (unsigned long long)numError, (unsigned long long)numError);
    ret = -3;
  }

  if (0 != hGenerateBuffer) {
    ADMXRC3_Unlock(hCard, hGenerateBuffer);
    hGenerateBuffer = 0;
  }
  if (0 != hCheckBuffer) {
    ADMXRC3_Unlock(hCard, hCheckBuffer);
    hCheckBuffer = 0;
  }
  if (NULL != pGenerateBuffer) {
    free(pGenerateBuffer);
    pGenerateBuffer = NULL;
  }
  if (NULL != pCheckBuffer) {
    free(pCheckBuffer);
    pCheckBuffer = NULL;
  }
  if (NULL != pFpgaSpace) {
    status = ADMXRC3_UnmapWindow(hCard, (void*)pFpgaSpace);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("+++ Failed to unmap FPGA space: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    }
    pFpgaSpace = NULL;
  }

  if (hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hCard);
    hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}
