--==============================================================================
--
--     File Name: adb3_ocp_ocp2ddr3_nb.vhd
--       Version: 0.2
--          Date: 07 Mar 2011
--         Model: syn (Synthesisable)
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: ADB3 OCP to DDR3 MIG user interface (non-blocking)
--
--   Description: Component giving OCP port access to DDR3 MIG user interface
--
--   Limitations: Up to 32 read bursts in flight
--
--         Notes: Non-blocking component
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 24 Sep 10: Initial design
--                0.2, KDC, 07 Mar 11:
--                       - Must ensure mdata FIFO is not empty during a write
--                         command before advancing cmd FIFO read pointer.
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library work;
use work.fifo.all;
use work.adb3_ocp.all;

entity adb3_ocp_ocp2ddr3_nb is
generic(
  app_row_width  : natural;
  app_col_width  : natural;
  app_bank_width : natural;
  app_addr_width : natural);  -- 4-byte address width
port(
  -- OCP Interface
  ocp_rst : in    std_logic;
  ocp_clk : in    std_logic;
  ocp_m2s : in    adb3_ocp_m2sT;
  ocp_s2m : out   adb3_ocp_s2mT;
  -- DDR3 MIG user interface
  mig_rst           : in    std_logic;
  mig_clk           : in    std_logic;
  phy_init_done     : in    std_logic;
  app_rdy           : in    std_logic;
  app_wdf_rdy       : in    std_logic;
  app_rd_data       : in    adb3_ocp_data_s;
  app_rd_data_valid : in    std_logic;
  app_en            : out   std_logic;
  app_cmd           : out   std_logic_vector(2 downto 0);
  app_addr          : out   std_logic_vector(app_addr_width-1 downto 0);
  app_sz            : out   std_logic;
  app_wdf_wren      : out   std_logic;
  app_wdf_data      : out   adb3_ocp_data_s;
  app_wdf_mask      : out   adb3_ocp_be_s;
  app_wdf_end       : out   std_logic;
  -- Debug
  debug_info        : out   std_logic_vector(31 downto 0));
end;

architecture synth of adb3_ocp_ocp2ddr3_nb is

  -- Define depth of FIFOs = 2^FIFO_DEPTH
  -- Tag FIFO needs to keep on accepting read commands until data is received back from RAM
  -- sdata FIFO needs to have space for data in flight = 2 x read commands for BL8
  constant CMD_FIFO_DEPTH  : natural := 4;
  constant TAG_FIFO_DEPTH  : natural := 5;
  constant RESP_FIFO_DEPTH : natural := 6;

  -- Command FIFO address width (4-byte addressing)
  constant CMD_ADDR_WIDTH : natural := app_bank_width + app_row_width + app_col_width;

  -- Define FIFO widths
  constant CMD_WIDTH   : natural := 1 + ADB3_OCP_TAG_WIDTH + CMD_ADDR_WIDTH + 2;
  constant MDATA_WIDTH : natural := ADB3_OCP_DATA_WIDTH + ADB3_OCP_DATA_WIDTH/8;
  constant SDATA_WIDTH : natural := ADB3_OCP_DATA_WIDTH + ADB3_OCP_RESP_WIDTH + ADB3_OCP_TAG_WIDTH;
  constant TAG_WIDTH   : natural := 1 + ADB3_OCP_TAG_WIDTH;

  signal ocp_m2s_int  : adb3_ocp_m2sT;
  signal ocp_s2m_int  : adb3_ocp_s2mT;

  signal writing_m2s      : std_logic := '0';
  signal reading_m2s      : std_logic := '0';
  signal wburst_m2s       : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0) := (others => '0');
  signal wburst_aligned   : std_logic := '0';
  signal rburst_m2s       : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0) := (others => '0');
--  signal rburst_aligned   : std_logic := '0';
  signal wburst_m2s_first : std_logic := '0';
  signal wdata_last_m2s   : std_logic := '0';
  signal rdata_last_m2s   : std_logic := '0';

  signal en_bit : std_logic;
  signal en_cmd : ocp_CmdT;
  signal en_wr  : std_logic;

  -- cmd FIFO signals
  signal cmd_fifo_data_in  : std_logic_vector(CMD_WIDTH-1 downto 0);
  signal cmd_fifo_data_out : std_logic_vector(CMD_WIDTH-1 downto 0);

  signal cmd_fifo_bl8_in   : std_logic := '0';
  signal cmd_fifo_w_in     : std_logic;
  signal cmd_fifo_r_in     : std_logic;
  signal cmd_fifo_addr_in  : std_logic_vector(CMD_ADDR_WIDTH-1 downto 0);
  signal cmd_fifo_tag_in   : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
  signal cmd_fifo_addr_inc : natural;

  signal cmd_fifo_bl8_out  : std_logic;
  signal cmd_fifo_w_out    : std_logic;
  signal cmd_fifo_r_out    : std_logic;
  signal cmd_fifo_addr_out : std_logic_vector(CMD_ADDR_WIDTH-1 downto 0);
  signal cmd_fifo_tag_out  : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
  signal cmd_fifo_w_last   : std_logic := '0';

  signal cmd_fifo_ready    : std_logic := '0';
  signal cmd_fifo_wadv     : std_logic := '0';
  signal cmd_fifo_radv     : std_logic := '0';
  signal cmd_fifo_empty    : std_logic;
  signal cmd_fifo_nfull    : std_logic := '0';
  signal cmd_fifo_full     : std_logic := '0';
  signal cmd_ok            : std_logic := '0';
  signal phy_init_done_mig : std_logic := '0';

  signal app_bank          : std_logic_vector(app_bank_width-1 downto 0);
  signal app_row           : std_logic_vector(app_row_width-1 downto 0);
  signal app_col           : std_logic_vector(app_col_width-1 downto 0);

  -- mdata FIFO signals
  signal mdata_fifo_data_in  : std_logic_vector(MDATA_WIDTH-1 downto 0);
  signal mdata_fifo_data_out : std_logic_vector(MDATA_WIDTH-1 downto 0);

  signal mdata_fifo_d_out  : std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0);
  signal mdata_fifo_be_out : std_logic_vector((ADB3_OCP_DATA_WIDTH/8)-1 downto 0);

  signal mdata_fifo_ready : std_logic := '0';
  signal mdata_fifo_wadv  : std_logic := '0';
  signal mdata_fifo_radv  : std_logic := '0';
  signal mdata_fifo_empty : std_logic;
  signal mdata_fifo_nfull : std_logic := '0';
  signal mdata_fifo_full  : std_logic := '0';

  -- sdata FIFO signals
  signal sdata_fifo_data_in  : std_logic_vector(SDATA_WIDTH-1 downto 0);
  signal sdata_fifo_data_out : std_logic_vector(SDATA_WIDTH-1 downto 0);

  signal sdata_fifo_wadv   : std_logic := '0';
  signal sdata_fifo_radv   : std_logic := '0';
  signal sdata_fifo_empty  : std_logic;
  signal sdata_fifo_nfull  : std_logic := '0';
  signal sdata_fifo_full   : std_logic := '0';

  signal sdata_fifo_tag_in  : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
  signal sdata_fifo_resp_in : ocp_RespT;
  signal sdata_fifo_d_in    : std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0);

  -- tag FIFO signals
  signal tag_fifo_data_in  : std_logic_vector(TAG_WIDTH-1 downto 0);
  signal tag_fifo_data_out : std_logic_vector(TAG_WIDTH-1 downto 0);

  signal tag_fifo_bl8_out : std_logic;
  signal tag_fifo_tag_out : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);

  signal tag_fifo_wadv    : std_logic := '0';
  signal tag_fifo_radv    : std_logic := '0';
  signal tag_fifo_empty   : std_logic;
  signal tag_fifo_nfull   : std_logic := '0';
  signal tag_fifo_full    : std_logic := '0';
  signal tag_fifo_bl_cnt  : std_logic := '0';
  signal tag_fifo_bl_comp : std_logic := '0';
  signal tag_rd_empty_err : std_logic := '0';

  -- FIFO levels
  signal cmd_rlevel       : std_logic_vector(15 downto 0);
  signal cmd_wlevel       : std_logic_vector(15 downto 0);
  signal mdata_rlevel     : std_logic_vector(15 downto 0);
  signal mdata_wlevel     : std_logic_vector(15 downto 0);
  signal sdata_rlevel     : std_logic_vector(15 downto 0);
  signal sdata_wlevel     : std_logic_vector(15 downto 0);
  signal tag_rlevel       : std_logic_vector(15 downto 0);

  signal max_tag_rlevel   : std_logic_vector(7 downto 0) := (others => '0');
  signal max_sdata_rlevel : std_logic_vector(7 downto 0) := (others => '0');

  signal sdata_fifo_space     : std_logic_vector(RESP_FIFO_DEPTH-1 downto 0);
  signal tag_fifo_in_flight   : std_logic_vector(TAG_FIFO_DEPTH downto 0);
  signal sdata_full_in_flight : std_logic := '0';
  signal sdata_ovf_in_flight  : std_logic := '0';
  signal sdata_wr_full_err    : std_logic := '0';
  signal tag_wr_full_err      : std_logic := '0';
  signal cmd_wr_rd_err        : std_logic := '0';

  signal sdata_wr_full_cnt : std_logic_vector(7 downto 0) := (others => '0');
  signal tag_wr_full_cnt   : std_logic_vector(7 downto 0) := (others => '0');
  signal cmd_wr_rd_cnt     : std_logic_vector(7 downto 0) := (others => '0');

  -- MIG
  subtype mig_cmd_s is std_logic_vector(2 downto 0);

  constant MIG_WR_CMD : mig_cmd_s := "000";
  constant MIG_RD_CMD : mig_cmd_s := "001";

begin

  -- Command group m2s - Qualify command with not writing/reading
  ocp_m2s_int.Cmd         <= ocp_m2s.Cmd and en_cmd;
  ocp_m2s_int.Addr        <= ocp_m2s.Addr;
  ocp_m2s_int.BurstLength <= ocp_m2s.BurstLength;
  ocp_m2s_int.Tag         <= ocp_m2s.Tag;

  -- Command group s2m - Qualify command accept with not writing/reading
  ocp_s2m.CmdAccept <= ocp_s2m_int.CmdAccept and en_bit;

  -- Data write group m2s - Qualify data valid with command accept or writing
  ocp_m2s_int.Data       <= ocp_m2s.Data;
  ocp_m2s_int.DataByteEn <= ocp_m2s.DataByteEn;
  ocp_m2s_int.DataValid  <= ocp_m2s.DataValid and en_wr;

  -- Data write group s2m - Qualify data accept with command accept or writing
  ocp_s2m.DataAccept <= ocp_s2m_int.DataAccept and en_wr;

  -- Data read group s2m
  ocp_s2m.Data <= ocp_s2m_int.Data;
  ocp_s2m.Tag  <= ocp_s2m_int.Tag;
  ocp_s2m.Resp <= ocp_s2m_int.Resp;

  -- Data read group m2s
  ocp_m2s_int.RespAccept <= ocp_m2s.RespAccept;


  -- OCP cmd/write data end-point processing block
  -- Controls command/write data acceptance to re-generate their correct relative order
  process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      -- Assign default values
      cmd_fifo_wadv   <= '0';
      cmd_fifo_bl8_in <= '0';
      cmd_fifo_w_in   <= '0';
      cmd_fifo_r_in   <= '0';
      mdata_fifo_wadv <= '0';

      -- Process m2s write data
      -- Count each burst of OCP write data
      -- Push a write command into the command FIFO for each single (BC4)/double (BL8) burst of OCP write data
      -- Mark the write command with a BC4/BL8 indicator
      -- Cmd qualified by writing/reading
      -- DataValid qualified by writing
      -- Determine whether first burst is BC4 or BL8 using size of burstlength eg. 1=>BC4, 2=>BL8, 3=>BL8, etc.
      -- Determine whether first burst is BC4 or BL8 using 32-byte alignment eg. 32-byte aligned=>BL8, 16-byte aligned=>BC4
      -- Determine whether suceeding bursts are BC4 or BL8 using size/32-byte alignment of wburst_m2s
      if (ocp_m2s_int.Cmd = OCP_CMD_WRITE)and(ocp_s2m_int.CmdAccept = '1') then
        -- Data valid and accepted in same cycle as command
        if (ocp_m2s_int.DataValid = '1')and(ocp_s2m_int.DataAccept = '1') then
          wburst_m2s <= ocp_m2s_int.BurstLength-1;
          wburst_aligned <= ocp_m2s_int.Addr(4);
          wburst_m2s_first <= '0';
          cmd_fifo_addr_inc <= 4;  -- 4-byte addressing
          cmd_fifo_wadv <= '1';

          -- There are 3 cases:
          -- 32-byte aligned, burst count > 1: BL8 write first cycle
          -- 32-byte aligned, burst count = 1: BC4 write cycle
          -- Not 32-byte aligned, First burst: BC4 write cycle
          if (OR_Reduce(ocp_m2s_int.BurstLength(ADB3_OCP_BURST_WIDTH-1 downto 1)) = '1')and(ocp_m2s_int.Addr(4) = '0') then
            cmd_fifo_bl8_in <= '1'; -- BL8 write first
          else
            cmd_fifo_bl8_in <= '0'; -- BC4 write
          end if;
          cmd_fifo_w_in <= '1';

          mdata_fifo_data_in <= ocp_m2s_int.DataByteEn & ocp_m2s_int.Data;
          mdata_fifo_wadv <= '1';

          -- If burst length of 1 then write burst complete else ongoing
          if (ocp_m2s_int.BurstLength = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH))) then
            writing_m2s <= '0';
          else
            writing_m2s <= '1';
          end if;
        else
          -- Data not valid or accepted in same cycle as command
          wburst_m2s <= ocp_m2s_int.BurstLength;
          wburst_aligned <= not(ocp_m2s_int.Addr(4));
          wburst_m2s_first <= '1';
          writing_m2s <= '1';
        end if;
      else
        if (ocp_m2s_int.DataValid = '1')and(ocp_s2m_int.DataAccept = '1') then
          wburst_m2s <= wburst_m2s-1;
          wburst_aligned <= not(wburst_aligned);
          wburst_m2s_first <= '0';
          cmd_fifo_addr_inc <= 4;  -- 4-byte addressing
          cmd_fifo_wadv <= '1';

          -- There are 4 cases:
          -- 32-byte aligned, burst count > 1: BL8 write first cycle
          -- 32-byte aligned, burst count = 1: BC4 write cycle
          -- Not 32-byte aligned, First burst: BC4 write cycle
          -- Not 32-byte aligned, Not first burst: BL8 write second cycle
          if (wburst_aligned = '1') then
            if (OR_Reduce(wburst_m2s(ADB3_OCP_BURST_WIDTH-1 downto 1)) = '1') then
              cmd_fifo_bl8_in <= '1'; -- BL8 write first
              cmd_fifo_w_in <= '1';
            else
              cmd_fifo_bl8_in <= '0'; -- BC4 write
              cmd_fifo_w_in <= '1';
            end if;
          else
            if (wburst_m2s_first = '1') then
              cmd_fifo_bl8_in <= '0'; -- BC4 write
              cmd_fifo_w_in <= '1';
            else
              cmd_fifo_bl8_in <= '0'; -- BL8 write second
              cmd_fifo_w_in <= '0';
            end if;
          end if;

          mdata_fifo_data_in <= ocp_m2s_int.DataByteEn & ocp_m2s_int.Data;
          mdata_fifo_wadv <= '1';

          -- Last data burst clears writing
          if (wdata_last_m2s = '1') then
            writing_m2s <= '0';
          end if;
        end if;
      end if;


      -- Process m2s read command
      -- Push a single burst read command into the command FIFO for each burst of OCP read command
      -- Cmd qualified by writing/reading
      if (ocp_m2s_int.Cmd = OCP_CMD_READ)and(ocp_s2m_int.CmdAccept = '1') then
        -- There are 3 cases:
        -- 32-byte aligned, burst count > 1 => BL8 read cycle => still aligned
        -- 32-byte aligned, burst count = 1 => BC4 read cycle => end of burst
        -- Not 32-byte aligned              => BC4 read cycle => now aligned
        if (OR_Reduce(ocp_m2s_int.BurstLength(ADB3_OCP_BURST_WIDTH-1 downto 1)) = '1')and(ocp_m2s_int.Addr(4) = '0') then
          rburst_m2s <= ocp_m2s_int.BurstLength-2;
          cmd_fifo_bl8_in <= '1';  -- BL8 read
          cmd_fifo_addr_inc <= 8;  -- 4-byte addressing
          -- If burst length of 1 then BL8 read burst complete else ongoing
          if (ocp_m2s_int.BurstLength = std_logic_vector(to_unsigned(2,ADB3_OCP_BURST_WIDTH))) then
            reading_m2s <= '0';
          else
            reading_m2s <= '1';
          end if;
        else
          rburst_m2s <= ocp_m2s_int.BurstLength-1;
          cmd_fifo_bl8_in <= '0';  -- BC4 read
          cmd_fifo_addr_inc <= 4;  -- 4-byte addressing
          -- If burst length of 1 then BC4 read burst complete else ongoing
          if (ocp_m2s_int.BurstLength = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH))) then
            reading_m2s <= '0';
          else
            reading_m2s <= '1';
          end if;
        end if;
        cmd_fifo_r_in <= '1';
        cmd_fifo_wadv <= '1';
      else
        if (OR_reduce(rburst_m2s) = '1')and(ocp_s2m_int.CmdAccept = '1') then
          -- There are 2 cases:
          -- 32-byte aligned, burst count > 1 => BL8 read cycle => still aligned
          -- 32-byte aligned, burst count = 1 => BC4 read cycle => end of burst
          if (OR_Reduce(rburst_m2s(ADB3_OCP_BURST_WIDTH-1 downto 1)) = '1') then
            rburst_m2s <= rburst_m2s-2;
            cmd_fifo_bl8_in <= '1';  -- BL8 read
            cmd_fifo_addr_inc <= 8;  -- 4-byte addressing
          else
            rburst_m2s <= rburst_m2s-1;
            cmd_fifo_bl8_in <= '0';  -- BC4 read
            cmd_fifo_addr_inc <= 4;  -- 4-byte addressing
          end if;
          cmd_fifo_r_in <= '1';
          cmd_fifo_wadv <= '1';
        end if;

        -- Last data burst clears reading
        if (rdata_last_m2s = '1')and(ocp_s2m_int.CmdAccept = '1') then
          reading_m2s <= '0';
        end if;
      end if;

      -- Store cmd data on OCP command and increment address on cmd_fifo_wadv
      -- Prioritise the store over the increment if they are both actve
      -- Cmd qualified by writing/reading
      -- Convert OCP btye address to MIG 4-byte address
      if ((ocp_m2s_int.Cmd = OCP_CMD_WRITE)or(ocp_m2s_int.Cmd = OCP_CMD_READ))and(ocp_s2m_int.CmdAccept = '1') then
        cmd_fifo_tag_in  <= ocp_m2s_int.Tag;
        cmd_fifo_addr_in <= ocp_m2s_int.Addr(CMD_ADDR_WIDTH+1 downto 2);  -- 4-byte addressing
      elsif (cmd_fifo_wadv = '1') then
        cmd_fifo_addr_in <= cmd_fifo_addr_in + cmd_fifo_addr_inc;  -- 4-byte addressing
      end if;
    end if;
  end process;

  en_bit <= not(writing_m2s or reading_m2s);
  en_cmd <= en_bit & en_bit & en_bit;
--  en_wr  <= '1' when ((ocp_m2s.Cmd = OCP_CMD_WRITE)and(ocp_s2m_int.CmdAccept = '1')and(en_bit = '1'))or(writing_m2s = '1') else '0';
  en_wr  <= writing_m2s;

  -- We have reached the last data in the burst
  wdata_last_m2s <= '1' when (wburst_m2s = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH))) else '0';
  rdata_last_m2s <= '1' when (rburst_m2s = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH)))or
                             (rburst_m2s = std_logic_vector(to_unsigned(2,ADB3_OCP_BURST_WIDTH))) else '0';


  -- Connect s2m accepts
  -- Use cmd FIFO full for data accept as mdata FIFO is never more full than cmd FIFO
  ocp_s2m_int.CmdAccept  <= not(cmd_fifo_full);
  ocp_s2m_int.DataAccept <= not(cmd_fifo_full);

  -- Connect s2m read data
  ocp_s2m_int.Data <= sdata_fifo_data_out(ADB3_OCP_DATA_WIDTH-1 downto 0);
  ocp_s2m_int.Resp <= sdata_fifo_data_out(ADB3_OCP_DATA_WIDTH+1 downto ADB3_OCP_DATA_WIDTH) when sdata_fifo_empty = '0' else OCP_RESP_NONE;
  ocp_s2m_int.Tag  <= sdata_fifo_data_out(SDATA_WIDTH-1 downto ADB3_OCP_DATA_WIDTH+2);


  cmd_fifo_data_in <= cmd_fifo_bl8_in & cmd_fifo_tag_in & cmd_fifo_addr_in & cmd_fifo_r_in & cmd_fifo_w_in;

  cmd_full_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      -- mdata FIFO full is unused as cmd FIFO will always be full first
      cmd_fifo_full <= cmd_fifo_nfull;
    end if;
  end process;

  -- OCP Command FIFO
  cmd_fifo_i : afifo
    generic map(
      width  => CMD_WIDTH,
      order  => CMD_FIFO_DEPTH,
      wpfl   => 4,
      wpfl2  => 0,
      rpel   => 0,
      family => family_virtex5,
      safety => false,
      fflush => false)
    port map(
      rst    => ocp_rst,   -- Asynchronous to wr and rd clocks
      wk     => ocp_clk,
      wd     => cmd_fifo_data_in,
      wadv   => cmd_fifo_wadv,
      we     => open,
      wpf    => cmd_fifo_nfull,
      wpf2   => open,
      wf     => open,
      wovf   => open,
      wlevel => cmd_wlevel,
      rk     => mig_clk,
      rq     => cmd_fifo_data_out,
      radv   => cmd_fifo_radv,
      rflush => '0',
      re     => cmd_fifo_empty,
      rpe    => open,
      rf     => open,
      rudf   => open,
      rlevel => cmd_rlevel);

  -- Advance cmd read pointer if:
  -- MIG initialisation is complete
  -- cmd FIFO is not empty and
  -- mdata FIFO is not empty during a write and
  -- MIG is ready to receive cmd and data and
  -- tag FIFO is not full and
  -- sdata FIFO will not be filled by in-flight data
  stop_cmd_p : process(mig_clk)
  begin
    if rising_edge(mig_clk) then
      phy_init_done_mig <= phy_init_done;
      cmd_ok <= not(tag_fifo_full) and not(sdata_full_in_flight);
      cmd_wr_rd_err <= (cmd_fifo_w_out and cmd_fifo_r_out)or((not(cmd_fifo_w_out or cmd_fifo_r_out))and(cmd_fifo_bl8_out));
    end if;
  end process;

  cmd_fifo_ready <= app_rdy and app_wdf_rdy and not(cmd_fifo_empty) and phy_init_done_mig;
  cmd_fifo_radv  <= cmd_fifo_ready and not(mdata_fifo_empty and cmd_fifo_w_out) and cmd_ok;

  -- Generate MIG command from FIFO
  cmd_fifo_w_out    <= cmd_fifo_data_out(0);
  cmd_fifo_r_out    <= cmd_fifo_data_out(1);
  cmd_fifo_addr_out <= cmd_fifo_data_out(CMD_ADDR_WIDTH+1 downto 2);
  cmd_fifo_tag_out  <= cmd_fifo_data_out(CMD_WIDTH-2 downto CMD_ADDR_WIDTH+2);
  cmd_fifo_bl8_out  <= cmd_fifo_data_out(CMD_WIDTH-1);

  -- MIG command outputs
  -- Convert OCP byte address to MIG Bank/Row/Column address
  -- Byte addresses 00..03 => Col 0 32-bits
  -- Byte addresses 04..07 => Col 1 32-bits
  -- Byte addresses 08..11 => Col 2 32-bits
  -- Byte addresses 12..15 => Col 3 32-bits etc.
  -- Enable command on BC4 data or first cycle of BL8 data
  app_en  <= cmd_fifo_radv and (cmd_fifo_w_out or cmd_fifo_r_out);
  app_cmd <= MIG_WR_CMD when (cmd_fifo_w_out = '1') else MIG_RD_CMD;
  app_sz  <= cmd_fifo_bl8_out;

  -- ROW address width will change depending on density of memory (1Gb/2Gb).
  -- ROW address bits need to be the MS bits of OCP 4-byte address so MIG core for 2Gb will cope with both
  -- So OCP 4-byte address is organised as ROW, BANK, COL
  app_row  <= cmd_fifo_addr_out(CMD_ADDR_WIDTH-1 downto app_bank_width+app_col_width);
  app_bank <= cmd_fifo_addr_out(app_bank_width+app_col_width-1 downto app_col_width);
  app_col  <= cmd_fifo_addr_out(app_col_width-1 downto 0);

  -- MIG address app_addr is organised as RANK (unused), BANK, ROW, COL
--  app_addr <= '0' & app_bank & app_row & app_col;
  app_addr <= '0' & cmd_fifo_addr_out;

  -- OCP write data to FIFO
  wr_data_full_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      -- mdata FIFO full is unused as cmd FIFO will always be full first
      mdata_fifo_full <= mdata_fifo_nfull;
    end if;
  end process;

  -- OCP write data FIFO
  mdata_fifo_i : afifo
    generic map(
      width  => MDATA_WIDTH,
      order  => CMD_FIFO_DEPTH,
      wpfl   => 4,
      wpfl2  => 0,
      rpel   => 0,
      family => family_virtex5,
      safety => false,
      fflush => false)
    port map(
      rst    => ocp_rst,   -- Asynchronous to wr and rd clocks
      wk     => ocp_clk,
      wd     => mdata_fifo_data_in,
      wadv   => mdata_fifo_wadv,
      we     => open,
      wpf    => mdata_fifo_nfull,
      wpf2   => open,
      wf     => open,
      wovf   => open,
      wlevel => mdata_wlevel,
      rk     => mig_clk,
      rq     => mdata_fifo_data_out,
      radv   => mdata_fifo_radv,
      rflush => '0',
      re     => mdata_fifo_empty,
      rpe    => open,
      rf     => open,
      rudf   => open,
      rlevel => mdata_rlevel);

  -- OCP write data to FIFO
  w_last_p : process(mig_clk)
  begin
    if rising_edge(mig_clk) then
      if (mdata_fifo_radv = '1') then
        -- Indicates BL8 second cycle
        cmd_fifo_w_last <= cmd_fifo_w_out and cmd_fifo_bl8_out;
      end if;
    end if;
  end process;

  -- Advance mdata read pointer if:
  -- MIG initialisation is complete
  -- data FIFO is not empty and
  -- MIG is ready to receive cmd and data and
  mdata_fifo_ready <= app_rdy and app_wdf_rdy and not(mdata_fifo_empty) and phy_init_done_mig;
  mdata_fifo_radv  <= mdata_fifo_ready when (cmd_fifo_w_out = '1')or(cmd_fifo_w_last = '1') else '0';

  -- Generate MIG write data from FIFO
  mdata_fifo_be_out <= mdata_fifo_data_out(MDATA_WIDTH-1 downto ADB3_OCP_DATA_WIDTH);
  mdata_fifo_d_out  <= mdata_fifo_data_out(ADB3_OCP_DATA_WIDTH-1 downto 0);

  app_wdf_wren <= mdata_fifo_radv;
  app_wdf_data <= mdata_fifo_d_out;
  app_wdf_mask <= not(mdata_fifo_be_out);   -- Active low at MIG
  --  Enable end on BC4 data or second cycle of BL8 data
  app_wdf_end  <= mdata_fifo_radv and not(cmd_fifo_bl8_out);


  -- OCP read data from MIG
  sdata_fifo_tag_in  <= tag_fifo_tag_out;
  sdata_fifo_resp_in <= OCP_RESP_VALID;
  sdata_fifo_d_in    <= app_rd_data;

  sdata_fifo_data_in <= sdata_fifo_tag_in & sdata_fifo_resp_in & sdata_fifo_d_in;

  -- MIG write data to FIFO
  rd_data_full_p : process(mig_clk)
  begin
    if rising_edge(mig_clk) then
      sdata_fifo_full <= sdata_fifo_nfull;
      sdata_wr_full_err <= app_rd_data_valid and sdata_fifo_full and phy_init_done_mig;
    end if;
  end process;

  sdata_fifo_wadv <= app_rd_data_valid and not(sdata_fifo_full) and phy_init_done_mig;

  -- OCP read data FIFO
  -- Order = 5 => 32 read data results can be stored
  -- If there is less space in the FIFO than the (number of reads in flight)*2
  -- (assume all reads = BL8) then we must stop issuing read commands
  sdata_fifo_i : afifo
    generic map(
      width  => SDATA_WIDTH,
      order  => RESP_FIFO_DEPTH,
      wpfl   => 4,
      wpfl2  => 0,
      rpel   => 0,
      family => family_virtex5,
      safety => false,
      fflush => false)
    port map(
      rst    => ocp_rst,   -- Asynchronous to wr and rd clocks
      wk     => mig_clk,
      wd     => sdata_fifo_data_in,
      wadv   => sdata_fifo_wadv,
      we     => open,
      wpf    => sdata_fifo_nfull,
      wpf2   => open,
      wf     => open,
      wovf   => open,
      wlevel => sdata_wlevel,
      rk     => ocp_clk,
      rq     => sdata_fifo_data_out,
      radv   => sdata_fifo_radv,
      rflush => '0',
      re     => sdata_fifo_empty,
      rpe    => open,
      rf     => open,
      rudf   => open,
      rlevel => sdata_rlevel);

  -- OCP read data from FIFO
  sdata_fifo_radv <= ocp_m2s.RespAccept and not(sdata_fifo_empty);


  -- Push BC4/BL8 indicator and read command tag value into tag FIFO on read command acceptance
  tag_fifo_data_in <= cmd_fifo_bl8_out & cmd_fifo_tag_out;

  tag_fifo_wadv <= cmd_fifo_radv when (cmd_fifo_r_out = '1')and(tag_fifo_full = '0') else '0';

  -- read tag FIFO
  -- Order = 5 => 32 read command tags can be stored
  -- Max MIG read commands that may be issued before first returned data is 32
  -- Otherwise FIFO will become full and MIG read commands will be stopped
  -- If higher latency is required, then increase size of FIFO
  tag_fifo_i : sfifo
  generic map(
    width  => TAG_WIDTH,      -- natural
    order  => TAG_FIFO_DEPTH, -- natural
    wpfl   => 4,              -- natural
    rpel   => 0,              -- natural
    family => family_virtex5, -- work.fifo.family_t
    safety => false)          -- boolean)
  port map(
    rst    => mig_rst,           -- in    std_logic
    k      => mig_clk,           -- in    std_logic
    wd     => tag_fifo_data_in,  -- in    std_logic_vector(width - 1 downto 0)
    wadv   => tag_fifo_wadv,     -- in    std_logic
    we     => open,              -- out   std_logic
    wpf    => tag_fifo_nfull,    -- out   std_logic
    wf     => open,              -- out   std_logic
    wovf   => open,              -- out   std_logic
    rq     => tag_fifo_data_out, -- out   std_logic_vector(width - 1 downto 0)
    radv   => tag_fifo_radv,     -- in    std_logic
    rflush => '0',               -- in    std_logic
    re     => tag_fifo_empty,    -- out   std_logic
    rpe    => open,              -- out   std_logic
    rf     => open,              -- out   std_logic
    rudf   => tag_rd_empty_err,  -- out   std_logic
    rlevel => tag_rlevel);       -- out   std_logic_vector(15 downto 0))


  -- MIG read command controls write side of FIFO
  tag_p : process(mig_clk)
  begin
    if rising_edge(mig_clk) then
      tag_fifo_full <= tag_fifo_nfull;
      if (cmd_fifo_r_out = '1')and(tag_fifo_full = '1') then
        tag_wr_full_err <= cmd_fifo_radv;
      else
        tag_wr_full_err <= '0';
      end if;
    end if;
  end process;

  tag_fifo_radv <= tag_fifo_bl_comp and sdata_fifo_wadv;

  -- tag_fifo_bl8_out: next read data burst length 0=BC4, 1=BL8
  -- tag_fifo_tag_out: next read data tag
  tag_fifo_bl8_out <= tag_fifo_data_out(TAG_WIDTH-1);
  tag_fifo_tag_out <= tag_fifo_data_out(ADB3_OCP_TAG_WIDTH-1 downto 0);

  -- Read next tag value when BC4 or BL8 data finished and tag FIFO not empty
  -- Count valid read data in BC4 (1) or BL8 (2) until we are finished with tag
  tag_bl_cnt_p : process(mig_clk)
  begin
    if rising_edge(mig_clk) then
      if (tag_fifo_radv = '1') then
        tag_fifo_bl_cnt <= '0';
      elsif (sdata_fifo_wadv = '1') then
        tag_fifo_bl_cnt <= '1';
      end if;
    end if;
  end process;

  tag_fifo_bl_comp <= '1' when (tag_fifo_bl_cnt = tag_fifo_bl8_out) else '0';

  -- Compare sdata FIFO space (sdata_wlevel) with read data in flight (tag_rlevel)
  -- Assume every read command is BL8 and so double tag_rlevel
  sdata_fifo_space   <= sdata_wlevel(RESP_FIFO_DEPTH-1 downto 0);
  tag_fifo_in_flight <= tag_rlevel(TAG_FIFO_DEPTH-1 downto 0) & '0';

  -- Must account for sdata FIFO full on wlevel = 4, and 2 cycle delay from level counts to stopping
  -- read commands, by adding 4+2x2 = 8 to tag_fifo_in_flight
  in_flight_p : process(mig_clk)
  begin
    if rising_edge(mig_clk) then
      if (sdata_fifo_space > (tag_fifo_in_flight + 10)) then
        sdata_full_in_flight <= '0';
      else
        sdata_full_in_flight <= '1';
      end if;
      if (sdata_fifo_space >= tag_fifo_in_flight + 3) then
        sdata_ovf_in_flight <= '0';
      else
        sdata_ovf_in_flight <= '1';
      end if;
    end if;
  end process;


  -- Error tracking
  count_errors_p : process(mig_clk)
  begin
    if rising_edge(mig_clk) then
      -- Count number of cycles of cmd_wr_rd_err
      -- That is cmd FIFO indicates write and read
      if (cmd_wr_rd_err = '1') then
        cmd_wr_rd_cnt <= cmd_wr_rd_cnt + 1;
        assert false
          report "(adb3_ocp_ocp2ddr3): cmd FIFO indicates write and read."
          severity error;
      end if;

      -- Count number of cycles of sdata_wr_full_err
      -- That is sdata FIFO is full and data is lost
      if (sdata_wr_full_err = '1') then
        sdata_wr_full_cnt <= sdata_wr_full_cnt + 1;
        assert false
          report "(adb3_ocp_ocp2ddr3): sdata FIFO full: DDR3 read data lost."
          severity error;
      end if;

      -- Count number of cycles of tag_wr_full_err
      -- That is tag FIFO is full and read data tags are lost
      if (tag_wr_full_err = '1') then
        tag_wr_full_cnt <= tag_wr_full_cnt + 1;
        assert false
          report "(adb3_ocp_ocp2ddr3): Tag FIFO full: OCP read command tag ignored."
          severity error;
      end if;
    end if;
  end process;

  max_levels_p : process(mig_clk)
  begin
    if rising_edge(mig_clk) then
      -- Store Max tag FIFO level
      if (tag_rlevel(TAG_FIFO_DEPTH-1 downto 0) > max_tag_rlevel(TAG_FIFO_DEPTH-1 downto 0)) then
        max_tag_rlevel(TAG_FIFO_DEPTH-1 downto 0) <= tag_rlevel(TAG_FIFO_DEPTH-1 downto 0);
      end if;

      -- Store Max sdata FIFO level
      if (sdata_rlevel(RESP_FIFO_DEPTH-1 downto 0) > max_sdata_rlevel(RESP_FIFO_DEPTH-1 downto 0)) then
        max_sdata_rlevel(RESP_FIFO_DEPTH-1 downto 0) <= sdata_rlevel(RESP_FIFO_DEPTH-1 downto 0);
      end if;
    end if;
  end process;

  debug_info <= max_tag_rlevel & max_sdata_rlevel & tag_wr_full_cnt & sdata_wr_full_cnt;
end;