--==============================================================================
--
--     File Name: adb3_ocp_pkg.vhd
--       Version: 0.2
--          Date: 09 Feb 2011
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: Package adb3_ocp_pkg
--
--   Description: Package defining ADB3 OCP profile.
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2009-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 06 Nov 09: Initial design
--                0.2, TW,  09 Feb 2011: Added some more useful constants
--                  and types.
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;

package adb3_ocp is

  -- Define ADB3 OCP widths
  constant ADB3_OCP_DATA_WIDTH      : integer := 128;
  constant ADB3_OCP_DATA_WIDTH_LOG2 : integer := 7;
  constant ADB3_OCP_BE_WIDTH        : integer := 16;
  constant ADB3_OCP_BE_WIDTH_LOG2   : integer := 4;
  constant ADB3_OCP_ADDR_WIDTH      : integer := 64;
  constant ADB3_OCP_BURST_WIDTH     : integer := 12;
  constant ADB3_OCP_TAG_WIDTH       : integer := 8;
  constant ADB3_OCP_CMD_WIDTH       : integer := 3;
  constant ADB3_OCP_RESP_WIDTH      : integer := 2;

  -- Define ADB3 OCP subtypes
  subtype adb3_ocp_data_s  is std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0);
  subtype adb3_ocp_be_s    is std_logic_vector(ADB3_OCP_BE_WIDTH-1 downto 0);
  subtype adb3_ocp_addr_s  is std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0);
  subtype adb3_ocp_burst_s is std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
  subtype adb3_ocp_tag_s   is std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);

  subtype ocp_CmdT is std_logic_vector(ADB3_OCP_CMD_WIDTH-1 downto 0);
  constant OCP_CMD_IDLE         : ocp_CmdT := "000";
  constant OCP_CMD_WRITE        : ocp_CmdT := "001";
  constant OCP_CMD_READ         : ocp_CmdT := "010";
  constant OCP_CMD_WRITENONPOST : ocp_CmdT := "101";

  subtype ocp_RespT is std_logic_vector(ADB3_OCP_RESP_WIDTH-1 downto 0);
  constant OCP_RESP_NONE   : ocp_RespT := "00";
  constant OCP_RESP_VALID  : ocp_RespT := "01";
  constant OCP_RESP_FAILED : ocp_RespT := "10";
  constant OCP_RESP_ERROR  : ocp_RespT := "11";

  constant adb3_ocp_m2sT_slv_length  : natural := ADB3_OCP_CMD_WIDTH + ADB3_OCP_ADDR_WIDTH + ADB3_OCP_BURST_WIDTH + ADB3_OCP_DATA_WIDTH + ADB3_OCP_BE_WIDTH + 1 + 1 + ADB3_OCP_TAG_WIDTH;
  constant adb3_ocp_s2mT_slv_length  : natural := 1 + 1 + ADB3_OCP_DATA_WIDTH + ADB3_OCP_RESP_WIDTH + ADB3_OCP_TAG_WIDTH;

  -- Burst Capable bus for high performance data flows
  -- Block Data Flow from Master to Slave
  type adb3_ocp_m2sT is record
    -- Request group
    Cmd         : ocp_CmdT;   -- Command
    Addr        : std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0);
    BurstLength : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
    -- DataHandshake group
    Data        : std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0);
    DataByteEn  : std_logic_vector(ADB3_OCP_BE_WIDTH-1 downto 0);
    DataValid   : std_logic;  -- Qualifies Data and the other signals in bundle
    -- Response group
    RespAccept  : std_logic;  -- Master indication read data was accepted
    -- Tag group
    Tag         : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
  end record adb3_ocp_m2sT;

  -- Block Data Flow from Slave to Master
  type adb3_ocp_s2mT is record
    -- Request group
    CmdAccept  : std_logic;   -- Slave indication command/request accepted
    -- DataHandshake group
    DataAccept : std_logic;   -- Slave indication write data accepted
    -- Response group
    Data       : std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0);
    Resp       : ocp_RespT;   -- Slave read response
    -- Tag group
    Tag        : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
  end record adb3_ocp_s2mT;

  type adb3_ocp_s2mT_array is array (natural range <>) of adb3_ocp_s2mT;
  type adb3_ocp_m2sT_array is array (natural range <>) of adb3_ocp_m2sT;


  -- Define type defaults
  constant ADB3_OCP_DEFAULT_M2S : adb3_ocp_m2sT :=
    (Cmd         => OCP_CMD_IDLE,
     Addr        => (others => '0'),
     BurstLength => (others => '0'),
     Data        => (others => '0'),
     DataByteEn  => (others => '0'),
     DataValid   => '0',
     RespAccept  => '0',
     Tag         => (others => '0'));

  constant ADB3_OCP_DEFAULT_S2M : adb3_ocp_s2mT :=
    (CmdAccept  => '0',
     DataAccept => '0',
     Data       => (others => '0'),
     Resp       => OCP_RESP_NONE,
     Tag        => (others => '0'));

  -- Define type tri-states
  constant ADB3_OCP_Z_M2S : adb3_ocp_m2sT :=
    (Cmd         => (others => 'Z'),
     Addr        => (others => 'Z'),
     BurstLength => (others => 'Z'),
     Data        => (others => 'Z'),
     DataByteEn  => (others => 'Z'),
     DataValid   => 'Z',
     RespAccept  => 'Z',
     Tag         => (others => 'Z'));

  constant ADB3_OCP_Z_S2M : adb3_ocp_s2mT :=
    (CmdAccept  => 'Z',
     DataAccept => 'Z',
     Data       => (others => 'Z'),
     Resp       => (others => 'Z'),
     Tag        => (others => 'Z'));

  function to_std_logic(bool : boolean) return std_logic;
  function to_std_logic_vector(x : adb3_ocp_m2sT) return std_logic_vector;
  function to_std_logic_vector(x : adb3_ocp_s2mT) return std_logic_vector;
  function to_std_logic_vector(x : adb3_ocp_m2sT_array) return std_logic_vector;
  function to_std_logic_vector(x : adb3_ocp_s2mT_array) return std_logic_vector;
  function to_adb3_ocp_m2sT(x : std_logic_vector) return adb3_ocp_m2sT;
  function to_adb3_ocp_s2mT(x : std_logic_vector) return adb3_ocp_s2mT;
  function to_adb3_ocp_m2sT_array(x : std_logic_vector) return adb3_ocp_m2sT_array;
  function to_adb3_ocp_s2mT_array(x : std_logic_vector) return adb3_ocp_s2mT_array;

end package adb3_ocp;

package body adb3_ocp is

  function to_std_logic (bool : boolean) return std_logic is
  begin
    if (bool) then
      return '1';
    else
      return '0';
    end if;
  end function to_std_logic;


  function to_std_logic_vector(x : adb3_ocp_m2sT) return std_logic_vector is
    variable y : std_logic_vector(adb3_ocp_m2sT_slv_length-1 downto 0);
  begin
    y(2 downto 0) := x.Cmd;
    y(ADB3_OCP_ADDR_WIDTH-1 + 3 downto 3) := x.Addr;
    y(ADB3_OCP_BURST_WIDTH-1 + ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_ADDR_WIDTH+3) := x.BurstLength;
    y(ADB3_OCP_DATA_WIDTH-1 + ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.Data;
    y(ADB3_OCP_BE_WIDTH-1 + ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.DataByteEn;
    y(ADB3_OCP_BE_WIDTH+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.DataValid;
    y(1+ADB3_OCP_BE_WIDTH+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.RespAccept;
    y(ADB3_OCP_TAG_WIDTH-1 + 1+1+ADB3_OCP_BE_WIDTH+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto 1+1+ADB3_OCP_BE_WIDTH+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.Tag;
    return y;
  end function to_std_logic_vector;

  function to_std_logic_vector(x : adb3_ocp_s2mT) return std_logic_vector is
    variable y : std_logic_vector(adb3_ocp_s2mT_slv_length-1 downto 0);
  begin
    y(0) := x.CmdAccept;
    y(1) := x.DataAccept;
    y(ADB3_OCP_DATA_WIDTH-1+2 downto 2) := x.Data;
    y(1 + ADB3_OCP_DATA_WIDTH+2 downto ADB3_OCP_DATA_WIDTH+2) := x.Resp;
    y(ADB3_OCP_TAG_WIDTH-1 + 2+ADB3_OCP_DATA_WIDTH+2 downto 2+ADB3_OCP_DATA_WIDTH+2) := x.Tag;
    return y;
  end function to_std_logic_vector;

  function to_std_logic_vector(x : adb3_ocp_m2sT_array) return std_logic_vector is
    variable y : std_logic_vector(adb3_ocp_m2sT_slv_length*x'length-1 downto 0);
  begin
    for i in 0 to x'length-1 loop
      y((i+1)*adb3_ocp_m2sT_slv_length-1 downto i*adb3_ocp_m2sT_slv_length) := to_std_logic_vector(x(i));
    end loop;  -- i
    return y;
  end function to_std_logic_vector;


  function to_std_logic_vector(x : adb3_ocp_s2mT_array) return std_logic_vector is
    variable y : std_logic_vector(adb3_ocp_s2mT_slv_length*x'length-1 downto 0);
  begin
    for i in 0 to x'length-1 loop
      y((i+1)*adb3_ocp_s2mT_slv_length-1 downto i*adb3_ocp_s2mT_slv_length) := to_std_logic_vector(x(i));
    end loop;  -- i
    return y;
  end function to_std_logic_vector;


  function to_adb3_ocp_m2sT(x : std_logic_vector) return adb3_ocp_m2sT is
    variable y : adb3_ocp_m2sT;
  begin
    y.Cmd := x(2 downto 0);
    y.Addr := x(ADB3_OCP_ADDR_WIDTH-1 + 3 downto 3);
    y.BurstLength := x(ADB3_OCP_BURST_WIDTH-1 + ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_ADDR_WIDTH+3);
    y.Data := x(ADB3_OCP_DATA_WIDTH-1 + ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    y.DataByteEn := x(ADB3_OCP_BE_WIDTH-1 + ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    y.DataValid := x(ADB3_OCP_BE_WIDTH+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    y.RespAccept := x(1+ADB3_OCP_BE_WIDTH+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    y.Tag := x(ADB3_OCP_TAG_WIDTH-1 + 1+1+ADB3_OCP_BE_WIDTH+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto 1+1+ADB3_OCP_BE_WIDTH+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    return y;
  end function to_adb3_ocp_m2sT;

  function to_adb3_ocp_s2mT(x : std_logic_vector) return adb3_ocp_s2mT is
    variable y : adb3_ocp_s2mT;
  begin
    y.CmdAccept := x(0);
    y.DataAccept := x(1);
    y.Data := x(ADB3_OCP_DATA_WIDTH-1+2 downto 2);
    y.Resp := x(1 + ADB3_OCP_DATA_WIDTH+2 downto ADB3_OCP_DATA_WIDTH+2);
    y.Tag := x(ADB3_OCP_TAG_WIDTH-1 + 2+ADB3_OCP_DATA_WIDTH+2 downto 2+ADB3_OCP_DATA_WIDTH+2);
    return y;
  end function to_adb3_ocp_s2mT;

  function to_adb3_ocp_m2sT_array(x : std_logic_vector) return adb3_ocp_m2sT_array is
    variable y : adb3_ocp_m2sT_array(0 to (x'length/adb3_ocp_m2sT_slv_length) -1);
    variable x_i : std_logic_vector(adb3_ocp_m2sT_slv_length-1 downto 0);
  begin
    for i in 0 to (x'length/adb3_ocp_m2sT_slv_length)-1 loop
      x_i := x((i+1)*adb3_ocp_m2sT_slv_length-1 downto i*adb3_ocp_m2sT_slv_length);
      y(i) := to_adb3_ocp_m2sT(x_i);
    end loop;  -- i
    return y;
  end function to_adb3_ocp_m2sT_array;

  function to_adb3_ocp_s2mT_array(x : std_logic_vector) return adb3_ocp_s2mT_array is
    variable y : adb3_ocp_s2mT_array(0 to (x'length/adb3_ocp_s2mT_slv_length) -1);
    variable x_i : std_logic_vector(adb3_ocp_s2mT_slv_length-1 downto 0);
  begin
    for i in 0 to (x'length/adb3_ocp_s2mT_slv_length)-1 loop
      x_i := x((i+1)*adb3_ocp_s2mT_slv_length-1 downto i*adb3_ocp_s2mT_slv_length);
      y(i) := to_adb3_ocp_s2mT(x_i);
    end loop;  -- i
    return y;
  end function to_adb3_ocp_s2mT_array;

end package body adb3_ocp;

