--==============================================================================
--
--     File Name: adb3_ocp_retime_nb.vhd
--       Version: 0.3
--          Date: 09 Feb 2011
--         Model: synth
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: ADB3 OCP re-time (non-blocking)
--
--   Description: Component for re-time of OCP port using asynchronous FIFOs
--                constructed from SRL16s
--
--   Limitations:
--
--         Notes: Non-blocking component
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 04 Nov 2010: Initial design
--                0.2, KDC  09 Feb 2011: Replaced AFIFO design with SRL16 design
--                0.3, TW   09 Feb 2011: Replaced FDCE primitives with inferred
--                  flip-flops and applied SHREG_EXTRACT = "no" attributes, to
--                  eliminate unused output warnings when outputs of FDCEs are
--                  optimized away.
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;

library unisim;
use unisim.vcomponents.all;

entity adb3_ocp_retime_nb is
port(
  ocp_rst    : in  std_logic;
  ocp_clk    : in  std_logic;
  -- OCP primary port
  slave_m2s  : in  adb3_ocp_m2sT;
  slave_s2m  : out adb3_ocp_s2mT;
  -- OCP secondary port
  master_m2s : out adb3_ocp_m2sT;
  master_s2m : in  adb3_ocp_s2mT);
end;

architecture mixed of adb3_ocp_retime_nb is

--  signal cmdaccept_reg  : std_logic := '0';
--  signal dataaccept_reg : std_logic := '0';
--  signal respaccept_reg : std_logic := '0';

  signal slave_s2m_i   : adb3_ocp_s2mT := ADB3_OCP_DEFAULT_S2M;
  signal master_m2s_i  : adb3_ocp_m2sT := ADB3_OCP_DEFAULT_M2S;

  signal slave_s2m_ii   : adb3_ocp_s2mT := ADB3_OCP_DEFAULT_S2M;
  signal master_m2s_ii  : adb3_ocp_m2sT := ADB3_OCP_DEFAULT_M2S;

  signal srl_a_cmdpath  : std_logic_vector(3 downto 0) := "0000";
  signal srl_a_datapath : std_logic_vector(3 downto 0) := "0000";
  signal srl_a_resppath : std_logic_vector(3 downto 0) := "0000";

  signal srl_ce_cmdpath  : std_logic := '1';
  signal srl_ce_datapath : std_logic := '1';
  signal srl_ce_resppath : std_logic := '1';

  signal reg_ce_cmdpath  : std_logic := '1';
  signal reg_ce_datapath : std_logic := '1';
  signal reg_ce_resppath : std_logic := '1';

  attribute SHREG_EXTRACT : string;
  attribute SHREG_EXTRACT of master_m2s_ii : signal is "no";
  attribute SHREG_EXTRACT of slave_s2m_ii  : signal is "no";

begin

  slave_s2m <= slave_s2m_ii;
  master_m2s  <= master_m2s_ii;

  reg_ce_cmdpath  <= '1' when (master_s2m.CmdAccept = '1')  or (master_m2s_ii.Cmd = OCP_CMD_IDLE)  else '0';
--  reg_ce_datapath <= master_s2m.DataAccept or not master_m2s_ii.DataValid;
  reg_ce_datapath <= '1' when (master_s2m.DataAccept = '1') or (master_m2s_ii.DataValid = '0')     else '0';
  reg_ce_resppath <= '1' when (slave_m2s.RespAccept = '1')  or (slave_s2m_ii.Resp = OCP_RESP_NONE) else '0';

  -- Cmd Group
  srl_cmd : for i in 0 to 2 generate
    SRL16E_inst : SRL16E
      generic map (
        INIT => X"0000")
      port map (
        Q   => master_m2s_i.Cmd(i),
        A0  => srl_a_cmdpath(0),
        A1  => srl_a_cmdpath(1),
        A2  => srl_a_cmdpath(2),
        A3  => srl_a_cmdpath(3),
        CE  => srl_ce_cmdpath,
        CLK => ocp_clk,
        D   => slave_m2s.Cmd(i));

    FDCE0 : process(ocp_clk)
    begin
      if rising_edge(ocp_clk) then
        if reg_ce_cmdpath = '1' then
          master_m2s_ii.Cmd(i) <= master_m2s_i.Cmd(i);
        end if;
      end if;
    end process;
  end generate srl_cmd;

  srl_addr : for i in 0 to ADB3_OCP_ADDR_WIDTH-1 generate
    SRL16E_inst : SRL16E
      generic map (
        INIT => X"0000")
      port map (
        Q   => master_m2s_i.Addr(i),
        A0  => srl_a_cmdpath(0),
        A1  => srl_a_cmdpath(1),
        A2  => srl_a_cmdpath(2),
        A3  => srl_a_cmdpath(3),
        CE  => srl_ce_cmdpath,
        CLK => ocp_clk,
        D   => slave_m2s.Addr(i));

    FDCE0 : process(ocp_clk)
    begin
      if rising_edge(ocp_clk) then
        if reg_ce_cmdpath = '1' then
          master_m2s_ii.Addr(i) <= master_m2s_i.Addr(i);
        end if;
      end if;
    end process;
  end generate srl_addr;

  srl_burst : for i in 0 to ADB3_OCP_BURST_WIDTH-1 generate
    SRL16E_inst : SRL16E
      generic map (
        INIT => X"0000")
      port map (
        Q   => master_m2s_i.BurstLength(i),
        A0  => srl_a_cmdpath(0),
        A1  => srl_a_cmdpath(1),
        A2  => srl_a_cmdpath(2),
        A3  => srl_a_cmdpath(3),
        CE  => srl_ce_cmdpath,
        CLK => ocp_clk,
        D   => slave_m2s.BurstLength(i));

    FDCE0 : process(ocp_clk)
    begin
      if rising_edge(ocp_clk) then
        if reg_ce_cmdpath = '1' then
          master_m2s_ii.BurstLength(i) <= master_m2s_i.BurstLength(i);
        end if;
      end if;
    end process;
  end generate srl_burst;

  srl_mtag : for i in 0 to ADB3_OCP_TAG_WIDTH-1 generate
    SRL16E_inst : SRL16E
      generic map (
        INIT => X"0000")
      port map (
        Q   => master_m2s_i.Tag(i),
        A0  => srl_a_cmdpath(0),
        A1  => srl_a_cmdpath(1),
        A2  => srl_a_cmdpath(2),
        A3  => srl_a_cmdpath(3),
        CE  => srl_ce_cmdpath,
        CLK => ocp_clk,
        D   => slave_m2s.Tag(i));

    FDCE0 : process(ocp_clk)
    begin
      if rising_edge(ocp_clk) then
        if reg_ce_cmdpath = '1' then
          master_m2s_ii.Tag(i) <= master_m2s_i.Tag(i);
        end if;
      end if;
    end process;
  end generate srl_mtag;


  -- Write Data Group
  srl_data : for i in 0 to ADB3_OCP_DATA_WIDTH-1 generate
    SRL16E_inst : SRL16E
      generic map (
        INIT => X"0000")
      port map (
        Q   => master_m2s_i.Data(i),
        A0  => srl_a_datapath(0),
        A1  => srl_a_datapath(1),
        A2  => srl_a_datapath(2),
        A3  => srl_a_datapath(3),
        CE  => srl_ce_datapath,
        CLK => ocp_clk,
        D   => slave_m2s.Data(i));

    FDCE0 : process(ocp_clk)
    begin
      if rising_edge(ocp_clk) then
        if reg_ce_datapath = '1' then
          master_m2s_ii.Data(i) <= master_m2s_i.Data(i);
        end if;
      end if;
    end process;
  end generate srl_data;

  srl_databe : for i in 0 to ADB3_OCP_DATA_WIDTH/8-1 generate
    SRL16E_inst : SRL16E
      generic map (
        INIT => X"0000")
      port map (
        Q   => master_m2s_i.DataByteEn(i),
        A0  => srl_a_datapath(0),
        A1  => srl_a_datapath(1),
        A2  => srl_a_datapath(2),
        A3  => srl_a_datapath(3),
        CE  => srl_ce_datapath,
        CLK => ocp_clk,
        D   => slave_m2s.DataByteEn(i));

    FDCE0 : process(ocp_clk)
    begin
      if rising_edge(ocp_clk) then
        if reg_ce_datapath = '1' then
          master_m2s_ii.DataByteEn(i) <= master_m2s_i.DataByteEn(i);
        end if;
      end if;
    end process;
  end generate srl_databe;

  SRL16E_datavalid : SRL16E
    generic map (
      INIT => X"0000")
    port map (
      Q   => master_m2s_i.DataValid,
      A0  => srl_a_datapath(0),
      A1  => srl_a_datapath(1),
      A2  => srl_a_datapath(2),
      A3  => srl_a_datapath(3),
      CE  => srl_ce_datapath,
      CLK => ocp_clk,
      D   => slave_m2s.DataValid);

  FDCE_datavalid : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      if reg_ce_datapath = '1' then
        master_m2s_ii.DataValid <= master_m2s_i.DataValid;
      end if;
    end if;
  end process;

  -- Response Group
  srl_rdata : for i in 0 to ADB3_OCP_DATA_WIDTH-1 generate
    SRL16E_inst : SRL16E
      generic map (
        INIT => X"0000")
      port map (
        Q   => slave_s2m_i.Data(i),
        A0  => srl_a_resppath(0),
        A1  => srl_a_resppath(1),
        A2  => srl_a_resppath(2),
        A3  => srl_a_resppath(3),
        CE  => srl_ce_resppath,
        CLK => ocp_clk,
        D   => master_s2m.Data(i));

    FDCE0 : process(ocp_clk)
    begin
      if rising_edge(ocp_clk) then
        if reg_ce_resppath = '1' then
          slave_s2m_ii.Data(i) <= slave_s2m_i.Data(i);
        end if;
      end if;
    end process;
  end generate srl_rdata;

  srl_resp : for i in 0 to 1 generate
    SRL16E_inst : SRL16E
      generic map (
        INIT => X"0000")
      port map (
        Q   => slave_s2m_i.Resp(i),
        A0  => srl_a_resppath(0),
        A1  => srl_a_resppath(1),
        A2  => srl_a_resppath(2),
        A3  => srl_a_resppath(3),
        CE  => srl_ce_resppath,
        CLK => ocp_clk,
        D   => master_s2m.Resp(i));

    FDCE0 : process(ocp_clk)
    begin
      if rising_edge(ocp_clk) then
        if reg_ce_resppath = '1' then
          slave_s2m_ii.Resp(i) <= slave_s2m_i.Resp(i);
        end if;
      end if;
    end process;
  end generate srl_resp;

  srl_rtag : for i in 0 to ADB3_OCP_TAG_WIDTH-1 generate
    SRL16E_inst : SRL16E
      generic map (
        INIT => X"0000")
      port map (
        Q   => slave_s2m_i.Tag(i),
        A0  => srl_a_resppath(0),
        A1  => srl_a_resppath(1),
        A2  => srl_a_resppath(2),
        A3  => srl_a_resppath(3),
        CE  => srl_ce_resppath,
        CLK => ocp_clk,
        D   => master_s2m.Tag(i));

    FDCE0 : process(ocp_clk)
    begin
      if rising_edge(ocp_clk) then
        if reg_ce_resppath = '1' then
          slave_s2m_ii.Tag(i) <= slave_s2m_i.Tag(i);
        end if;
      end if;
    end process;
  end generate srl_rtag;


  slave_s2m_ii.CmdAccept    <= srl_ce_cmdpath;
  slave_s2m_ii.DataAccept   <= srl_ce_datapath;
  master_m2s_ii.RespAccept  <= srl_ce_resppath;

  reg_ocp0 : process(ocp_clk)
  begin
    if ocp_clk'event and ocp_clk = '1' then     -- rising clock edge

      -- Cmd Group
      if srl_ce_cmdpath = '1' then
        if master_m2s_i.Cmd /= OCP_CMD_IDLE and reg_ce_cmdpath = '0' then
          srl_ce_cmdpath <= '0';
          srl_a_cmdpath  <= "0001";
        end if;
      else
        if srl_a_cmdpath /= "0000" then
          if master_m2s_i.Cmd = OCP_CMD_IDLE or reg_ce_cmdpath = '1' then
            srl_ce_cmdpath <= '1';  -- Added
            srl_a_cmdpath  <= "0000";
          end if;
--        else
--          if master_m2s_i.Cmd = OCP_CMD_IDLE or reg_ce_cmdpath = '1' then
--            srl_ce_cmdpath <= '1';
--          end if;
        end if;
      end if;

      -- Write Data Group

--    Able to accept more data
      if srl_ce_datapath = '1' then
--      Select SR(0) and accept more data until full
--      SR is full
        if reg_ce_datapath = '0' and master_m2s_i.DataValid = '1' then
          srl_ce_datapath <= '0';
          srl_a_datapath  <= "0001";
        end if;
--    Unable to accept more data
      else
        -- SR(1) selected
        if srl_a_datapath /= "0000" then
--        SR(1) read => select SR(0) else still select SR(1)
          if reg_ce_datapath = '1' or master_m2s_i.DataValid = '0' then
            srl_ce_datapath <= '1';
            srl_a_datapath  <= "0000";
          end if;
        -- SR(0) selected
--        else
--        SR(0) read => select SR(0) and accept more data
--          if reg_ce_datapath = '1' or master_m2s_i.DataValid = '0' then
--            srl_ce_datapath <= '1';
--          end if;
        end if;
      end if;

      -- Read Data Group
      if srl_ce_resppath = '1' then
        if reg_ce_resppath = '0' and slave_s2m_i.Resp /= OCP_RESP_NONE then
          srl_ce_resppath <= '0';
          srl_a_resppath  <= "0001";
        end if;
      else
        if srl_a_resppath /= "0000" then
          if reg_ce_resppath = '1' or slave_s2m_i.Resp = OCP_RESP_NONE then
            srl_ce_resppath <= '1';  -- added
            srl_a_resppath  <= "0000";
          end if;
--        else
--          if reg_ce_resppath = '1' or slave_s2m_i.Resp = OCP_RESP_NONE then
--            srl_ce_resppath <= '1';
--          end if;
        end if;
      end if;
    end if;
  end process;

end;