--==============================================================================
--
--     File Name: adb3_target_tb_pkg.vhd
--       Version: 0.1
--          Date: 19 Oct 2009
--         Model: Synthesisable
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Package adb3_target_tb_pkg
--
--   Description: Package of components, procedures and functions for applying
--                testbench stimulus to a target FPGA using ADB3.
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2009 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 19 Oct 09: Initial design
--
--==============================================================================

--
-- Exported data types:
--
--   byte_enable_t
--   byte_t
--   byte_vector_t

--
-- Exported constants:
--

-- Exported routines (parenthesis indicated number of overloads):
--
--   conv_byte_vector
--   conv_byte_enable
--   conv_vector
--   conv_string_hex
--   conv_string (2)
--   adb3_target_sim_read32
--   adb3_target_sim_write32
--   adb3_target_sim_read64
--   adb3_target_sim_write64
--   adb3_target_sim_read
--   adb3_target_sim_write
--   adb3_wait_cycles

--
-- Exported components:
--
--   mptl_if_bridge_wrap
--   mptl_if_bridge_sim
--   mptl64par_interface_bridge_6tl_top
--   mptl64par_interface_bridge_6tl
--   mptl128_interface_bridge_6t1_top
--   mptl128_interface_bridge_6t1
--


library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;
use work.adb3_target_inc_pkg.all;

package adb3_target_tb_pkg is

  --
  -- A byte datatype
  --
  subtype byte_t is std_logic_vector(7 downto 0);

  --
  -- An array of std_logic, intended to hold byte enables in ACTIVE HIGH
  -- representation
  --
  type byte_enable_t is array(natural range <>) of std_logic;

  --
  -- An array of bytes, intended to hold data to be written to the
  -- local bus or read from the local bus.
  --
  type byte_vector_t is array(natural range <>) of byte_t;


  --
  -- Utility functions for converting between std_logic_vector, byte_vector
  -- and integer datatypes.
  --
  function conv_byte_vector(
    constant x : in std_logic_vector)
  return byte_vector_t;

  function conv_byte_enable(
    constant x : in std_logic_vector)
  return byte_enable_t;

  function conv_vector(
    constant x : in byte_vector_t)
  return std_logic_vector;

  function conv_string_hex(
    constant x : in byte_vector_t)
  return string;

  function conv_string_hex(
    constant x : in std_logic_vector)
  return string;

  function conv_string(
    constant x : in byte_enable_t)
  return string;

  function conv_string(
    constant x : in std_logic_vector)
  return string;


  -- ------------------------------------------------------------
  -- Public routines for performing ADB3 transfers
  -- ------------------------------------------------------------

  -- Perform an ADB3 OCP interface direct slave read.
  --
  --     address     Byte address.
  --
  --     tag         Identifier for transfer
  --
  --     data        Data to be read. data'length indicates
  --                 desired number of bytes to read.
  --
  --     nxfered     Returns the actual number of bytes transferred.

  procedure adb3_target_sim_read_reg32(
           address      : in    std_logic_vector;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT);

  procedure adb3_target_sim_read_reg64(
           address      : in    std_logic_vector;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT);

  -- Blocking read procedure
  -- Waits until all data has been read before completing
  procedure adb3_target_sim_read(
           address      : in    std_logic_vector;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT);

  -- Non-Blocking read command procedure
  -- Produces read commands only
  procedure adb3_target_sim_read_cmd(
           address      : in    std_logic_vector;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT);

  -- Non-Blocking read response procedure
  -- waits for read response only
  procedure adb3_target_sim_read_resp(
           data         : out   byte_vector_t;
           tag          : out   natural;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT);


  -- Perform an ADB3 OCP interface direct slave write.
  --
  --     address     Byte address.
  --
  --     tag         Identifier for transfer.
  --
  --     be          Byte enables. be'length must be the same as
  --                 data'length. Byte enables used are limited
  --                 to data bus width (16) as they are ignored
  --                 thereafter.
  --
  --     data        Data to be written. data'length indicates
  --                 desired number of bytes to write.
  --
  --     nxfered     Returns the actual number of bytes transferred.

  procedure adb3_target_sim_write_reg32(
           address      : in    std_logic_vector;
           tag          : in    natural;
           be           : in    byte_enable_t;
           data         : in    byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT);

  procedure adb3_target_sim_write_reg64(
           address      : in    std_logic_vector;
           tag          : in    natural;
           be           : in    byte_enable_t;
           data         : in    byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT);

  procedure adb3_target_sim_write(
           address      : in    std_logic_vector;
           tag          : in    natural;
           be           : in    byte_enable_t;
           data         : in    byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT);


  -- Delay for a specified number of clock cycles
  --
  --     cycles      Number of clock cycles of delay.
  --     clk         Clock signal to use for delay

  procedure adb3_wait_cycles(
           cycles : in    natural;
    signal clk    : in    std_logic);


  component mptl_if_bridge_wrap
  port(
    -- OCP Interface
    ocp_clk_in               : in    std_logic;
    direct_slave_m2s         : in    adb3_ocp_m2sT;
    direct_slave_s2m         : out   adb3_ocp_s2mT;
    dma_channels_m2s         : in    adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
    dma_channels_s2m         : out   adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
    direct_masters_m2s       : out   adb3_ocp_m2sT_array(0 to DM_CHANNELS-1);
    direct_masters_s2m       : in    adb3_ocp_s2mT_array(0 to DM_CHANNELS-1);
    dma_abort                : out   std_logic_vector(DMA_CHANNELS-1 downto 0);
    ocp_clk_out              : out   std_logic;
    -- MPTL Interface
    mptl_t2b                 : in    mptl_pins_t;
    mptl_b2t                 : out   mptl_pins_t;
    mptl_clk                 : in    std_logic_dbl_t;
    -- MPTL Initialisation
    mptl_sb_t2b              : in    mptl_sb_t2b_t;
    mptl_sb_b2t              : out   mptl_sb_b2t_t;
    -- MPTL Sideband
    gpio_b2t                 : in    std_logic_vector(15 downto 0);
    gpio_t2b                 : out   std_logic_vector(15 downto 0);
    mptl_status              : out   std_logic_vector(63 downto 0);
    mptl_config              : in    std_logic_vector(31 downto 0));
  end component;

  component mptl_if_bridge_sim
  port(
    -- OCP Interface
    ocp_clk_in         : in    std_logic;
    direct_slave_m2s   : in    adb3_ocp_m2sT;
    direct_slave_s2m   : out   adb3_ocp_s2mT;
    dma_channels_m2s   : in    adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
    dma_channels_s2m   : out   adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
    direct_masters_m2s : out   adb3_ocp_m2sT_array(0 to DM_CHANNELS-1);
    direct_masters_s2m : in    adb3_ocp_s2mT_array(0 to DM_CHANNELS-1);
    dma_abort          : out   std_logic_vector(DMA_CHANNELS-1 downto 0);
    ocp_clk_out        : out   std_logic;
    -- GTP Interface
    mptl_t2b           : in    mptl_pins_t;
    mptl_b2t           : out   mptl_pins_t;
    mptl_clk           : in    std_logic_dbl_t;
    -- MPTL Initialisation
    mptl_sb_t2b        : in    mptl_sb_t2b_t;
    mptl_sb_b2t        : out   mptl_sb_b2t_t;
    -- MPTL Sideband
    gpio_b2t           : in    std_logic_vector(15 downto 0);
    gpio_t2b           : out   std_logic_vector(15 downto 0);
    mptl_status        : out   std_logic_vector(63 downto 0);
    mptl_config        : in    std_logic_vector(31 downto 0));
  end component;

  component mptl64par_interface_bridge_6tl_top
  port(
    -- OCP Interface
    ocp_clk_in         : in    std_logic;
    direct_slave_m2s   : in    adb3_ocp_m2sT;
    direct_slave_s2m   : out   adb3_ocp_s2mT;
    dma_channels_m2s   : in    adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
    dma_channels_s2m   : out   adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
    direct_masters_m2s : out   adb3_ocp_m2sT_array(0 to DM_CHANNELS-1);
    direct_masters_s2m : in    adb3_ocp_s2mT_array(0 to DM_CHANNELS-1);
    dma_abort          : out   std_logic_vector(DMA_CHANNELS-1 downto 0);
    ocp_clk_out        : out   std_logic;
    -- GTP Interface
    mptl_t2b           : in    mptl_pins_t;
    mptl_b2t           : out   mptl_pins_t;
    mptl_clk           : in    std_logic_dbl_t;
    -- MPTL Initialisation
    mptl_sb_t2b        : in    mptl_sb_t2b_t;
    mptl_sb_b2t        : out   mptl_sb_b2t_t;
    -- MPTL Sideband
    gpio_b2t           : in    std_logic_vector(15 downto 0);
    gpio_t2b           : out   std_logic_vector(15 downto 0);
    mptl_status        : out   std_logic_vector(63 downto 0);
    mptl_config        : in    std_logic_vector(31 downto 0));
  end component;

  component mptl64par_interface_bridge_6tl
  port(
    -- OCP Interface
    ocp_clk            : in  std_logic;
    direct_slave_m2s   : in  adb3_ocp_m2sT;
    direct_slave_s2m   : out adb3_ocp_s2mT;
    dma_channels_m2s   : in  adb3_ocp_m2sT_array(0 to 1);
    dma_channels_s2m   : out adb3_ocp_s2mT_array(0 to 1);
    direct_master_m2s  : out adb3_ocp_m2sT;
    direct_master_s2m  : in  adb3_ocp_s2mT;
    -- MPTL Interface
    mptl_tx            : out mptl_pins_t;
    mptl_rx            : in  mptl_pins_t;
    mptl_clk           : in  std_logic_dbl_t;
    -- Bridge to Target Sideband Signals
    target_configured  : in  std_logic;
    target_gtp_online  : in  std_logic;
    bridge_gtp_online  : out std_logic;
    -- MPTL64 Sideband Interface
    gpio_b2t           : in  std_logic_vector(15 downto 0);
    gpio_t2b           : out std_logic_vector(15 downto 0);
    mptl_status        : out std_logic_vector(63 downto 0);
    debug              : out std_logic_vector(255 downto 0));
  end component;

  component mptl128_interface_bridge_6t1_top
  port(
    -- OCP Interface
    ocp_clk_in         : in    std_logic;
    direct_slave_m2s   : in    adb3_ocp_m2sT;
    direct_slave_s2m   : out   adb3_ocp_s2mT;
    dma_channels_m2s   : in    adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
    dma_channels_s2m   : out   adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
    direct_masters_m2s : out   adb3_ocp_m2sT_array(0 to DM_CHANNELS-1);
    direct_masters_s2m : in    adb3_ocp_s2mT_array(0 to DM_CHANNELS-1);
    dma_abort          : out   std_logic_vector(DMA_CHANNELS-1 downto 0);
    ocp_clk_out        : out   std_logic;
    -- GTP Interface
    mptl_t2b           : in    mptl_pins_t;
    mptl_b2t           : out   mptl_pins_t;
    mptl_clk           : in    std_logic_dbl_t;
    -- MPTL Initialisation
    mptl_sb_t2b        : in    mptl_sb_t2b_t;
    mptl_sb_b2t        : out   mptl_sb_b2t_t;
    -- MPTL Sideband
    gpio_b2t           : in    std_logic_vector(15 downto 0);
    gpio_t2b           : out   std_logic_vector(15 downto 0);
    mptl_status        : out   std_logic_vector(63 downto 0);
    mptl_config        : in    std_logic_vector(31 downto 0));
  end component;

  component mptl128_interface_bridge_6t1
  port(
    -- OCP Interface
    ocp_clk            : in  std_logic;
    direct_slave_m2s   : in  adb3_ocp_m2sT;
    direct_slave_s2m   : out adb3_ocp_s2mT;
    dma_channels_m2s   : in  adb3_ocp_m2sT_array(0 to 3);
    dma_channels_s2m   : out adb3_ocp_s2mT_array(0 to 3);
    direct_masters_m2s : out adb3_ocp_m2sT_array(0 to 2);
    direct_masters_s2m : in  adb3_ocp_s2mT_array(0 to 2);
    dma_abort          : out std_logic_vector(3 downto 0);
    -- MPTL Interface
    mptl_tx            : out mptl_pins_t;
    mptl_rx            : in  mptl_pins_t;
    mptl_clk           : in  std_logic_dbl_t;
    -- Bridge to Target Sideband Signals
    target_configured  : in  std_logic;
    target_gtp_online  : in  std_logic;
    bridge_gtp_online  : out std_logic;
    -- MPTL64 Sideband Interface
    gpio_in            : in  std_logic_vector(15 downto 0);
    gpio_out           : out std_logic_vector(15 downto 0);
    mptl_status        : out std_logic_vector(63 downto 0);
    mptl_config        : in  std_logic_vector(31 downto 0);
    -- Debug
    debug              : out std_logic_vector(255 downto 0));
  end component;

end;


package body adb3_target_tb_pkg is

  -- ---------------------------------------------------------------
  -- Private
  -- ---------------------------------------------------------------

  function conv_string_hexdigit(
    constant val : in std_logic_vector)
  return character is
    variable tmp : std_logic_vector(3 downto 0);
  begin
    assert not(val'length /= 4)
      report "(conv_string_hex_digit): input has invalid length " & integer'image(val'length) & " /= 4"
      severity failure;

    tmp := val;
    case tmp is
      when "0000" => return '0';
      when "0001" => return '1';
      when "0010" => return '2';
      when "0011" => return '3';
      when "0100" => return '4';
      when "0101" => return '5';
      when "0110" => return '6';
      when "0111" => return '7';
      when "1000" => return '8';
      when "1001" => return '9';
      when "1010" => return 'A';
      when "1011" => return 'B';
      when "1100" => return 'C';
      when "1101" => return 'D';
      when "1110" => return 'E';
      when "1111" => return 'F';
      when others => return 'X';
    end case;
  end;


  function conv_std_logic_char(
    constant val: in std_logic)
  return character is
  begin
    case val is
      when 'U' => return 'U';
      when 'X' => return 'X';
      when '0' => return '0';
      when '1' => return '1';
      when 'Z' => return 'Z';
      when 'W' => return 'W';
      when 'L' => return 'L';
      when 'H' => return 'H';
      when '-' => return '-';
    end case;
  end;


  -- ---------------------------------------------------------------
  -- Private ADB3 routines
  -- ---------------------------------------------------------------

  procedure adb3_target_sim_burst_read(
           order        : in    natural;
           address      : in    std_logic_vector;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
    constant OCP_WIDTH    : natural := 16;                          -- Number of bytes in OCP data
    constant BST_MAX      : natural := 2**ADB3_OCP_BURST_WIDTH - 1;

    variable addr_lo      : natural;
    variable data_offset  : natural;
    variable remaining_d  : natural;
    variable burst_len    : natural range 0 to BST_MAX;
    variable m_d          : natural range 0 to OCP_WIDTH;

    type burst_state_type is (ocp_cmd,ocp_acc,ocp_finished);
    variable state, n_state : burst_state_type;

    variable n_cmd         : ocp_CmdT;
    variable n_addr        : std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0);
    variable n_burstlength : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
    variable n_datavalid   : std_logic;
    variable n_tag         : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
    variable n_respaccept  : std_logic;
    variable n_data        : std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0);

  begin
    if (order > 0) then
      addr_lo := TO_INTEGER(UNSIGNED(address(order-1 downto 0)));
      assert not(addr_lo /= 0)
        report "(adb3_target_sim_burst_read): non-zero address bits (" & integer'image(order-1) &
               ":0) will be ignored for " & integer'image(2**order) & "-byte register read"
        severity warning;
    end if;

    state        := ocp_cmd;
    data_offset  := data'low;
    remaining_d  := data'length;

    -- remaining =  1..16 => burst_len = 1
    -- remaining = 17..32 => burst_len = 2 etc.
    burst_len := (data'length + OCP_WIDTH - 1) / OCP_WIDTH;

    while (remaining_d /= 0) loop
      wait until rising_edge(ocp_clk);
      n_datavalid := '0';

      case state is
        when ocp_cmd =>
          n_cmd         := OCP_CMD_READ;
          n_addr        := std_logic_vector(RESIZE(unsigned(address),ADB3_OCP_ADDR_WIDTH));
          n_burstlength := std_logic_vector(TO_UNSIGNED(burst_len,ADB3_OCP_BURST_WIDTH));
          n_tag         := std_logic_vector(TO_UNSIGNED(tag,ADB3_OCP_TAG_WIDTH));
          n_respaccept  := '1';
          n_state       := ocp_acc;

        when ocp_acc =>
          if (adb3_ocp_s2m.CmdAccept = '1') then
            n_cmd := OCP_CMD_IDLE;
            n_state := ocp_finished;
          else
            n_state := ocp_acc;
          end if;

        when ocp_finished =>
          n_state := ocp_finished;
      end case;

      state := n_state;

      if (adb3_ocp_s2m.Resp = OCP_RESP_VALID) then
        m_d := OCP_WIDTH;
        if (m_d > remaining_d) then
          m_d := remaining_d;
        end if;

        for i in 0 to (OCP_WIDTH - 1) loop
          if (m_d > 0) then
            data(data_offset) := adb3_ocp_s2m.Data(8 * (i + 1) - 1 downto 8 * i);
            data_offset := data_offset + 1;
            remaining_d := remaining_d - 1;
            m_d := m_d - 1;
--            assert false report "Read adb3_ocp_s2m.Data - Remaining: " & integer'image(remaining_d) severity warning;
          end if;
        end loop;
      end if;

      adb3_ocp_m2s.Cmd         <= n_cmd;
      adb3_ocp_m2s.Addr        <= n_addr;
      adb3_ocp_m2s.BurstLength <= n_burstlength;
      adb3_ocp_m2s.DataValid   <= n_datavalid;
      adb3_ocp_m2s.Tag         <= n_tag;
      adb3_ocp_m2s.RespAccept  <= n_respaccept;
      adb3_ocp_m2s.DataByteEn  <= (others => '0');
      adb3_ocp_m2s.RespAccept  <= n_respaccept;
      wait until falling_edge(ocp_clk);
    end loop;

    n_respaccept := '0';
    adb3_ocp_m2s.RespAccept  <= n_respaccept;
    nxfered := data'length - remaining_d;
  end;


  procedure adb3_target_sim_burst_read_cmd(
           order        : in    natural;
           address      : in    std_logic_vector;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
    constant OCP_WIDTH    : natural := 16;                          -- Number of bytes in OCP data
    constant BST_MAX      : natural := 2**ADB3_OCP_BURST_WIDTH - 1;

    variable addr_lo      : natural;
    variable data_offset  : natural;
    variable burst_len    : natural range 0 to BST_MAX;

    type burst_state_type is (ocp_cmd,ocp_acc,ocp_finished);
    variable state, n_state : burst_state_type;

    variable n_cmd         : ocp_CmdT;
    variable n_addr        : std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0);
    variable n_burstlength : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
    variable n_datavalid   : std_logic;
    variable n_tag         : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);

  begin
    if (order > 0) then
      addr_lo := TO_INTEGER(UNSIGNED(address(order-1 downto 0)));
      assert not(addr_lo /= 0)
        report "(adb3_target_sim_burst_read): non-zero address bits (" & integer'image(order-1) &
               ":0) will be ignored for " & integer'image(2**order) & "-byte register read"
        severity warning;
    end if;

    state        := ocp_cmd;
    data_offset  := data'low;

    -- remaining =  1..16 => burst_len = 1
    -- remaining = 17..32 => burst_len = 2 etc.
    burst_len := (data'length + OCP_WIDTH - 1) / OCP_WIDTH;

    while (state /= ocp_finished) loop
      wait until rising_edge(ocp_clk);
      n_datavalid := '0';

      case state is
        when ocp_cmd =>
          n_cmd         := OCP_CMD_READ;
          n_addr        := std_logic_vector(RESIZE(unsigned(address),ADB3_OCP_ADDR_WIDTH));
          n_burstlength := std_logic_vector(TO_UNSIGNED(burst_len,ADB3_OCP_BURST_WIDTH));
          n_tag         := std_logic_vector(TO_UNSIGNED(tag,ADB3_OCP_TAG_WIDTH));
          n_state       := ocp_acc;

        when ocp_acc =>
          if (adb3_ocp_s2m.CmdAccept = '1') then
            n_cmd := OCP_CMD_IDLE;
            n_state := ocp_finished;
          else
            n_state := ocp_acc;
          end if;

        when ocp_finished =>
          n_state := ocp_finished;
      end case;

      state := n_state;

      adb3_ocp_m2s.Cmd         <= n_cmd;
      adb3_ocp_m2s.Addr        <= n_addr;
      adb3_ocp_m2s.BurstLength <= n_burstlength;
      adb3_ocp_m2s.DataValid   <= n_datavalid;
      adb3_ocp_m2s.Tag         <= n_tag;
      adb3_ocp_m2s.DataByteEn  <= (others => '0');
      wait until falling_edge(ocp_clk);
    end loop;

    nxfered := data'length;
  end;


  procedure adb3_target_sim_burst_read_resp(
           data         : out   byte_vector_t;
           tag          : out   natural;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
    constant OCP_WIDTH    : natural := 16;                          -- Number of bytes in OCP data
    constant BST_MAX      : natural := 2**ADB3_OCP_BURST_WIDTH - 1;

    variable addr_lo      : natural;
    variable data_offset  : natural;
    variable remaining_d  : natural;
    variable burst_len    : natural range 0 to BST_MAX;
    variable m_d          : natural range 0 to OCP_WIDTH;

    variable n_respaccept  : std_logic;

  begin
    data_offset  := data'low;
    remaining_d  := data'length;

    -- remaining =  1..16 => burst_len = 1
    -- remaining = 17..32 => burst_len = 2 etc.
    burst_len := (data'length + OCP_WIDTH - 1) / OCP_WIDTH;
    adb3_ocp_m2s.RespAccept <= '1';

    while (remaining_d /= 0) loop
      wait until rising_edge(ocp_clk);

      if (adb3_ocp_s2m.Resp = OCP_RESP_VALID) then
        m_d := OCP_WIDTH;
        if (m_d > remaining_d) then
          m_d := remaining_d;
        end if;

        for i in 0 to (OCP_WIDTH - 1) loop
          if (m_d > 0) then
            data(data_offset) := adb3_ocp_s2m.Data(8 * (i + 1) - 1 downto 8 * i);
            data_offset := data_offset + 1;
            remaining_d := remaining_d - 1;
            m_d := m_d - 1;
--            assert false report "Read adb3_ocp_s2m.Data - Remaining: " & integer'image(remaining_d) severity warning;
          end if;
        end loop;
        tag := TO_INTEGER(UNSIGNED(adb3_ocp_s2m.Tag));
      end if;

      wait until falling_edge(ocp_clk);
    end loop;

    adb3_ocp_m2s.RespAccept  <= '0';
    nxfered := data'length - remaining_d;
  end;


  procedure adb3_target_sim_burst_write(
           order        : in    natural;
           address      : in    std_logic_vector;
           tag          : in    natural;
           be           : in    byte_enable_t;
           data         : in    byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
    constant OCP_WIDTH    : natural := 16;                          -- Number of bytes in OCP data
    constant BST_MAX      : natural := 2**ADB3_OCP_BURST_WIDTH - 1;

    variable addr_lo      : natural;
    variable i            : natural;
    variable beat_one     : boolean;
    variable data_offset  : natural;
    variable be_offset    : natural;
    variable remaining    : natural;
    variable next_be      : boolean;
    variable burst_len    : natural range 0 to BST_MAX;
    variable m            : natural range 0 to OCP_WIDTH;

    type burst_state_type is (ocp_cmd,ocp_data,ocp_finished);
    variable state, n_state : burst_state_type;

    variable n_cmd         : ocp_CmdT;
    variable n_addr        : std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0);
    variable n_burstlength : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
    variable n_datavalid   : std_logic;
    variable n_tag         : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
    variable n_respaccept  : std_logic;
    variable n_data        : std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0);
    variable n_databyteen  : std_logic_vector(ADB3_OCP_DATA_WIDTH/8-1 downto 0);

  begin
    assert not(be'length /= data'length)
      report "(adb3_target_sim_burst_write): byte enable length (" & integer'image(be'length) &
             ") does not match data length (" & integer'image(data'length) & ")"
      severity failure;

    if (order > 0) then
      addr_lo := TO_INTEGER(UNSIGNED(address(order-1 downto 0)));
      assert not(addr_lo /= 0)
        report "(adb3_target_sim_burst_write): non-zero address bits (" & integer'image(order-1) &
               ":0) will be ignored for " & integer'image(2**order) & "-byte register write"
        severity warning;
    end if;

    -- If burst transfer is larger than OCP bus width, then byte enables are ignored for all beats
    -- after the initial one. Check to verify we are not ignoring active enables.
    for i in be'low to be'high loop
      assert not((i >= OCP_WIDTH)and(be(i)='0'))
        report "(adb3_target_sim_burst_write): active byte enable (be(" & integer'image(i) &
               ") = " & std_logic'image(be(i)) &
               ") will be ignored during writes with bytes > " & integer'image(OCP_WIDTH)
        severity warning;
    end loop;

    state       := ocp_cmd;
    beat_one    := true;
    data_offset := data'low;
    be_offset   := be'low;
    remaining   := data'length;

    -- remaining =  1..16 => burst_len = 1
    -- remaining = 17..32 => burst_len = 2 etc.
    burst_len := (data'length + OCP_WIDTH - 1) / OCP_WIDTH;

    while (state /= ocp_finished) loop
      wait until rising_edge(ocp_clk);
      n_respaccept := '0';
      next_be      := false;

      case state is
        when ocp_cmd =>
          n_cmd         := OCP_CMD_WRITE;
          n_addr        := std_logic_vector(RESIZE(unsigned(address),ADB3_OCP_ADDR_WIDTH));
          n_burstlength := std_logic_vector(TO_UNSIGNED(burst_len,ADB3_OCP_BURST_WIDTH));
          n_datavalid   := '1';
          n_tag         := std_logic_vector(TO_UNSIGNED(tag,ADB3_OCP_TAG_WIDTH));
          next_be       := true;
          n_state       := ocp_data;

        when ocp_data =>
          n_state := ocp_data;
          if (adb3_ocp_s2m.CmdAccept = '1') then
            n_cmd   := OCP_CMD_IDLE;
          end if;
          if (adb3_ocp_s2m.DataAccept = '1')and(n_cmd = OCP_CMD_IDLE) then
            if (remaining /= 0) then
              next_be := true;
            else
              n_datavalid := '0';
              n_state := ocp_finished;
            end if;
          end if;

        when ocp_finished =>    -- State is never executed - used in while test only
          null;
      end case;

      state := n_state;

      if next_be then
        m := OCP_WIDTH;
        if (m > remaining) then
          m := remaining;
        end if;

        -- Bytes > m are set to zero
        for i in 0 to (OCP_WIDTH - 1) loop
          if (m = 0) then
            if (beat_one) then
              n_databyteen(i) := '0';  -- Byte enable used if first beat
            else
              n_databyteen(i) := '1';  -- Byte enables active on subsequent beats
            end if;
            n_data(8 * (i + 1) - 1 downto 8 * i) := (others => '0');
          else
            if (beat_one) then
              n_databyteen(i) := be(be_offset);  -- Byte enable used if first beat
            else
              n_databyteen(i) := '1';            -- Byte enables active on subsequent beats
            end if;
            be_offset := be_offset + 1;
            n_data(8 * (i + 1) - 1 downto 8 * i) := data(data_offset);
            data_offset := data_offset + 1;
            remaining := remaining - 1;
            m := m - 1;
          end if;
        end loop;

        beat_one := false;
      end if;

      adb3_ocp_m2s.Cmd         <= n_cmd;
      adb3_ocp_m2s.Addr        <= n_addr;
      adb3_ocp_m2s.BurstLength <= n_burstlength;
      adb3_ocp_m2s.DataValid   <= n_datavalid;
      adb3_ocp_m2s.Tag         <= n_tag;
      adb3_ocp_m2s.RespAccept  <= n_respaccept;
      adb3_ocp_m2s.Data        <= n_data;
      adb3_ocp_m2s.DataByteEn  <= n_databyteen;
    end loop;

    nxfered := data'length - remaining;
  end;


  -- ---------------------------------------------------------------
  -- Public
  -- ---------------------------------------------------------------

  function conv_byte_vector(
    constant x : in std_logic_vector)
  return byte_vector_t is
    variable n : natural := (x'length + 7) / 8;         -- Size of x in bytes (min size of 1)
    variable m : std_logic_vector(8 * n - 1 downto 0);  -- New vector of size n bytes
    variable t : byte_vector_t(0 to n - 1);             -- New byte_vector of size n bytes
  begin
    m := std_logic_vector(RESIZE(unsigned(x), 8 * n));

    for i in 0 to n - 1 loop
      t(i) := m(8 * (i + 1) - 1 downto 8 * i);
    end loop;

    return t;
  end;

  function conv_byte_enable(
    constant x : in std_logic_vector)
  return byte_enable_t is
    variable n : natural := x'length;               -- Size of x in bits (min size of 1)
    variable m : std_logic_vector(n - 1 downto 0);  -- New vector of size n bits
    variable t : byte_enable_t(0 to n - 1);         -- New byte_enable of size n bits
  begin
    m := x;

    for i in 0 to n - 1 loop
      t(i) := m(i);
    end loop;

    return t;
  end;

  function conv_vector(
    constant x : in byte_vector_t)
  return std_logic_vector is
    variable v          : std_logic_vector((8*x'length)-1 downto 0);
    variable next_byte  : byte_t;
    variable j          : natural;
  begin
    for i in x'range loop
      next_byte := x(i);
      j := i - x'low;
      v(j*8+7 downto j*8) := next_byte;
    end loop;

    return v;
  end;

  function conv_string_hex(
    constant x : in byte_vector_t)
  return string is
    variable j         : natural := 2 * x'length;
    variable s         : string(1 to j);
    variable next_byte : byte_t;
  begin
    for i in x'range loop
      next_byte := x(i);
      s(j)      := conv_string_hexdigit(next_byte(3 downto 0));
      s(j-1)    := conv_string_hexdigit(next_byte(7 downto 4));
      j         := j - 2;
    end loop;

    return s;
  end;

  function conv_string_hex(
    constant x : in std_logic_vector)
  return string is
    variable i : natural := x'length mod 4;
    variable j : natural := x'length + (4-i) mod 4;
    variable s : string(1 to j/4);
    variable y : std_logic_vector(j-1 downto 0) := (others => '0');
  begin
    y(x'length - 1 downto 0) := x;

    for k in (j/4)-1 downto 0 loop
      s(j/4 - k) := conv_string_hexdigit(y((k*4)+3 downto k*4));
    end loop;

    return s;
  end;

  function conv_string(
    constant x : in byte_enable_t)
  return string is
    variable j        : natural;
    variable s        : string(1 to x'length);
    variable next_bit : std_logic;
  begin
    j := x'length;

    for i in x'range loop
      next_bit := x(i);
      s(j) := conv_std_logic_char(next_bit);
      j    := j - 1;
    end loop;

    return s;
  end;


  -- Calculate the string width required to represent
  -- a decimal number in decimal format
  function conv_dec_width(
    decimal : in natural)
  return natural is
  begin
    assert not(decimal < 0)
      report "(conv_dec_width): Cannot use 32-bit natural"
      severity failure;

    case decimal is
      when 0 to 9 => return 1;
      when 10 to 99 => return 2;
      when 100 to 999 => return 3;
      when 1000 to 9999 => return 4;
      when 10000 to 99999 => return 5;
      when 100000 to 999999 => return 6;
      when 1000000 to 9999999 => return 7;
      when 10000000 to 99999999 => return 8;
      when 100000000 to 999999999 => return 9;
      when 1000000000 to integer'high => return 10;
    end case;
  end function;


  function conv_string(
    constant x : in std_logic_vector)
  return string is
    variable j        : natural := TO_INTEGER(unsigned(x));
    variable s        : string(1 to conv_dec_width(j));
  begin
    s := integer'image(j);

    return s;
  end;


  -- ---------------------------------------------------------------
  -- Public ADB3 routines
  -- ---------------------------------------------------------------

  procedure adb3_target_sim_read_reg32(
           address      : in    std_logic_vector;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
  begin
    adb3_target_sim_burst_read(2,address, tag, data, nxfered, ocp_clk, adb3_ocp_m2s, adb3_ocp_s2m);
  end;

  procedure adb3_target_sim_write_reg32(
           address      : in    std_logic_vector;
           tag          : in    natural;
           be           : in    byte_enable_t;
           data         : in    byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
  begin
    adb3_target_sim_burst_write(2,address, tag, be, data, nxfered, ocp_clk, adb3_ocp_m2s, adb3_ocp_s2m);
  end;


  procedure adb3_target_sim_read_reg64(
           address      : in    std_logic_vector;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
  begin
    adb3_target_sim_burst_read(3,address, tag, data, nxfered, ocp_clk, adb3_ocp_m2s, adb3_ocp_s2m);
  end;

  procedure adb3_target_sim_write_reg64(
           address      : in    std_logic_vector;
           tag          : in    natural;
           be           : in    byte_enable_t;
           data         : in    byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
  begin
    adb3_target_sim_burst_write(3,address, tag, be, data, nxfered, ocp_clk, adb3_ocp_m2s, adb3_ocp_s2m);
  end;

  -- Blocking read procedure
  -- Waits until all data has been read before completing
  procedure adb3_target_sim_read(
           address      : in    std_logic_vector;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
  begin
    adb3_target_sim_burst_read(0,address, tag, data, nxfered, ocp_clk, adb3_ocp_m2s, adb3_ocp_s2m);
  end;

  -- Non-Blocking read command procedure
  -- Produces read commands only
  procedure adb3_target_sim_read_cmd(
           address      : in    std_logic_vector;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
  begin
    adb3_target_sim_burst_read_cmd(0,address, tag, data, nxfered, ocp_clk, adb3_ocp_m2s, adb3_ocp_s2m);
  end;

  -- Non-Blocking read response procedure
  -- waits for read response only
  procedure adb3_target_sim_read_resp(
           data         : out   byte_vector_t;
           tag          : out   natural;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
  begin
    adb3_target_sim_burst_read_resp(data, tag, nxfered, ocp_clk, adb3_ocp_m2s, adb3_ocp_s2m);
  end;

  -- Write procedure
  procedure adb3_target_sim_write(
           address      : in    std_logic_vector;
           tag          : in    natural;
           be           : in    byte_enable_t;
           data         : in    byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
  begin
    adb3_target_sim_burst_write(0,address, tag, be, data, nxfered, ocp_clk, adb3_ocp_m2s, adb3_ocp_s2m);
  end;


  procedure adb3_wait_cycles(
           cycles : in    natural;
    signal clk    : in    std_logic)
  is
  begin
    for i in 0 to (cycles-1) loop
      wait until rising_edge(clk);
    end loop;
  end;

end;
