--==============================================================================
--
--     File Name: adb3_target_inc_syn_ngc_6t1_pkg.vhd
--       Version: 0.1
--          Date: 01 Jun 2010
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: Package adb3_target_inc_pkg
--
--   Description: Package used in target FPGAs during synthesis using .NGC cores
--
--   Limitations: Applicable to ADM-XRC-6T1 synthesis only
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 01 Jun 10: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;

library work;
use work.adb3_ocp.all;
use work.adb3_target_types_pkg.all;


package adb3_target_inc_pkg is

  -------------------------
  -- General definitions --
  -------------------------

  -- Select board type
  constant BOARD_TYPE : board_type_t := ADM_XRC_6T1;

  -- Select target use
  -- SIM_OCP:  Generate simulation code. Direct OCP to OCP connections between bridge (TB) and target (ie bypass MPTL)
  -- SIM_MPTL: Generate simulation code. Full OCP to MPTL to OCP connections between bridge (TB) and target
  -- SYN_NGC:  Generate synthesis code.  Instantiate .ngc core for OCP-MPTL interface to bridge
  constant TARGET_USE : target_use_t := SYN_NGC;


  -----------------------
  -- Clock definitions --
  -----------------------

  -- Target FPGA clock inputs
  constant REF_CLK_VALID      : std_logic := '1';
  constant LCLK_VALID         : std_logic := '0';
  constant XRM_GCLK_M2C_VALID : std_logic := '1';

  -- Target FPGA MGT clock inputs
  -- Enable  MGT113 CLK0 = 250MHz oscillator reference clock
  -- Enable  MGT113 CLK1 = 250MHz PCIe reference clock (derived from 100MHz input on P5)
  -- Enable  MGT114 CLK0 = 250MHz PCIe reference clock (derived from 100MHz input on P5)
  -- Disable MGT115 CLK0 = 250MHz oscillator reference clock (reserved for MPTL interface use)
  -- Enable  MGT116 CLK0 = 250MHz oscillator reference clock
  -- Enable  MGT117 CLK0 = XRM reference clock
  --                                                         876543210
  constant MGT_CLK0_VALID : std_logic_vector(8 downto 0) := "011011000"; -- MGT118-110 clk 0 valid
  constant MGT_CLK1_VALID : std_logic_vector(8 downto 0) := "000001000"; -- NGT118-110 clk 1 valid

  -- Target FPGA clock outputs
  constant XRM_MGTCLK_C2M_VALID : std_logic := '1';

  -- Board reference clock frequency
  constant REF_CLK_FREQ_HZ : natural := 200000000;


  ----------------------
  -- GPIO definitions --
  ----------------------

  -- Target FPGA XRM interface
  constant XRM_GPIO_VALID : boolean := true;
  constant XRM_GPIO_WIDTH : natural range 1 to MAX_XRM_GPIO_WIDTH := 146;
  constant XRM_MGT_WIDTH  : natural range 1 to MAX_XRM_MGT_WIDTH  := 8;

  -- Target FPGA PN4 interface
  constant PN4_GPIO_VALID : boolean := true;
  constant PN4_GPIO_WIDTH : natural range 1 to MAX_PN4_GPIO_WIDTH := 32;

  -- Target FPGA PN6 interface
  constant PN6_GPIO_VALID : boolean := true;
  constant PN6_GPIO_WIDTH : natural range 1 to MAX_PN6_GPIO_WIDTH := 38;
  constant PN6_MGT_WIDTH  : natural range 1 to MAX_PN6_MGT_WIDTH  := 8;


  ------------------------
  -- Memory definitions --
  ------------------------

  -- Target FPGA DDR3 SDRAM interface
  constant DDR3_VALID : boolean := true;
  constant DDR3_BANKS : natural range 0 to MAX_MEM_BANKS := 4;

  -- DDR3 memory interface widths
  constant DDR3_BANK_ROW_WIDTH : natural := 13; -- 1 Gb Part = 8K x 16 x 8 banks
--  constant DDR3_BANK_ROW_WIDTH : natural := 14; -- 2 Gb Part = 16K x 16 x 8 banks
  constant DDR3_BANK_DATA_WIDTH : natural := 32; -- 2 off 16-bit parts

  -- MIG interface address width = Bank(3)+Row(13/14)+Col(10) and convert to byte addressing (+2)
  constant DDR3_BYTE_ADDR_WIDTH : natural := 3 + DDR3_BANK_ROW_WIDTH + 10 + 2;

  -- Convert to 16-byte addressing
  constant DDR3_16_BYTE_ADDR_WIDTH : natural := DDR3_BYTE_ADDR_WIDTH - 4;


  -- Target FPGA memory interface
  constant MEM_VALID : boolean := DDR3_VALID;
  constant MEM_BANKS : natural range 0 to MAX_MEM_BANKS := DDR3_BANKS;


  --------------------------------
  -- MPTL interface definitions --
  --------------------------------

  -- Target FPGA MPTL interface OCP channels
  constant DS_CHANNELS  : natural range 1 to MAX_DS_CHANNELS  := 1;
  constant DMA_CHANNELS : natural range 1 to MAX_DMA_CHANNELS := 4;
  constant DM_CHANNELS  : natural range 1 to MAX_DM_CHANNELS  := 1;

  -- Target FPGA MPTL interface OCP channel address widths
  -- Direct Slave:      22 bits =   4 MiB (expandable using bridge paging)
  -- DMA address space: 39 bits = 512 GiB
  -- Direct Master:     32 bits =   4 GiB
  constant DS_ADDR_WIDTH  : natural range 1 to ADB3_OCP_ADDR_WIDTH := 22;
  constant DMA_ADDR_WIDTH : natural range 1 to ADB3_OCP_ADDR_WIDTH := 39;
  constant DM_ADDR_WIDTH  : natural range 1 to ADB3_OCP_ADDR_WIDTH := 32;

  -- Target FPGA MPTL interface serial width
  constant MPTL_SER_WIDTH : natural range 1 to MAX_MPTL_SER_WIDTH := 4;

  -- Record type for double ended signals
  type std_logic_dbl_t is record
    p : std_logic;
    n : std_logic;
  end record;

  -- MPTL data interface
  type mptl_pins_t is record
    p : std_logic_vector(MPTL_SER_WIDTH-1 downto 0);
    n : std_logic_vector(MPTL_SER_WIDTH-1 downto 0);
  end record;

  -- MPTL sideband interface inputs
  type mptl_sb_b2t_t is record
    mptl_bridge_gtp_online_l : std_logic;
  end record;

  -- MPTL sideband interface outputs
  type mptl_sb_t2b_t is record
    mptl_target_configured_l : std_logic;
    mptl_target_gtp_online_l : std_logic;
  end record;

end;


package body adb3_target_inc_pkg is

end;