--==============================================================================
--
--     File Name: blk_mem_test.vhd
--       Version: 0.2
--          Date: 09 Dec 2010
--         Model: synth
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: On-board memory bank test
--
--   Description: Module to perform a test of a bank of memory
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 29 Oct 2010: Initial design
--                0.2, TW,  09 Dec 2010: Changed interface to be OCP
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_arith.all;
use work.adb3_ocp.all;

entity blk_mem_test is
generic(
    a_width       : in    natural;                                -- Number of logical address bits in memory port
    d_width       : in    natural;                                -- Number of logical bits in a memory port word
    rd_width      : in    natural;                                -- Number of physical data pins on memory bank
    tag_base      : in    std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);  -- Tag base value
    tag_incr      : in    std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);  -- Tag value increment
    tag_mask      : in    std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0)); -- Which tag bits to check
port(
    -- OCP Interface, to/from memory interface
    ocp_rst       : in    std_logic;                              -- OCP reset
    ocp_clk       : in    std_logic;                              -- OCP clock
    ocp_m2s       : out   adb3_ocp_m2sT;                          -- OCP master to slave signal bundle
    ocp_s2m       : in    adb3_ocp_s2mT;                          -- OCP slave to master signal bundle
    -- Control interface
    go            : in    std_logic;                              -- Pulse high for at least one cycle to initiate test
    offset        : in    std_logic_vector(a_width - 1 downto 0); -- Starting logical address of test
    length        : in    std_logic_vector(a_width - 1 downto 0); -- Number of logical words to test, minus 1
    done          : out   std_logic;                              -- High indicates finished or idle
    error         : out   std_logic;                              -- Qualified by 'done', high indicates at least one error ocurred
    eaddr         : out   std_logic_vector(a_width - 1 downto 0); -- Qualified by 'done and error', gives word address of first error
    ephase        : out   std_logic_vector(2 downto 0));          -- Qualified by 'done and error', gives phase in which first error occurred
end entity;

architecture synth of blk_mem_test is

    -- LFSR with polynomial x^64 + x^4 + x^3 + 1
    function lfsr64(
        x : in std_logic_vector(63 downto 0))
    return std_logic_vector is
        variable y : std_logic_vector(63 downto 0);
    begin
        y(0) := x(63) xor x(3) xor x(2);
        y(63 downto 1) := x(62 downto 0);
        return y;
    end;

    -- Number of physical data words that make up a logical data word
    constant num_physical_word : natural := d_width / rd_width;

    -- Number of 32-bit words that make up a logical data word
    constant num_32bit_word : natural := d_width / 32;

    -- OCP preferred burst length
    constant ocp_burst_length : natural := 32;
    constant ocp_burst_order : natural := 5;

    -- Log2 of ocp data width
    constant ocp_data_order : natural := 4;

    constant phase_idle : std_logic_vector(2 downto 0) := "---";
    constant phase_55 : std_logic_vector(2 downto 0) := "000";
    constant phase_aa : std_logic_vector(2 downto 0) := "001";
    constant phase_55_aa : std_logic_vector(2 downto 0) := "010";
    constant phase_addr : std_logic_vector(2 downto 0) := "011";
    constant phase_rev_addr : std_logic_vector(2 downto 0) := "100";
    constant phase_random : std_logic_vector(2 downto 0) := "101";
    constant final_phase : std_logic_vector(2 downto 0) := phase_random;

    -- Local reset, generated from 'ocp_rst'
    signal lsr0, lsr1 : std_logic;
    signal lsr : std_logic;

    -- Latched versions of parameters, in case they change after 'go' pulse
    signal offset_q : std_logic_vector(a_width - 1 downto 0);
    signal length_q : std_logic_vector(a_width downto 0);

    -- Internal version of 'ocp_m2s'
    signal i_ocp_m2s, n_ocp_m2s : adb3_ocp_m2sT := ADB3_OCP_DEFAULT_M2S;

    -- Executive state machine
    type state1_t is (s1_idle, s1_doing);
    signal state1, n_state1 : state1_t;
    signal i_done, n_done : std_logic;
    signal phase, n_phase : std_logic_vector(2 downto 0);
    signal do_phase, n_do_phase : std_logic;
    signal done_phase_q, n_done_phase_q : std_logic_vector(1 downto 0);

    -- OCP command machine
    type state2_t is (s2_idle,
                      s2_write_init1, s2_write_init2, s2_write_wait, s2_write,
                      s2_read_init1, s2_read_init2, s2_read_wait, s2_read,
                      s2_done);
    signal state2, n_state2 : state2_t;
    signal cmd_accepted, n_cmd_accepted : std_logic_vector(1 downto 0);
    signal cmd_done, n_cmd_done : std_logic;
    signal cmd_init, n_cmd_init : std_logic;
    signal cmd_count : std_logic_vector(1 downto 0);
    signal cmd_remaining : std_logic_vector(a_width downto 0);
    signal cmd_chunk, cmd_blen : std_logic_vector(ocp_burst_order downto 0) := (others => '0');
    signal cmd_final : std_logic;
    signal cmd_issue, cmd_retire : std_logic;

    -- OCP data transfer state machine
    type state3_t is (s3_idle,
                      s3_write_init, s3_write_data,
                      s3_read_init, s3_read_data,
                      s3_done);
    signal state3, n_state3 : state3_t;
    signal data_done, n_data_done : std_logic;
    signal data_init, n_data_init : std_logic;
    signal data_offset : std_logic_vector(a_width - 1 downto 0);
    signal data_remaining : std_logic_vector(a_width downto 0);
    signal data_final : std_logic;
    signal data_init_q : std_logic;

    -- Helper signals to detect when data transfer takes place on OCP
    signal transfer : std_logic;
    signal read : std_logic;
    signal write : std_logic;

    -- Linear feedback shift registers, used for data generation
    type lfsreg_vector_t is array(0 to num_32bit_word - 1) of std_logic_vector(63 downto 0);
    signal lfsreg : lfsreg_vector_t;

    -- Data generation
    signal next_data : std_logic;
    signal phase_1h : std_logic_vector(5 downto 0);
    signal data_55 : std_logic_vector(d_width - 1 downto 0);
    signal data_aa : std_logic_vector(d_width - 1 downto 0);
    signal data_55_aa : std_logic_vector(d_width - 1 downto 0);
    signal data_addr : std_logic_vector(d_width - 1 downto 0);
    signal data_rev_addr : std_logic_vector(d_width - 1 downto 0);
    signal data_random : std_logic_vector(d_width - 1 downto 0);

    -- Data verification
    signal qq_valid : std_logic;
    signal qq : std_logic_vector(d_width - 1 downto 0);
    signal ok : std_logic;
    signal ok_valid : std_logic;
    signal error_detected : std_logic;
    signal check_addr : std_logic_vector(a_width - 1 downto 0);
    signal i_error : std_logic;

begin

    -- synopsys translate_off
    assert d_width mod rd_width = 0 report "memory_test : d_width not an integer multiple of rd_width" severity failure;
    -- synopsys translate_on

    -- synopsys translate_off
    assert d_width mod 8 = 0 report "memory_test : d_width not an integer multiple of 8" severity failure;
    -- synopsys translate_on

    -- synopsys translate_off
    assert d_width >= 32 report "memory_test : d_width less than 32" severity failure;
    -- synopsys translate_on

    -- synopsys translate_off
    assert d_width mod 32 = 0 report "memory_test : d_width not an integer multiple of 32" severity failure;
    -- synopsys translate_on

    -- Output internal versions of 'done' and 'error'
    done <= i_done;
    error <= i_error;

    -- Output internal version of 'ocp_m2s'
    ocp_m2s <= i_ocp_m2s;

    -- Generate a local reset from 'ocp_rst'.
    gen_lsr : process(ocp_clk, ocp_rst)
    begin
        if ocp_rst = '1' then
            lsr0 <= '1';
            lsr1 <= '1';
            lsr <= ocp_rst;
        elsif ocp_clk'event and ocp_clk = '1' then
            lsr0 <= '0';
            lsr1 <= lsr0;
            lsr <= lsr1;
        end if;
    end process;

    -- Remember parameters, in case they change after initial 'go' pulse.
    remember_param : process(ocp_clk)
    begin
        if ocp_clk'event and ocp_clk = '1' then
            -- Latch that is transparent when idle
            if i_done = '1' then
                offset_q <= offset;
                length_q <= EXT(length, a_width + 1) + 1;
            end if;
        end if;
    end process;

    -----------------------------------------------------------------------------
    -- Executive state machine, responsible for initiating a number of test
    -- phases in the subordinate state machines.
    -----------------------------------------------------------------------------

    gen_state1 : process(ocp_clk)
    begin
        if ocp_clk'event and ocp_clk = '1' then
            if lsr = '1' then
                state1 <= s1_idle;
                i_done <= '1';
                phase <= (others => '-');
                do_phase <= '0';
                done_phase_q <= "00";
            else
                state1 <= n_state1;
                i_done <= n_done;
                phase <= n_phase;
                do_phase <= n_do_phase;
                done_phase_q <= n_done_phase_q;
            end if;
        end if;
    end process;

    gen_n_state1 : process(
        state1,
        go,
        phase,
        done_phase_q,
        cmd_done,
        data_done)
    begin
        case state1 is
            when s1_idle =>
                if go = '1' then
                    n_state1 <= s1_doing;
                    n_done <= '0';
                    n_phase <= phase_55;
                    n_do_phase <= '1';
                    n_done_phase_q <= (others => '0');
                else
                    n_state1 <= s1_idle;
                    n_done <= '1';
                    n_phase <= phase_idle;
                    n_do_phase <= '0';
                    n_done_phase_q <= (others => '-');
                end if;

            when s1_doing =>
                if AND_reduce(done_phase_q) = '1' then
                    if phase = final_phase then
                        n_state1 <= s1_idle;
                        n_done <= '0';
                        n_phase <= phase_idle;
                        n_do_phase <= '0';
                        n_done_phase_q <= (others => '-');
                    else
                        n_state1 <= s1_doing;
                        n_done <= '0';
                        n_phase <= phase + 1;
                        n_do_phase <= '1';
                        n_done_phase_q <= (others => '0');
                    end if;
                else
                    n_state1 <= s1_doing;
                    n_done <= '0';
                    n_phase <= phase;
                    n_do_phase <= '0';
                    n_done_phase_q <= done_phase_q or (cmd_done & data_done);
                end if;
        end case;
    end process;

    -----------------------------------------------------------------------------
    -- OCP command issue state machine, responsible for issuing OCP commands
    -- during each test phase.
    -----------------------------------------------------------------------------

    gen_tag : process(ocp_clk)
    begin
        if ocp_clk'event and ocp_clk = '1' then
            if lsr = '1' then
                i_ocp_m2s.Tag <= tag_base;
            else
                if cmd_issue = '1' then
                    i_ocp_m2s.Tag <= (i_ocp_m2s.Tag + tag_incr) and tag_mask;
                end if;
            end if;
        end if;
    end process;

    -- Count OCP commands in progress
    count_commands : process(ocp_clk)
    begin
        if ocp_clk'event and ocp_clk = '1' then
            if lsr = '1' then
                cmd_count <= (others => '0');
            else
                if cmd_issue = '1' then
                    if cmd_retire = '0' then
                        cmd_count <= cmd_count + 1;
                    end if;
                else
                    if cmd_retire = '1' then
                        cmd_count <= cmd_count - 1;
                    end if;
                end if;
            end if;
        end if;
    end process;

    i_ocp_m2s.BurstLength <= EXT(cmd_blen, ADB3_OCP_BURST_WIDTH);

    -- Non-word bits of OCP address are always 0
    i_ocp_m2s.Addr(ocp_data_order - 1 downto 0) <= (others => '0');

    -- Generate OCP burst lengths and addresses
    gen_burst_length : process(ocp_clk)
    begin
        if ocp_clk'event and ocp_clk = '1' then
            if i_ocp_m2s.Cmd /= OCP_CMD_IDLE and ocp_s2m.CmdAccept = '1' then
                cmd_issue <= '1';
            else
                cmd_issue <= '0';
            end if;

            if cmd_init = '1' then
                cmd_remaining <= length_q;
                cmd_chunk <= CONV_STD_LOGIC_VECTOR(ocp_burst_length, ocp_burst_order + 1) - ("0" & offset_q(ocp_burst_order - 1 downto 0));
                i_ocp_m2s.Addr(ADB3_OCP_ADDR_WIDTH - 1 downto ocp_data_order) <= EXT(offset_q, ADB3_OCP_ADDR_WIDTH - ocp_data_order);
            else
                if cmd_issue = '1' then
                    cmd_remaining <= cmd_remaining - cmd_blen;
                    cmd_chunk <= CONV_STD_LOGIC_VECTOR(ocp_burst_length, ocp_burst_order + 1);
                    i_ocp_m2s.Addr(ADB3_OCP_ADDR_WIDTH - 1 downto ocp_data_order) <= EXT(i_ocp_m2s.Addr(ocp_data_order + a_width - 1 downto ocp_data_order) + cmd_blen, ADB3_OCP_ADDR_WIDTH - ocp_data_order);
                end if;
            end if;

            if cmd_chunk >= cmd_remaining then
                cmd_blen <= cmd_remaining(ocp_burst_order downto 0);
                cmd_final <= '1';
            else
                cmd_blen <= cmd_chunk;
                cmd_final <= '0';
            end if;
        end if;
    end process;

    -- State machine to generate OCP commands
    gen_state2 : process(ocp_clk)
    begin
        if ocp_clk'event and ocp_clk = '1' then
            if lsr = '1' then
                state2 <= s2_idle;
                cmd_done <= '0';
                cmd_init <= '0';
                i_ocp_m2s.Cmd <= OCP_CMD_IDLE;
            else
                state2 <= n_state2;
                cmd_done <= n_cmd_done;
                cmd_init <= n_cmd_init;
                i_ocp_m2s.Cmd <= n_ocp_m2s.Cmd;
            end if;
        end if;
    end process;

    gen_n_state2 : process(
        state2,
        cmd_count,
        cmd_final,
        do_phase,
        i_ocp_m2s,
        ocp_s2m)
    begin
        case state2 is
            when s2_idle =>
                if do_phase = '1' then
                    n_state2 <= s2_write_init1;
                    n_cmd_done <= '0';
                    n_cmd_init <= '1';
                    n_ocp_m2s.Cmd <= OCP_CMD_IDLE;
                else
                    n_state2 <= s2_idle;
                    n_cmd_done <= '0';
                    n_cmd_init <= '0';
                    n_ocp_m2s.Cmd <= OCP_CMD_IDLE;
                end if;

            when s2_write_init1 =>
                n_state2 <= s2_write_init2;
                n_cmd_done <= '0';
                n_cmd_init <= '0';
                n_ocp_m2s.Cmd <= OCP_CMD_IDLE;

            when s2_write_init2 =>
                n_state2 <= s2_write_wait;
                n_cmd_done <= '0';
                n_cmd_init <= '0';
                n_ocp_m2s.Cmd <= OCP_CMD_IDLE;

            when s2_write_wait =>
                if AND_reduce(cmd_count) = '0' then
                    n_state2 <= s2_write;
                    n_cmd_done <= '0';
                    n_cmd_init <= '0';
                    n_ocp_m2s.Cmd <= OCP_CMD_WRITE;
                else
                    n_state2 <= s2_write_wait;
                    n_cmd_done <= '0';
                    n_cmd_init <= '0';
                    n_ocp_m2s.Cmd <= OCP_CMD_IDLE;
                end if;

            when s2_write =>
                if ocp_s2m.CmdAccept = '1' then
                    if cmd_final = '1' then
                        n_state2 <= s2_read_init1;
                        n_cmd_done <= '0';
                        n_cmd_init <= '1';
                        n_ocp_m2s.Cmd <= OCP_CMD_IDLE;
                    else
                        n_state2 <= s2_write_init2;
                        n_cmd_done <= '0';
                        n_cmd_init <= '0';
                        n_ocp_m2s.Cmd <= OCP_CMD_IDLE;
                    end if;
                else
                    n_state2 <= s2_write;
                    n_cmd_done <= '0';
                    n_cmd_init <= '0';
                    n_ocp_m2s.Cmd <= OCP_CMD_WRITE;
                end if;

            when s2_read_init1 =>
                n_state2 <= s2_read_init2;
                n_cmd_done <= '0';
                n_cmd_init <= '0';
                n_ocp_m2s.Cmd <= OCP_CMD_IDLE;

            when s2_read_init2 =>
                n_state2 <= s2_read_wait;
                n_cmd_done <= '0';
                n_cmd_init <= '0';
                n_ocp_m2s.Cmd <= OCP_CMD_IDLE;

            when s2_read_wait =>
                if AND_reduce(cmd_count) = '0' then
                    n_state2 <= s2_read;
                    n_cmd_done <= '0';
                    n_cmd_init <= '0';
                    n_ocp_m2s.Cmd <= OCP_CMD_READ;
                else
                    n_state2 <= s2_read_wait;
                    n_cmd_done <= '0';
                    n_cmd_init <= '0';
                    n_ocp_m2s.Cmd <= OCP_CMD_IDLE;
                end if;

            when s2_read =>
                if ocp_s2m.CmdAccept = '1' then
                    if cmd_final = '1' then
                        n_state2 <= s2_done;
                        n_cmd_done <= '0';
                        n_cmd_init <= '0';
                        n_ocp_m2s.Cmd <= OCP_CMD_IDLE;
                    else
                        n_state2 <= s2_read_init2;
                        n_cmd_done <= '0';
                        n_cmd_init <= '0';
                        n_ocp_m2s.Cmd <= OCP_CMD_IDLE;
                    end if;
                else
                    n_state2 <= s2_read;
                    n_cmd_done <= '0';
                    n_cmd_init <= '0';
                    n_ocp_m2s.Cmd <= OCP_CMD_READ;
                end if;

            when s2_done =>
                n_state2 <= s2_idle;
                n_cmd_done <= '1';
                n_cmd_init <= '0';
                n_ocp_m2s.Cmd <= OCP_CMD_IDLE;
        end case;
    end process;

    -----------------------------------------------------------------------------
    -- OCP data transfer state machine, responsible for transferring data to and
    -- from the memory port during each test phase.
    -----------------------------------------------------------------------------

    -- Helper signals indicating when data is actually transferred via OCP
    read     <= '1' when ocp_s2m.Resp = OCP_RESP_VALID and i_ocp_m2s.RespAccept = '1' else '0';
    write    <= i_ocp_m2s.DataValid and ocp_s2m.DataAccept;
    transfer <= read or write;

    --
    -- Track test phase address and remaining word count, in order to figure out when each OCP
    -- command is considered completed (from this module's point of view).
    --
    gen_data_remaining : process(ocp_clk)
        variable increment : boolean;
    begin
        if ocp_clk'event and ocp_clk = '1' then
            data_init_q <= data_init;

            if data_init = '1' then
                data_offset <= offset_q;
            else
                if transfer = '1' then
                    data_offset <= data_offset + 1;
               end if;
            end if;

            if data_init = '1' then
                data_remaining <= length_q;
                if length_q(a_width downto 1) = 0 then
                    data_final <= '1';
                else
                    data_final <= '0';
                end if;
            else
                if transfer = '1' then
                    data_remaining <= data_remaining - 1;
                    if data_remaining(a_width downto 0) = 2 then
                        data_final <= '1';
                    else
                        data_final <= '0';
                    end if;
                end if;
            end if;

            cmd_retire <= transfer and (AND_reduce(data_offset(ocp_burst_order - 1 downto 0)) or data_final);
        end if;
    end process;

    gen_state3 : process(ocp_rst, ocp_clk)
    begin
        if ocp_clk'event and ocp_clk = '1' then
            if lsr = '1' then
                state3 <= s3_idle;
                data_done <= '0';
                data_init <= '0';
                i_ocp_m2s.DataValid <= '0';
                i_ocp_m2s.RespAccept <= '0';
            else
                state3 <= n_state3;
                data_done <= n_data_done;
                data_init <= n_data_init;
                i_ocp_m2s.DataValid <= n_ocp_m2s.DataValid;
                i_ocp_m2s.RespAccept <= n_ocp_m2s.RespAccept;
            end if;
        end if;
    end process;

    gen_n_state3 : process(
        state3,
        read,
        write,
        data_final,
        do_phase)
    begin
        case state3 is
            when s3_idle =>
                if do_phase = '1' then
                    n_state3 <= s3_write_init;
                    n_data_done <= '0';
                    n_data_init <= '1';
                    n_ocp_m2s.DataValid <= '0';
                    n_ocp_m2s.RespAccept <= '0';
                else
                    n_state3 <= s3_idle;
                    n_data_done <= '0';
                    n_data_init <= '0';
                    n_ocp_m2s.DataValid <= '0';
                    n_ocp_m2s.RespAccept <= '0';
                end if;

            when s3_write_init =>
                n_state3 <= s3_write_data;
                n_data_done <= '0';
                n_data_init <= '0';
                n_ocp_m2s.DataValid <= '0';
                n_ocp_m2s.RespAccept <= '0';

            when s3_write_data =>
                if write = '1' then
                    if data_final = '1' then
                        n_state3 <= s3_read_init;
                        n_data_done <= '0';
                        n_data_init <= '1';
                        n_ocp_m2s.DataValid <= '0';
                        n_ocp_m2s.RespAccept <= '0';
                    else
                        n_state3 <= s3_write_data;
                        n_data_done <= '0';
                        n_data_init <= '0';
                        n_ocp_m2s.DataValid <= '1';
                        n_ocp_m2s.RespAccept <= '0';
                    end if;
                else
                    n_state3 <= s3_write_data;
                    n_data_done <= '0';
                    n_data_init <= '0';
                    n_ocp_m2s.DataValid <= '1';
                    n_ocp_m2s.RespAccept <= '0';
                end if;

            when s3_read_init =>
                n_state3 <= s3_read_data;
                n_data_done <= '0';
                n_data_init <= '0';
                n_ocp_m2s.DataValid <= '0';
                n_ocp_m2s.RespAccept <= '0';

            when s3_read_data =>
                if read = '1' then
                    if data_final = '1' then
                        n_state3 <= s3_done;
                        n_data_done <= '0';
                        n_data_init <= '0';
                        n_ocp_m2s.DataValid <= '0';
                        n_ocp_m2s.RespAccept <= '0';
                    else
                        n_state3 <= s3_read_data;
                        n_data_done <= '0';
                        n_data_init <= '0';
                        n_ocp_m2s.DataValid <= '0';
                        n_ocp_m2s.RespAccept <= '1';
                    end if;
                else
                    n_state3 <= s3_read_data;
                    n_data_done <= '0';
                    n_data_init <= '0';
                    n_ocp_m2s.DataValid <= '0';
                    n_ocp_m2s.RespAccept <= '1';
                end if;

            when s3_done =>
                n_state3 <= s3_idle;
                n_data_done <= '1';
                n_data_init <= '0';
                n_ocp_m2s.DataValid <= '0';
                n_ocp_m2s.RespAccept <= '0';
        end case;
    end process;

    -----------------------------------------------------------------------------
    -- Data generation
    -----------------------------------------------------------------------------

    gen_phase_1h : process(ocp_clk)
    begin
        if ocp_clk'event and ocp_clk = '1' then
            for i in phase_1h'range loop
               if phase = i then
                   phase_1h(i) <= '1';
               else
                   phase_1h(i) <= '0';
               end if;
            end loop;
        end if;
    end process;

    next_data <= data_init_q or write or qq_valid;

    -- Constant 0x55 pattern
    gen_data_55 : for i in 0 to d_width / 8 - 1 generate
        data_55(8 * i + 7 downto 8 * i) <= X"55";
    end generate;

    -- Constant 0xAA pattern
    gen_data_aa : for i in 0 to d_width / 8 - 1 generate
        data_aa(8 * i + 7 downto 8 * i) <= X"aa";
    end generate;

    -- Alternating 0x55/0xAA pattern
    gen_not_ddr : if num_physical_word = 1 generate
        -- This memory type must be non-DDR, so we need to alternate between
        -- between 55 and aa on successive writes in order to force all data bits
        -- to transition on every clock cycle.
        gen_data_55_aa : process(ocp_clk)
        begin
            if ocp_clk'event and ocp_clk = '1' then
                if data_init = '1' then
                    data_55_aa <= data_55;
                else
                    if next_data = '1' and phase_1h(2) = '1' then
                        data_55_aa <= not data_55_aa;
                    end if;
                end if;
            end if;
        end process;
    end generate;

    gen_ddr : if num_physical_word /= 1 generate
        -- This memory type must be some kind of DDR, so we need to alternate
        -- between 55 and aa within a logical memory word in order to force all
        -- data bits to transition on every clock cycle.
        gen_data_55_aa : for i in 0 to d_width - 1 generate
            data_55_aa(i) <= data_55(i) when ((i / rd_width) mod 2) = 0 else data_aa(i);
        end generate;
    end generate;

    -- Own address
    gen_data_addr : process(ocp_clk)
    begin
        if ocp_clk'event and ocp_clk = '1' then
            if data_init = '1' then
                for i in 0 to num_32bit_word - 1 loop
                    data_addr(32 * (i + 1) - 1 downto 32 * i) <= conv_std_logic_vector(i, 32);
                end loop;
            else
                if next_data = '1' and OR_reduce(phase_1h(4 downto 3)) = '1' then
                    for i in 0 to num_32bit_word - 1 loop
                        data_addr(32 * (i + 1) - 1 downto 32 * i) <=
                            data_addr(32 * (i + 1) - 1 downto 32 * i) + num_32bit_word;
                    end loop;
                end if;
            end if;
        end if;
    end process;

    -- Bit-reversed own address
    gen_data_rev_addr : process(
        data_addr)
        variable j : natural;
    begin
        for i in 0 to num_32bit_word - 1 loop
            for j in 0 to 31 loop
                data_rev_addr(32 * i + j) <= data_addr(32 * i + 31 - j);
            end loop;
        end loop;
    end process;

    -- Random data using 64-bit LFSR
    gen_lfsreg : process(ocp_clk)
    begin
        if ocp_clk'event and ocp_clk = '1' then
            if data_init = '1' then
                for i in 0 to num_32bit_word - 1 loop
                    lfsreg(i) <= conv_std_logic_vector(i + 1, 64);
                end loop;
            else
                if next_data = '1' and phase_1h(5) = '1' then
                    for i in 0 to num_32bit_word - 1 loop
                        lfsreg(i) <= lfsr64(lfsreg(i));
                    end loop;
                end if;
            end if;
        end if;
    end process;

    gen_data_random : for i in 0 to num_32bit_word - 1 generate
        data_random(32 * (i + 1) - 1 downto 32 * i) <= lfsreg(i)(31 downto 0);
    end generate;

    -- Select type of test pattern according to which test phase we are in
    gen_data : process(ocp_clk)
    begin
        if ocp_clk'event and ocp_clk = '1' then
            if next_data = '1' then
                case phase is
                    when phase_55 =>
                        i_ocp_m2s.Data <= data_55;
                    when phase_aa =>
                        i_ocp_m2s.Data <= data_aa;
                    when phase_55_aa =>
                        i_ocp_m2s.Data <= data_55_aa;
                    when phase_addr =>
                        i_ocp_m2s.Data <= data_addr;
                    when phase_rev_addr =>
                        i_ocp_m2s.Data <= data_rev_addr;
                    when phase_random =>
                        i_ocp_m2s.Data <= data_random;
                    when others =>
                        i_ocp_m2s.Data <= (others => '-');
                end case;
            end if;
        end if;
    end process;

    i_ocp_m2s.DataByteEn <= (others => '1');

    -----------------------------------------------------------------------------
    -- Data verification
    -----------------------------------------------------------------------------

    error_detected <= ok_valid and not ok;

    check_q : process(ocp_clk)
    begin
        if ocp_clk'event and ocp_clk = '1' then
            if read = '1' then
                qq <= ocp_s2m.Data;
            end if;

            qq_valid <= read;

            if qq_valid = '1' then
                ok <= AND_reduce(i_ocp_m2s.Data xnor qq);
            end if;

            ok_valid <= qq_valid;

            if data_init = '1' then
                check_addr <= offset_q;
            else
                if ok_valid = '1' then
                    check_addr <= check_addr + 1;
                end if;
            end if;

            if i_done = '0' and i_error = '0' and error_detected = '1' then
                eaddr <= check_addr;
                ephase <= phase;
            end if;
        end if;
    end process;

    gen_error : process(ocp_clk)
    begin
        if ocp_clk'event and ocp_clk = '1' then
            if lsr = '1' then
                i_error <= '0';
            else
                if i_done = '1' then
                    if go = '1' then
                        i_error <= '0';
                end if;
                else
                    i_error <= i_error or error_detected;
                end if;
            end if;
        end if;
    end process;

end architecture;
