--==============================================================================
--
--     File Name: ddr3_sdram.vhd
--       Version: 0.1
--          Date: 28 Dec 2009
--         Model: Behavioral (NOT SYNTHESISABLE)
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Generic DDR3 SDRAM model
--
--   Description:
--
--   Limitations: OCD Calibration not supported.
--                Reads interrupted by reads not supported.
--                Writes interrupted by writes not supported.
--                ODT checks are incomplete.
--                Burst Chop not supported (A12/BC4)
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2009 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 28 Dec 09: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library std;
use std.textio.all;

library work;
use work.ddr3_sdram_pkg.all;
use work.ddr3_sdram_int_pkg.all;


entity ddr3_sdram is
  generic(
    message_level  : natural range 0 to 7 := 0;
    part           : part_t;
    short_init_dly : boolean);
  port(
    ck            : in    std_logic;  -- Double ended input: pos
    ck_l          : in    std_logic;  -- Double ended input: Neg
    reset_l       : in    std_logic;
    cke           : in    std_logic;
    cs_l          : in    std_logic;
    ras_l         : in    std_logic;
    cas_l         : in    std_logic;
    we_l          : in    std_logic;
    odt           : in    std_logic;
    dm            : in    std_logic_vector(part.part_size.data_bytes-1 downto 0);
    ba            : in    std_logic_vector(part.part_size.bank_width-1 downto 0);
    a             : in    std_logic_vector(part.part_size.arow_width-1 downto 0);
    dq            : inout std_logic_vector(part.part_size.data_width-1 downto 0);
    dqs           : inout std_logic_vector(part.part_size.data_bytes-1 downto 0);
    dqs_l         : inout std_logic_vector(part.part_size.data_bytes-1 downto 0);
    init_start    : in    boolean := false;
    init_filename : in    string := "init.txt";
    log_start     : in    boolean := false;
    log_filename  : in    string := "log.txt");
  end ddr3_sdram;


architecture behavioural of ddr3_sdram is

  -- Message reporting (warnings, errors, failures are always reported)
  --
  -- 7 Write and read and info additional messages
  -- 6 Read and info additional messages
  -- 5 Write and info additional messages
  -- 4 Write and read additional messages
  -- 3 Info additional messages only
  -- 2 Read additional messages only
  -- 1 Write additional messages only
  -- 0 No additional messages

  constant MESSAGE_WR : boolean := (message_level=1)or(message_level=4)or(message_level=5)or(message_level=7);
  constant MESSAGE_RD : boolean := (message_level=2)or(message_level=4)or(message_level=6)or(message_level=7);
  constant MESSAGE_IN : boolean := (message_level=3)or(message_level=5)or(message_level=6)or(message_level=7);

  -- Define constants for part size parameters for convenience
  constant PART_WIDTH : natural := part.part_size.sdram_bits;
  constant DATA_WIDTH : natural := part.part_size.data_width;
  constant BYTE_WIDTH : natural := part.part_size.byte_width;
  constant AROW_WIDTH : natural := part.part_size.arow_width;
  constant ACOL_WIDTH : natural := part.part_size.acol_width;
  constant DATA_BYTES : natural := part.part_size.data_bytes;
  constant BANK_WIDTH : natural := part.part_size.bank_width;

  constant NUM_ROWS  : natural := 2**AROW_WIDTH;
  constant NUM_COLS  : natural := 2**ACOL_WIDTH;
  constant NUM_BANKS : natural := 2**BANK_WIDTH;

  constant CHECK_TIMING : boolean := part.check_timing;

  subtype data_s is std_logic_vector(DATA_WIDTH-1 downto 0);
  subtype arow_s is std_logic_vector(AROW_WIDTH-1 downto 0);
  subtype acol_s is std_logic_vector(ACOL_WIDTH-1 downto 0);
  subtype bank_s is std_logic_vector(BANK_WIDTH-1 downto 0);
  subtype byte_s is std_logic_vector(DATA_BYTES-1 downto 0);

  type byte_n_s is array  (natural range 0 to (DATA_BYTES-1)) of natural;

  -- Define SRAM col type = col array of data_s
  type col_array_t is array (natural range 0 to (NUM_COLS - 1)) of data_s;

  -- Define pointer type to memory row
  type row_pointer_t is access col_array_t;
  shared variable row_pointer : row_pointer_t;

  -- Define SRAM bank type = row array of col_array_t
  type bank_array_t is array (natural range 0 to (NUM_ROWS - 1)) of row_pointer_t;

  -- Define SRAM memory type = bank array of bank_array_t
  type data_array_t is array (natural range 0 to (NUM_BANKS - 1)) of bank_array_t;

  shared variable memory : data_array_t;

  -- Define array to hold bank active row addresses
  type bank_act_row_t is array (natural range 0 to (NUM_BANKS - 1)) of arow_s;
  constant BANK_ACT_ROW_EMPTY : bank_act_row_t := (others => (others => 'U'));
  signal bank_act_row : bank_act_row_t;

  -- Define counter for number of active memory rows
  shared variable memory_row_cnt : natural := 0;

  -- Inputs that have been converted to X01
  signal ck_con      : std_logic;
  signal ck_l_con    : std_logic;
  signal a_con       : arow_s;
  signal ba_con      : bank_s;
  signal cs_l_con    : std_logic;
  signal ras_l_con   : std_logic;
  signal cas_l_con   : std_logic;
  signal we_l_con    : std_logic;
  signal odt_con     : std_logic;
  signal reset_l_con : std_logic;
  signal cke_con     : std_logic;
  signal dqs_con     : byte_s;
  signal dqs_l_con   : byte_s;
  signal dm_con      : byte_s;
  signal dq_con      : data_s;

  -- Retimed inputs
  signal odt_d1    : std_logic := '0';
  signal cke_d1    : std_logic := '0';

  -- Command decoding
  signal in_cmd           : command_s;
  signal in_addr_bit_10   : std_logic;
  signal in_addr_bit_12   : std_logic;
  signal in_mr1_dll_en_l  : std_logic;
--  signal in_mr1_ocd       : ocd_mode_s;
  signal in_mr0_dll_rst   : std_logic;
  signal in_command       : command_t;
  signal in_bank          : bank_full_s;
  signal in_bank_num      : natural range 0 to NUM_BANKS;
  signal in_addr          : arow_s;

  -- Mode registers
  signal reg_mr0  : arow_s;
  signal reg_mr1  : arow_s;
  signal reg_mr2  : arow_s;
  signal reg_mr3  : arow_s;

  -- MR0 register bits
  signal mr0_wr_cyc       : natural range WR_MIN to WR_MAX := WR_MIN;
  signal mr0_bl_cyc       : natural range BL_MIN to BL_MAX := BL_MIN;
  signal write_mr0_bl_cyc : natural range BL_MIN to BL_MAX := BL_MIN;
  signal read_mr0_bl_cyc  : natural range BL_MIN to BL_MAX := BL_MIN;
  signal mr0_cl_cyc       : natural range CL_MIN to CL_MAX := CL_MIN;
  signal mr0_burst_len    : bl_s;
  signal mr0_burst_typ    : std_logic;
  signal mr0_dll_reset    : std_logic;
  signal mr0_pd_fast_exit : boolean;

  -- MR1 register bits
  signal mr1_al_cyc      : natural range AL_MIN to AL_MAX := AL_MIN;
  signal mr1_dll_en_l    : std_logic;
  signal mr1_ods         : ods_type_s;
  signal mr1_r_tt_nom    : odt_nom_s;
  signal mr1_wl          : std_logic;
  signal mr1_tdqs_en     : std_logic;
  signal mr1_q_off       : boolean;

  -- MR2 register bits
  signal mr2_asr_en     : std_logic;
  signal mr2_srt_ext    : std_logic;
  signal mr2_r_tt_wr    : odt_wr_s;
--  signal mr2_sp_bits    : mr2_sp_s;
  signal mr2_cwl_cyc    : natural range CWL_MIN to CWL_MAX := CWL_MIN;

  -- MR3 register bits
  signal mr3_mpr_en     : std_logic;
  signal mr3_mpr_mode   : mpr_rd_s;

  -- Init FSM outputs
  signal init_complete     : boolean;
  signal delay_500us_over  : boolean;
  signal ck_stopped_error  : boolean := true;
  signal timing_msg_en     : boolean;

  -- Main FSM outputs
  signal load_mr0        : boolean;
  signal load_mr1        : boolean;
  signal load_mr2        : boolean;
  signal load_mr3        : boolean;
  signal auto_refresh    : boolean;
  signal self_refresh    : boolean;
  signal self_refresh_d1 : boolean;

  -- Bank FSM inputs
--  type bank_num_array_t is array (natural range 0 to (NUM_BANKS - 1)) of bank_full_s;
--  constant bank_num_array : bank_num_array_t := (BANK0,BANK1,BANK2,BANK3,BANK4,BANK5,BANK6,BANK7);

  -- Bank FSM outputs
  signal bank_act_any   : boolean;
  signal bank_read_any  : boolean;
  signal bank_write_any : boolean;
  signal bank_pre_any   : boolean;
  signal bank_pre_all   : boolean;

  -- Memory write/read preparation
  type rl_array_t is array (natural range RL_MAX downto 1) of boolean;
  signal bank_read_dly    : rl_array_t;
  signal bank_pre_dly     : rl_array_t;
  signal bank_pre_all_dly : rl_array_t;

  type wl_array_t is array (natural range WL_MAX downto 1) of boolean;
  signal bank_write_dly   : wl_array_t;
  signal bank_act_dly     : wl_array_t;

  type rl_addr_array_t is array (natural range RL_MAX downto 1) of arow_s;
  signal in_addr_read_dly : rl_addr_array_t;
  type wl_addr_array_t is array (natural range WL_MAX downto 1) of arow_s;
  signal in_addr_write_dly : wl_addr_array_t;

  type rl_bl8_array_t is array (natural range RL_MAX downto 1) of boolean;
  signal in_bl8_read_dly : rl_bl8_array_t;
  type wl_bl8_array_t is array (natural range WL_MAX downto 1) of boolean;
  signal in_bl8_write_dly : wl_bl8_array_t;

  type rl_bank_array_t is array (natural range RL_MAX downto 1) of natural;
  signal in_bank_read_dly : rl_bank_array_t;
  type wl_bank_array_t is array (natural range WL_MAX downto 1) of natural;
  signal in_bank_write_dly : wl_bank_array_t;

  signal bank_act_wl      : boolean;
  signal bank_pre_rl      : boolean;
  signal bank_pre_all_rl  : boolean;
  signal bank_read_rl     : boolean;
  signal bank_write_wl    : boolean;
  signal bank_addr_wl     : arow_s;
  signal bank_bl8_wl      : boolean;
  signal bank_addr_rl     : arow_s;
  signal bank_bl8_rl      : boolean;
  signal bank_num_wl      : natural;
  signal bank_num_rl      : natural;

  -- Addresses used to generate data output on DQ from memory array
  -- '_r' suffix means centred on rising edge of 'ck'
  -- '_f' suffix means centred on falling edge of 'ck'
  signal read_bank_r  : natural range 0 to NUM_BANKS;
  signal read_bank_f  : natural range 0 to NUM_BANKS;
  signal read_row_r   : natural;
  signal read_row_f   : natural;
  signal read_col_r   : natural;
  signal read_col_f   : natural;
  signal read_burst_r : natural;
  signal read_burst_f : natural;

  -- Burst index counter for reads. These are used to index a lookup
  -- table that has burst address offsets.
  -- '_r' suffix means centred on rising edge of 'ck'
  -- '_f' suffix means centred on falling edge of 'ck'
  signal read_burst_cnt_r : natural;
  signal read_burst_cnt_f : natural;

  -- Addresses used when storing DQ data in the memory array
  -- '_r' suffix means centred on rising edge of 'ck'
  -- '_f' suffix means centred on falling edge of 'ck'
  signal write_bank_r  : natural range 0 to NUM_BANKS;
  signal write_bank_f  : natural range 0 to NUM_BANKS;
  signal write_row_r   : natural;
  signal write_row_f   : natural;
  signal write_col_r   : natural;
  signal write_col_f   : natural;
  signal write_burst_r : natural;
  signal write_burst_f : natural;

  -- Burst index counter for writes. These are used to index a lookup
  -- table that has burst address offsets.
  -- '_r' suffix means centred on rising edge of 'ck'
  -- '_f' suffix means centred on falling edge of 'ck'
  signal write_burst_cnt_r : natural;
  signal write_burst_cnt_f : natural;

  -- Burst counter and write/read enables
  signal data_en_d0       : boolean;
  signal data_en_d1       : boolean;
  signal data_en_d2       : boolean;
  signal data_en_d3       : boolean;
  signal data_en_d4       : boolean;
  signal data_en_bl4      : boolean;
  signal data_en_bl8      : boolean;
  signal data_en          : boolean;
  signal data_st_en_bl4   : boolean;
  signal data_st_en_bl8   : boolean;
  signal data_st_en       : boolean;
  signal burst_rd_en_l_d1 : boolean;
  signal burst_rd_en_l_d2 : boolean;
  signal burst_rd_en_l_d3 : boolean;
  signal burst_rd_en_l_d4 : boolean;
  signal read_en_bl4      : boolean;
  signal read_en_bl8      : boolean;
  -- These signals cause data to be read from memory array and output on DQ,
  -- and are also used to generate DQS
  -- '_r' suffix means centred on rising edge of 'ck'
  -- '_f' suffix means centred on falling edge of 'ck'
  signal read_en_r        : boolean;
  signal read_en_f        : boolean;

  -- Signals that facilitate checking of writes in 'ddr3_sdram_chk'.
  signal burst_wr_en_d1   : boolean;
  signal burst_wr_en_d2   : boolean;
  signal burst_wr_en_d3   : boolean;
  signal burst_wr_en_d4   : boolean;
  signal write_en_r_bl4   : boolean := false;
  signal write_en_r_bl8   : boolean := false;
  -- These signals cause DQ data to be commited to memory array on DQS edges
  -- '_r' suffix means centred on rising edge of 'ck'
  -- '_f' suffix means centred on falling edge of 'ck'
  signal write_en_r       : boolean := false;
  signal write_en_f       : boolean := false;

  -- dq bi-directional bus
  signal dq_oe      : boolean;
  signal dq_out_mux : data_s;
  signal dq_out     : data_s;

  -- dqs/dqs_l bi-directional bus
  signal dqs_oe    : boolean;
  signal dqs_out   : byte_s;

  -- Control enable of parameter based timing checks
  signal timing_enabled : boolean;

  -- Clock period measurement variables
  -- Initialise to sensible values before measurement occurs
  shared variable ck_rising : time := 0 ns;
  shared variable ck_period : time := 0 ns;
  signal ck_period_sig : time := 0 ns;

  type byte_array_time_t is array (natural range 0 to (DATA_BYTES - 1)) of time;
  shared variable dqs_in_hold   : byte_array_time_t := (others => 0 ns);

  signal dqs_in_hold_sig  : byte_array_time_t := (others => 0 ns);
  signal dqs_before_ck    : std_logic_vector((DATA_BYTES - 1) downto 0) := (others => '0');

  signal t_ck_min : time := 0 ns;
  signal t_ck_max : time := 0 ns;

  -- Bank Activation t_faw check variables
  shared variable bank_act_time5 : time := 0 ns;
  shared variable bank_act_time4 : time := 0 ns;
  shared variable bank_act_time3 : time := 0 ns;
  shared variable bank_act_time2 : time := 0 ns;
  shared variable bank_act_time1 : time := 0 ns;

  -- Auto-Refresh interval
  signal ref_average_d1 : time := 0 ns;

  -- DLL Reset delay counter
  signal dll_delay    : natural;
  signal dll_delay_go : boolean;
  signal dll_delay_tc : boolean;

  -- Timing derived from registers
  signal rl_cyc       : natural range RL_MIN to RL_MAX := RL_MIN;
  signal wl_cyc       : natural range WL_MIN to WL_MAX := WL_MIN;

  -- Timing parameter constants converted to cycle counts
  signal t_rfc_min      : time;
  signal t_rp_cyc       : natural := 1;
  signal t_rpa_cyc      : natural := 1;
  signal t_rfc_cyc      : natural := 1;
  signal t_refi_cyc     : natural := 1;
  signal t_mrd_cyc      : natural := 1;
  signal t_ccd_cyc      : natural := 1;
  signal t_xs_cyc       : natural := 1;
  signal t_xsdll_cyc    : natural := 1;
  signal t_cke_cyc      : natural := 1;
  signal t_ras_cyc      : natural := 1;
  signal t_rc_cyc       : natural := 1;
  signal t_mod_cyc      : natural := 1;
  signal t_rcd_cyc      : natural := 1;
  signal t_rrd_cyc      : natural := 1;
  signal t_rtp_cyc      : natural := 1;
  signal t_wr_cyc       : natural := 1;
  signal t_wtr_cyc      : natural := 1;
  signal t_xp_cyc       : natural := 1;
  signal t_xpdll_cyc    : natural := 1;
  signal t_xpr_cyc      : natural := 1;
  signal t_zqoper_cyc   : natural := 1;
  signal t_zqcs_cyc     : natural := 1;
  signal t_faw          : time := 1 ns;
  signal t_dqss         : time := 1 ps;
  signal t_aofd         : time := 1 ns;
  signal t_mod          : time := 1 ns;
  signal t_refi         : time := 1 ns;

  -- Input signal qualifiers
  signal cke_setup   : boolean;
  signal cs_l_setup  : boolean;
  signal ras_l_setup : boolean;
  signal cas_l_setup : boolean;
  signal we_l_setup  : boolean;
  signal odt_setup   : boolean;
  signal ba_setup    : boolean;
  signal a_setup     : boolean;
  signal a10_setup   : boolean;
  signal dm_setup_r  : boolean;
  signal dm_setup_f  : boolean;

  -- Write Levelling
  signal wr_lvl_q : byte_s;
  signal dq_wrlvl : data_s;

  -- Test data write
  type mem_data_array_t is array (natural range 0 to (DATA_BYTES-1)) of std_logic_vector(BYTE_WIDTH-1 downto 0);
  signal memory_wr_bank : byte_n_s;
  signal memory_wr_row  : byte_n_s;
  signal memory_wr_col  : byte_n_s;
  signal memory_wr      : byte_s;
  signal memory_wr_data : mem_data_array_t;
  signal memory_rd_bank : natural;
  signal memory_rd_row  : natural;
  signal memory_rd_col  : natural;
  signal memory_rd      : std_logic;
  signal memory_rd_data : data_s;


begin

  ------------------- Input Retime Section ---------------------

  -- Convert all inputs to versions that are X01
  ck_con      <= to_X01(ck);
  ck_l_con    <= to_X01(ck_l);
  a_con       <= to_X01(a);
  ba_con      <= to_X01(ba);
  reset_l_con <= to_X01(reset_l);
  cke_con     <= to_X01(cke);
  cs_l_con    <= to_X01(cs_l);
  ras_l_con   <= to_X01(ras_l);
  cas_l_con   <= to_X01(cas_l);
  we_l_con    <= to_X01(we_l);
  odt_con     <= to_X01(odt);
  dm_con      <= to_X01(dm);
  dq_con      <= to_X01(dq);
  dqs_con     <= to_X01(dqs);
  dqs_l_con   <= to_X01(dqs_l);

  -- Sample control, address inputs on rising edge of ck
  retime_inputs : process(ck_con)
  begin
    if (ck_con'event and ck_con = '1') then
      odt_d1 <= odt_con;
      cke_d1 <= cke_con;
    end if;
  end process;

  -- Form in_cmd from retimed inputs
  -- Use cke from previous and present cycles
  in_cmd  <= (cke_d1 & cke_con & cs_l_con & ras_l_con & cas_l_con & we_l_con);

  in_addr <= a_con;

  -- Use a_con(10) as single bank/all banks select for Precharge
  -- Use a_con(10) as with/without auto-precharge select for Read or Write
  in_addr_bit_10 <= a_con(ADDR_SEL_BIT);

  -- Use a_con(12) as BC4/BL8 on the fly select for writes/reads
  in_addr_bit_12 <= a_con(ADDR_OTF_BIT);

  -- Use a_con(0) in emr1 as dll enable (active low)
  in_mr1_dll_en_l <= a_con(0);

  -- Use a_con(8) in mreg as dll reset
  in_mr0_dll_rst <= a_con(8);

  in_bank <= ba_con        when (BANK_WIDTH = 3) else
             '0' & ba_con  when (BANK_WIDTH = 2) else
             (others => 'U');

  in_bank_num <= 0 when (in_bank = BANK0) else
                 1 when (in_bank = BANK1) else
                 2 when (in_bank = BANK2) else
                 3 when (in_bank = BANK3) else
                 4 when (in_bank = BANK4) else
                 5 when (in_bank = BANK5) else
                 6 when (in_bank = BANK6) else
                 7 when (in_bank = BANK7) else
                 8; -- illegal bank number when undefined

  -- Decode in_cmd to command type
  in_command <= MRS0 when (std_match(in_cmd,C_LDMR))and(in_bank = BANK_MR0) else         -- Mode Register Set 0
                MRS1 when (std_match(in_cmd,C_LDMR))and(in_bank = BANK_MR1) else         -- Mode Register Set 1
                MRS2 when (std_match(in_cmd,C_LDMR))and(in_bank = BANK_MR2) else         -- Mode Register Set 2
                MRS3 when (std_match(in_cmd,C_LDMR))and(in_bank = BANK_MR3) else         -- Mode Register Set 3
                AREF when (std_match(in_cmd,C_AREF)) else                                -- Auto REFresh
                SPRE when (std_match(in_cmd,C_BPRE))and(in_addr_bit_10 = '0') else       -- Single Bank PREcharge (A10)
                APRE when (std_match(in_cmd,C_BPRE))and(in_addr_bit_10 = '1') else       -- All Bank PREcharge (A10)
                BACT when (std_match(in_cmd,C_BACT)) else                                -- Bank ACTivate
                WRTE when (std_match(in_cmd,C_WRTE))and(in_addr_bit_10 = '0') else       -- WRiTE without auto-precharge (A10)
                WRAP when (std_match(in_cmd,C_WRTE))and(in_addr_bit_10 = '1') else       -- WRite with Auto-Precharge (A10)
                READ when (std_match(in_cmd,C_READ))and(in_addr_bit_10 = '0') else       -- READ without auto-precharge (A10)
                RDAP when (std_match(in_cmd,C_READ))and(in_addr_bit_10 = '1') else       -- ReaD with Auto-Precharge (A10)
                SREN when (std_match(in_cmd,C_SREN)) else                                -- Self Refresh ENtry
                PDEN when (std_match(in_cmd,C_PDEN1))or(std_match(in_cmd,C_PDEN2)) else  -- Power Down ENtry (PDEN1 = NOOP)
                SPEX when (std_match(in_cmd,C_SREX1))or(std_match(in_cmd,C_SREX2))or     -- Self refresh or Power down EXit
                          (std_match(in_cmd,C_PDEX1))or(std_match(in_cmd,C_PDEX2)) else
                NOOP when (std_match(in_cmd,C_NOOP)) else                                -- NO OPeration
                DDES when (std_match(in_cmd,C_DDES)) else                                -- Device DESelect
                ZQCL when (std_match(in_cmd,C_ZQCAL))and(in_addr_bit_10 = '1') else      -- ZQ Calibration Long (A10)
                ZQCS when (std_match(in_cmd,C_ZQCAL))and(in_addr_bit_10 = '0') else      -- ZQ Calibration Short (A10)
                ILGL;                                                                    -- Illegal command

  ------------------- FSM Control Section ---------------------

  -- Initialisation FSM
  ddr3_sdram_init_fsm_i : ddr3_sdram_init_fsm
    generic map(
      part             => part)
    port map(
      ck               => ck_con,
      ck_stopped_error => ck_stopped_error,
      reset_l          => reset_l_con,
      cke              => cke_con,
      in_command       => in_command,
      in_addr          => in_addr,
      short_init_dly   => short_init_dly,
      t_xpr_cyc        => t_xpr_cyc,
      init_complete    => init_complete,
      delay_500us_over => delay_500us_over,
      timing_msg_en    => timing_msg_en);


  mr0_bl_cyc <= 4 when (mr0_burst_len = BL8)or((mr0_burst_len = OTF)and(in_addr_bit_12 = '1')) else 2;

  -- Main FSM
  ddr3_sdram_main_fsm_i : ddr3_sdram_main_fsm
    generic map(
      message_in => MESSAGE_IN,
      num_banks  => NUM_BANKS)
    port map(
      ck               => ck_con,
      reset_l          => reset_l_con,
      cke              => cke_con,
      cke_d1           => cke_d1,
      odt              => odt_con,
      odt_d1           => odt_d1,
      cs_l             => cs_l_con,
      in_command       => in_command,
      in_bank_num      => in_bank_num,
      delay_500us_over => delay_500us_over,
      pd_fast_exit     => mr0_pd_fast_exit,
      al_cyc           => mr1_al_cyc,
      cl_cyc           => mr0_cl_cyc,
      bl_cyc           => mr0_bl_cyc,
      wl_cyc           => wl_cyc,
      wr_cyc           => mr0_wr_cyc,
      t_aofd           => t_aofd,
      t_mod            => t_mod,
      t_wtr_cyc        => t_wtr_cyc,
      t_rfc_cyc        => t_rfc_cyc,
      t_xs_cyc         => t_xs_cyc,
      t_xsdll_cyc      => t_xsdll_cyc,
      t_xp_cyc         => t_xp_cyc,
      t_xpdll_cyc      => t_xpdll_cyc,
      t_cke_cyc        => t_cke_cyc,
      t_mrd_cyc        => t_mrd_cyc,
      t_ccd_cyc        => t_ccd_cyc,
      t_ras_cyc        => t_ras_cyc,
      t_rp_cyc         => t_rp_cyc,
      t_rpa_cyc        => t_rpa_cyc,
      t_rc_cyc         => t_rc_cyc,
      t_rrd_cyc        => t_rrd_cyc,
      t_rcd_cyc        => t_rcd_cyc,
      t_rtp_cyc        => t_rtp_cyc,
      t_wr_cyc         => t_wr_cyc,
      t_zqoper_cyc     => t_zqoper_cyc,
      t_zqcs_cyc       => t_zqcs_cyc,
      t_ck_min         => t_ck_min,
      dll_delay_tc     => dll_delay_tc,
      ck_stopped_error => ck_stopped_error,
      load_mr0         => load_mr0,
      load_mr1         => load_mr1,
      load_mr2         => load_mr2,
      load_mr3         => load_mr3,
      auto_refresh     => auto_refresh,
      self_refresh     => self_refresh,
      bank_act_any     => bank_act_any,
      bank_read_any    => bank_read_any,
      bank_write_any   => bank_write_any,
      bank_pre_any     => bank_pre_any,
      bank_pre_all     => bank_pre_all,
      cke_setup        => cke_setup,
      a_setup          => a_setup,
      a10_setup        => a10_setup,
      ba_setup         => ba_setup,
      odt_setup        => odt_setup,
      ras_l_setup      => ras_l_setup,
      cas_l_setup      => cas_l_setup,
      cs_l_setup       => cs_l_setup,
      we_l_setup       => we_l_setup);

  ------------------- Mode Register Section ---------------------

  -- Check write to MR0
  -- Determine t_ck_min and t_ck_max from updated MR register cl bits
  -- Detect if cl bits are out of allowed range
  -- Check for unused bits not zeroed
  -- Check for unused cl values for this speed grade
  -- Set DLL reset on exit from self refresh
  check_mr0 : process(ck_con)
    variable mr0         : arow_s;
    variable burst_len   : bl_s;
    variable sp_bit2     : std_logic;
    variable sp_bit7     : std_logic;
    variable sp_bit13    : std_logic;
    variable cas_lat     : cl_s;
    variable wr_rec      : wr_s;
  begin
    if (ck_con'event and ck_con = '1') then
      if load_mr0 then
        mr0 := a_con;

        burst_len := mr0(1 downto 0);
        sp_bit2   := mr0(2);           -- Must be zero
        cas_lat   := mr0(6 downto 4);
        sp_bit7   := mr0(7);           -- Must be zero
        wr_rec    := mr0(11 downto 9);

        assert not(MESSAGE_IN)
          report "DDR3 SDRAM Top Level (01): Load MR0"
          severity note;

        case burst_len is
          when OTF    => null;
          when BC4    => null;
          when BL8    => null;
          when others => assert false
                           report "DDR3 SDRAM Top Level (02): BL in MR0 out of range " & integer'image(BL_MIN) &
                                  " to "  & integer'image(BL_MAX) & " - Set to " & integer'image(BL_MIN)
                           severity error;
        end case;

        case cas_lat is
          when CL5    => mr0_cl_cyc <= 5;
          when CL6    => mr0_cl_cyc <= 6;
          when CL7    => mr0_cl_cyc <= 7;
          when CL8    => mr0_cl_cyc <= 8;
          when CL9    => mr0_cl_cyc <= 9;
          when CL10   => mr0_cl_cyc <= 10;
          when CL11   => mr0_cl_cyc <= 11;
          when others => mr0_cl_cyc <= 5;
                         assert false
                           report "DDR3 SDRAM Top Level (03): CL in MR0 out of range " & integer'image(CL_MIN) &
                                  " to "  & integer'image(CL_MAX) & " - Set to " & integer'image(CL_MIN)
                           severity error;
        end case;

        case wr_rec is
          when WR5    => mr0_wr_cyc <= 5;
          when WR6    => mr0_wr_cyc <= 6;
          when WR7    => mr0_wr_cyc <= 7;
          when WR8    => mr0_wr_cyc <= 8;
          when WR10   => mr0_wr_cyc <= 10;
          when WR12   => mr0_wr_cyc <= 12;
          when others => mr0_wr_cyc <= 5;
                         assert false
                           report "DDR3 SDRAM Top Level (04): WR in MR0 out of range " & integer'image(WR_MIN) &
                                  " to " & integer'image(WR_MAX) & " - Set to " &  integer'image(WR_MIN)
                           severity error;
        end case;

        assert not((sp_bit2 /= '0')or(sp_bit7 /= '0'))
          report "DDR3 SDRAM Top Level (05): Unused bits in MR0 (7,2) are not set to zero"
          severity warning;

        if (AROW_WIDTH = 14) then
          sp_bit13 := mr0(AROW_WIDTH-1);          -- Must be zero
          assert not((sp_bit13 /= '0'))
            report "DDR3 SDRAM Top Level (06): Unused bits in MR0 (13) are not set to zero"
            severity warning;
        end if;

        reg_mr0 <= mr0;
      elsif not(self_refresh)and(self_refresh_d1) then  -- exit from self refresh
        reg_mr0(8) <= '1';   -- Set DLL Reset bit
      else
        reg_mr0(8) <= '0';   -- Clear DLL Reset bit
      end if;
    end if;
  end process;

  -- Decompose MR0 into fields that are used elsewhere in this model
  mr0_burst_len    <= reg_mr0(1 downto 0);
  mr0_burst_typ    <= reg_mr0(3);
  mr0_dll_reset    <= reg_mr0(8);
  mr0_pd_fast_exit <= true when (reg_mr0(12) = '1') else false;


  ------------------- MR1 Section ---------------------

  -- Check write to MR1
  check_mr1 : process(ck_con)
    variable mr1       : arow_s;
    variable add_lat   : al_s;
    variable sp_bit8   : std_logic;
    variable sp_bit10  : std_logic;
    variable sp_bit13  : std_logic;
  begin
    if (ck_con'event and ck_con = '1') then
      if load_mr1 then
        mr1 := a_con;

        add_lat   := mr1(4 downto 3);
        sp_bit8   := mr1(8);           -- Must be zero
        sp_bit10  := mr1(10);          -- Must be zero

        assert not(MESSAGE_IN)
          report "DDR3 SDRAM Top Level (07): Load MR1"
          severity note;

        case add_lat is
          when AL0    => mr1_al_cyc <= 0;               -- AL = 0
          when AL_CL1 => mr1_al_cyc <= mr0_cl_cyc - 1;  -- AL = CL - 1
          when AL_CL2 => mr1_al_cyc <= mr0_cl_cyc - 2;  -- AL = CL - 2
          when others => mr1_al_cyc <= 0;
                         assert false
                           report "DDR3 SDRAM Top Level (08): AL in MR1 out of range " & integer'image(AL_MIN) &
                                  " to " & integer'image(AL_MAX) & " - Set to " & integer'image(AL_MIN)
                           severity error;
        end case;

        assert not((sp_bit8 /= '0')or(sp_bit10 /= '0'))
          report "DDR3 SDRAM Top Level (09): Unused bits in MR1 (10,8) are not set to zero"
          severity warning;

        if (AROW_WIDTH = 14) then
          sp_bit13 := mr1(AROW_WIDTH-1);          -- Must be zero
          assert not((sp_bit13 /= '0'))
            report "DDR3 SDRAM Top Level (10): Unused bits in MR1 (13) are not set to zero"
            severity warning;
        end if;

        assert not((mr1_wl='0')and(mr1(7)='1')and(MESSAGE_IN))
          report "DDR3 SDRAM Top Level (11): Entering Write Leveling mode"
          severity note;

        assert not((mr1_wl='1')and(mr1(7)='0')and(MESSAGE_IN))
          report "DDR3 SDRAM Top Level (12): Leaving Write Leveling mode"
          severity note;

        reg_mr1 <= mr1;
      end if;

      self_refresh_d1 <= self_refresh;

      if not(self_refresh_d1)and(self_refresh) then  -- entry to self refresh
        reg_mr1(0) <= '1';   -- Disable dll
      end if;
      if not(self_refresh)and(self_refresh_d1) then  -- exit from self refresh
        reg_mr1(0) <= '0';   -- Enable dll
      end if;
    end if;
  end process;

  -- Decompose MR1 into fields that are used elsewhere in this model
  mr1_dll_en_l    <= reg_mr1(0);
  mr1_ods         <= reg_mr1(5) & reg_mr1(1);                    -- Not used by model
  mr1_r_tt_nom    <= (reg_mr1(9) & reg_mr1(6) & reg_mr1(2)) when (odt_d1 = '1') else ODT_NOM_DIS;
  mr1_wl          <= reg_mr1(7);                                 -- Not used by model
  mr1_tdqs_en     <= reg_mr1(11);                                -- Not used by model
  mr1_q_off       <= true when (reg_mr1(12) = '1') else false;


  ------------------- MR2 Section ---------------------

  -- Check write to MR2
  check_mr2 : process(ck_con)
    variable mr2       : arow_s;
    variable sp_bits_a : std_logic_vector(2 downto 0);
    variable cas_wl    : cwl_s;
    variable sp_bit8   : std_logic;
    variable sp_bits_b : std_logic_vector(1 downto 0);
    variable sp_bits_c : std_logic_vector(2 downto 0);
  begin
    if (ck_con'event and ck_con = '1') then
      if load_mr2 then
        mr2 := a_con;

        sp_bits_a := mr2(2 downto 0);    -- Must be zero
        cas_wl    := mr2(5 downto 3);
        sp_bit8   := mr2(8);             -- Must be zero

        assert not(MESSAGE_IN)
          report "DDR3 SDRAM Top Level (13): Load MR2"
          severity note;

        case cas_wl is
          when CWL5   => mr2_cwl_cyc <= 5;
          when CWL6   => mr2_cwl_cyc <= 6;
          when CWL7   => mr2_cwl_cyc <= 7;
          when others => mr2_cwl_cyc <= 5;
                         assert false
                           report "DDR3 SDRAM Top Level (14): CWL in MR2 out of range " & integer'image(CWL_MIN) &
                                  " to " & integer'image(CWL_MAX) & " - Set to " & integer'image(CWL_MIN)
                           severity error;
        end case;

        assert not((sp_bits_a /= "000")or(sp_bit8 /= '0'))
          report "DDR3 SDRAM Top Level (15): Unused bits in MR2 (0:2,8) are not set to zero"
          severity warning;

        if (AROW_WIDTH = 13) then
          sp_bits_b := mr2(AROW_WIDTH-1 downto 11);  -- Must be zero
          assert not((sp_bits_b /= "00"))
            report "DDR3 SDRAM Top Level (16): Unused bits in MR2 (12:11) are not set to zero"
            severity warning;
        end if;

        if (AROW_WIDTH = 14) then
          sp_bits_c := mr2(AROW_WIDTH-1 downto 11);  -- Must be zero
          assert not((sp_bits_c /= "000"))
            report "DDR3 SDRAM Top Level (17): Unused bits in MR2 (13:11) are not set to zero"
            severity warning;
        end if;

        reg_mr2 <= mr2;
      end if;
    end if;
  end process;

  -- Update MR2 register bits
  mr2_asr_en      <= reg_mr2(6);                                 -- Not used by model
  mr2_srt_ext     <= reg_mr2(7);                                 -- Not used by model
  mr2_r_tt_wr     <= (reg_mr2(10) & reg_mr2(9)) when (odt_d1 = '1') else ODT_WR_DIS;


  ------------------- MR3 Section ---------------------

  -- Check write to MR3
  check_mr3 : process(ck_con)
    variable mr3       : arow_s;
    variable sp_bits_a : std_logic_vector(9 downto 0);
    variable sp_bits_b : std_logic_vector(10 downto 0);
  begin
    if (ck_con'event and ck_con = '1') then
      if load_mr3 then
        mr3 := a_con;

        assert not(MESSAGE_IN)
          report "DDR3 SDRAM Top Level (18): Load MR3"
          severity note;

        if (AROW_WIDTH = 13) then
          sp_bits_a := mr3(AROW_WIDTH-1 downto 3);  -- Must be zero
          assert not((sp_bits_a /= "0000000000"))
            report "DDR3 SDRAM Top Level (19): Unused bits in MR3 (12:3) are not set to zero"
            severity warning;
        end if;

        if (AROW_WIDTH = 14) then
          sp_bits_b := mr3(AROW_WIDTH-1 downto 3);  -- Must be zero
          assert not((sp_bits_b /= "00000000000"))
            report "DDR3 SDRAM Top Level (20): Unused bits in MR3 (13:3) are not set to zero"
            severity warning;
        end if;

        reg_mr3 <= mr3;
      end if;
    end if;
  end process;

  -- Update MR3 register bits
  mr3_mpr_en      <= reg_mr3(2);          -- Not used by model
  mr3_mpr_mode    <= reg_mr3(1 downto 0); -- Not used by model

  ------------------- t_ck Section ---------------------

  -- Update t_ck_min and t_ck_max
  update_t_ck : process(mr0_cl_cyc,mr2_cwl_cyc)
    variable cl_t_ck_min : time;
    variable cl_t_ck_max : time;
  begin
    cl_t_ck_min := part.speed_grade_cl_cwl_min(mr0_cl_cyc,mr2_cwl_cyc);
    cl_t_ck_max := part.speed_grade_cl_cwl_max(mr0_cl_cyc,mr2_cwl_cyc);

    assert not((cl_t_ck_min = 0 ns)and(cl_t_ck_max = 0 ns))
      report "DDR3 SDRAM Top Level (21): CL = " & integer'image(mr0_cl_cyc) &
             " with CWL = " & integer'image(mr2_cwl_cyc) &
             " is not valid for this speed grade"
      severity error;

    t_ck_min <= cl_t_ck_min;
    t_ck_max <= cl_t_ck_max;
  end process;

  ------------------- Memory Write/Read Preparation Section ---------------------

  -- Create POSITIVE edge delayed bank read signals by up to RL_MAX
  -- Delay precharge and read flags
  -- Delay bank and row/col address for reads
  rl_delay : process(ck_con)
    variable i : natural;
  begin
    if (ck_con'event and ck_con = '1') then
      bank_pre_dly(1)     <= bank_pre_any;
      bank_pre_all_dly(1) <= bank_pre_all;
      bank_read_dly(1)    <= bank_read_any;
      in_addr_read_dly(1) <= in_addr;
      in_bl8_read_dly(1)  <= (mr0_burst_len = BL8)or((mr0_burst_len = OTF)and(in_addr_bit_12 = '1'));
      in_bank_read_dly(1) <= in_bank_num;
      for i in 2 to RL_MAX loop
        bank_pre_dly(i)     <= bank_pre_dly(i-1);
        bank_pre_all_dly(i) <= bank_pre_all_dly(i-1);
        bank_read_dly(i)    <= bank_read_dly(i-1);
        in_addr_read_dly(i) <= in_addr_read_dly(i-1);
        in_bl8_read_dly(i)  <= in_bl8_read_dly(i-1);
        in_bank_read_dly(i) <= in_bank_read_dly(i-1);
      end loop;
    end if;
  end process;

  -- Create POSITIVE edge delayed bank write signals by up to WL_MAX
  -- Delay active and write flags
  -- Delay bank and row/col address for writes
  wl_delay : process(ck_con)
    variable i : natural;
  begin
    if (ck_con'event and ck_con = '1') then
      bank_act_dly(1)      <= bank_act_any;
      bank_write_dly(1)    <= bank_write_any;
      in_addr_write_dly(1) <= in_addr;
      in_bl8_write_dly(1)  <= (mr0_burst_len = BL8)or((mr0_burst_len = OTF)and(in_addr_bit_12 = '1'));
      in_bank_write_dly(1) <= in_bank_num;
      for i in 2 to WL_MAX loop
        bank_act_dly(i)      <= bank_act_dly(i-1);
        bank_write_dly(i)    <= bank_write_dly(i-1);
        in_addr_write_dly(i) <= in_addr_write_dly(i-1);
        in_bl8_write_dly(i)  <= in_bl8_write_dly(i-1);
        in_bank_write_dly(i) <= in_bank_write_dly(i-1);
      end loop;
    end if;
  end process;

  -- Avoid assigning a negative number during initialisation
  rl_cyc <= RL_MIN when (mr1_al_cyc + mr0_cl_cyc < 1) else (mr1_al_cyc + mr0_cl_cyc);
  wl_cyc <= WL_MIN when (mr1_al_cyc + mr0_cl_cyc < 2) else (mr1_al_cyc + mr0_cl_cyc - 1);

  -- Minimum value of rl_cyc is RL_MIN = 3
  bank_pre_rl     <= bank_pre_dly(rl_cyc);
  bank_pre_all_rl <= bank_pre_all_dly(rl_cyc);
  bank_read_rl    <= bank_read_dly(rl_cyc);
  bank_addr_rl    <= in_addr_read_dly(rl_cyc);
  bank_bl8_rl     <= in_bl8_read_dly(rl_cyc);
  bank_num_rl     <= in_bank_read_dly(rl_cyc);

  -- Minimum value of wl_cyc is WL_MIN = 2
  bank_act_wl     <= bank_act_dly(wl_cyc);
  bank_write_wl   <= bank_write_dly(wl_cyc);
  bank_addr_wl    <= in_addr_write_dly(wl_cyc);
  bank_bl8_wl     <= in_bl8_write_dly(wl_cyc);
  bank_num_wl     <= in_bank_write_dly(wl_cyc);

  -- Retime write signals on positive edge of clock
  -- Retime read signals on negative edge of clock
  latch_bank_write : process(ck_con)
    variable burst_bits : natural;
    variable burst_mask : acol_s;
  begin
    -- Update bank_act_row during Activate and Precharge
    -- Perform checks to see if bank number is in range
    if (ck_con'event and ck_con = '1') then
      if (bank_act_wl) then
        if (bank_num_wl < NUM_BANKS) then
          bank_act_row(bank_num_wl) <= bank_addr_wl;  -- store activated row address in selected bank
        else
          assert false
            report "DDR3 SDRAM Top Level (22): Bank number " & integer'image(bank_num_wl) & " out of range during ACTIVATE bank"
            severity error;
        end if;
      end if;

      if (bank_pre_rl) then
        if bank_pre_all_rl then
          bank_act_row <= BANK_ACT_ROW_EMPTY;              -- remove de-activated row addresses in all banks
        else
          if (bank_num_rl < NUM_BANKS) then
            bank_act_row(bank_num_rl) <= (others => 'U');  -- remove de-activated row address in selected bank
          else
            assert false
              report "DDR3 SDRAM Top Level (23): Bank number " & integer'image(bank_num_rl) & " out of range during PRECHARGE bank"
              severity error;
          end if;
        end if;
      end if;
    end if;


    -- Select between BC4 and BL8 on the fly using a(12)
    if (ck_con'event and ck_con = '1') then
      if (bank_write_wl) then
        if bank_bl8_wl then
          burst_bits := 3;
          burst_mask := std_logic_vector(resize(unsigned'("111"),ACOL_WIDTH));
          write_mr0_bl_cyc <= 4;
        else
          burst_bits := 2;
          burst_mask := std_logic_vector(resize(unsigned'("11"),ACOL_WIDTH));
          write_mr0_bl_cyc <= 2;
        end if;
      end if;

      if (bank_read_rl) then
        if bank_bl8_rl then
          burst_bits := 3;
          burst_mask := std_logic_vector(resize(unsigned'("111"),ACOL_WIDTH));
          read_mr0_bl_cyc <= 4;
        else
          burst_bits := 2;
          burst_mask := std_logic_vector(resize(unsigned'("11"),ACOL_WIDTH));
          read_mr0_bl_cyc <= 2;
        end if;
      end if;
    end if;


    -- Generate write addresses for burst with valid window centred on 'ck' rising edge
    if (ck_con'event and ck_con = '0') then
      if (bank_write_wl) then
        write_bank_r <= bank_num_wl;                                -- bank number for write command
        write_row_r  <= CONV_INTEGER(bank_act_row(bank_num_wl));    -- activated row number in selected bank for write command

        if (ACOL_WIDTH > ADDR_SEL_BIT) then                         -- col number for write command
          -- Use address bits above and below ADDR_SEL_BIT
          write_col_r <= CONV_INTEGER(bank_addr_wl(ACOL_WIDTH downto (ADDR_SEL_BIT + 1)) & bank_addr_wl((ADDR_SEL_BIT - 1) downto 0) and not burst_mask);
        else
          -- Use address bits below ADDR_SEL_BIT only
          write_col_r <= CONV_INTEGER(bank_addr_wl(ACOL_WIDTH-1 downto 0) and not burst_mask);
        end if;

        write_burst_r <= CONV_INTEGER(bank_addr_wl(burst_bits - 1 downto 0));    -- col address for write command
      end if;
    end if;

    -- Generate write addresses for burst with valid window centred on 'ck' falling edge
    if (ck_con'event and ck_con = '1') then
      write_bank_f  <= write_bank_r;
      write_row_f   <= write_row_r;
      write_col_f   <= write_col_r;
      write_burst_f <= write_burst_r;
    end if;

    -- Generate read addresses for burst with valid window centred on 'ck' rising edge
    if (ck_con'event and ck_con = '0') then
      if (bank_read_rl) then
        read_bank_r <= bank_num_rl;                  -- bank number for read command
        read_row_r  <= CONV_INTEGER(bank_act_row(bank_num_rl));    -- activated row number in selected bank for read command

        if (ACOL_WIDTH > ADDR_SEL_BIT) then       -- col number for read command
          -- Use address bits above and below ADDR_SEL_BIT
          read_col_r <= CONV_INTEGER(bank_addr_rl(ACOL_WIDTH downto (ADDR_SEL_BIT + 1)) & bank_addr_rl((ADDR_SEL_BIT - 1) downto 0) and not burst_mask);
        else
          -- Use address bits below ADDR_SEL_BIT only
          read_col_r <= CONV_INTEGER(bank_addr_rl(ACOL_WIDTH-1 downto 0) and not burst_mask);
        end if;

        read_burst_r <= CONV_INTEGER(bank_addr_rl(burst_bits - 1 downto 0));    -- col address for read command
      end if;
    end if;

    -- Generate read addresses for burst with valid window centred on 'ck' falling edge
    if (ck_con'event and ck_con = '1') then
      read_bank_f  <= read_bank_r;
      read_row_f   <= read_row_r;
      read_col_f   <= read_col_r;
      read_burst_f <= read_burst_r;
    end if;
  end process;

  -- Burst counter, added to address for writes
  burst_counter_writes : process(ck_con)
  begin
    if falling_edge(ck_con) then
      if bank_write_wl then
        write_burst_cnt_r <= 0;
      else
        if write_en_f then
          if write_burst_cnt_r < (2*write_mr0_bl_cyc - 2) then
            write_burst_cnt_r <= write_burst_cnt_r + 2;  -- increment if less than max
          else
            write_burst_cnt_r <= 0;          -- zero if reached max or undefined
          end if;
        end if;
      end if;
    end if;

    if rising_edge(ck_con) then
      write_burst_cnt_f <= write_burst_cnt_r;
    end if;
  end process;

  -- Burst counter, added to address for reads
  burst_counter_reads : process(ck_con)
  begin
    if falling_edge(ck_con) then
      if bank_read_rl then
        read_burst_cnt_r <= 0;
      else
        if read_en_f then
          if read_burst_cnt_r < (2*read_mr0_bl_cyc - 2) then
            read_burst_cnt_r <= read_burst_cnt_r + 2;  -- increment if less than max
          else
            read_burst_cnt_r <= 0;          -- zero if reached max or undefined
          end if;
        end if;
      end if;
    end if;

    if rising_edge(ck_con) then
      read_burst_cnt_f <= read_burst_cnt_r;
    end if;
  end process;

  -- Produce dq/dqs output enable signals at 'ck' rising edge
  data_enable_delay : process(ck_con)
  begin
    if (ck_con'event and ck_con = '1') then
      data_en_d1     <= bank_read_rl;
      data_en_d2     <= data_en_d1;
      data_en_d3     <= data_en_d2;
      data_en_d4     <= data_en_d3;
    end if;
  end process;

  data_en_d0 <= bank_read_rl;

  -- Output enable for dq bus
  data_en_bl4 <= data_en_d1 or data_en_d2;
  data_en_bl8 <= data_en_d1 or data_en_d2 or data_en_d3 or data_en_d4;
  data_en <= data_en_bl8 when (read_mr0_bl_cyc = 4) else data_en_bl4;

  -- Output enable for dqs bus
  data_st_en_bl4 <= data_en_d0 or data_en_d1 or data_en_d2;
  data_st_en_bl8 <= data_en_d0 or data_en_d1 or data_en_d2 or data_en_d3 or data_en_d4;
  data_st_en <= data_st_en_bl8 when (read_mr0_bl_cyc = 4) else data_st_en_bl4;

  -- Generete delayed burst write enable signal at falling edge of 'ck'
  write_enable_delay : process(ck_con)
  begin
    if falling_edge(ck_con) then
      burst_wr_en_d1 <= bank_write_wl;
      burst_wr_en_d2 <= burst_wr_en_d1;
      burst_wr_en_d3 <= burst_wr_en_d2;
      burst_wr_en_d4 <= burst_wr_en_d3;
    end if;
  end process;

  write_en_r_bl4 <= burst_wr_en_d1 or burst_wr_en_d2;
  write_en_r_bl8 <= burst_wr_en_d1 or burst_wr_en_d2 or burst_wr_en_d3 or burst_wr_en_d4;

  -- Produce burst write enable signal with valid window centred on falling edge of 'ck'
  gen_write_en_f : process(ck_con)
  begin
    if rising_edge(ck_con) then
      write_en_f <= write_en_r;
    end if;
  end process;

  -- Produce burst write enable signal with valid window centred on rising edge of 'ck'
  write_en_r <= write_en_r_bl8 when (write_mr0_bl_cyc = 4) else write_en_r_bl4;

  read_enable_delay : process(ck_con)
  begin
    if falling_edge(ck_con) then
      burst_rd_en_l_d1 <= bank_read_rl;
      burst_rd_en_l_d2 <= burst_rd_en_l_d1;
      burst_rd_en_l_d3 <= burst_rd_en_l_d2;
      burst_rd_en_l_d4 <= burst_rd_en_l_d3;
    end if;
  end process;

  read_en_bl4 <= burst_rd_en_l_d1 or burst_rd_en_l_d2;
  read_en_bl8 <= burst_rd_en_l_d1 or burst_rd_en_l_d2 or burst_rd_en_l_d3 or burst_rd_en_l_d4;

  -- Produce burst read enable signal with valid window centred on falling edge of 'ck'
  gen_read_en_f : process(ck_con)
  begin
    if rising_edge(ck_con) then
      read_en_f <= read_en_r;
    end if;
  end process;

  -- Produce burst read enable signal with valid window centred on rising edge of 'ck'
  read_en_r <= read_en_bl8 when (read_mr0_bl_cyc = 4) else read_en_bl4;

  ------------------- Memory Clear Section ---------------------

  clear_memory : process(ck_stopped_error)
    variable bank_i : natural;
    variable row_i  : natural;
  begin
    if (ck_stopped_error) then
      for bank_i in 0 to (NUM_BANKS - 1) loop
        for row_i in 0 to (NUM_ROWS - 1) loop
          deallocate(memory(bank_i)(row_i));
        end loop;
      end loop;
    end if;
  end process;

  ------------------- Memory Log Section ----------------------

  log_memory : process(log_start)
    file     log_file  : text;
    variable file_line : line;
    variable temp_line : line;

    variable row_dec_width : natural;
    variable col_dec_width : natural;
    variable field_width   : natural;

    variable bank_i     : natural;
    variable row_i      : natural;
    variable col_i      : natural;

    variable byte_act   : natural;
    variable byte_undef : boolean;
    variable byte       : std_logic_vector(BYTE_WIDTH-1 downto 0);
  begin
    row_dec_width := conv_dec_width(AROW_WIDTH);
    col_dec_width := conv_dec_width(ACOL_WIDTH);

    -- Round up vector length to next nearest nibble
    if (BYTE_WIDTH mod 4 = 0) then
      field_width := BYTE_WIDTH/4;
    else
      field_width := BYTE_WIDTH/4 + 1;
    end if;
    field_width := 2 + field_width + 1;

    if (log_start) then
      open_file(log_filename,WRITE_MODE,log_file);
      assert false
        report "DDR3 SDRAM Top Level (24): Logging memory contents to " & log_filename & "..."
        severity note;

      byte_act := 0;

      for bank_i in 0 to (NUM_BANKS-1) loop
        report "Bank: " & integer'image(bank_i);
        for row_i in 0 to (NUM_ROWS-1) loop
          if (memory(bank_i)(row_i) /= NULL) then   -- Only process rows that are defined
            for col_i in 0 to (NUM_COLS-1) loop
              for byte_i in 0 to (DATA_BYTES-1) loop
                if (byte_act = 0) then              -- Start of new line
                  write(file_line,bank_i,right,1);            write(file_line,string'(" ")); -- Write start bank (decimal)
                  write(file_line,row_i,right,row_dec_width); write(file_line,string'(" ")); -- Write start row (decimal)
                  write(file_line,col_i,right,col_dec_width); write(file_line,string'(" ")); -- Write start col (decimal)
                  write(file_line,byte_i,right,1);            write(file_line,string'(" ")); -- Write start byte (decimal)
                end if;

                byte := memory(bank_i)(row_i)(col_i)(((byte_i+1)*BYTE_WIDTH)-1 downto byte_i*BYTE_WIDTH);
                hex_write(temp_line,byte,left,field_width,byte_undef);  -- Procedure defined in package

                if not(byte_undef) then                   -- All bits valid 0 or 1
                  byte_act := byte_act + 1;
                  write(file_line,temp_line.all);
                else                                 -- All bits undefined
                  if (byte_act > 0) then             -- Write line if it contains valid data
                    writeline(log_file,file_line);
                  else                               -- Delete line if it does not
                    deallocate(file_line);
                  end if;
                  byte_act := 0;
                end if;

                if (byte_act = BYTES_PER_LINE) then  -- Write new line after 16 bytes output
                  writeline(log_file,file_line);
                  byte_act := 0;
                end if;
              end loop;
            end loop;
            -- Either write line or discard at end of active row
            if (byte_act > 0) then             -- Write line if it contains valid data
              writeline(log_file,file_line);
            else                               -- Delete line if it does not
              deallocate(file_line);
            end if;
            byte_act := 0;
          end if;
        end loop;
      end loop;

      file_close(log_file);
      assert false
        report "DDR3 SDRAM Top Level (25): Logging complete to " & log_filename
        severity note;
    end if;
  end process;

  ------------------- Memory Init Section ----------------------

  init_start_memory : process(init_start)
    file     init_file : text;
    variable file_line : line;

    variable read_ok        : boolean;
    variable bank_start_ok  : boolean;
    variable row_start_ok   : boolean;
    variable col_start_ok   : boolean;
    variable byte_start_ok  : boolean;
    variable end_of_line    : boolean;
    variable width_error    : boolean;

    variable col_cnt  : natural;
    variable byte_cnt : natural;

    variable bank_start_val : natural;
    variable row_start_val  : natural;
    variable col_start_val  : natural;
    variable byte_start_val : natural;

    variable byte : std_logic_vector(BYTE_WIDTH-1 downto 0);
  begin
    if (init_start) then
      open_file(init_filename,READ_MODE,init_file);
      assert false
        report "DDR3 SDRAM Top Level (26): Initialising memory contents from " & init_filename & "..."
        severity note;

      while not endfile(init_file) loop
        readline(init_file,file_line);

        read(file_line,bank_start_val,read_ok);
        if (read_ok)and(bank_start_val >= 0)and(bank_start_val < NUM_BANKS) then
          bank_start_ok := true;
        else
          bank_start_ok := false;
          assert false
            report "DDR3 SDRAM Top Level (27): Illegal BANK start value " & integer'image(bank_start_val)
            severity error;
        end if;

        read(file_line,row_start_val,read_ok);
        if (read_ok)and(row_start_val >= 0)and(row_start_val < NUM_ROWS) then
          row_start_ok := true;
        else
          row_start_ok := false;
          assert false
            report "DDR3 SDRAM Top Level (28): Illegal ROW start value " & integer'image(row_start_val)
            severity error;
        end if;

        read(file_line,col_start_val,read_ok);
        if (read_ok)and(col_start_val >= 0)and(col_start_val < NUM_COLS) then
          col_start_ok := true;
          col_cnt := col_start_val;
        else
          col_start_ok := false;
          assert false
            report "DDR3 SDRAM Top Level (29): Illegal COL start value " & integer'image(col_start_val)
            severity error;
        end if;

        read(file_line,byte_start_val,read_ok);
        if (read_ok)and(byte_start_val >= 0)and(byte_start_val < DATA_BYTES) then
          byte_start_ok := true;
          byte_cnt := byte_start_val;
        else
          byte_start_ok := false;
          assert false
            report "DDR3 SDRAM Top Level (30): Illegal BYTE start value " & integer'image(byte_start_val)
            severity error;
        end if;

        if (bank_start_ok and row_start_ok and col_start_ok and byte_start_ok) then
          byte_read(file_line,byte,end_of_line,width_error);
          while not(end_of_line) loop
            if (width_error) then
              assert false
                report "DDR3 SDRAM Top Level (31): Illegal BYTE width at position (BANK|ROW|COL|BYTE) " &
                integer'image(bank_start_val) & " " &  integer'image(row_start_val) & " " &
                integer'image(col_cnt) & " " & integer'image(byte_cnt)
                severity error;
            else
              -- If row doesn't exist then create pointer to new row ready for writing
              if (memory(bank_start_val)(row_start_val) = NULL) then
                memory(bank_start_val)(row_start_val) := new col_array_t;
                memory_row_cnt := memory_row_cnt + 1;
              end if;
              -- Initialise byte contents from file
              memory(bank_start_val)(row_start_val)(col_cnt)(((byte_cnt+1)*BYTE_WIDTH)-1 downto byte_cnt*BYTE_WIDTH) := byte;
              if (byte_cnt = (DATA_BYTES-1)) then
                byte_cnt := 0;
                col_cnt := col_cnt + 1;
              else
                byte_cnt := byte_cnt + 1;
              end if;
            end if;
            byte_read(file_line,byte,end_of_line,width_error);
          end loop;
        end if;
      end loop;

      file_close(init_file);
      assert false
        report "DDR3 SDRAM Top Level (32): Init complete from " & init_filename-- & " with " &
        severity note;
    end if;
  end process;

  ------------------- Memory Write Section ---------------------

  -- Perform write burst to memory
  -- Only perform write if we are in a valid FSM write cycle
  -- Only write to bytes enabled by data mask dm
  -- Write burst order determined by burst counter
  write_data : for i in dqs_con'range generate
    U0 : process(dqs_con(i))
      variable bank_i     : natural;
      variable row_i      : natural;
      variable col_i      : natural;
      variable byte_i_top : natural;
      variable byte_i_bot : natural;
      variable ok         : boolean;
      variable data_hex   : string(1 to (BYTE_WIDTH/4));
    begin
      if dqs_con(i)'event then
        ok := false;
        if rising_edge(dqs_con(i)) and write_en_r then
          bank_i := write_bank_r;
          row_i  := write_row_r;
          col_i  := write_col_r + burst_offset(write_burst_cnt_r, write_burst_r, mr0_burst_typ = BT_SEQ);
          ok := true;
        end if;
        if falling_edge(dqs_con(i)) and write_en_f then
          bank_i := write_bank_f;
          row_i  := write_row_f;
          col_i  := write_col_f + burst_offset(write_burst_cnt_f + 1, write_burst_f, mr0_burst_typ = BT_SEQ);
          ok := true;
        end if;

        if ok then
          -- Create pointer to new row ready for writing
          if (memory(bank_i)(row_i) = NULL) then
            memory(bank_i)(row_i) := new col_array_t;
            memory_row_cnt := memory_row_cnt + 1;
          end if;

          -- Only write data byte if appropriate data mask bit is active
          byte_i_top := BYTE_WIDTH * (i + 1) - 1;
          byte_i_bot := BYTE_WIDTH * i;

          if (dm_con(i) = '0') then
            memory(bank_i)(row_i)(col_i)(byte_i_top downto byte_i_bot) := dq_con(byte_i_top downto byte_i_bot);

            memory_wr(i)      <= '1';
            memory_wr_data(i) <= dq_con(byte_i_top downto byte_i_bot);
            memory_wr_bank(i) <= bank_i;
            memory_wr_row(i)  <= row_i;
            memory_wr_col(i)  <= col_i;

            for n in 0 to (BYTE_WIDTH/4)-1 loop
              data_hex((BYTE_WIDTH/4)-n) := conv_nibble_hex(dq_con((byte_i_bot+(n*4)+3) downto (byte_i_bot+(n*4))));
            end loop;
            assert not(MESSAGE_WR)
              report "DDR3 SDRAM Write : Bank = " & integer'image(bank_i) &
                                        " Row = " & integer'image(row_i) &
                                        " Col = " & integer'image(col_i) &
                                         " Data(" & integer'image(byte_i_top) & ":" & integer'image(byte_i_bot) & ")" &
                                          " = 0x" & data_hex
              severity note;
          end if;
        else
          memory_wr(i)      <= '0';
          memory_wr_data(i) <= (others => 'U');
        end if;
      end if;
    end process;
  end generate;


  ------------------- Memory Read Section ----------------------

  -- Perform read burst from memory controlled by the burst counter
  -- Only perform read if we are in a valid FSM read cycle
  -- Add timing to dq_out values
  generate_dq_out : process(ck_con)
    variable bank_i   : natural;
    variable row_i    : natural;
    variable col_i    : natural;
    variable data_out : data_s;
    variable data_hex : string(1 to (DATA_WIDTH/4));
  begin

    if ck_con'event then
      if rising_edge(ck_con) then
        -- Rising edge of ck outputs even data words
        if read_en_r then
          bank_i   := read_bank_r;
          row_i    := read_row_r;
          col_i    := read_col_r + burst_offset(read_burst_cnt_r, read_burst_r, mr0_burst_typ = BT_SEQ);

          row_pointer := memory(bank_i)(row_i);
          if (row_pointer = null) then
            data_out := (others => 'U');                         -- Non-existent row - data_out undefined
          else
            data_out := memory(bank_i)(row_i)(col_i);            -- Existing row - access pointer
          end if;
          dq_out  <= data_out;

          for n in 0 to (DATA_WIDTH/4)-1 loop
            data_hex((DATA_WIDTH/4)-n) := conv_nibble_hex(data_out(((n*4)+3) downto (n*4)));
          end loop;

          memory_rd      <= '1';
          memory_rd_data <= data_out;
          memory_rd_bank <= bank_i;
          memory_rd_row  <= row_i;
          memory_rd_col  <= col_i;

          assert not(MESSAGE_RD)
            report "DDR3 SDRAM Read  : Bank = " & integer'image(bank_i) &
                                      " Row = " & integer'image(row_i) &
                                      " Col = " & integer'image(col_i) &
                                   " Data = 0x" & data_hex
            severity note;
        else
          dq_out         <= (others => 'U');
          memory_rd      <= '0';
          memory_rd_data <= (others => 'U');
        end if;

        if read_en_r then
          dqs_out <= (others => '1');    -- dqs_out -> 1 for 1/2 'ck' period at rising 'ck' edges for read
        else
          dqs_out <= (others => '0');
        end if;
      else
        -- Falling edge of ck outputs odd data words
        if read_en_f then
          bank_i   := read_bank_f;
          row_i    := read_row_f;
          col_i    := read_col_f + burst_offset(read_burst_cnt_f + 1, read_burst_f, mr0_burst_typ = BT_SEQ);

          row_pointer := memory(bank_i)(row_i);
          if (row_pointer = null) then
            data_out := (others => 'U');                         -- Non-existent row - data_out undefined
          else
            data_out := memory(bank_i)(row_i)(col_i);            -- Existing row - access pointer
          end if;
          dq_out  <= data_out;

          for n in 0 to (DATA_WIDTH/4)-1 loop
            data_hex((DATA_WIDTH/4)-n) := conv_nibble_hex(data_out(((n*4)+3) downto (n*4)));
          end loop;

          memory_rd      <= '1';
          memory_rd_data <= data_out;
          memory_rd_bank <= bank_i;
          memory_rd_row  <= row_i;
          memory_rd_col  <= col_i;

          assert not(MESSAGE_RD)
            report "DDR3 SDRAM Read  : Bank = " & integer'image(bank_i) &
                                      " Row = " & integer'image(row_i) &
                                      " Col = " & integer'image(col_i) &
                                   " Data = 0x" & data_hex
            severity note;
        else
          dq_out         <= (others => 'U');
          memory_rd      <= '0';
          memory_rd_data <= (others => 'U');
        end if;

        dqs_out <= (others => '0');
      end if;
    end if;
  end process;


  ------------------- Bi-Directional Data Bus Section ----------------------

  -- Add timing to dq_oe value
  dq_oe  <= (data_en or (mr1_wl='1')) and not(mr1_q_off);

  -- Bidirectional data bus
  dq_out_mux <= dq_wrlvl when (mr1_wl='1') else dq_out;
  dq         <= dq_out_mux when dq_oe else (others => 'Z');

  -- Add timing to dqs_oe value
  dqs_oe <= (data_st_en and (mr1_wl='0')) and not(mr1_q_off);

  -- Bidirectional data bus
  dqs    <= dqs_out when dqs_oe else (others => 'Z');
  dqs_l  <= not(dqs_out) when dqs_oe else (others => 'Z');


  ------------------------ Write Levelling Section -------------------------

  -- Use ck_con as data input to write levelling register
  -- Use dqs as clock input to write levelling register

  gen_wrlvl_dqs_g : for i in 0 to DATA_BYTES-1 generate
    write_levelling_p : process(dqs(i))
    begin
      if rising_edge(dqs(i)) then
        wr_lvl_q(i) <= ck_con;
      end if;
    end process;

    dq_wrlvl(i*BYTE_WIDTH) <= wr_lvl_q(i);
    dq_wrlvl(((i+1)*BYTE_WIDTH)-1 downto (i*BYTE_WIDTH)+1) <= (others => '0');
  end generate;

  ------------------- Clock Period Measurements ----------------------------

  gen_ck_period_sig : process(delay_500us_over)
  begin
    if delay_500us_over'event and delay_500us_over then
      ck_period_sig <= ck_period;
    end if;
  end process;

  -- Determine ck period
  get_ck_period : process(ck_con)
  begin
    if (ck_con'event and ck_con = '1') then
      ck_period := now - ck_rising;
      ck_rising := now;
    end if;
  end process;

  -- Determing hold time between ck and dqs
  generate_dqs_timing : for i in 0 to (DATA_BYTES - 1) generate
    get_dqs_setup : process(dqs_con(i),ck_con)
    begin
      if (dqs_con(i)'event and dqs_con(i) = '1') then
        dqs_in_hold(i)   := now - ck_rising;

        if (dqs_in_hold(i) >= (ck_period - t_dqss))and(not(dqs_oe)) then   -- dqs before ck
          dqs_before_ck(i) <= '1';
        else                                                               -- dqs on or after ck
          dqs_before_ck(i) <= '0';
        end if;
      end if;
      dqs_in_hold_sig(i) <= dqs_in_hold(i);
    end process;
  end generate generate_dqs_timing;

  -- Check if t_faw is violated for 8 bank devices
  -- Allowed no more than 4 banks open in t_faw window
  get_bank_act_any_time : process(bank_act_any)
    variable bank_act_time_dif : time;
  begin
    if (bank_act_any'event and bank_act_any = true) then
      bank_act_time5 := bank_act_time4;
      bank_act_time4 := bank_act_time3;
      bank_act_time3 := bank_act_time2;
      bank_act_time2 := bank_act_time1;
      bank_act_time1 := now;

      bank_act_time_dif := bank_act_time1 - bank_act_time5;

      if (NUM_BANKS > 4) then                    -- Only test in 8 bank devices
        assert not(bank_act_time_dif < t_faw)
          report "DDR3 SDRAM Top Level (33): T_FAW delay violation during 5th bank ACTIVATE"
          severity error;
      end if;
    end if;
  end process;


  ------------------- Auto-Refresh Interval Check ----------------------------

  -- Count consecutive auto-refresh commands
  -- Max of 8 consecutive commands can be used for refresh
  count_arf : process(ck_con)
    variable aref_con_cnt : natural := 0;
  begin
    if (ck_con'event and ck_con = '1') then
      if (init_complete)and((in_command = AREF)or(in_command = NOOP)or(in_command = DDES)) then
        if (auto_refresh) then
          aref_con_cnt := aref_con_cnt + 1;
          assert not(aref_con_cnt = 8)
            report "DDR3 SDRAM Top Level (34): Cumulative consecutive auto refresh limit (8) exceeded"
            severity warning;
        end if;
      else
        aref_con_cnt := 0;
      end if;
    end if;
  end process;

  -- Auto refresh average interval calculation
  update_ref_int : process(ck_con)
    variable ref_start_time : time;
    variable ref_average    : time := 0 ns;
    variable ref_count      : natural;
  begin
    if (ck_con'event and ck_con = '1') then
      -- Record initial time for average interval calculation
      -- Zero auto refresh counter
      if timing_msg_en then
        ref_start_time := now;
        ref_count := 1;
      end if;

      -- Calculate average interval on completion of initialisation
      if init_complete then
        if auto_refresh then
          ref_count := ref_count + 1;
        end if;

        ref_average := (now - ref_start_time)/ref_count;
        ref_average_d1 <= ref_average;

        -- Generate warning when ref_average crosses the t_refi threshold
        assert not(init_complete and (ref_average > t_refi)and(ref_average_d1 <= t_refi))
          report "DDR3 SDRAM Top Level (35): Auto Refresh interval running average " & time'image(ref_average) &
                 " > T_REFI (" & time'image(t_refi) & ")"
          severity warning;
      end if;
    end if;
  end process;


  -- Initiate DLL Reset delay during normal operation
  dll_reset_count : process(ck_con)
  begin
    if (ck_con'event and ck_con = '1') then
      dll_delay <= t_xsdll_cyc;

      if (init_complete and (mr0_dll_reset = '1')) then
        dll_delay_go <= true;
      else
        dll_delay_go <= false;
      end if;
    end if;
  end process;

  -- DLL Reset delay counter
  -- Smallest delay generated is 1 cycle
  ddr3_sdram_dly_cnt_i1 : ddr3_sdram_dly_cnt
    port map(
      ck       => ck_con,
      delay_go => dll_delay_go,
      delay    => dll_delay,
      delay_tc => dll_delay_tc);


  ------------------- Timing parameter generation ----------------------------

  -- Assign values to timing parameters declared in package
  t_mrd_cyc    <= part.speed_grade.t_mrd_min;
  t_ccd_cyc    <= part.speed_grade.t_ccd_min;
  t_cke_cyc    <= greater_of(part.speed_grade.t_cke_min1,part.speed_grade.t_cke_min2,ck_period_sig);

  t_faw        <= part.speed_grade.t_faw_2k_min when (ACOL_WIDTH = 11) else
                  part.speed_grade.t_faw_1k_min;

  -- Assign cycle values to timing parameters declared in package
  t_rp_cyc     <= conv_time_cyc(part.speed_grade.t_rp_min,ck_period_sig);
  t_refi_cyc   <= conv_time_cyc(part.speed_grade.t_refi_max,ck_period_sig);
  t_ras_cyc    <= conv_time_cyc(part.speed_grade.t_ras_min,ck_period_sig);
  t_rc_cyc     <= conv_time_cyc(part.speed_grade.t_rc_min,ck_period_sig);
  t_mod_cyc    <= greater_of(part.speed_grade.t_mod_min1,part.speed_grade.t_mod_min2,ck_period_sig);
  t_rcd_cyc    <= conv_time_cyc(part.speed_grade.t_rcd_min,ck_period_sig);
  t_rtp_cyc    <= greater_of(part.speed_grade.t_rtp_min1,part.speed_grade.t_rtp_min2,ck_period_sig);
  t_wr_cyc     <= conv_time_cyc(part.speed_grade.t_wr_min,ck_period_sig);
  t_wtr_cyc    <= greater_of(part.speed_grade.t_wtr_min1,part.speed_grade.t_wtr_min2,ck_period_sig);
  t_rrd_cyc    <= greater_of(part.speed_grade.t_rrd_2k_min1,part.speed_grade.t_rrd_2k_min2,ck_period_sig) when (ACOL_WIDTH = 11) else
                  greater_of(part.speed_grade.t_rrd_1k_min1,part.speed_grade.t_rrd_1k_min2,ck_period_sig);

  with PART_WIDTH select t_rfc_min <=
    part.speed_grade.t_rfc_1g_min   when 30,     -- 1Gb
    part.speed_grade.t_rfc_2g_min   when 31,     -- 2Gb
    part.speed_grade.t_rfc_4g_min   when others; -- 4Gb

  t_xpr_cyc      <= greater_of(5,t_rfc_min + 10 ns,ck_period_sig);

  t_rfc_cyc      <= conv_time_cyc((t_rfc_min),ck_period_sig);

  t_xs_cyc       <= greater_of(part.speed_grade.t_xs_min1,(t_rfc_min + part.speed_grade.t_xs_min2),ck_period_sig);
  t_xsdll_cyc    <= part.speed_grade.t_dllk_min;

  t_xp_cyc       <= greater_of(part.speed_grade.t_xp_min1,part.speed_grade.t_xp_min2,ck_period_sig);
  t_xpdll_cyc    <= greater_of(part.speed_grade.t_xpdll_min1,part.speed_grade.t_xpdll_min2,ck_period_sig);

  t_zqoper_cyc  <= part.speed_grade.t_zqoper_min;
  t_zqcs_cyc    <= part.speed_grade.t_zqcs_min;

  -- time <= time * real
  t_dqss         <= ck_period_sig * part.speed_grade.t_dqss;

  -- time
  t_mod          <= ck_period_sig * t_mod_cyc;
  t_refi         <= part.speed_grade.t_refi_max;

  -------------------------------- Timing Checks ----------------------------

  -- Only enable parameter based timing checks if CHECK_TIMING is
  -- true for part and the 500us delay is over
  timing_enabled <= (CHECK_TIMING and delay_500us_over);

  dm_setup_r <= write_en_r;
  dm_setup_f <= write_en_f;

  ddr3_sdram_timing_chk_i : ddr3_sdram_timing_chk
    generic map(
      part            => part)
    port map(
      timing_enabled  => timing_enabled,
      bank_read_any   => bank_read_any,
      t_ck_min        => t_ck_min,
      t_ck_max        => t_ck_max,
      ck_period_sig   => ck_period_sig,
      cke_setup       => cke_setup,
      cs_l_setup      => cs_l_setup,
      cas_l_setup     => cas_l_setup,
      ras_l_setup     => ras_l_setup,
      we_l_setup      => we_l_setup,
      odt_setup       => odt_setup,
      a_setup         => a_setup,
      a10_setup       => a10_setup,
      ba_setup        => ba_setup,
      dm_setup_r      => dm_setup_r,
      dm_setup_f      => dm_setup_f,
      ck              => ck_con,
      ck_l            => ck_l_con,
      cke             => cke_con,
      cs_l            => cs_l_con,
      ras_l           => ras_l_con,
      cas_l           => cas_l_con,
      we_l            => we_l_con,
      odt             => odt_con,
      dm              => dm_con,
      ba              => ba_con,
      a               => a_con,
      dq              => dq_con,
      dqs             => dqs_con,
      dqs_l           => dqs_l_con);

  -- Produce message at end of initialisation period if timing disabled
  assert not((not CHECK_TIMING)and(timing_msg_en))
    report "DDR3 SDRAM Top Level (36): Model timing checks disabled"
    severity warning;

end behavioural;
