--==============================================================================
--
--     File Name: ddr3_sdram_int_pkg.vhd
--       Version: 0.1
--          Date: 28 Dec 2009
--         Model: Behavioral (NOT SYNTHESISABLE)
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Generic DDR3 SDRAM model
--
--   Description: Model package
--                Definition of internal types and constants
--
--   Limitations: Precharge Power Down and Active Power Down not supported
--                OCD Calibration not supported
--                Reads interrupted by reads not supported
--                Writes interrupted by writes not supported
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2009 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 28 Dec 09: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library std;
use std.textio.all;

library work;
use work.ddr3_sdram_pkg.all;


package ddr3_sdram_int_pkg is

  -- Define commands made up from input signals
  -- cke(prev) & cke(curr) & cs_l & ras_l & cas_l & we_l
  subtype command_s is std_logic_vector(5 downto 0);

  -- Define arrays commonly used by modules
  type boolean_array_t is array (natural range <>) of boolean;
  type natural_array_t is array (natural range <>) of natural;

  constant C_DDES  : command_s := "1-1---"; -- Device DESelect
  constant C_NOOP  : command_s := "1-0111"; -- NO OPeration
  constant C_READ  : command_s := "110101"; -- READ (with/without auto-precharge A10)
  constant C_WRTE  : command_s := "110100"; -- WRITE (with/without auto-precharge A10)
  constant C_BACT  : command_s := "110011"; -- Bank ACTivate
  constant C_BPRE  : command_s := "110010"; -- Bank PREcharge (single/all A10)
  constant C_AREF  : command_s := "110001"; -- Auto REFresh
  constant C_LDMR  : command_s := "110000"; -- LoaD Mode Register
  constant C_SREN  : command_s := "100001"; -- Self Refresh ENtry cke 1->0
  constant C_PDEN1 : command_s := "100111"; -- Power Down ENtry
  constant C_PDEN2 : command_s := "101---"; -- Power Down ENtry
  constant C_SREX1 : command_s := "010111"; -- Self Refresh EXit (Same as Power Down Exit1)
  constant C_SREX2 : command_s := "011---"; -- Self Refresh EXit (Same as Power Down Exit2)
  constant C_PDEX1 : command_s := "010111"; -- Power Down EXit (Same as self refresh Exit1)
  constant C_PDEX2 : command_s := "011---"; -- Power Down EXit (Same as self refresh Exit2)
  constant C_ZQCAL : command_s := "110110"; -- ZQ Calibration (long/short A10)

  -- Define bank registers made up from input signals ba(2:0) or '0' & ba(1:0)
  subtype bank_full_s is std_logic_vector(2 downto 0);

  constant BANK_MR0 : bank_full_s := b"000";
  constant BANK_MR1 : bank_full_s := b"001";
  constant BANK_MR2 : bank_full_s := b"010";
  constant BANK_MR3 : bank_full_s := b"011";

  constant BANK0 : bank_full_s := b"000";
  constant BANK1 : bank_full_s := b"001";
  constant BANK2 : bank_full_s := b"010";
  constant BANK3 : bank_full_s := b"011";
  constant BANK4 : bank_full_s := b"100";
  constant BANK5 : bank_full_s := b"101";
  constant BANK6 : bank_full_s := b"110";
  constant BANK7 : bank_full_s := b"111";

  -- Define which bit of address bus is used to select between
  -- normal/auto-precharge read/write and single/all precharge
  constant ADDR_SEL_BIT : natural := 10;

  -- Define which bit of address bus is used to select between
  -- BC4 and BL8 on the fly
  constant ADDR_OTF_BIT : natural := 12;

  -- Define full/short time delays for first stages of initialisation
  constant DELAY200SHRT : time := 2 ns;
  constant DELAY200FULL : time := 200 us;
  constant DELAY500SHRT : time := 5 ns;
  constant DELAY500FULL : time := 500 us;

  -- Define min time delay for reset active
  constant DELAY100NS : time := 100 ns;

  -- Define min time delay for cke 0 before reset inactive
  constant DELAY10NS : time := 10 ns;

  -- Define width of mode registers
  constant MREG_WIDTH : natural := 14;
--  constant MR0_WIDTH : natural := 14;
--  constant MR1_WIDTH : natural := 14;
--  constant MR2_WIDTH : natural := 14;
--  constant MR3_WIDTH : natural := 14;

  -- Define MR CL constants
  subtype cl_s is std_logic_vector(2 downto 0);
  constant CL_MAX : natural := 11;
  constant CL_MIN : natural := 5;  -- Values < 2 will cause pipeline delay problems
  constant CL5  : cl_s := b"001";
  constant CL6  : cl_s := b"010";
  constant CL7  : cl_s := b"011";
  constant CL8  : cl_s := b"100";
  constant CL9  : cl_s := b"101";
  constant CL10 : cl_s := b"110";
  constant CL11 : cl_s := b"111";

  -- Define MR Burst Length constants
  subtype bl_s is std_logic_vector(1 downto 0);
  constant BL_MAX : natural := 4;
  constant BL_MIN : natural := 2;
  constant BL8 : bl_s := b"00";  -- Fixed BL8
  constant OTF : bl_s := b"01";  -- Selectable on the fly using A12
  constant BC4 : bl_s := b"10";  -- Fixed BC4

  -- Define MR Write Recovery constants
  subtype wr_s is std_logic_vector(2 downto 0);
  constant WR_MAX : natural := 12;
  constant WR_MIN : natural := 5;
  constant WR5  : cl_s := b"001";
  constant WR6  : cl_s := b"010";
  constant WR7  : cl_s := b"011";
  constant WR8  : cl_s := b"100";
  constant WR10 : cl_s := b"101";
  constant WR12 : cl_s := b"110";

  -- Define MR Burst Type constants
  constant BT_SEQ : std_logic := '0';
  constant BT_INT : std_logic := '1';

  -- Define MR1 Output Drive Strength (ODS) constants
  subtype ods_type_s is std_logic_vector(1 downto 0);
  constant ODS_40 : ods_type_s := b"00";
  constant ODS_34 : ods_type_s := b"01";

  -- Define MR1 OCD Calibration Program constants
--  subtype ocd_mode_s is std_logic_vector(2 downto 0);
--  constant OCD_CAL_EXIT : ocd_mode_s := b"000";
--  constant OCD_DRIVE1   : ocd_mode_s := b"001";
--  constant OCD_DRIVE2   : ocd_mode_s := b"010";
--  constant OCD_ADJUST   : ocd_mode_s := b"100";
--  constant OCD_CAL_DEF  : ocd_mode_s := b"111";

  -- Define MR1 Additive Latency
  subtype al_s is std_logic_vector(1 downto 0);
  constant AL_MAX : natural := 2;
  constant AL_MIN : natural := 0;
  constant AL0    : al_s := b"00";
  constant AL_CL1 : al_s := b"01";
  constant AL_CL2 : al_s := b"10";

  -- Define MR1 ODT
  subtype odt_nom_s is std_logic_vector(2 downto 0);
  constant ODT_NOM_DIS : odt_nom_s := b"000";
  constant ODT_NOM_60  : odt_nom_s := b"001";
  constant ODT_NOM_120 : odt_nom_s := b"010";
  constant ODT_NOM_40  : odt_nom_s := b"011";
  constant ODT_NOM_20  : odt_nom_s := b"100";
  constant ODT_NOM_30  : odt_nom_s := b"101";

  -- Define MR2 CAS Write Latency (CWL)
  subtype cwl_s is std_logic_vector(2 downto 0);
  constant CWL_MAX : natural := 8;
  constant CWL_MIN : natural := 5;
  constant CWL5 : cwl_s := b"000";
  constant CWL6 : cwl_s := b"001";
  constant CWL7 : cwl_s := b"010";
  constant CWL8 : cwl_s := b"011";

  -- Define MR2 ODT
  subtype odt_wr_s is std_logic_vector(1 downto 0);
  constant ODT_WR_DIS : odt_wr_s := b"00";
  constant ODT_WR_60  : odt_wr_s := b"01";
  constant ODT_WR_120 : odt_wr_s := b"10";

  -- Define MR3 MPR read
  subtype mpr_rd_s is std_logic_vector(1 downto 0);
  constant MPR_RD_PRE : mpr_rd_s := b"00";

  -- Max/Min write/read cycle delays
  constant RL_MAX : natural := CL_MAX + AL_MAX;
  constant RL_MIN : natural := CL_MIN + AL_MIN;
  constant WL_MAX : natural := CL_MAX + AL_MAX - 1;
  constant WL_MIN : natural := CL_MIN + AL_MIN - 1;

  -- Define command types
  type command_t is (
    MRS0, MRS1, MRS2, MRS3, -- Mode Register Set
    AREF,                   -- Auto REFresh
    SPRE, APRE,             -- Single/All bank PREcharge (A10)
    BACT,                   -- Bank ACTivate
    WRTE, WRAP,             -- WRITE without/with Auto-Precharge (A10)
    READ, RDAP,             -- READ without/with Auto-Precharge (A10)
    NOOP,                   -- NO OPeration
    DDES,                   -- Device DESelect
    PDEN,                   -- Power Down ENtry
    SREN,                   -- Self Refresh ENtry
    SPEX,                   -- Self refresh/Power down EXit
    ZQCL, ZQCS,             -- ZQ Calibration Long/Short (A10)
    ILGL);                  -- ILleGaL command

  -- Define array for interleaved and sequential burst sequences
  -- Array indexed by start value and position in sequence
  subtype burst_cnt_s is std_logic_vector(2 downto 0);
  type burst_array_pos_t is array (natural range 0 to 7) of burst_cnt_s;
  type burst_array_t is array (natural range 0 to 7) of burst_array_pos_t;
  type burst_int_array_t is array (natural range 0 to 7, natural range 0 to 7) of natural;

  constant BURST_SEQ_ARRAY : burst_array_t := (("000","001","010","011","100","101","110","111"),  -- 0 = 0,1,2,3,4,5,6,7
                                               ("001","010","011","000","101","110","111","100"),  -- 1 = 1,2,3,0,5,6,7,4
                                               ("010","011","000","001","110","111","100","101"),  -- 2 = 2,3,0,1,6,7,4,5
                                               ("011","000","001","010","111","100","101","110"),  -- 3 = 3,0,1,2,7,4,5,6
                                               ("100","101","110","111","000","001","010","011"),  -- 4 = 4,5,6,7,0,1,2,3
                                               ("101","110","111","100","001","010","011","000"),  -- 5 = 5,6,7,4,1,2,3,0
                                               ("110","111","100","101","010","011","000","001"),  -- 6 = 6,7,4,5,2,3,0,1
                                               ("111","100","101","110","011","000","001","010")); -- 7 = 7,4,5,6,3,0,1,2

  constant BURST_INT_ARRAY : burst_array_t := (("000","001","010","011","100","101","110","111"),  -- 0 = 0,1,2,3,4,5,6,7
                                               ("001","000","011","010","101","100","111","110"),  -- 1 = 1,0,3,2,5,4,7,6
                                               ("010","011","000","001","110","111","100","101"),  -- 2 = 2,3,0,1,6,7,4,5
                                               ("011","010","001","000","111","110","101","100"),  -- 3 = 3,2,1,0,7,6,5,4
                                               ("100","101","110","111","000","001","010","011"),  -- 4 = 4,5,6,7,0,1,2,3
                                               ("101","100","111","110","001","000","011","010"),  -- 5 = 5,4,7,6,1,0,3,2
                                               ("110","111","100","101","010","011","000","001"),  -- 6 = 6,7,4,5,2,3,0,1
                                               ("111","110","101","100","011","010","001","000")); -- 7 = 7,6,5,4,3,2,1,0

  constant BURST_SEQ_INT_ARRAY : burst_int_array_t := ((0,1,2,3,4,5,6,7),
                                                       (1,2,3,0,5,6,7,4),
                                                       (2,3,0,1,6,7,4,5),
                                                       (3,0,1,2,7,4,5,6),
                                                       (4,5,6,7,0,1,2,3),
                                                       (5,6,7,4,1,2,3,0),
                                                       (6,7,4,5,2,3,0,1),
                                                       (7,4,5,6,3,0,1,2));

  constant BURST_INT_INT_ARRAY : burst_int_array_t := ((0,1,2,3,4,5,6,7),
                                                       (1,0,3,2,5,4,7,6),
                                                       (2,3,0,1,6,7,4,5),
                                                       (3,2,1,0,7,6,5,4),
                                                       (4,5,6,7,0,1,2,3),
                                                       (5,4,7,6,1,0,3,2),
                                                       (6,7,4,5,2,3,0,1),
                                                       (7,6,5,4,3,2,1,0));

  constant BYTES_PER_LINE : natural := 16;


  -- Function definitions
  function burst_offset(
    pos : in natural;
    start : in natural;
    sequential : in boolean)
  return natural;

  function conv_time_cyc(
    val : in time;
    per : in time)
  return natural;

  function conv_dec_width(
    bits : in natural range 0 to 31)
  return natural;

  function conv_nibble_hex(
    nibble : in std_logic_vector(3 downto 0))
  return character;

  function conv_hex_nibble(
    hex : in character)
  return std_logic_vector;

  function greater_of(
    val1 : in natural;
    val2 : in time;
    per  : in time)
  return natural;


  -- Procedure definitions
  procedure open_file(
    file_name    : in string;
    file_mode    : in file_open_kind;
    file file_id : text);

  procedure close_file(
    file file_id : text);

  procedure hex_write(
    hex_line  : inout line;
    vector    : in    std_logic_vector;
    justified : in    side;
    field     : in    natural;
    undefined : out   boolean);

  procedure byte_read(
    byte_line   : inout line;
    vector      : out   std_logic_vector;
    end_of_line : out   boolean;
    width_error : out   boolean);

  procedure hex_read(
    hex_line    : inout line;
    vector      : out   std_logic_vector;
    width_error : out   boolean);


  -- Component definitions
  component ddr3_sdram_dly_cnt
    port(
      ck       : in    std_logic;
      delay_go : in    boolean;
      delay    : in    natural;
      delay_tc : out   boolean);
  end component;

  component ddr3_sdram_init_fsm
    generic(
      part : part_t);
    port(
      ck               : in    std_logic;
      reset_l          : in    std_logic;
      ck_stopped_error : in    boolean;
      cke              : in    std_logic;
      in_command       : in    command_t;
      in_addr          : in    std_logic_vector(part.part_size.arow_width-1 downto 0);
      short_init_dly   : in    boolean;
      t_xpr_cyc        : in    natural;
      init_complete    : out   boolean;
      delay_500us_over : out   boolean;
      timing_msg_en    : out   boolean);
  end component;

  component ddr3_sdram_main_fsm
    generic(
      message_in : boolean;
      num_banks  : natural);
    port(
      ck               : in    std_logic;      -- Clock
      reset_l          : in    std_logic;      -- Reset
      cke              : in    std_logic;      -- Current CKE level on pins
      cke_d1           : in    std_logic;      -- As 'cke', but delayed by one cycle
      odt              : in    std_logic;      -- Current ODT pin level
      odt_d1           : in    std_logic;      -- As 'odt', but delayed by one cycle
      cs_l             : in    std_logic;      -- Current CS# pin level
      in_command       : in    command_t;      -- Decoded command currently on pins
      in_bank_num      : in    natural;        -- Bank index currently on pins
      delay_500us_over : in    boolean;        -- TRUE => initial 500 us delay is over
      pd_fast_exit     : in    boolean;        -- TRUE => fast exit power-down exit, from MREG[12]
      al_cyc           : in    natural;        -- Additive Latency from EMR1
      bl_cyc           : in    natural;        -- Burst Length from MR
      cl_cyc           : in    natural;        -- CAS latency from MR
      wl_cyc           : in    natural;        -- Write Length = RL - 1
      wr_cyc           : in    natural;        -- Write recovery from MR
      t_aofd           : in    time;           -- ODT turn-off delay
      t_mod            : in    time;           -- EMR1 to ODT turn-on delay
      t_wtr_cyc        : in    natural;        -- Write to read delay
      t_rfc_cyc        : in    natural;        -- Auto-refresh command period
      t_xsdll_cyc      : in    natural;        -- Exit self-refresh to command delay
      t_xs_cyc         : in    natural;        -- Exit self-refresh to command delay
      t_xp_cyc         : in    natural;        -- Exit power-down (fast) to command delay
      t_xpdll_cyc      : in    natural;        -- Exit power-down (slow) to command delay
      t_cke_cyc        : in    natural;        -- Clock enable low minimum clock cycles
      t_mrd_cyc        : in    natural;        -- MREG/EMR1/EMR2/EMR3 set command period
      t_ccd_cyc        : in    natural;        -- CAS-to-CAS delay between any two banks
      t_ras_cyc        : in    natural;        -- Minimum row open period
      t_rp_cyc         : in    natural;        -- Precharge single bank command period
      t_rpa_cyc        : in    natural;        -- Precharge all banks command period
      t_rc_cyc         : in    natural;        -- Activate-to-activate delay, same bank
      t_rrd_cyc        : in    natural;        -- Activate-to-activate delay between any two banks
      t_rcd_cyc        : in    natural;        -- RAS-to-CAS delay for a particular bank
      t_rtp_cyc        : in    natural;        -- Read to precharge delay for a particular bank
      t_wr_cyc         : in    natural;        -- Write recovery time, in cycles
      t_zqoper_cyc     : in    natural;        -- ZQ Calibration Long delay
      t_zqcs_cyc       : in    natural;        -- ZQ Calibration Short delay
--      t_anpd_cyc       : in    natural;        -- ODP high/low to power down entry period
--      t_axpd_cyc       : in    natural;        -- Power doen exit to ODP high/low period
      t_ck_min         : in    time;           -- Clock period (minimum)
      dll_delay_tc     : in    boolean;        -- DLL reset to first read delay counter terminal count
      ck_stopped_error : out   boolean;        -- TRUE => Clock has gone out of spec (not in self refresh)
      load_mr0         : out   boolean;        -- TRUE => current command is MR0 set
      load_mr1         : out   boolean;        -- TRUE => current command is MR1 set
      load_mr2         : out   boolean;        -- TRUE => current command is MR2 set
      load_mr3         : out   boolean;        -- TRUE => current command is MR3 set
      auto_refresh     : out   boolean;        -- TRUE => current command is auto-refresh
      self_refresh     : out   boolean;        -- TRUE => currently in self-refresh mode
      bank_act_any     : out   boolean;        -- TRUE => bank activate (any bank)
      bank_read_any    : out   boolean;        -- TRUE => bank read (any bank)
      bank_write_any   : out   boolean;        -- TRUE => bank write (any bank)
      bank_pre_any     : out   boolean;        -- TRUE => bank pre-charge (any bank)
      bank_pre_all     : out   boolean;        -- TRUE => bank pre-charge all
      -- These outputs indicate when certain DDR2 SDRAM inputs must be valid
      cke_setup        : out   boolean;        -- TRUE => CKE should be checked for setup/hold/validity
      cs_l_setup       : out   boolean;        -- TRUE => CS# should be checked for setup/hold/validity
      ras_l_setup      : out   boolean;        -- TRUE => RAS# should be checked for setup/hold/validity
      cas_l_setup      : out   boolean;        -- TRUE => CAS# should be checked for setup/hold/validity
      we_l_setup       : out   boolean;        -- TRUE => WE# should be checked for setup/hold/validity
      odt_setup        : out   boolean;        -- TRUE => ODT should be checked for setup/hold/validity
      ba_setup         : out   boolean;        -- TRUE => BA should be checked for setup/hold/validity
      a_setup          : out   boolean;        -- TRUE => A except A[10] should be checked for setup/hold/validity
      a10_setup        : out   boolean);       -- TRUE => A[10] should be checked for setup/hold/validity
    end component;

  component ddr3_sdram_timing_chk
    generic(
      part : part_t);
    port(
      timing_enabled : in    boolean;
      bank_read_any  : in    boolean;  -- TRUE => current command is a READ/RDAP
      cke_setup      : in    boolean;  -- TRUE => CKE should be checked for setup/hold/validity
      cs_l_setup     : in    boolean;  -- TRUE => CS# should be checked for setup/hold/validity
      ras_l_setup    : in    boolean;  -- TRUE => RAS# should be checked for setup/hold/validity
      cas_l_setup    : in    boolean;  -- TRUE => CAS# should be checked for setup/hold/validity
      we_l_setup     : in    boolean;  -- TRUE => WE# should be checked for setup/hold/validity
      odt_setup      : in    boolean;  -- TRUE => ODT should be checked for setup/hold/validity
      ba_setup       : in    boolean;  -- TRUE => BA should be checked for setup/hold/validity
      a_setup        : in    boolean;  -- TRUE => A except A[10] should be checked for setup/hold/validity
      a10_setup      : in    boolean;  -- TRUE => A[10] should be checked for setup/hold/validity
      dm_setup_r     : in    boolean;  -- TRUE => DM should be checked for setup/hold/validity at CK rising
      dm_setup_f     : in    boolean;  -- TRUE => DM should be checked for setup/hold/validity at CK falling
      t_ck_min       : in    time;
      t_ck_max       : in    time;
      ck_period_sig  : in    time;
      ck             : in    std_logic;
      ck_l           : in    std_logic;
      cke            : in    std_logic;
      cs_l           : in    std_logic;
      ras_l          : in    std_logic;
      cas_l          : in    std_logic;
      we_l           : in    std_logic;
      odt            : in    std_logic;
      dm             : in    std_logic_vector(part.part_size.data_bytes-1 downto 0);
      ba             : in    std_logic_vector(part.part_size.bank_width-1 downto 0);
      a              : in    std_logic_vector(part.part_size.arow_width-1 downto 0);
      dq             : in    std_logic_vector(part.part_size.data_width-1 downto 0);
      dqs            : in    std_logic_vector(part.part_size.data_bytes-1 downto 0);
      dqs_l          : in    std_logic_vector(part.part_size.data_bytes-1 downto 0));
  end component;

end ddr3_sdram_int_pkg;


package body ddr3_sdram_int_pkg is

  -- Get burst address
  function burst_offset(
    pos        : in natural;
    start      : in natural;
    sequential : in boolean)
  return natural is
  begin
    if sequential then
      return BURST_SEQ_INT_ARRAY(start,pos); -- Sequential
    else
      return BURST_INT_INT_ARRAY(start,pos); -- Interleaved
    end if;
  end;


  -- Divide parameter value (time) val by clock period (time) per
  -- and round up to nearest integer number of cycles
  function conv_time_cyc(
    val : in time;
    per : in time)
  return natural is
    variable cyc_cnt : natural;
    variable val_rem : natural;
    variable val_ps  : natural;
    variable per_ps  : natural;
  begin
    val_ps := val / 1 ps;
    per_ps := per / 1 ps;

    cyc_cnt := 0;
    if (per_ps = 0) then
      return 0;
    else
      val_rem := val_ps;
      while (val_rem > per_ps) loop
        cyc_cnt := cyc_cnt + 1;
        val_rem := val_rem - per_ps;
      end loop;
      return (cyc_cnt + 1);
    end if;
  end function;


  -- Select greater of (val1 * clock period) and (val2)
  function greater_of(
    val1 : in natural;
    val2 : in time;
    per  : in time)
  return natural is
    variable val2_cyc : natural;
  begin
    val2_cyc := conv_time_cyc(val2,per);

    if (val2_cyc > val1) then
      return val2_cyc;
    else
      return val1;
    end if;
  end function;


  -- Calculate the string width required to represent
  -- a number of binary width 'bits' in decimal format
  function conv_dec_width(
    bits : in natural range 0 to 31)
  return natural is
    variable max_val : natural;
  begin
    max_val := (2**bits)-1;

    case max_val is
      when 0 to 9 => return 1;
      when 10 to 99 => return 2;
      when 100 to 999 => return 3;
      when 1000 to 9999 => return 4;
      when 10000 to 99999 => return 5;
      when 100000 to 999999 => return 6;
      when 1000000 to 9999999 => return 7;
      when 10000000 to 99999999 => return 8;
      when 100000000 to 999999999 => return 9;
      when 1000000000 to integer'high => return 10;
    end case;
  end function;


  -- Calculate the string equivalent of
  -- a std_logic_vector nibble
  function conv_nibble_hex(
    nibble : in std_logic_vector(3 downto 0))
  return character is
    variable found_w : boolean := false;
    variable found_z : boolean := false;
    variable found_h : boolean := false;
    variable found_l : boolean := false;
    variable found_x : boolean := false;
    variable found_u : boolean := false;
    variable found_d : boolean := false;
  begin
    -- Detect presence of non '0','1' values
    -- In order of importance
    for i in 0 to 3 loop
      case nibble(i) is
        when 'U' => found_u := true;
        when 'X' => found_x := true;
        when '0' => null;
        when '1' => null;
        when 'Z' => found_z := true;
        when 'W' => found_w := true;
        when 'L' => found_l := true;
        when 'H' => found_h := true;
        when '-' => found_d := true;
      end case;
    end loop;

    -- Assign character for all nibble values
    -- Non-hex character results are prioritised U,X,Z,W,L,H,-
    case nibble is
      when "0000" => return '0';
      when "0001" => return '1';
      when "0010" => return '2';
      when "0011" => return '3';
      when "0100" => return '4';
      when "0101" => return '5';
      when "0110" => return '6';
      when "0111" => return '7';
      when "1000" => return '8';
      when "1001" => return '9';
      when "1010" => return 'A';
      when "1011" => return 'B';
      when "1100" => return 'C';
      when "1101" => return 'D';
      when "1110" => return 'E';
      when "1111" => return 'F';
      when others => if    (found_u) then return 'U';
                     elsif (found_x) then return 'X';
                     elsif (found_z) then return 'Z';
                     elsif (found_w) then return 'W';
                     elsif (found_l) then return 'L';
                     elsif (found_h) then return 'H';
                     elsif (found_d) then return '-';
                     else return '?';
                     end if;
    end case;
  end function;


  -- Calculate the string equivalent of
  -- a std_logic_vector nibble
  function conv_hex_nibble(
    hex : in character)
  return std_logic_vector is
  begin
    case hex is
      when '0' => return "0000";
      when '1' => return "0001";
      when '2' => return "0010";
      when '3' => return "0011";
      when '4' => return "0100";
      when '5' => return "0101";
      when '6' => return "0110";
      when '7' => return "0111";
      when '8' => return "1000";
      when '9' => return "1001";
      when 'A' => return "1010";
      when 'B' => return "1011";
      when 'C' => return "1100";
      when 'D' => return "1101";
      when 'E' => return "1110";
      when 'F' => return "1111";
      when 'U' => return "UUUU";
      when 'X' => return "XXXX";
      when 'Z' => return "ZZZZ";
      when 'W' => return "WWWW";
      when 'L' => return "LLLL";
      when 'H' => return "HHHH";
      when '-' => return "----";
      when others => return "UUUU";
                     assert false
                       report "Function conv_hex_nibble error - Illegal char value: " & character'image(hex)
                       severity error;
    end case;
  end function;


  -- Convert the std_logic_vector input 'vector' to hex character
  -- format eg 0x55, and write to line 'hex_line'. Use justification
  -- and field size specified and return boolean 'undefined' indicating
  -- if complete byte is undefined
  procedure hex_write(
    hex_line  : inout line;
    vector    : in    std_logic_vector;
    justified : in    side;
    field     : in    natural;
    undefined : out   boolean)
  is
    variable full_nibbles  : natural;
    variable vector_space  : integer;
    variable nibble_top    : natural;
    variable vector_top    : natural;
    variable vector_bot    : natural;
    variable vector_low    : natural;
    variable vector_len    : natural;
    variable nibble        : std_logic_vector(3 downto 0);
    variable nibble_hex    : character;
    variable nibble_undef  : boolean;
  begin
    deallocate(hex_line);

    -- Get lower bound and length of vector input
    vector_low := vector'low;
    vector_len := vector'length;

    -- Round up vector length to next nearest nibble
    if (vector_len mod 4 = 0) then
      full_nibbles := vector_len/4;
    else
      full_nibbles := vector_len/4 + 1;
    end if;

    -- Calculate difference between field size and length of hex string
    vector_space := field-(2+full_nibbles);

    -- Insert space before hex string if appropriate
    if (vector_space > 0)and(justified=right) then
      for i in 1 to vector_space loop
        write(hex_line,string'(" "));
      end loop;
    end if;

    -- Process bits one at a time starting at msb
    write(hex_line,string'("0x"));

    undefined := true;
    for i in (full_nibbles-1) downto 0 loop
      vector_bot := i*4 + vector_low;
      if (i = (full_nibbles-1)) then                      -- Write top nibble bits (may be < 4 bits)
        nibble_top := (vector'length-1);
        vector_top := nibble_top + vector_low;
        nibble     := "0000";
        nibble((nibble_top rem 4) downto 0) := vector(vector_top downto vector_bot);
      else                                                -- Write remaining nibble bits (4 bits)
        vector_top := ((i+1)*4)-1 + vector_low;
        nibble     := vector(vector_top downto vector_bot);
      end if;
      nibble_hex := conv_nibble_hex(nibble);
      write(hex_line,nibble_hex);
      if (nibble_hex /= 'U') then
        undefined := false;
      end if;
    end loop;

    -- Insert space after hex string if appropriate
    if (vector_space > 0)and(justified=left) then
      for i in 1 to vector_space loop
        write(hex_line,string'(" "));
      end loop;
    end if;
  end procedure;


  -- Read one bytes worth of characters from 'byte_line'
  -- Byte will be in hex format eg 0x55 or in decimal eg 85
  -- In both cases convert characters to std_logic_vector
  -- and return indicating if there is a width error or if
  -- there are no more byte characters to read
  procedure byte_read(
    byte_line   : inout line;
    vector      : out   std_logic_vector;
    end_of_line : out   boolean;
    width_error : out   boolean)
  is
    variable line_char     : character;
    variable char_ok       : boolean;
    variable temp_line     : line;
    variable dec_value     : natural;
  begin
    -- Ignore white space before byte
    read(byte_line,line_char,char_ok);
    while (line_char = ' ')and(char_ok) loop
      read(byte_line,line_char,char_ok);
    end loop;

    -- Write byte characters to temp_line
    if (char_ok) then
      deallocate(temp_line);
      write(temp_line,line_char);
      read(byte_line,line_char,char_ok);
      while (line_char /= ' ')and(char_ok) loop
        write(temp_line,line_char);
        read(byte_line,line_char,char_ok);
      end loop;
      end_of_line := false;

      if (temp_line(1 to 2) = "0x") then
        hex_read(temp_line,vector,width_error);
      else
        read(temp_line,dec_value);
        vector := STD_LOGIC_VECTOR(TO_UNSIGNED(dec_value,vector'length));
      end if;
    else
      end_of_line := true;
    end if;
  end procedure;


  -- Generate the std_logic_vector output 'vector' from hex character
  -- format eg 0x55, read from line 'hex_line'. Return boolean 'width_error'
  -- indicating if number of characters in 'hex_line' does not match with
  -- number of bits in 'vector'
  procedure hex_read(
    hex_line    : inout line;
    vector      : out   std_logic_vector;
    width_error : out   boolean)
  is
    variable vector_len    : natural;
    variable full_nibbles  : natural;
    variable nibble        : std_logic_vector(3 downto 0);
    variable vector_top    : natural;
    variable vector_bot    : natural;
  begin
    -- Get length of vector output
    vector_len := vector'length;

    -- Round up vector length to next nearest nibble
    if (vector_len mod 4 = 0) then
      full_nibbles := vector_len/4;
    else
      full_nibbles := vector_len/4 + 1;
    end if;

    if (((hex_line'length)-2) = full_nibbles) then          -- Number of nibbles matches vector length
      for i in 0 to full_nibbles-1 loop
        nibble := conv_hex_nibble(hex_line(3+(full_nibbles-1)-i));
        vector_bot := i*4;
        if (i = (full_nibbles-1)) then                      -- Write vector top nibble bits (may be < 4 bits)
          vector_top := vector'length-1;
          vector(vector_top downto vector_bot) := nibble((vector_top rem 4) downto 0);
        else                                                -- Write vector remaining nibble bits (4 bits)
          vector_top := ((i+1)*4)-1;
          vector(vector_top downto vector_bot) := nibble;
        end if;
      end loop;
      width_error := false;
    else                                                    -- Not enough nibbles to fill vector
      vector(vector'length-1 downto 0) := (others => 'U');
      width_error := true;
    end if;
  end procedure;


  ------------------------------------------------------------------------------
  -- Procedure Name: open_file
  -- Dependencies: package std.textio
  -- Procedure to open file file_name in mode file_mode
  -- Tests file_mode for validity and asserts failure if invalid
  -- Decodes file_status and asserts failure in all cases except:
  --   OPEN_OK      : file was opened successfully for first time
  --   STATUS_ERROR : file was previously opened (warning)
  -- Requires file definition in calling process or architecture:
  --   file <file_id_name> : text;
  -- Procedure call:
  --   open_file(<file_name>,<file_mode>,<file_id_name>);
  ------------------------------------------------------------------------------

  procedure open_file(
    file_name : in string;
    file_mode : in file_open_kind;
    file file_id : text)
  is
    variable file_status : file_open_status;
    variable mode_line   : line;
    variable status_line : line;
    variable report_line : line;
  begin
    case file_mode is
      when READ_MODE   => write(mode_line,string'("'Read Mode'"));
      when WRITE_MODE  => write(mode_line,string'("'Write Mode'"));
      when APPEND_MODE => write(mode_line,string'("'Append Mode'"));
      when others      => assert false report "Illegal file mode" severity failure;
    end case;

    file_open(file_status,file_id,file_name,file_mode);

    case file_status is
      when OPEN_OK      => write(status_line,string'("'File opened successfully'"));
      when STATUS_ERROR => write(status_line,string'("'File previously opened'"));
      when NAME_ERROR   => if (file_mode = WRITE_MODE) then
                             write(status_line,string'("'Cannot create file'"));
                           else
                             write(status_line,string'("'File does not exist'"));
                           end if;
      when MODE_ERROR   => write(status_line,string'("'File exists but cannot be opened in specific mode'"));
      when others       => write(status_line,string'("'Illegal file status'"));
    end case;

    write(report_line,string'("Opened file '" & file_name & "' in " & mode_line.all & ", Status: " & status_line.all));

    case file_status is
      when OPEN_OK      => assert false report report_line.all severity note;
      when STATUS_ERROR => assert false report report_line.all severity warning;
      when NAME_ERROR   => assert false report report_line.all severity failure;
      when MODE_ERROR   => assert false report report_line.all severity failure;
      when others       => assert false report report_line.all severity failure;
    end case;
  end;

  procedure close_file(
    file file_id : text)
  is
  begin
    file_close(file_id);
  end;

end ddr3_sdram_int_pkg;
