--==============================================================================
--
--     File Name: ddr3_sdram_pkg.vhd
--       Version: 0.1
--          Date: 28 Dec 2009
--         Model: Behavioral (NOT SYNTHESISABLE)
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Generic DDR3 SDRAM model
--
--   Description: Model package
--                Definition of part types and constants
--                Definition of part timing parameters
--
--   Limitations: Precharge Power Down and Active Power Down not supported
--                OCD Calibration not supported
--                Reads interrupted by reads not supported
--                Writes interrupted by writes not supported
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2009 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 10 Mar 09: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;


package ddr3_sdram_pkg is

  -- Record type for different part sizes
  type part_size_t is
    record
      sdram_bits : natural;
      data_width : natural range 4 to 32;
      byte_width : natural range 4 to 8;
      data_bytes : natural range 1 to 2;
      bank_width : natural range 3 to 3;
      arow_width : natural range 13 to 16;
      acol_width : natural range 10 to 11;
    end record;

  -- Array type for timing parameters which vary with speed grade, CL, and CWL
  subtype cl_index_s  is natural range 5 to 11;
  subtype cwl_index_s is natural range 5 to 8;
  type speed_grade_cl_cwl_t is array(cl_index_s,cwl_index_s) of time;


  -- Record type for timing parameters which vary with speed grade
  type speed_grade_t is
    record
      t_mrd_min      : natural;
      t_ccd_min      : natural;
      t_dllk_min     : natural;
      t_xs_min1      : natural;  t_xs_min2      : time;   -- greater of appliesgreater of applies (t_rfc_min1 or t_rfc_min2+t_rfc)
--      t_xsdll_min1   : natural;  t_xsdll_min2   : time;   -- greater of applies
      t_xp_min1      : natural;  t_xp_min2      : time;   -- greater of applies
      t_xpdll_min1   : natural;  t_xpdll_min2   : time;   -- greater of applies
      t_cke_min1     : natural;  t_cke_min2     : time;   -- Greater of applies
      t_mod_min1     : natural;  t_mod_min2     : time;   -- greater of applies
      t_zqinit_min   : natural;
      t_zqoper_min   : natural;
      t_zqcs_min     : natural;
      t_cl_min       : real;     t_cl_max       : real;
      t_ch_min       : real;     t_ch_max       : real;
      t_wpre_min     : real;
      t_dqsl_min     : real;
      t_dqsh_min     : real;
      t_dqss         : real;
      t_dss_min      : real;
      t_dsh_min      : real;
--      t_anpd_min     : natural;
--      t_axpd_min     : natural;
--      t_aofd         : real;
      t_rcd_min      : time;
      t_rc_min       : time;
      t_ras_min      : time;     t_ras_max      : time;
      t_rp_min       : time;
--      t_rpa_00_min   : time;     t_rpa_12_min   : time;
      t_rtp_min1     : natural;  t_rtp_min2     : time;   -- greater of applies
      t_wr_min       : time;
      t_wtr_min1     : natural;  t_wtr_min2     : time;   -- greater of applies
      t_faw_1k_min   : time;     t_faw_2k_min   : time;
      t_rrd_1k_min1  : natural;  t_rrd_1k_min2  : time;   -- greater of applies
      t_rrd_2k_min1  : natural;  t_rrd_2k_min2  : time;   -- greater of applies
--      t_rfc_256m_min : time;
--      t_rfc_512m_min : time;
      t_rfc_1g_min   : time; t_rfc_1g_max   : time;
      t_rfc_2g_min   : time; t_rfc_2g_max   : time;
      t_rfc_4g_min   : time; t_rfc_4g_max   : time;
      t_refi_max     : time;
      t_is_min       : time;
      t_ih_min       : time;
      t_ds_min       : time;
      t_dh_min       : time;
    end record;

  -- Record type for overall part used by generic model
  type part_t is
    record
      part_size              : part_size_t;
      speed_grade_cl_cwl_min : speed_grade_cl_cwl_t;
      speed_grade_cl_cwl_max : speed_grade_cl_cwl_t;
      speed_grade            : speed_grade_t;
      check_timing           : boolean;
    end record;


  -------------------------------------------------------
  -- Constants to define sizes of parts being modelled --
  -------------------------------------------------------

  -- Micron parts

  --  8Mb Array x 8 banks x 16 data bits = 1Gb part
  -- 16Mb Array x 8 banks x 8  data bits = 1Gb part
  -- 32Mb Array x 8 banks x 4  data bits = 1Gb part

  -- 8Mb Array x 8 banks x 16 data bits = 1Gb part
  constant M8_X_B8_X_D16 : part_size_t := (
    sdram_bits => 30,  -- 2**30 bits = 1Gb part
    data_width => 16,  -- 16-bit data
    byte_width => 8,   -- 8-bit data bytes (individually masked)
    data_bytes => 2,   -- 2 bytes in data
    bank_width => 3,   -- 2**3  = 8 banks
    arow_width => 13,  -- 2**13 = 8K row address (Min 13 = 8K)
    acol_width => 10); -- 2**10 = 1K col address (A10 not used in address)

  -- 64Mb Array x 8 banks x 4  data bits = 2Gb part
  constant M64_X_B8_X_D4 : part_size_t := (
    sdram_bits => 31,  -- 2**31 bits = 2Gb part
    data_width => 4,   -- 4-bit data
    byte_width => 4,   -- 4-bit data bytes (individually masked)
    data_bytes => 1,   -- 1 bytes in data
    bank_width => 3,   -- 2**3  =   8 banks
    arow_width => 15,  -- 2**15 = 32K row address (Min 13 = 8K)
    acol_width => 11); -- 2**11 =  2K col address (A10 not used in address)

  -- 32Mb Array x 8 banks x 8  data bits = 2Gb part

  -- 16Mb Array x 8 banks x 16 data bits = 2Gb part
  constant M16_X_B8_X_D16 : part_size_t := (
    sdram_bits => 31,  -- 2**31 bits = 2Gb part
    data_width => 16,  -- 16-bit data
    byte_width => 8,   -- 8-bit data bytes (individually masked)
    data_bytes => 2,   -- 2 bytes in data
    bank_width => 3,   -- 2**3  =   8 banks
    arow_width => 14,  -- 2**14 = 16K row address (Min 13 = 8K)
    acol_width => 10); -- 2**10 =  1K col address (A10 not used in address)


  --------------------------------------------------------------------------
  -- Constants to define clock timing for each CL of parts being modelled --
  --------------------------------------------------------------------------

  -- Micron parts

  constant MT41J_15E_CL_CWL_MIN : speed_grade_cl_cwl_t := (
  -- CWL5      CWL6      CWL7      CWL8
    (3.000 ns,     0 ns,     0 ns,     0 ns),  -- CL5
    (2.500 ns,     0 ns,     0 ns,     0 ns),  -- CL6
    (    0 ns, 1.875 ns,     0 ns,     0 ns),  -- CL7
    (    0 ns, 1.875 ns,     0 ns,     0 ns),  -- CL8
    (    0 ns,     0 ns, 1.500 ns,     0 ns),  -- CL9
    (    0 ns,     0 ns, 1.500 ns,     0 ns),  -- CL10
    (    0 ns,     0 ns,     0 ns,     0 ns)); -- CL11

  constant MT41J_15E_CL_CWL_MAX : speed_grade_cl_cwl_t := (
  -- CWL5      CWL6      CWL7      CWL8
    (3.300 ns,     0 ns,     0 ns,     0 ns),  -- CL5
    (3.300 ns,     0 ns,     0 ns,     0 ns),  -- CL6
    (    0 ns, 2.500 ns,     0 ns,     0 ns),  -- CL7
    (    0 ns, 2.500 ns,     0 ns,     0 ns),  -- CL8
    (    0 ns,     0 ns, 1.875 ns,     0 ns),  -- CL9
    (    0 ns,     0 ns, 1.875 ns,     0 ns),  -- CL10
    (    0 ns,     0 ns,     0 ns,     0 ns)); -- CL11


  constant MT41J_187E_CL_CWL_MIN : speed_grade_cl_cwl_t := (
  -- CWL5      CWL6      CWL7      CWL8
    (3.000 ns,     0 ns,     0 ns,     0 ns),  -- CL5
    (2.500 ns,     0 ns,     0 ns,     0 ns),  -- CL6
    (    0 ns, 1.875 ns,     0 ns,     0 ns),  -- CL7
    (    0 ns, 1.875 ns,     0 ns,     0 ns),  -- CL8
    (    0 ns,     0 ns,     0 ns,     0 ns),  -- CL9
    (    0 ns,     0 ns,     0 ns,     0 ns),  -- CL10
    (    0 ns,     0 ns,     0 ns,     0 ns)); -- CL11

  constant MT41J_187E_CL_CWL_MAX : speed_grade_cl_cwl_t := (
  -- CWL5      CWL6      CWL7      CWL8
    (3.300 ns,     0 ns,     0 ns,     0 ns),  -- CL5
    (3.300 ns,     0 ns,     0 ns,     0 ns),  -- CL6
    (    0 ns, 2.500 ns,     0 ns,     0 ns),  -- CL7
    (    0 ns, 2.500 ns,     0 ns,     0 ns),  -- CL8
    (    0 ns,     0 ns,     0 ns,     0 ns),  -- CL9
    (    0 ns,     0 ns,     0 ns,     0 ns),  -- CL10
    (    0 ns,     0 ns,     0 ns,     0 ns)); -- CL11


  constant MT41J_25E_CL_CWL_MIN : speed_grade_cl_cwl_t := (
  -- CWL5      CWL6      CWL7      CWL8
    (2.500 ns,     0 ns,     0 ns,     0 ns),  -- CL5
    (2.500 ns,     0 ns,     0 ns,     0 ns),  -- CL6
    (    0 ns,     0 ns,     0 ns,     0 ns),  -- CL7
    (    0 ns,     0 ns,     0 ns,     0 ns),  -- CL8
    (    0 ns,     0 ns,     0 ns,     0 ns),  -- CL9
    (    0 ns,     0 ns,     0 ns,     0 ns),  -- CL10
    (    0 ns,     0 ns,     0 ns,     0 ns)); -- CL11

  constant MT41J_25E_CL_CWL_MAX : speed_grade_cl_cwl_t := (
  -- CWL5      CWL6      CWL7      CWL8
    (3.300 ns,     0 ns,     0 ns,     0 ns),  -- CL5
    (3.300 ns,     0 ns,     0 ns,     0 ns),  -- CL6
    (    0 ns,     0 ns,     0 ns,     0 ns),  -- CL7
    (    0 ns,     0 ns,     0 ns,     0 ns),  -- CL8
    (    0 ns,     0 ns,     0 ns,     0 ns),  -- CL9
    (    0 ns,     0 ns,     0 ns,     0 ns),  -- CL10
    (    0 ns,     0 ns,     0 ns,     0 ns)); -- CL11



  --------------------------------------------------------------
  -- Constants to define speed grades of parts being modelled --
  --------------------------------------------------------------

  -- Micron parts

  constant MT41J_15E : speed_grade_t := (
    t_mrd_min      =>     4,
    t_ccd_min      =>     4,
    t_dllk_min     =>   512,
    t_xs_min1      =>     5,   t_xs_min2      =>  10 ns,   -- greater of applies (t_rfc_min1 or t_rfc_min2+t_rfc)
--    t_xsdll_min1   =>     5,   t_xsdll_min2   =>  10 ns,   -- greater of applies
    t_xp_min1      =>     3,   t_xp_min2      => 7.5 ns,   -- greater of applies
    t_xpdll_min1   =>    10,   t_xpdll_min2   =>  24 ns,   -- greater of applies
    t_cke_min1     =>     3,   t_cke_min2     => 7.5 ns,   -- greater of applies
    t_mod_min1     =>    12,   t_mod_min2     =>  15 ns,   -- greater of applies
    t_zqinit_min   =>   512,
    t_zqoper_min   =>   256,
    t_zqcs_min     =>    64,
    t_cl_min       =>  0.47,   t_cl_max       => 0.53,
    t_ch_min       =>  0.47,   t_ch_max       => 0.53,
    t_wpre_min     =>  0.90,
    t_dqsl_min     =>  0.45,
    t_dqsh_min     =>  0.45,
    t_dqss         =>  0.25,                               -- dqs rising to ck rising +- value
    t_dss_min      =>  0.20,                               -- dqs falling to ck rising (Setup)
    t_dsh_min      =>  0.20,                               -- dqs falling from ck rising (Hold)
--    t_anpd_min     =>     3,
--    t_axpd_min     =>     8,
--    t_aofd         =>   2.5,
    t_rcd_min      =>  13.5 ns,
    t_rc_min       =>  49.5 ns,
    t_ras_min      =>  36.0 ns, t_ras_max      => 70200 ns,
    t_rp_min       =>  13.5 ns,
--    t_rpa_00_min   =>  15.0 ns, t_rpa_12_min   =>  18.75 ns, -- 256Mb/512Mb, 1Gb/2Gb parts
    t_rtp_min1     =>        4, t_rtp_min2     =>   7.5 ns,
    t_wr_min       =>  15.0 ns,
    t_wtr_min1     =>        4, t_wtr_min2     =>   7.5 ns,    -- greater of applies
    t_faw_1k_min   =>  40.0 ns, t_faw_2k_min   =>  50.0 ns,    -- 1KB, 2KB page size
    t_rrd_1k_min1  =>        4, t_rrd_1k_min2  =>  10.0 ns,    -- 1KB page size, greater of applies
    t_rrd_2k_min1  =>        4, t_rrd_2k_min2  =>  10.0 ns,    -- 2KB page size, greater of applies
--    t_rfc_256m_min =>  75.0 ns,                               -- 256Mb parts
--    t_rfc_512m_min => 125.0 ns,                               -- 512Mb parts
    t_rfc_1g_min   => 110.0 ns, t_rfc_1g_max   => 70200 ns,    -- 1Gb parts
    t_rfc_2g_min   => 160.0 ns, t_rfc_2g_max   => 70200 ns,    -- 2Gb parts
    t_rfc_4g_min   => 300.0 ns, t_rfc_4g_max   => 70200 ns,    -- 4Gb parts
    t_refi_max     => 64 ms/8192,                              -- 7.8125 us,
    t_is_min       => 200.0 ps,
    t_ih_min       => 275.0 ps,
    t_ds_min       =>  75.0 ps,
    t_dh_min       => 150.0 ps);

  constant MT41J_187E : speed_grade_t := (
    t_mrd_min      =>     4,
    t_ccd_min      =>     4,
    t_dllk_min     =>   512,
    t_xs_min1      =>     5,   t_xs_min2      => 10 ns,      -- greater of applies (t_rfc_min1 or t_rfc_min2+t_rfc)
--     t_xsdll_min1   =>     5,   t_xsdll_min2   =>  10 ns,     -- greater of applies
    t_xp_min1      =>     3,   t_xp_min2      =>  7.500 ns,  -- greater of applies
    t_xpdll_min1   =>    10,   t_xpdll_min2   => 24 ns,      -- greater of applies
    t_cke_min1     =>     3,   t_cke_min2     =>  5.625 ns,  -- greater of applies
    t_mod_min1     =>    12,   t_mod_min2     => 15 ns,      -- greater of applies
    t_zqinit_min   =>   512,
    t_zqoper_min   =>   256,
    t_zqcs_min     =>    64,
    t_cl_min       =>  0.47,   t_cl_max       => 0.53,
    t_ch_min       =>  0.47,   t_ch_max       => 0.53,
    t_wpre_min     =>  0.90,
    t_dqsl_min     =>  0.45,
    t_dqsh_min     =>  0.45,
    t_dqss         =>  0.25,                               -- dqs rising to ck rising +- value
    t_dss_min      =>  0.20,                               -- dqs falling to ck rising (Setup)
    t_dsh_min      =>  0.20,                               -- dqs falling from ck rising (Hold)
--    t_anpd_min     =>     3,
--    t_axpd_min     =>     8,
--    t_aofd         =>   2.5,
    t_rcd_min      =>  13.125 ns,
    t_rc_min       =>  50.625 ns,
    t_ras_min      =>  37.500 ns, t_ras_max      => 70200 ns,
    t_rp_min       =>  13.125 ns,
--    t_rpa_00_min   =>  15.0 ns, t_rpa_12_min   =>  18.75 ns, -- 256Mb/512Mb, 1Gb/2Gb parts
    t_rtp_min1     =>        4, t_rtp_min2     =>   7.5 ns,
    t_wr_min       =>  15.0 ns,
    t_wtr_min1     =>        4, t_wtr_min2     =>   7.5 ns,    -- greater of applies
    t_faw_1k_min   =>  37.5 ns, t_faw_2k_min   =>  50.0 ns,    -- 1KB, 2KB page size
    t_rrd_1k_min1  =>        4, t_rrd_1k_min2  =>   7.5 ns,    -- 1KB page size, greater of applies
    t_rrd_2k_min1  =>        4, t_rrd_2k_min2  =>  10.0 ns,    -- 2KB page size, greater of applies
--    t_rfc_256m_min =>  75.0 ns,                               -- 256Mb parts
--    t_rfc_512m_min => 125.0 ns,                               -- 512Mb parts
    t_rfc_1g_min   => 110.0 ns, t_rfc_1g_max   => 70200 ns,    -- 1Gb parts
    t_rfc_2g_min   => 160.0 ns, t_rfc_2g_max   => 70200 ns,    -- 2Gb parts
    t_rfc_4g_min   => 300.0 ns, t_rfc_4g_max   => 70200 ns,    -- 4Gb parts
    t_refi_max     => 64 ms/8192,                              -- 7.8125 us,
    t_is_min       => 125.0 ps,
    t_ih_min       => 200.0 ps,
    t_ds_min       =>  25.0 ps,
    t_dh_min       => 100.0 ps);

  constant MT41J_25E : speed_grade_t := (
    t_mrd_min      =>     4,
    t_ccd_min      =>     4,
    t_dllk_min     =>   512,
    t_xs_min1      =>     5,   t_xs_min2      =>  10 ns,   -- greater of applies (t_rfc_min1 or t_rfc_min2+t_rfc)
--    t_xsdll_min1   =>     5,   t_xsdll_min2   =>  10 ns,   -- greater of applies
    t_xp_min1      =>     3,   t_xp_min2      => 7.5 ns,   -- greater of applies
    t_xpdll_min1   =>    10,   t_xpdll_min2   =>  24 ns,   -- greater of applies
    t_cke_min1     =>     3,   t_cke_min2     => 7.5 ns,   -- greater of applies
    t_mod_min1     =>    12,   t_mod_min2     =>  15 ns,   -- greater of applies
    t_zqinit_min   =>   512,
    t_zqoper_min   =>   256,
    t_zqcs_min     =>    64,
    t_cl_min       =>  0.47,   t_cl_max       => 0.53,
    t_ch_min       =>  0.47,   t_ch_max       => 0.53,
    t_wpre_min     =>  0.90,
    t_dqsl_min     =>  0.45,
    t_dqsh_min     =>  0.45,
    t_dqss         =>  0.25,                               -- dqs rising to ck rising +- value
    t_dss_min      =>  0.20,                               -- dqs falling to ck rising (Setup)
    t_dsh_min      =>  0.20,                               -- dqs falling from ck rising (Hold)
--    t_anpd_min     =>     3,
--    t_axpd_min     =>     8,
--    t_aofd         =>   2.5,
    t_rcd_min      =>  12.5 ns,
    t_rc_min       =>  50.0 ns,
    t_ras_min      =>  37.5 ns, t_ras_max      => 70200 ns,
    t_rp_min       =>  12.5 ns,
--    t_rpa_00_min   =>  15.0 ns, t_rpa_12_min   =>  18.75 ns, -- 256Mb/512Mb, 1Gb/2Gb parts
    t_rtp_min1     =>        4, t_rtp_min2     =>   7.5 ns,
    t_wr_min       =>  15.0 ns,
    t_wtr_min1     =>        4, t_wtr_min2     =>   7.5 ns,    -- greater of applies
    t_faw_1k_min   =>  40.0 ns, t_faw_2k_min   =>  50.0 ns,    -- 1KB, 2KB page size
    t_rrd_1k_min1  =>        4, t_rrd_1k_min2  =>  10.0 ns,    -- 1KB page size, greater of applies
    t_rrd_2k_min1  =>        4, t_rrd_2k_min2  =>  10.0 ns,    -- 2KB page size, greater of applies
--    t_rfc_256m_min =>  75.0 ns,                               -- 256Mb parts
--    t_rfc_512m_min => 125.0 ns,                               -- 512Mb parts
    t_rfc_1g_min   => 110.0 ns, t_rfc_1g_max   => 70200 ns,    -- 1Gb parts
    t_rfc_2g_min   => 160.0 ns, t_rfc_2g_max   => 70200 ns,    -- 2Gb parts
    t_rfc_4g_min   => 300.0 ns, t_rfc_4g_max   => 70200 ns,    -- 4Gb parts
    t_refi_max     => 64 ms/8192,                              -- 7.8125 us,
    t_is_min       => 200.0 ps,
    t_ih_min       => 275.0 ps,
    t_ds_min       =>  75.0 ps,
    t_dh_min       => 150.0 ps);




  ----------------------------------------------
  -- Constants to define parts being modelled --
  ----------------------------------------------

  -- Micron parts

  constant MT41J64M16_15E : part_t := (
    part_size              => M8_X_B8_X_D16,         -- 1Gb part
    speed_grade_cl_cwl_min => MT41J_15E_CL_CWL_MIN,  -- 15E with CL=5,6,7,8,9,10
    speed_grade_cl_cwl_max => MT41J_15E_CL_CWL_MAX,  -- 15E with CL=5,6,7,8,9,10
    speed_grade            => MT41J_15E,             -- 15E
    check_timing           => true);

  constant MT41J64M16_187E : part_t := (
    part_size              => M8_X_B8_X_D16,         -- 1Gb part
    speed_grade_cl_cwl_min => MT41J_187E_CL_CWL_MIN, -- 187E with CL=5,6,7,8
    speed_grade_cl_cwl_max => MT41J_187E_CL_CWL_MAX, -- 187E with CL=5,6,7,8
    speed_grade            => MT41J_187E,            -- 187E
    check_timing           => true);

  constant MT41J64M16_25E : part_t := (
    part_size              => M8_X_B8_X_D16,         -- 1Gb part
    speed_grade_cl_cwl_min => MT41J_25E_CL_CWL_MIN,  -- 25E with CL=5,6
    speed_grade_cl_cwl_max => MT41J_25E_CL_CWL_MAX,  -- 25E with CL=5,6
    speed_grade            => MT41J_25E,             -- 25E
    check_timing           => true);

--  constant MT47H64M16_3E : part_t := (
--    part_size      => M8_X_B8_X_D16,     -- 1Gb part
--    speed_grade_cl => MT47H_3E_CL,       -- 3E with CL=3,4,5
--    speed_grade    => MT47H_3E,          -- 3E
--    check_timing   => true);

--  constant MT47H512M4_3E : part_t := (
--    part_size      => M64_X_B8_X_D4,     -- 2Gb part
--    speed_grade_cl => MT47H_3E_CL,       -- 3E with CL=3,4,5
--    speed_grade    => MT47H_3E,          -- 3E
--    check_timing   => true);

  constant MT47J128M16_187E : part_t := (
    part_size              => M16_X_B8_X_D16,        -- 2Gb part
    speed_grade_cl_cwl_min => MT41J_187E_CL_CWL_MIN, -- 187E with CL=5,6,7,8
    speed_grade_cl_cwl_max => MT41J_187E_CL_CWL_MAX, -- 187E with CL=5,6,7,8
    speed_grade            => MT41J_187E,            -- 187E
    check_timing           => true);


  ----------------------------
  -- Component Declarations --
  ----------------------------

  component ddr3_sdram
    generic(
      message_level  : natural range 0 to 7 := 0;
      part           : part_t;
      short_init_dly : boolean);
    port(
      ck            : in    std_logic;
      ck_l          : in    std_logic;
      reset_l       : in    std_logic;
      cke           : in    std_logic;
      cs_l          : in    std_logic;
      ras_l         : in    std_logic;
      cas_l         : in    std_logic;
      we_l          : in    std_logic;
      odt           : in    std_logic;
      dm            : in    std_logic_vector(part.part_size.data_bytes-1 downto 0);
      ba            : in    std_logic_vector(part.part_size.bank_width-1 downto 0);
      a             : in    std_logic_vector(part.part_size.arow_width-1 downto 0);
      dq            : inout std_logic_vector(part.part_size.data_width-1 downto 0);
      dqs           : inout std_logic_vector(part.part_size.data_bytes-1 downto 0);
      dqs_l         : inout std_logic_vector(part.part_size.data_bytes-1 downto 0);
      init_start    : in    boolean := false;
      init_filename : in    string  := "init.txt";
      log_start     : in    boolean := false;
      log_filename  : in    string  := "log.txt");
  end component;

end ddr3_sdram_pkg;


package body ddr3_sdram_pkg is

end ddr3_sdram_pkg;