--==============================================================================
--
--     File Name: mptl_if_target_sim.vhd
--       Version: 0.1
--          Date: 23 Mar 2010
--         Model: sim (Not Synthesisable)
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Target MPTL if for simulation only
--
--   Description: Simulation version of MPTL to OCP interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 23 Mar 10: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

library work;
use work.adb3_ocp.all;
use work.adb3_target_types_pkg.all;
use work.adb3_target_inc_pkg.all;
use work.adb3_target_pkg.all;

entity mptl_if_target_sim is
port (
  -- OCP Interface
  ocp_clk                  : in    std_logic;
  direct_slave_m2s         : out   adb3_ocp_m2sT;
  direct_slave_s2m         : in    adb3_ocp_s2mT;
  dma_channels_m2s         : out   adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
  dma_channels_s2m         : in    adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
  direct_masters_m2s       : in    adb3_ocp_m2sT_array(0 to DM_CHANNELS-1);
  direct_masters_s2m       : out   adb3_ocp_s2mT_array(0 to DM_CHANNELS-1);
  dma_abort                : in    std_logic_vector(DMA_CHANNELS-1 downto 0);
  -- MPTL Interface
  mptl_b2t                 : in    mptl_pins_t;
  mptl_t2b                 : out   mptl_pins_t;
  mptl_clk                 : in    std_logic_dbl_t;
  mptl_clk_out             : out   std_logic;
  -- MPTL Initialisation
  ocp_ready                : in    std_logic;
  mptl_sb_b2t              : in    mptl_sb_b2t_t;
  mptl_sb_t2b              : out   mptl_sb_t2b_t;
  -- MPTL Sideband
  gpio_t2b                 : in    std_logic_vector(15 downto 0);
  gpio_b2t                 : out   std_logic_vector(15 downto 0));
end;


architecture sim of mptl_if_target_sim is

  signal online_count : std_logic_vector(5 downto 0) := "111111";
  signal online_cnt_0 : boolean := false;

  signal target_configured_l : std_logic;
  signal target_gtp_online_l : std_logic;
  signal bridge_gtp_online_l : std_logic;

  signal mptl_clk_div2 : std_logic := '0';

begin

  -- Hold off MPTL target configured until all OCP clocks are stable (resets inactive)
  mptl_sb_t2b.mptl_target_configured_l <= target_configured_l or not(ocp_ready);

  -- Connect MPTL initialisation signals
  mptl_sb_t2b.mptl_target_gtp_online_l <= target_gtp_online_l;
  bridge_gtp_online_l <= mptl_sb_b2t.mptl_bridge_gtp_online_l;


  -- Simulation interface inputs
  direct_slave_m2s   <= mptl_b2t.direct_slave_m2s;
  dma_channels_m2s   <= mptl_b2t.dma_channels_m2s;
  direct_masters_s2m <= mptl_b2t.direct_masters_s2m;
  gpio_b2t           <= mptl_b2t.gpio_b2t;

  -- Simulation interface outputs
  mptl_t2b.target_ocp_clk     <= ocp_clk;
  mptl_t2b.direct_slave_s2m   <= direct_slave_s2m;
  mptl_t2b.dma_channels_s2m   <= dma_channels_s2m;
  mptl_t2b.direct_masters_m2s <= direct_masters_m2s;
  mptl_t2b.gpio_t2b           <= gpio_t2b;

  -- Configuration outputs
  target_configured_l <= '0';

  -- Online outputs
  online_p : process(ocp_clk)
  begin
    if ocp_clk'event and ocp_clk = '1' then
      if online_count = "000000" then
        online_cnt_0 <= true;
      else
        online_count <= online_count-1;
        online_cnt_0 <= false;
      end if;

      if (online_cnt_0) then
        target_gtp_online_l <= bridge_gtp_online_l;
      else
        target_gtp_online_l <= '1';
      end if;
    end if;
  end process;

  -- MPTL clock div 2
  mptl_clk_out_p : process(mptl_clk.p)
  begin
    if rising_edge(mptl_clk.p) then
      mptl_clk_div2 <= not(mptl_clk_div2);
    end if;
  end process;

  -- 6T1 specific outputs
  -- dma_abort output uses MPTL
  -- mptl_clk_out = half mptl_clk frequency
  inst_6t1_g : if (BOARD_TYPE = ADM_XRC_6T1) generate
    mptl_t2b.dma_abort <= dma_abort;
    mptl_clk_out <= mptl_clk_div2;
  end generate;

  -- 6TL specific outputs
  -- dma_abort output directly on pins
  -- mptl_clk_out = mptl_clk
  inst_6tl_g : if (BOARD_TYPE = ADM_XRC_6TL) generate
    mptl_sb_t2b.mptl_dma_abort_l <= not(dma_abort);
    mptl_clk_out <= mptl_clk.p;
  end generate;

end;