--==============================================================================
--
--     File Name: blk_clock_freq.vhd
--       Version: 0.1
--          Date: 17 Feb 2011
--         Model: RTL (SYNTHESISABLE)
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Clock Frequency measurement
--
--   Description: Measures the number of rising edges on 'smp_clk' during
--                a measurement period that is 'ref_clk_tcval' cycles
--                of 'ref_clk'.
--
--   Limitations: Accuracy is limited mainly by
--
--                  (a) Accuracy of 'ref_clk', and
--
--                  (b) The number of division stages for 'smp_clk', as
--                      specified by the 'smp_clk_div_stages' generic,
--                      which causes the low 'smp_clk_div_stages' bits of
--                      'count' to hold no useful information.
--
--                Accuracy of 'ref_clk' is usually the dominant source of
--                inaccuracy (e.g. 25 ppm crystal oscillator).
--
--                In order for aliasing of the sampled clock to be avoided,
--                which would result in incorrect measurements, the following
--                relationship must hold between the frequencies of 'ref_clk'
--                and 'smp_clk':
--
--                  ref_clk_freq > smp_clk_freq*2 / (2**smp_clk_div_stages)
--
--         Notes: 1. It is legal to tie 'do' to 1, or to tie 'idle' to 'do',
--                   in order to have this module perform continual
--                   measurements.
--
--                2. The clocks 'read_clk', 'ref_clk' and 'smp_clk' need not
--                   be related in any way, other than the above frequency
--                   relationship.
--
--                3. The 'count' output always indicates the elapsed number
--                   of cycles of 'smp_clk' and does not need to be adjusted
--                   to take account of 'smp_clk_div_stages'.
--                   However, the low 'smp_clk_div_stages' bits of 'count'
--                   are always 0, and represent uncertainty introduced by
--                   dividing 'smp_clk' by 2**smp_clk_div_stages.
--
--                4. 'done' pulses for 1 cycle when a measurement is
--                   completed, where 'idle' is asserted when no measurement
--                   is in progress. 'idle' is guaranteed to be asserted in
--                   the same cycle as a 'done' pulse.
--
--                5. 'valid' transitions to one in the same cycle as the
--                   first 'done' pulse after reset. Only asserting 'rst'
--                   causes 'valid' to be deasserted.
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, TW, 17 Feb 2011: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

entity blk_clock_freq is
generic(
  ref_clk_tcval      : natural range 2 to 2147483647;   -- Measurement period, in 'ref_clk' cycles
  smp_clk_div_stages : natural range 0 to 30 := 0);     -- Number of ripple-divide stages for 'smp_clk'
port(
  rst             : in  std_logic;                      -- Asynchronous reset
  ref_clk         : in  std_logic;                      -- Reference clock
  smp_clk         : in  std_logic;                      -- Clock to be measured
  -- 'read_clk' domain
  read_clk        : in  std_logic;                      -- Output clock
  do              : in  std_logic;                      -- Pulse for at least 1 cycle to start a measurement
  count           : out std_logic_vector(31 downto 0);  -- Number of 'smp_clk' cycles counted, qualified by 'valid'
  running         : out std_logic;                      -- '1' => clock is running (i.e. count /= 0), qualified by 'valid'
  valid           : out std_logic;                      -- '1' => 'count' & 'running' are valid
  done            : out std_logic;                      -- '1' (pulsed) => measurement completed
  idle            : out std_logic);                     -- '1' => measurement not in progress
end entity;

architecture syn of blk_clock_freq is

  -- Return log2 of input value, rounded up to integer value
  function log2up(
    constant n : in natural)
  return natural is
  begin
    assert n /= 0 severity failure;
    for i in 0 to 30 loop 
      if n <= 2**i then
        return i;
      end if;
    end loop;
    assert false severity failure;
    return 0;
  end;

  constant REF_CLK_CNT_TCVAL_ORDER : natural := log2up(ref_clk_tcval - 1);
  constant REF_CLK_CNT_TCVAL : std_logic_vector(REF_CLK_CNT_TCVAL_ORDER - 1 downto 0) := std_logic_vector(to_unsigned(ref_clk_tcval - 1, REF_CLK_CNT_TCVAL_ORDER));

  attribute TIG : string; 

  -- Local reset ('ref_clk' domain)
  signal r_sr0 : std_logic;
  signal r_sr : std_logic;

  -- Ripple divider for 'smp_clk', where 'smp_clk_div(0)' is 'smp_clk'
  signal smp_clk_div : std_logic_vector(smp_clk_div_stages downto 0) := (others => '0');
  -- Put a TIG attribute on 'smp_clk_div', because only the output of the final flip-flop
  -- in the ripple chain is sampled. It is not sensitive to skew.
  attribute TIG of smp_clk_div : signal is "yes";

  -- Ripple-divided version of 'smp_clk'
  signal smp_clk_sel : std_logic;
  -- 'smp_clk_sel' sampled by 'ref_clk'
  signal smp_clk_sel_q : std_logic_vector(3 downto 0);

  -- Reference clock cycle counter ('ref_clk' domain)
  signal ref_clk_cnt : std_logic_vector(REF_CLK_CNT_TCVAL_ORDER - 1 downto 0);
  signal ref_clk_cnt_tc : std_logic;

  -- Counter of low -> high transitions of 'smp_clk_sel_q(3)' ('ref_clk' domain)
  signal smp_edge_cnt : std_logic_vector(REF_CLK_CNT_TCVAL_ORDER - 1 downto 0);

  -- Frequency measurement state machine ('ref_clk' domain)
  type r_state_t is (r_s_idle, r_s_running, r_s_transfer);
  signal r_state, n_r_state : r_state_t;
  signal r_cnt_sr, n_r_cnt_sr : std_logic;
  signal r_cnt_ce, n_r_cnt_ce : std_logic;
  signal r_done, n_r_done : std_logic;
  signal r_do, r_do0 : std_logic;

  -- Local reset ('read_clk' domain)
  signal e_sr0 : std_logic;
  signal e_sr : std_logic;

  -- Executive state machine ('read_clk' domain)
  type e_state_t is (e_s_idle, e_s_running, e_s_transfer, e_s_ack, e_s_pause);
  signal e_state, n_e_state : e_state_t;
  signal e_do, n_e_do : std_logic;
  signal e_ce, n_e_ce : std_logic;
  signal e_done, e_done0 : std_logic;
  signal i_idle, n_idle : std_logic;
  signal i_done, n_done : std_logic;
  signal i_valid, n_valid : std_logic;

  -- Registers for capturing measurement from 'ref_clk' domain ('read_clk' domain)
  signal i_count : std_logic_vector(31 downto 0);
  signal i_running : std_logic;

begin

  done <= i_done;
  idle <= i_idle;
  valid <= i_valid;
  count <= i_count;
  running <= i_running;

  -- ------------------------------------------------------------------------
  -- Ripple divider for 'smp_clk'
  -- ------------------------------------------------------------------------

  smp_clk_div(0) <= smp_clk;

  -- Implement ripple divider for 'smp_clk'
  gen_ripple_divider : for i in 1 to smp_clk_div_stages generate
    U0 : process(smp_clk_div)
    begin
      if rising_edge(smp_clk_div(i - 1)) then
        smp_clk_div(i) <= not smp_clk_div(i);
      end if;
    end process;
  end generate;

  -- ------------------------------------------------------------------------
  -- 'ref_clk' domain
  -- ------------------------------------------------------------------------

  -- Generate local reset of 'ref_clk' domain
  local_reset_r : process(rst, ref_clk)
  begin
    if rst = '1' then
      r_sr0 <= '1';
      r_sr <= '1';
    elsif rising_edge(ref_clk) then
      r_sr0 <= '0';
      r_sr <= r_sr0;
    end if;
  end process;

  -- Count cycles of 'ref_clk'
  gen_ref_clk_cnt : process(ref_clk)
  begin
    if rising_edge(ref_clk) then
      if r_cnt_sr = '1' then
        ref_clk_cnt <= (others => '0');
      else
        ref_clk_cnt <= ref_clk_cnt + 1;
      end if;
    end if;
  end process;

  smp_clk_sel <= smp_clk_div(smp_clk_div_stages);

  -- Count rising edges of 'smp_clk_sel'
  gen_smp_edge_cnt : process(ref_clk)
  begin
    if rising_edge(ref_clk) then
      smp_clk_sel_q <= smp_clk_sel_q(2 downto 0) & smp_clk_sel;
      -- Count rising edges of 'smp_clk_sel'
      if r_cnt_sr = '1' then
        smp_edge_cnt <= (others => '0');
      else
        if r_cnt_ce = '1' and smp_clk_sel_q(3) = '1' and smp_clk_sel_q(2) = '0' then
          smp_edge_cnt <= smp_edge_cnt + 1;
        end if;
      end if;
    end if;
  end process;

  ref_clk_cnt_tc <= '1' when ref_clk_cnt = REF_CLK_CNT_TCVAL else '0';

  -- Synchronise handshake from 'read_clk' domain
  gen_r_do : process(ref_clk)
  begin
    if rising_edge(ref_clk) then
      if r_sr = '1' then
        r_do0 <= '0';
        r_do <= '0';
      else
        r_do0 <= e_do;
        r_do <= r_do0;
      end if;
    end if;
  end process;

  -- Frequency measurement state machine
  gen_r_state : process(ref_clk)
  begin
    if rising_edge(ref_clk) then
      if r_sr = '1' then
        r_state <= r_s_idle;
        r_cnt_sr <= '1';
        r_cnt_ce <= '-';
        r_done <= '0';
      else
        r_state <= n_r_state;
        r_cnt_sr <= n_r_cnt_sr;
        r_cnt_ce <= n_r_cnt_ce;
        r_done <= n_r_done;
      end if;
    end if;
  end process;

  gen_n_r_state : process(
    r_state,
    r_do,
    ref_clk_cnt_tc)
  begin
    case r_state is
      when r_s_idle =>
        if r_do = '1' then
          n_r_state <= r_s_running;
          n_r_cnt_sr <= '0';
          n_r_cnt_ce <= '1';
          n_r_done <= '0';
        else
          n_r_state <= r_s_idle;
          n_r_cnt_sr <= '1';
          n_r_cnt_ce <= '-';
          n_r_done <= '0';
        end if;

      when r_s_running =>
        if ref_clk_cnt_tc = '1' then
          n_r_state <= r_s_transfer;
          n_r_cnt_sr <= '0';
          n_r_cnt_ce <= '0';
          n_r_done <= '1';
        else
          n_r_state <= r_s_running;
          n_r_cnt_sr <= '0';
          n_r_cnt_ce <= '1';
          n_r_done <= '0';
        end if;

      when r_s_transfer =>
        if r_do = '0' then
          n_r_state <= r_s_idle;
          n_r_cnt_sr <= '1';
          n_r_cnt_ce <= '-';
          n_r_done <= '0';
        else
          n_r_state <= r_s_transfer;
          n_r_cnt_sr <= '0';
          n_r_cnt_ce <= '0';
          n_r_done <= '1';
        end if;
    end case;
  end process;

  -- ------------------------------------------------------------------------
  -- 'read_clk' domain
  -- ------------------------------------------------------------------------

  -- Generate local reset for 'read_clk' domain
  local_reset_e : process(rst, read_clk)
  begin
    if rst = '1' then
      e_sr0 <= '1';
      e_sr <= '1';
    elsif rising_edge(read_clk) then
      e_sr0 <= '0';
      e_sr <= e_sr0;
    end if;
  end process;

  -- Transfer frequency measurement from 'ref_clk' domain into 'read_clk' domain
  gen_count : process(read_clk)
    variable tmp : std_logic_vector(REF_CLK_CNT_TCVAL_ORDER + smp_clk_div_stages - 1 downto 0);
  begin
    if rising_edge(read_clk) then
      if e_ce = '1' then
        if smp_clk_div_stages > 0 then
          tmp(smp_clk_div_stages - 1 downto 0) := (others => '0');
        end if;
        tmp(REF_CLK_CNT_TCVAL_ORDER + smp_clk_div_stages - 1 downto smp_clk_div_stages) := smp_edge_cnt;
        i_count <= std_logic_vector(RESIZE(unsigned(tmp), 32));
      end if;
      i_running <= OR_reduce(i_count);
    end if;
  end process;

  -- Synchronise handshake from 'ref_clk' domain
  gen_e_done : process(read_clk)
  begin
    if rising_edge(read_clk) then
      if e_sr = '1' then
        e_done0 <= '0';
        e_done <= '0';
      else
        e_done0 <= r_done;
        e_done <= e_done0;
      end if;
    end if;
  end process;

  -- Executive state machine
  gen_e_state : process(read_clk)
  begin
    if rising_edge(read_clk) then
      if e_sr = '1' then
        e_state <= e_s_idle;
        e_do <= '0';
        e_ce <= '0';
        i_done <= '0';
        i_idle <= '1';
        i_valid <= '0';
      else
        e_state <= n_e_state;
        e_do <= n_e_do;
        e_ce <= n_e_ce;
        i_done <= n_done;
        i_idle <= n_idle;
        i_valid <= n_valid;
      end if;
    end if;
  end process;

  gen_n_e_state : process(
    e_state,
    e_done,
    do,
    i_valid)
  begin
    case e_state is
      when e_s_idle =>
        if do = '1' then
          n_e_state <= e_s_running;
          n_e_do <= '1';
          n_e_ce <= '0';
          n_done <= '0';
          n_idle <= '0';
          n_valid <= i_valid;
        else
          n_e_state <= e_s_idle;
          n_e_do <= '0';
          n_e_ce <= '0';
          n_done <= '0';
          n_idle <= '1';
          n_valid <= i_valid;
        end if;

      when e_s_running =>
        if e_done = '1' then
          n_e_state <= e_s_transfer;
          n_e_do <= '1';
          n_e_ce <= '0';
          n_done <= '0';
          n_idle <= '0';
          n_valid <= i_valid;
        else
          n_e_state <= e_s_running;
          n_e_do <= '1';
          n_e_ce <= '0';
          n_done <= '0';
          n_idle <= '0';
          n_valid <= i_valid;
        end if;

      when e_s_transfer =>
        n_e_state <= e_s_ack;
        n_e_do <= '1';
        n_e_ce <= '1';
        n_done <= '0';
        n_idle <= '0';
        n_valid <= i_valid;

      when e_s_ack =>
        if e_done = '0' then
          n_e_state <= e_s_pause;
          n_e_do <= '0';
          n_e_ce <= '0';
          n_done <= '0';
          n_idle <= '0';
          n_valid <= i_valid;
        else
          n_e_state <= e_s_ack;
          n_e_do <= '0';
          n_e_ce <= '0';
          n_done <= '0';
          n_idle <= '0';
          n_valid <= i_valid;
        end if;

      when e_s_pause =>
        n_e_state <= e_s_idle;
        n_e_do <= '0';
        n_e_ce <= '0';
        n_done <= '1';
        n_idle <= '1';
        n_valid <= '1';
    end case;
  end process;

end;
