--==============================================================================
--
--     File Name: blk_clocks.vhd
--       Version: 0.6
--          Date: 09 Mar 2011
--         Model: synth
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: Clock generation block
--
--   Description: Connection and generation of clocks
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 12 Mar 2010: Initial design
--                0.2, KDC, 08 Apr 2010: Changed PLL_BASE to MMCM_BASE
--                0.3, KDC, 16 Sep 2010: Added pll_mem_rst/pll_mem_clk
--                0.4, TW,  09 Feb 2011:
--                  - Changed ref_clk buffer to IBUFGDS and used buffered signal
--                    to clock the MMCM only.
--                  - Now distributes resets as asynchronous resets (individual
--                    blocks can still generate local synchronous resets from
--                    them, if necessary), to improve ease of timing closure.
--                  - pll_ref_clk is now called pll_pri_clk (primary clock), and
--                    pll_ref_clk is now a new 1x (guaranteed) version of of
--                    the input reference clock.
--                  - Removed ref_clk port.
--                0.5, KDC, 10 Feb 2011:
--                  - Changed XRM clock buffer to IBUFGDS
--                  - combined individual resets to single rst output
--                0.6, KDC, 09 Mar 2011:
--                  - Replaced board specific clock enable with single enable
--                    defined in uber_pkg which uses adb3_target_inc_pkg.
--
--==============================================================================


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

library work;
use work.adb3_ocp.all;
use work.adb3_target_types_pkg.all;
use work.adb3_target_inc_pkg.all;
use work.uber_pkg.all;

library unisim;
use unisim.vcomponents.all;

entity blk_clocks is
port(
  -- Clock pins
  clks_in        : in    clks_in_t;
  clks_out       : out   clks_out_t;
  clks_mgt_in    : in    clks_mgt_in_t;
  -- Reset output
  pll_rst        : out   std_logic;
  -- Clock outputs
  pll_pri_clk    : out   std_logic;
  pll_reg_clk    : out   std_logic;
  pll_mem_clk    : out   std_logic;
  pll_ref_clk    : out   std_logic;
  clk_vec        : out   clk_vec_t;
  mptl_clk       : out   std_logic_dbl_t);
end;


architecture synth of blk_clocks is

  -- Reference clock signals
  signal ref_clk_ibuf   : std_logic;
  signal ref_clk_bufg   : std_logic;

  -- PLL signals
  signal pll_rst_in_cnt : std_logic_vector(4 downto 0) := "11111";
  signal pll_rst_in     : std_logic := '1';
  signal fb_clk         : std_logic;
  signal fb_clk_bufg    : std_logic;
  signal pll_locked     : std_logic;
  signal out_clk0       : std_logic;
  signal out_clk1       : std_logic;
  signal out_clk2       : std_logic;
  signal out_clk3       : std_logic;
  signal out_clk0_bufg  : std_logic;
  signal out_clk1_bufg  : std_logic;
  signal out_clk2_bufg  : std_logic;

  -- Non MGT Clock signals
  signal xrm_clkin_ibuf : std_logic;

  -- MGT clock signals
  signal mgt_clk_pin    : mgt_clk_pin_t;
  signal mgt_clk_ibuf   : mgt_clk_buf_t;
  signal clk_vec_bufg   : clk_vec_t;


begin

  ----------------------------------
  -- Design reference clock input --
  ----------------------------------

  -- Use 200MHz oscillator global clock as design reference
  ref_clk_dif_i : IBUFGDS
  port map(
    I  => clks_in.ref_clk.p,
    IB => clks_in.ref_clk.n,
    O  => ref_clk_ibuf);


  -- Generate PLL reset using buffered reference clock
  ref_clk_bufg_i : BUFG
  port map(
    I => ref_clk_ibuf,
    O => ref_clk_bufg);

  pll_ref_clk <= ref_clk_bufg;

  pll_rst_in_p : process(ref_clk_bufg)
  begin
    if rising_edge(ref_clk_bufg) then
      if pll_rst_in_cnt = "00000" then
        pll_rst_in <= '0';
      else
        pll_rst_in <= '1';
        pll_rst_in_cnt <= pll_rst_in_cnt-1;
      end if;
    end if;
  end process;


  -- Reference clock MMCM (CLKFBOUT range 600 MHz to 1200 MHz)
  -- CLKFBOUT = (CLKIN1/DIVCLK_DIVIDE) * CLKFBOUT_MULT_F
  -- CLKOUTn  = (CLKIN1/DIVCLK_DIVIDE) * CLKFBOUT_MULT_F/CLKOUTn_DIVIDE
  -- CLKFBOUT = (200 MHz/3) * 12.000       = 800 MHz
  -- CLKOUT0  = (200 MHz/3) * 12.000/4.000 = 200 MHz (pll_pri_clk)
  -- CLKOUT1  = (200 MHz/3) * 12.000/10    =  80 MHz (pll_reg_clk)
  -- CLKOUT2  = (200 MHz/3) * 12.000/2     = 400 MHz (pll_mem_clk)

  mmcm_ref_clk_i : MMCM_BASE
  generic map(
    BANDWIDTH          => "OPTIMIZED", -- string := "OPTIMIZED"
    CLKIN1_PERIOD      => 5.000,       -- real := 0.0
    DIVCLK_DIVIDE      => 3,           -- integer := 1 (1 to 128)
    CLKFBOUT_MULT_F    => 12.000,      -- real := 1.0  (5.0 to 64.0)
    CLKOUT0_DIVIDE_F   => 4.000,       -- real := 1.0  (1.0 to 128.0)
    CLKOUT1_DIVIDE     => 10,          -- integer := 1
    CLKOUT2_DIVIDE     => 2,           -- integer := 1
    CLKOUT3_DIVIDE     => 1,           -- integer := 1
    CLKOUT4_DIVIDE     => 1,           -- integer := 1
    CLKOUT5_DIVIDE     => 1,           -- integer := 1
    CLKOUT6_DIVIDE     => 1,           -- integer := 1
    CLKFBOUT_PHASE     => 0.000,       -- real := 0.0
    CLKOUT0_PHASE      => 0.000,       -- real := 0.0
    CLKOUT1_PHASE      => 0.000,       -- real := 0.0
    CLKOUT2_PHASE      => 0.000,       -- real := 0.0
    CLKOUT3_PHASE      => 0.000,       -- real := 0.0
    CLKOUT4_PHASE      => 0.000,       -- real := 0.0
    CLKOUT5_PHASE      => 0.000,       -- real := 0.0
    CLKOUT6_PHASE      => 0.000,       -- real := 0.0
    CLKOUT0_DUTY_CYCLE => 0.500,       -- real := 0.5
    CLKOUT1_DUTY_CYCLE => 0.500,       -- real := 0.5
    CLKOUT2_DUTY_CYCLE => 0.500,       -- real := 0.5
    CLKOUT3_DUTY_CYCLE => 0.500,       -- real := 0.5
    CLKOUT4_DUTY_CYCLE => 0.500,       -- real := 0.5
    CLKOUT5_DUTY_CYCLE => 0.500,       -- real := 0.5
    CLKOUT6_DUTY_CYCLE => 0.500,       -- real := 0.5
    CLKOUT4_CASCADE    => FALSE,       -- boolean := FALSE
    CLOCK_HOLD         => FALSE,       -- boolean := FALSE
    REF_JITTER1        => 0.0,         -- real := 0.0
    STARTUP_WAIT       => FALSE)       -- boolean := FALSE
  port map(
    RST       => pll_rst_in,   -- in std_ulogic;
    PWRDWN    => '0',          -- in std_ulogic;
    CLKIN1    => ref_clk_ibuf, -- in std_ulogic;
    CLKFBIN   => fb_clk_bufg,  -- in std_ulogic;
    CLKFBOUT  => fb_clk,       -- out std_ulogic;
    CLKFBOUTB => open,         -- out std_ulogic;
    CLKOUT0   => out_clk0,     -- out std_ulogic;
    CLKOUT0B  => open,         -- out std_ulogic;
    CLKOUT1   => out_clk1,     -- out std_ulogic;
    CLKOUT1B  => open,         -- out std_ulogic;
    CLKOUT2   => out_clk2,     -- out std_ulogic;
    CLKOUT2B  => open,         -- out std_ulogic;
    CLKOUT3   => open,         -- out std_ulogic;
    CLKOUT3B  => open,         -- out std_ulogic;
    CLKOUT4   => open,         -- out std_ulogic;
    CLKOUT5   => open,         -- out std_ulogic;
    CLKOUT6   => open,         -- out std_ulogic;
    LOCKED    => pll_locked);  -- out std_ulogic;

  -- Generate asynchronous reset
  pll_rst <= not(pll_locked);

  -- MMCM feedback (not using BUFG, because we don't care about phase compensation)
  fb_clk_bufg <= fb_clk;

  -- MMCM output clock0 => pll_pri_clk main OCP clock
  out_clk0_bufg_i : BUFG
  port map(
    I => out_clk0,
    O => out_clk0_bufg);

  pll_pri_clk <= out_clk0_bufg;

  -- MMCM output clock1 => pll_reg_clk register interface clock
  out_clk1_bufg_i : BUFG
  port map(
    I => out_clk1,
    O => out_clk1_bufg);

  pll_reg_clk <= out_clk1_bufg;

  -- MMCM output clock2 => pll_mem_clk memory interface clock
  out_clk2_bufg_i : BUFG
  port map(
    I => out_clk2,
    O => out_clk2_bufg);

  pll_mem_clk <= out_clk2_bufg;

  -- Add more user clocks here
--  out_clk3_bufg_i : BUFG
--  port map(
--    I => out_clk3,
--    O => out_clk3_bufg);

--  pll_usr_clk <= out_clk3_bufg;


  ----------------
  -- MGT clocks --
  ----------------

  -- Connect clks_mgt_in to mgt_clk_pin vector
  mgt_clk_pin(MGT112_CLK0_NUM) <= clks_mgt_in.mgt112_clk0;
  mgt_clk_pin(MGT113_CLK0_NUM) <= clks_mgt_in.mgt113_clk0;
  mgt_clk_pin(MGT113_CLK1_NUM) <= clks_mgt_in.mgt113_clk1;
  mgt_clk_pin(MGT114_CLK0_NUM) <= clks_mgt_in.mgt114_clk0;
  mgt_clk_pin(MGT115_CLK0_NUM) <= clks_mgt_in.mgt115_clk0;
  mgt_clk_pin(MGT116_CLK0_NUM) <= clks_mgt_in.mgt116_clk0;
  mgt_clk_pin(MGT117_CLK0_NUM) <= clks_mgt_in.mgt117_clk0;


  mgt_clk_loop_g : for n in MGT110_CLK0_NUM to MGT118_CLK1_NUM generate
    ibufds_gtxe1_g : if (CLKS_IN_VALID(n) = '1') generate
      -- Extract mgt clks and buffer
      mgt_clk_ibufds_gtxe1_i : IBUFDS_GTXE1
      port map(
        I     => mgt_clk_pin(n).p,
        IB    => mgt_clk_pin(n).n,
        CEB   => '0',
        O     => mgt_clk_ibuf(n),
        ODIV2 => open);

      -- Buffer MGT clocks and connect to clk_vec_bufg vector
      mgt_clk_bufg_i : BUFG
      port map(
        I => mgt_clk_ibuf(n),
        O => clk_vec_bufg(n));
    end generate;
  end generate;


  --------------------------
  -- Non MGT clock inputs --
  --------------------------

  clk_vec_bufg_lclk_s_g : if (LCLK_VALID = '1') generate
    -- Buffer lclk_s and connect to clk_vec_bufg vector
    lclk_s_bufg_i : BUFG
    port map(
      I => clks_in.lclk,
      O => clk_vec_bufg(LCLK_NUM));
  end generate;

  clk_vec_bufg_xrm_g : if (XRM_GCLK_M2C_VALID = '1') generate
    -- Extract xrm_clkin global clock and buffer
    xrm_clk_dif_i : IBUFGDS
    port map(
      I  => clks_in.xrm_gclk_m2c.p,
      IB => clks_in.xrm_gclk_m2c.n,
      O  => xrm_clkin_ibuf);

    -- Buffer xrm_clkin and connect to clk_vec_bufg vector
    xrm_clk_bufg_i : BUFG
    port map(
      I => xrm_clkin_ibuf,
      O => clk_vec_bufg(XRM_CLKIN_NUM));
  end generate;

  clk_vec <= clk_vec_bufg;


  ---------------------------
  -- Non MGT clock outputs --
  ---------------------------

  xrm_mgtclk_c2m_g : if (XRM_MGTCLK_C2M_VALID = '1') generate
    obuftds_i : OBUFTDS
    generic map(
      IOSTANDARD => "DEFAULT")
    port map(
      O   => clks_out.xrm_mgtclk_c2m.p,
      OB  => clks_out.xrm_mgtclk_c2m.n,
      I   => out_clk1_bufg,
      T   => '0');
  end generate;


  ---------------------------
  -- MPTL clock generation --
  ---------------------------

  mptl_clk_6tl_g : if (BOARD_TYPE = ADM_XRC_6TL) generate
    -- MPTL Parallel interface: Connect 250MHz PCIe input to mptl_clk
    mptl_clk.p <= clks_mgt_in.mgt114_clk0.p;
    mptl_clk.n <= clks_mgt_in.mgt114_clk0.n;
  end generate;

  mptl_clk_6t1_g : if (BOARD_TYPE = ADM_XRC_6T1) generate
    -- MPTL serial interface: Connect 250 MHz Osc input to mptl_clk
    mptl_clk.p <= clks_mgt_in.mgt115_clk0.p;
    mptl_clk.n <= clks_mgt_in.mgt115_clk0.n;
  end generate;

end;