--==============================================================================
--
--     File Name: blk_direct_slave.vhd
--       Version: 0.3
--          Date: 10 Feb 2011
--         Model: synth
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: Direct Slave functional block
--                Functions:
--                  Direct Slave clock domain transfer
--                  Direct Slave address space split
--                  simple test block
--                  clock frequency measurement block
--                  io test block
--                  interrupt test block
--                  info register block
--                  BRAM interface clock domain transfer
--
--   Description: Adds OCP Direct Slave interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 15 Mar 2010: Initial design
--                0.2, TW, 09 Feb 2011:
--                  - Now uses adb3_ocp_comp package instead of adb3_target_pkg.
--                  - Got rid of some unnecessary resets.
--                  - pll_ref_clk is now called pll_pri_clk (primary clock).
--                0.3, KDC, 10 Feb 2011:
--                  - Added new pll_ref_clk measurement based on ref_clk input
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_comp.all;
use work.adb3_target_pkg.all;
use work.adb3_target_inc_pkg.all;
use work.mem_if_pkg.all;
use work.uber_pkg.all;

entity blk_direct_slave is
port(
  -- Reset input
  rst                 : in    std_logic;
  -- Clock inputs
  pll_pri_clk         : in    std_logic;
  pll_reg_clk         : in    std_logic;
  pll_ref_clk         : in    std_logic;
  pll_mem_clk         : in    std_logic;
  -- Direct slave primary OCP interface
  direct_slave_m2s    : in    adb3_ocp_m2sT;
  direct_slave_s2m    : out   adb3_ocp_s2mT;
  -- Clock frequency measurement register interface
  clk_vec             : in    clk_vec_t;
  -- GPIO signal test register interface
  gpio_inout          : inout gpio_inout_t;
  -- Interrupt test register interface
  interrupt_l         : out   std_logic;
  -- Memory interface status
  mem_if_stat         : in    mem_if_stat_array_t;
  mem_if_err          : in    mem_if_err_array_t;
  -- Memory application control
  mem_app_go          : out   mem_app_go_array_t;
  mem_app_offset      : out   mem_app_offset_array_t;
  mem_app_length      : out   mem_app_length_array_t;
  -- Memory application status
  mem_app_done        : in    mem_app_done_array_t;
  mem_app_err         : in    mem_app_err_array_t;
  mem_app_err_ph      : in    mem_app_err_ph_array_t;
  mem_app_err_addr    : in    mem_app_err_addr_array_t;
  -- BRAM secondary OCP interface
  ds_bram_m2s         : out   adb3_ocp_m2sT;
  ds_bram_s2m         : in    adb3_ocp_s2mT;
  -- RAM secondary OCP interface
  ds_ram_m2s          : out   adb3_ocp_m2sT;
  ds_ram_s2m          : in    adb3_ocp_s2mT;
  -- Debug
  bram_mux_err_info   : in    std_logic_vector(31 downto 0);
  ram_mux_err_info    : in    mem_if_debug_array_t;
  mem_if_err_info     : in    mem_if_debug_array_t);
end;


architecture synth of blk_direct_slave is

  -- Size of bank register (DS) to select between banks of memory
  constant BANK_ADDR_WIDTH : natural := 3;

  -- Define bank address signal
  signal ds_bank : std_logic_vector(BANK_ADDR_WIDTH-1 downto 0);

  -- Size of DS memory window BYTE address range = 2M bytes (21 bits)
  constant WIND_ADDR_WIDTH : natural := mask_vec_width(RAM_WIN_ADDR_RANGE.addr_mask);

  -- Size of DS PAGE register required to augment direct slave address bits [20:0] to DMA_ADDR_WIDTH - BANK_ADDR_WIDTH bits
  constant PAGE_ADDR_WIDTH_DDR3 : natural := DMA_ADDR_WIDTH - BANK_ADDR_WIDTH - WIND_ADDR_WIDTH;

  -- Define page address signal
  signal ds_page_ddr3 : std_logic_vector(PAGE_ADDR_WIDTH_DDR3-1 downto 0);

  -- Pad unused OCP address above DMA_ADDR_WIDTH with zeros
  constant OCP_ADDR_PAD : std_logic_vector((ADB3_OCP_ADDR_WIDTH-DMA_ADDR_WIDTH)-1 downto 0) := (others => '0');

  -- On-Board RAM Window Direct Slave OCP Interface
  signal ds_ram_mod_m2s : adb3_ocp_m2sT;
  signal ds_ram_mod_s2m : adb3_ocp_s2mT := ADB3_OCP_DEFAULT_S2M;

  -- Obtain number of slaves from DS address range table
  constant DS_SLAVES : natural := DS_ADDR_RANGE_TABLE'length;

  -- Decode address widths of register blocks
  constant SIMPLE_ADDR_WIDTH   : natural := mask_vec_width(SIMPLE_ADDR_RANGE.addr_mask);
  constant CLK_READ_ADDR_WIDTH : natural := mask_vec_width(CLK_READ_ADDR_RANGE.addr_mask);
  constant GPIO_ADDR_WIDTH     : natural := mask_vec_width(GPIO_ADDR_RANGE.addr_mask);
  constant INT_ADDR_WIDTH      : natural := mask_vec_width(INT_ADDR_RANGE.addr_mask);
  constant RAM_REG_ADDR_WIDTH  : natural := mask_vec_width(RAM_REG_ADDR_RANGE.addr_mask);
  constant INFO_ADDR_WIDTH     : natural := mask_vec_width(INFO_ADDR_RANGE.addr_mask);

  -- Direct slave OCP interface in pll_pri_clk domain
  signal ds_m2s        : adb3_ocp_m2sT;
  signal ds_s2m        : adb3_ocp_s2mT;

  -- Direct slave OCP interface in pll_reg_clk domain
  signal ds_reg_m2s    : adb3_ocp_m2sT;
  signal ds_reg_s2m    : adb3_ocp_s2mT;

  -- OCP split signals
  signal ds_slaves_m2s     : adb3_ocp_m2sT_array(0 to DS_SLAVES-1);
  signal ds_slaves_s2m     : adb3_ocp_s2mT_array(0 to DS_SLAVES-1);
  signal ds_split_err_info : std_logic_vector(31 downto 0);

begin

  ds_m2s <= direct_slave_m2s;
  direct_slave_s2m <= ds_s2m;

  -- Transfer direct slave OCP interface to pll_reg_clk domain
  ds_cross_clk_dom_i : adb3_ocp_cross_clk_dom
  port map(
    -- OCP primary port
    slave_rst  => rst,          -- in    std_logic
    slave_clk  => pll_pri_clk,  -- in    std_logic
    slave_m2s  => ds_m2s,       -- in    adb3_ocp_m2sT
    slave_s2m  => ds_s2m,       -- out   adb3_ocp_s2mT
    -- OCP secondary port
    master_rst => rst,          -- in    std_logic
    master_clk => pll_reg_clk,  -- in    std_logic
    master_m2s => ds_reg_m2s,   -- out   adb3_ocp_m2sT
    master_s2m => ds_reg_s2m);  -- in    adb3_ocp_s2mT


  -- Split direct slave OCP interface according to function
  adb3_ocp_split_b_i : adb3_ocp_split_b
  generic map(
    addr_range_table => DS_ADDR_RANGE_TABLE)  -- addr_range_table_t)
  port map(
    ocp_rst     => rst,                -- in    std_logic
    ocp_clk     => pll_reg_clk,        -- in    std_logic
    -- OCP primary port
    slave_m2s   => ds_reg_m2s,         -- in    adb3_ocp_m2sT
    slave_s2m   => ds_reg_s2m,         -- out   adb3_ocp_s2mT
    -- OCP secondary ports
    masters_m2s => ds_slaves_m2s,      -- out   adb3_ocp_m2sT_array(0 to addr_range_table'length-1)
    masters_s2m => ds_slaves_s2m,      -- in    adb3_ocp_s2mT_array(0 to addr_range_table'length-1)
    -- Debug
    debug_info  => ds_split_err_info); -- out   std_logic_vector(31 downto 0))


  -----------------------
  -- Simple test block --
  -----------------------
  blk_ds_simple_test_i : blk_ds_simple_test
  generic map(
    addr_width => SIMPLE_ADDR_WIDTH)  -- natural range 1 to ADB3_OCP_ADDR_WIDTH)
  port map(
    -- OCP interface
    ocp_rst          => rst,               -- in    std_logic
    ocp_clk          => pll_reg_clk,       -- in    std_logic
    direct_slave_m2s => ds_slaves_m2s(0),  -- in    adb3_ocp_m2sT
    direct_slave_s2m => ds_slaves_s2m(0)); -- out   adb3_ocp_s2mT)


  ---------------------------
  -- Clock frequency block --
  ---------------------------
  blk_ds_clk_read_i : blk_ds_clk_read
  generic map(
    addr_width => CLK_READ_ADDR_WIDTH)  -- natural range 1 to ADB3_OCP_ADDR_WIDTH)
  port map(
    -- OCP interface
    ocp_rst          => rst,              -- in    std_logic
    ocp_clk          => pll_reg_clk,      -- in    std_logic
    direct_slave_m2s => ds_slaves_m2s(1), -- in    adb3_ocp_m2sT
    direct_slave_s2m => ds_slaves_s2m(1), -- out   adb3_ocp_s2mT
    -- Clock inputs
    pll_pri_clk      => pll_pri_clk,      -- in    std_logic
    pll_reg_clk      => pll_reg_clk,      -- in    std_logic
    pll_ref_clk      => pll_ref_clk,      -- in    std_logic
    pll_mem_clk      => pll_mem_clk,      -- in    std_logic
    clk_vec          => clk_vec);         -- in    clk_vec_t


  ------------------------------
  -- Front/Rear IO test block --
  ------------------------------
  blk_ds_io_test_i : blk_ds_io_test
  generic map(
    addr_width => GPIO_ADDR_WIDTH)  -- natural range 1 to ADB3_OCP_ADDR_WIDTH)
  port map(
    -- OCP interface
    ocp_rst          => rst,               -- in    std_logic
    ocp_clk          => pll_reg_clk,       -- in    std_logic
    direct_slave_m2s => ds_slaves_m2s(2),  -- in    adb3_ocp_m2sT
    direct_slave_s2m => ds_slaves_s2m(2),  -- out   adb3_ocp_s2mT
    -- IO interface
    gpio_inout       => gpio_inout);       -- inout gpio_inout_t


  --------------------------
  -- Interrupt test block --
  --------------------------
  blk_ds_int_test_i : blk_ds_int_test
  generic map(
    addr_width => INT_ADDR_WIDTH)  -- natural range 1 to ADB3_OCP_ADDR_WIDTH)
  port map(
    -- OCP interface
    ocp_rst          => rst,               -- in    std_logic
    ocp_clk          => pll_reg_clk,       -- in    std_logic
    direct_slave_m2s => ds_slaves_m2s(3),  -- in    adb3_ocp_m2sT
    direct_slave_s2m => ds_slaves_s2m(3),  -- out   adb3_ocp_s2mT
    -- Interrupt interface
    interrupt_l      => interrupt_l);      -- out   std_logic)


  ------------------------
  -- RAM register block --
  ------------------------
  blk_ds_ram_reg_i : blk_ds_mem_reg
  generic map(
    addr_width           => RAM_REG_ADDR_WIDTH,   -- natural range 1 to ADB3_OCP_ADDR_WIDTH)
    bank_addr_width      => BANK_ADDR_WIDTH,      -- natural
    page_addr_width_ddr3 => PAGE_ADDR_WIDTH_DDR3) -- natural
  port map(
    -- OCP interface
    ocp_rst          => rst,               -- in    std_logic
    ocp_clk          => pll_reg_clk,       -- in    std_logic
    direct_slave_m2s => ds_slaves_m2s(4),  -- in    adb3_ocp_m2sT
    direct_slave_s2m => ds_slaves_s2m(4),  -- out   adb3_ocp_s2mT
    -- DS RAM interface
    ds_bank          => ds_bank,           -- out   std_logic_vector(bank_addr_width-1 downto 0)
    ds_page_ddr3     => ds_page_ddr3,      -- out   std_logic_vector(page_addr_width_ddr3-1 downto 0)
    -- Memory interface status
    mem_if_stat      => mem_if_stat,       -- in    mem_if_stat_array_t
    mem_if_err       => mem_if_err,        -- in    mem_if_err_array_t
    -- Memory application control
    mem_app_go       => mem_app_go,        -- out   mem_app_go_array_t
    mem_app_offset   => mem_app_offset,    -- out   mem_app_offset_array_t
    mem_app_length   => mem_app_length,    -- out   mem_app_length_array_t
    -- Memory application status
    mem_app_done     => mem_app_done,      -- in    mem_app_done_array_t
    mem_app_err      => mem_app_err,       -- in    mem_app_err_array_t
    mem_app_err_ph   => mem_app_err_ph,    -- in    mem_app_err_ph_array_t
    mem_app_err_addr => mem_app_err_addr,  -- in    mem_app_err_addr_array_t
    -- Debug
    ram_mux_err_info => ram_mux_err_info,  -- in    mem_if_debug_array_t
    mem_if_err_info  => mem_if_err_info);  -- in    mem_if_debug_array_t


  ----------------
  -- Info block --
  ----------------
  blk_ds_info_i : blk_ds_info
  generic map(
    addr_width => INFO_ADDR_WIDTH)  -- natural range 1 to ADB3_OCP_ADDR_WIDTH)
  port map(
    -- OCP interface
    ocp_rst           => rst,                -- in    std_logic
    ocp_clk           => pll_reg_clk,        -- in    std_logic
    direct_slave_m2s  => ds_slaves_m2s(5),   -- in    adb3_ocp_m2sT
    direct_slave_s2m  => ds_slaves_s2m(5),   -- out   adb3_ocp_s2mT
    -- Debug
    ds_split_err_info => ds_split_err_info); -- in    std_logic_vector(31 downto 0))


  --------------------------
  -- BRAM block interface --
  --------------------------
  bram_cross_clk_dom_i : adb3_ocp_cross_clk_dom
  port map(
    -- OCP primary port
    slave_rst  => rst,              -- in    std_logic
    slave_clk  => pll_reg_clk,      -- in    std_logic
    slave_m2s  => ds_slaves_m2s(6), -- in    adb3_ocp_m2sT
    slave_s2m  => ds_slaves_s2m(6), -- out   adb3_ocp_s2mT
    -- OCP secondary port
    master_rst => rst,              -- in    std_logic
    master_clk => pll_pri_clk,      -- in    std_logic
    master_m2s => ds_bram_m2s,      -- out   adb3_ocp_m2sT
    master_s2m => ds_bram_s2m);     -- in    adb3_ocp_s2mT)

  --------------------------
  -- RAM block interface --
  --------------------------

  -- Connect DS memory window channel to DMA interface
  -- Concatenate DS window select address (DS_PAGE) with DS memory window address
  -- Use full ADB3_OCP_ADDR_WIDTH address range and pad unused bits with zeros
  ds_ram_mod_m2s.Cmd         <= ds_slaves_m2s(7).Cmd;
  ds_ram_mod_m2s.Addr        <= OCP_ADDR_PAD & ds_bank & ds_page_ddr3 & ds_slaves_m2s(7).Addr(WIND_ADDR_WIDTH-1 downto 0);
  ds_ram_mod_m2s.BurstLength <= ds_slaves_m2s(7).BurstLength;
  ds_ram_mod_m2s.Data        <= ds_slaves_m2s(7).Data;
  ds_ram_mod_m2s.DataByteEn  <= ds_slaves_m2s(7).DataByteEn;
  ds_ram_mod_m2s.DataValid   <= ds_slaves_m2s(7).DataValid;
  ds_ram_mod_m2s.RespAccept  <= ds_slaves_m2s(7).RespAccept;
  ds_ram_mod_m2s.Tag         <= ds_slaves_m2s(7).Tag;

  ds_slaves_s2m(7) <= ds_ram_mod_s2m;

  ram_cross_clk_dom_i : adb3_ocp_cross_clk_dom
  port map(
    -- OCP primary port
    slave_rst  => rst,            -- in    std_logic
    slave_clk  => pll_reg_clk,    -- in    std_logic
    slave_m2s  => ds_ram_mod_m2s, -- in    adb3_ocp_m2sT
    slave_s2m  => ds_ram_mod_s2m, -- out   adb3_ocp_s2mT
    -- OCP secondary port
    master_rst => rst,            -- in    std_logic
    master_clk => pll_pri_clk,    -- in    std_logic
    master_m2s => ds_ram_m2s,     -- out   adb3_ocp_m2sT
    master_s2m => ds_ram_s2m);    -- in    adb3_ocp_s2mT)

end;