--==============================================================================
--
--     File Name: blk_dma_switch.vhd
--       Version: 0.2
--          Date: 09 Feb 2011
--         Model: synth
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: OCP switch in DMA address space
--
--   Description: Switch OCP sources to OCP destinations in the OCP DMA address space
--                OCP sources are:
--                  DS BRAM
--                  DS RAM window
--                  DMA channels
--                  Memory application
--                OCP destinations are:
--                  BRAM
--                  On-board memory banks
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 11 Nov 2010: Initial design
--                0.2, TW, 09 Feb 2011:
--                  - Now uses adb3_ocp_comp package instead of adb3_target_pkg.
--                  - Deleted port 0 of adb3_ocp_split_nb_ds_i, since it was
--                    unused and unconnected, so that it now has only 4 master
--                    ports (instead of 5) that allow the 4 DDR3 memory banks to
--                    be accessed via direct slave.
--
--
--==============================================================================


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_comp.all;
use work.adb3_target_inc_pkg.all;
use work.mem_if_pkg.all;
use work.uber_pkg.all;

entity blk_dma_switch is
port(
  ocp_rst           : in    std_logic;
  ocp_clk           : in    std_logic;
  -- Source: BRAM direct slave OCP interface
  ds_bram_m2s       : in    adb3_ocp_m2sT;
  ds_bram_s2m       : out   adb3_ocp_s2mT;
  -- Source: On-board RAM DS window OCP interface
  ds_ram_m2s        : in    adb3_ocp_m2sT;
  ds_ram_s2m        : out   adb3_ocp_s2mT;
  -- Source: DMA OCP interface
  dma_channels_m2s  : in    adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
  dma_channels_s2m  : out   adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
  -- Source: On-board RAM app OCP interface
  mem_app_m2s       : in    adb3_ocp_m2sT_array(0 to MEM_BANKS-1);
  mem_app_s2m       : out   adb3_ocp_s2mT_array(0 to MEM_BANKS-1);
  -- Dest: BRAM OCP interface
  bram_if_m2s       : out   adb3_ocp_m2sT;
  bram_if_s2m       : in    adb3_ocp_s2mT;
  -- Dest: On-board RAM OCP interface
  ram_if_m2s        : out   adb3_ocp_m2sT_array(0 to MEM_BANKS-1);
  ram_if_s2m        : in    adb3_ocp_s2mT_array(0 to MEM_BANKS-1);
  -- Debug
  bram_mux_err_info : out   std_logic_vector(31 downto 0);
  ram_mux_err_info  : out   mem_if_debug_array_t);
end;


architecture synth of blk_dma_switch is

  -- Define number of multiplexed OCP channels connected to BRAM = DS + DMA
  constant BRAM_MUX_CH : natural := DMA_CHANNELS + 1;

  -- Define number of multiplexed OCP channels connected to on-board RAM = DS + DMA(0) + MEM_APP
  constant RAM_MUX_CH : natural := 3;

  -- Obtain number of slaves from DMA address range table
  constant DMA_SLAVES : natural := DMA_ADDR_RANGE_TABLE'length;

  -- OCP cross clock domain signals
  signal dma_retime_m2s : adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
  signal dma_retime_s2m : adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);

  -- OCP split output arrays
  -- Split into DMA_SLAVES address ranges specified by DMA_ADDR_RANGE_TABLE
  signal ds_ram_split_m2s : adb3_ocp_m2sT_array(0 to MEM_BANKS-1);
  signal ds_ram_split_s2m : adb3_ocp_s2mT_array(0 to MEM_BANKS-1);

  type dma_split_m2s_t is array (natural range <>) of adb3_ocp_m2sT_array(0 to DMA_SLAVES-1);
  type dma_split_s2m_t is array (natural range <>) of adb3_ocp_s2mT_array(0 to DMA_SLAVES-1);

  signal dma_split_m2s_array : dma_split_m2s_t(0 to DMA_CHANNELS-1);
  signal dma_split_s2m_array : dma_split_s2m_t(0 to DMA_CHANNELS-1);

  -- OCP mux input arrays
  signal ocp_bram_mux_m2s : adb3_ocp_m2sT_array(0 to BRAM_MUX_CH-1);
  signal ocp_bram_mux_s2m : adb3_ocp_s2mT_array(0 to BRAM_MUX_CH-1);

  type ocp_ram_mux_m2s_t is array (natural range <>) of adb3_ocp_m2sT_array(0 to RAM_MUX_CH-1);
  type ocp_ram_mux_s2m_t is array (natural range <>) of adb3_ocp_s2mT_array(0 to RAM_MUX_CH-1);

  signal ocp_ram_mux_m2s_array : ocp_ram_mux_m2s_t(0 to MEM_BANKS-1);
  signal ocp_ram_mux_s2m_array : ocp_ram_mux_s2m_t(0 to MEM_BANKS-1);

  -- Debug
  signal ds_split_err_info  : std_logic_vector(31 downto 0);
  signal dma_split_err_info : std_logic_vector(31 downto 0);

begin

  -- Re-time DMA OCP interface
  dma_channel_retime_g : for m in 0 to DMA_CHANNELS-1 generate
    adb3_ocp_retime_i : adb3_ocp_retime_nb
    port map(
      ocp_rst     => ocp_rst,             -- in    std_logic
      ocp_clk     => ocp_clk,             -- in    std_logic
      -- OCP slave port
      slave_m2s   => dma_channels_m2s(m), -- in    adb3_ocp_m2sT
      slave_s2m   => dma_channels_s2m(m), -- out   adb3_ocp_s2mT
      -- OCP master port
      master_m2s  => dma_retime_m2s(m),   -- out   adb3_ocp_m2sT
      master_s2m  => dma_retime_s2m(m));  -- in    adb3_ocp_s2mT)
    end generate;


  -- Split DS DDR3 window OCP interface according to function
  adb3_ocp_split_nb_ds_i : adb3_ocp_split_nb
  generic map(
    addr_range_table => DS_DDR3_ADDR_RANGE_TABLE)  -- addr_range_table_t
  port map(
    ocp_rst     => ocp_rst,            -- in    std_logic
    ocp_clk     => ocp_clk,            -- in    std_logic
    -- Master OCP interface
    slave_m2s   => ds_ram_m2s,         -- in    adb3_ocp_m2sT
    slave_s2m   => ds_ram_s2m,         -- out   adb3_ocp_s2mT
    -- Slave OCP interfaces
    masters_m2s => ds_ram_split_m2s,   -- out   adb3_ocp_m2sT_array(0 to addr_range_table'length-1)
    masters_s2m => ds_ram_split_s2m,   -- in    adb3_ocp_s2mT_array(0 to addr_range_table'length-1)
    -- Debug
    debug_info  => ds_split_err_info); -- out   std_logic_vector(31 downto 0))


  dma_split_g : for m in 0 to 0 generate
    -- Split DMA OCP interface according to function
    adb3_ocp_split_nb_dma_i : adb3_ocp_split_nb
    generic map(
      addr_range_table => DMA_ADDR_RANGE_TABLE)  -- addr_range_table_t
    port map(
      ocp_rst     => ocp_rst,                -- in    std_logic
      ocp_clk     => ocp_clk,                -- in    std_logic
      -- Master OCP interface
      slave_m2s   => dma_retime_m2s(m),      -- in    adb3_ocp_m2sT
      slave_s2m   => dma_retime_s2m(m),      -- out   adb3_ocp_s2mT
      -- Slave OCP interfaces
      masters_m2s => dma_split_m2s_array(m), -- out   adb3_ocp_m2sT_array(0 to addr_range_table'length-1)
      masters_s2m => dma_split_s2m_array(m), -- in    adb3_ocp_s2mT_array(0 to addr_range_table'length-1)
      -- Debug
      debug_info  => dma_split_err_info);    -- out   std_logic_vector(31 downto 0))
    end generate;


  ---------------------------------------------------------------
  -- Connect split OCP channels to BRAM mux                    --
  --                                                           --
  -- Mux input        | OCP channel                            --
  ---------------------------------------------------------------
  -- 0                | DS BRAM                                --
  -- 1                | DMA(0) split(0)                        --
  -- 2..BRAM_MUX_CH-1 | DMA(1..BRAM_MUX_CH-2)                  --
  --                                                           --
  ---------------------------------------------------------------

  -- Connect DS BRAM channel to BRAM mux input (0)
  ocp_bram_mux_m2s(0) <= ds_bram_m2s;
  ds_bram_s2m         <= ocp_bram_mux_s2m(0);

  -- Connect DMA channels split array (mux ch)(ram ch) to BRAM mux input (1)
  ocp_bram_mux_m2s(1)       <= dma_split_m2s_array(0)(0);
  dma_split_s2m_array(0)(0) <= ocp_bram_mux_s2m(1);

  -- Connect DMA channels array (1..DMA_SLAVES-1) to BRAM mux inputs (2..BRAM_MUX_CH-1)
  connect_dma_ram_g : for d in 2 to BRAM_MUX_CH-1 generate
    ocp_bram_mux_m2s(d) <= dma_retime_m2s(d-1);
    dma_retime_s2m(d-1) <= ocp_bram_mux_s2m(d);
  end generate;

  ---------------------------------------------------------------
  -- Connect split OCP channels to RAM muxes for each RAM bank --
  --                                                           --
  -- Mux input       | OCP channel                             --
  ---------------------------------------------------------------
  -- 0               | DS ram split(r)                         --
  -- 1               | Mem App(r)                              --
  -- 2..RAM_MUX_CH-1 | DMA(0..RAM_MUX_CH-3) split(r+1)         --
  --                                                           --
  ---------------------------------------------------------------

  connect_ram_mux_g : for r in 0 to MEM_BANKS-1 generate
    -- Connect DS RAM Window channel split (ram ch+1) to RAM mux input array (ram ch)(0)
    ocp_ram_mux_m2s_array(r)(0) <= ds_ram_split_m2s(r);
    ds_ram_split_s2m(r)         <= ocp_ram_mux_s2m_array(r)(0);

    -- Connect RAM app channel (ram ch) to RAM mux input array (ram ch)(2)
    ocp_ram_mux_m2s_array(r)(1) <= mem_app_m2s(r);
    mem_app_s2m(r)              <= ocp_ram_mux_s2m_array(r)(1);

    -- Connect DMA channels split array (mux ch)(ram ch) to RAM mux input array (ram ch)(mux ch)
    connect_dma_ram_g : for m in 2 to RAM_MUX_CH-1 generate
      ocp_ram_mux_m2s_array(r)(m)   <= dma_split_m2s_array(m-2)(r+1);
      dma_split_s2m_array(m-2)(r+1) <= ocp_ram_mux_s2m_array(r)(m);
    end generate;
  end generate;


  -----------------------------------------
  -- Multiplex OCP split channels - BRAM --
  -----------------------------------------

  adb3_ocp_mux_nb_bram_i : adb3_ocp_mux_nb
  generic map(
    mux_inputs    => BRAM_MUX_CH) -- natural := 2
  port map(
    ocp_rst    => ocp_rst,            -- in    std_logic
    ocp_clk    => ocp_clk,            -- in    std_logic
    -- OCP slave ports
    slaves_m2s => ocp_bram_mux_m2s,   -- in    adb3_ocp_m2sT_array(0 to mux_inputs-1)
    slaves_s2m => ocp_bram_mux_s2m,   -- out   adb3_ocp_s2mT_array(0 to mux_inputs-1)
    -- OCP master port
    master_m2s => bram_if_m2s,        -- out   adb3_ocp_m2sT
    master_s2m => bram_if_s2m,        -- in    adb3_ocp_s2mT
    -- Debug
    debug_info => bram_mux_err_info); -- out   std_logic_vector(31 downto 0));


  ----------------------------------------
  -- Multiplex OCP split channels - RAM --
  ----------------------------------------

  mux_ram_g : for m in 0 to MEM_BANKS-1 generate
    adb3_ocp_mux_nb_ram_i : adb3_ocp_mux_nb
    generic map(
      mux_inputs    => RAM_MUX_CH) -- natural := 2
    port map(
      ocp_rst    => ocp_rst,                   -- in    std_logic
      ocp_clk    => ocp_clk,                   -- in    std_logic
      -- OCP slave ports
      slaves_m2s => ocp_ram_mux_m2s_array(m),  -- in    adb3_ocp_m2sT_array(0 to mux_inputs-1)
      slaves_s2m => ocp_ram_mux_s2m_array(m),  -- out   adb3_ocp_s2mT_array(0 to mux_inputs-1)
      -- OCP master port
      master_m2s => ram_if_m2s(m),             -- out   adb3_ocp_m2sT
      master_s2m => ram_if_s2m(m),             -- in    adb3_ocp_s2mT
      -- Debug
      debug_info => ram_mux_err_info(m));      -- out   std_logic_vector(31 downto 0));
  end generate;

end;