--==============================================================================
--
--     File Name: blk_ds_info.vhd
--       Version: 0.3
--          Date: 16 Feb 2011
--         Model: synth
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: Info
--
--   Description: Provides misc info on DS OCP
--                Connected to Direct Slave OCP interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 11 Jun 2010: Initial design
--                0.2, TW, 09 Feb 2011:
--                  - Now references adb3_ocp_comp package instead of
--                    adb3_target_pkg.
--                  - Got rid of some unnecessary resets.
--                0.3, KDC, 16 Feb 2011:
--                  - Changed range of addr_width to match OCP profile
--
--==============================================================================


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_comp.all;
use work.adb3_target_inc_pkg.all;
use work.uber_pkg.all;
use work.today_pkg.all;

entity blk_ds_info is
generic(
  addr_width : natural range 1 to ADB3_OCP_ADDR_WIDTH);
port(
  -- OCP interface
  ocp_rst           : in    std_logic;
  ocp_clk           : in    std_logic;
  direct_slave_m2s  : in    adb3_ocp_m2sT;
  direct_slave_s2m  : out   adb3_ocp_s2mT;
  -- Debug
  ds_split_err_info : in    std_logic_vector(31 downto 0));
end;


architecture synth of blk_ds_info is

  constant REG_WIDTH : natural := 32;
  constant WE_WIDTH  : natural := REG_WIDTH/8;

  constant BANK_NIBBLE : std_logic_vector(3 downto 0) := std_logic_vector(to_unsigned(MEM_BANKS,4));

  -- Register interface signals
  signal ds_d  : std_logic_vector(REG_WIDTH-1 downto 0);
  signal ds_q  : std_logic_vector(REG_WIDTH-1 downto 0);
  signal ds_a  : std_logic_vector(addr_width-1 downto 0);
  signal ds_w  : std_logic;
  signal ds_r  : std_logic;
  signal ds_we : std_logic_vector(WE_WIDTH-1 downto 0);

begin

  --------------------------------
  -- Direct Slave OCP interface --
  --------------------------------

  adb3_ocp_simple_bus_if_i : adb3_ocp_simple_bus_if
  generic map(
    addr_width => addr_width,   -- natural range 1 to ADB3_OCP_ADDR_WIDTH
    data_width => REG_WIDTH)    -- natural range 8 to ADB3_OCP_DATA_WIDTH
  port map(
    -- OCP Interface
    ocp_rst => ocp_rst,           -- in    std_logic
    ocp_clk => ocp_clk,           -- in    std_logic
    ocp_m2s => direct_slave_m2s,  -- in    adb3_ocp_m2sT
    ocp_s2m => direct_slave_s2m,  -- out   adb3_ocp_s2mT
    -- Simple bus interface
    d       => ds_d,        -- out   std_logic_vector(data_width-1 downto 0)
    q       => ds_q,        -- in    std_logic_vector(data_width-1 downto 0)
    a       => ds_a,        -- out   std_logic_vector(addr_width-1 downto 0)
    w       => ds_w,        -- out   std_logic
    r       => ds_r,        -- out   std_logic
    we      => ds_we);      -- out   std_logic_vector((data_width/8)-1 downto 0))


  -- Multiplex read registers
  rd_regs_p : process(ocp_clk)
    variable ds_q_var : std_logic_vector(REG_WIDTH-1 downto 0);
  begin
    if rising_edge(ocp_clk) then
      ds_q_var := (others => '-');

      if (ds_a = INFO_DATE_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := TODAYS_DATE;
      end if;
      if (ds_a = INFO_TIME_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := TODAYS_TIME;
      end if;
      if (ds_a = INFO_SPLIT_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := ds_split_err_info;
      end if;
      if (ds_a = INFO_BRAM_BASE_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var(REG_WIDTH-1 downto DS_ADDR_WIDTH) := (others => '0');
        ds_q_var(DS_ADDR_WIDTH-1 downto 0) := BRAM_ADDR_BASE(DS_ADDR_WIDTH-1 downto 0);
      end if;
      if (ds_a = INFO_BRAM_MASK_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var(REG_WIDTH-1 downto DS_ADDR_WIDTH) := (others => '0');
        ds_q_var(DS_ADDR_WIDTH-1 downto 0) := BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0);
      end if;
      if (ds_a = INFO_RAM_BASE_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var(REG_WIDTH-1 downto DS_ADDR_WIDTH) := (others => '0');
        ds_q_var(DS_ADDR_WIDTH-1 downto 0) := RAM_WIN_ADDR_BASE(DS_ADDR_WIDTH-1 downto 0);
      end if;
      if (ds_a = INFO_RAM_MASK_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var(REG_WIDTH-1 downto DS_ADDR_WIDTH) := (others => '0');
        ds_q_var(DS_ADDR_WIDTH-1 downto 0) := RAM_WIN_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0);
      end if;
      if (ds_a = INFO_RAM_INFO_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var(3 downto 0) := BANK_NIBBLE;
      end if;

      ds_q <= ds_q_var;
    end if;
  end process;

end;