--==============================================================================
--
--     File Name: blk_ds_int_test.vhd
--       Version: 0.3
--          Date: 16 Feb 2011
--         Model: synth
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: Interrupt test
--
--   Description: Interrupt test
--                Connected to Direct Slave OCP interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 21 Apr 2010: Initial design
--                0.2, TW, 09 Feb 2011:
--                  - Now references adb3_ocp_comp package instead of
--                    adb3_target_pkg.
--                  - Got rid of some unnecessary resets.
--                  - Added an IOB attribute to interrupt output, and made it
--                    active low to facilitate absorbtion into an IOB.
--                0.3, KDC, 16 Feb 2011:
--                  - Changed range of addr_width to match OCP profile
--
--==============================================================================


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_comp.all;
use work.uber_pkg.all;

entity blk_ds_int_test is
generic(
  addr_width : natural range 1 to ADB3_OCP_ADDR_WIDTH);
port(
  -- OCP interface
  ocp_rst          : in    std_logic;
  ocp_clk          : in    std_logic;
  direct_slave_m2s : in    adb3_ocp_m2sT;
  direct_slave_s2m : out   adb3_ocp_s2mT;
  -- Interrupt interface
  interrupt_l      : out   std_logic);
end;


architecture synth of blk_ds_int_test is

  constant REG_WIDTH : natural := 32;
  constant WE_WIDTH  : natural := REG_WIDTH/8;

  -- Register interface signals
  signal ds_d  : std_logic_vector(REG_WIDTH-1 downto 0);
  signal ds_q  : std_logic_vector(REG_WIDTH-1 downto 0);
  signal ds_a  : std_logic_vector(addr_width-1 downto 0);
  signal ds_w  : std_logic;
  signal ds_r  : std_logic;
  signal ds_we : std_logic_vector(WE_WIDTH-1 downto 0);

  -- Interrupt test
  signal stat_reg        : std_logic_vector(REG_WIDTH-1 downto 0);
  signal mask_reg        : std_logic_vector(REG_WIDTH-1 downto 0);
  signal wcount_reg      : std_logic_vector(REG_WIDTH-1 downto 0);
  signal wcount_reg_wr   : std_logic_vector(WE_WIDTH-1 downto 0);
  signal wcount_wr       : std_logic;
  signal count_reg       : std_logic_vector(REG_WIDTH-1 downto 0);
  signal arm_reg_wr      : std_logic_vector(WE_WIDTH-1 downto 0);
  signal arm_wr          : std_logic;
  signal interrupt_int_l : std_logic;

  -- Xilinx XST attributes
  attribute IOB : string;
  attribute IOB of interrupt_int_l : signal is "TRUE";

begin

  -- Output the internal version of 'interrupt'
  interrupt_l <= interrupt_int_l;

  --------------------------------
  -- Direct Slave OCP interface --
  --------------------------------

  adb3_ocp_simple_bus_if_i : adb3_ocp_simple_bus_if
  generic map(
    addr_width => addr_width,   -- natural range 1 to ADB3_OCP_ADDR_WIDTH
    data_width => REG_WIDTH)    -- natural range 8 to ADB3_OCP_DATA_WIDTH
  port map(
    -- OCP Interface
    ocp_rst => ocp_rst,           -- in    std_logic
    ocp_clk => ocp_clk,           -- in    std_logic
    ocp_m2s => direct_slave_m2s,  -- in    adb3_ocp_m2sT
    ocp_s2m => direct_slave_s2m,  -- out   adb3_ocp_s2mT
    -- Simple bus interface
    d       => ds_d,        -- out   std_logic_vector(data_width-1 downto 0)
    q       => ds_q,        -- in    std_logic_vector(data_width-1 downto 0)
    a       => ds_a,        -- out   std_logic_vector(addr_width-1 downto 0)
    w       => ds_w,        -- out   std_logic
    r       => ds_r,        -- out   std_logic
    we      => ds_we);      -- out   std_logic_vector((data_width/8)-1 downto 0))


  -- Update write registers
  wr_regs_p : process(ocp_rst,ocp_clk)
    variable u,l : natural;
  begin
    if (ocp_rst = '1') then
      stat_reg      <= (others => '0');
      mask_reg      <= (others => '1');
      arm_reg_wr    <= (others => '0');
      wcount_reg    <= (others => '0');
      wcount_reg_wr <= (others => '0');
    elsif rising_edge(ocp_clk) then
      arm_reg_wr    <= (others => '0');
      wcount_reg_wr <= (others => '0');
      for i in 0 to WE_WIDTH-1 loop
        u := 8*(i+1)-1;
        l := 8*i;
        if (ds_we(i)='1') then
          if (ds_a = INT_SET_REG_ADDR(addr_width-1 downto 0)) then
            stat_reg(u downto l) <= stat_reg(u downto l) or ds_d(u downto l);
          end if;
          if (ds_a = INT_CLEAR_REG_ADDR(addr_width-1 downto 0)) then
            stat_reg(u downto l) <= stat_reg(u downto l) and not(ds_d(u downto l));
          end if;
          if (ds_a = INT_WMASK_REG_ADDR(addr_width-1 downto 0)) then
            mask_reg(u downto l) <= ds_d(u downto l);
          end if;
          if (ds_a = INT_ARM_REG_ADDR(addr_width-1 downto 0)) then
            arm_reg_wr(i) <= '1';
          end if;
          if (ds_a = INT_WCOUNT_REG_ADDR(addr_width-1 downto 0)) then
            wcount_reg(u downto l) <= ds_d(u downto l);
            wcount_reg_wr(i) <= '1';
          end if;
        end if;
      end loop;
    end if;
  end process;

  arm_wr    <= AND_reduce(arm_reg_wr);
  wcount_wr <= AND_reduce(wcount_reg_wr);


  -- Multiplex read registers
  rd_regs_p : process(ocp_clk)
    variable ds_q_var : std_logic_vector(REG_WIDTH-1 downto 0);
  begin
    if rising_edge(ocp_clk) then
      ds_q_var := (others => '-');

      if (ds_a = INT_STAT_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := stat_reg;
      end if;
      if (ds_a = INT_RMASK_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := mask_reg;
      end if;
      if (ds_a = INT_RCOUNT_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := count_reg;
      end if;

      ds_q <= ds_q_var;
    end if;
  end process;


  -- Generate Elapsed cycle counter and interrupt
  int_p : process(ocp_rst,ocp_clk)
  begin
    if (ocp_rst = '1') then
      count_reg <= (others => '0');
      interrupt_int_l <= '1';
    elsif rising_edge(ocp_clk) then
      -- Load if status cleared, increment if status set
      if (OR_reduce(stat_reg) = '0') then
        if (wcount_wr = '1') then
          count_reg <= wcount_reg;
        end if;
      else
        count_reg <= count_reg + 1;
      end if;

      -- Set interrupt if unmasked bit set in status register
      -- Clear interrupt on int_arm
      if (OR_reduce(not(mask_reg) and stat_reg) = '0') or (arm_wr = '1') then
        interrupt_int_l <= '1';
      else
        interrupt_int_l <= '0';
      end if;
    end if;
  end process;

end;