--==============================================================================
--
--     File Name: test_uber_ds.vhd
--       Version: 0.1
--          Date: 17 Mar 2010
--         Model: Behavioral (NOT SYNTHESISABLE)
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: ADB3 Target Design Test: Direct Slave
--
--   Description: Provides bridge Direct Slave interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 17 Mar 09: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_comp.all;
use work.adb3_target_types_pkg.all;
use work.adb3_target_inc_pkg.all;
use work.adb3_target_pkg.all;
use work.adb3_target_tb_pkg.all;
use work.uber_pkg.all;
use work.today_pkg.all;
use work.uber_tb_pkg.all;

entity test_uber_ds is
port(
  -- OCP Interface
  ocp_clk          : in    std_logic;
  direct_slave_m2s : out   adb3_ocp_m2sT;
  direct_slave_s2m : in    adb3_ocp_s2mT;
  -- Test interface
  mptl_online_long : in    std_logic;
  ds_comp          : out   ds_comp_t;
  ds_pass          : out   ds_pass_t);
end test_uber_ds;

architecture test of test_uber_ds is

  -- Define test frequencies for Uber clocks
  constant PLL_REG_CLK_FREQ : natural := CLK_80_FREQ;
  constant PLL_PRI_CLK_FREQ : natural := CLK_200_FREQ;
  -- Define +/- acceptable error for clock frequency measurement
  constant CLK_TEST_DIFF : natural := 2;

  constant PN4_DATA : std_logic_vector(31 downto 0) := X"AABBCCDD";
  constant PN6_DATA : std_logic_vector(63 downto 0) := X"AAAABBBBCCCCDDDD";

  constant DS_WIN_RAM_BANK     : std_logic_vector(31 downto 0) := X"00000001";
  constant DS_WIN_RAM_PAGE_BOT : std_logic_vector(31 downto 0) := X"00000000";
  constant DS_WIN_RAM_PAGE_TOP : std_logic_vector(31 downto 0) := X"0000007F";

  -- RAM bank used for testing
  constant TEST_RAM_BANK : natural range 0 to MEM_BANKS-1 := 1;

  -- Words-1 (16-byte words) of each phase of the internal RAM test
  constant RAM_TEST_LENGTH : natural := 256 - 1;
  constant RAM_TEST_LEN : std_logic_vector(31 downto 0) := std_logic_vector(to_unsigned(RAM_TEST_LENGTH,32));

  -- Top address (16-byte words) of RAM
  constant RAM_TEST_TOP : natural := (2**(TB_DDR3_16_BYTE_ADDR_WIDTH)) - 1;

  -- Start address (16-byte words) of each phase of the internal RAM test
  -- Set to top address - length of test - top word used by DS part of test
  constant RAM_TEST_START : natural := RAM_TEST_TOP - RAM_TEST_LENGTH - 1;
  constant RAM_TEST_OFF : std_logic_vector(31 downto 0) := std_logic_vector(to_unsigned(RAM_TEST_START,32));

  constant DO_INT_RAM_TEST  : boolean := true;
  constant DO_INTERRUPT_NUM : natural range 1 to 32 := 32;

  signal ds_m2s : adb3_ocp_m2sT := ADB3_OCP_DEFAULT_M2S;
  signal ds_s2m : adb3_ocp_s2mT;

begin
  ---------------------------------
  -- MPTL Direct Slave interface --
  ---------------------------------

  -- Connect direct slave input
  ds_s2m <= direct_slave_s2m;

  direct_slave_p : process
    variable reg_data_wr     : byte_vector_t(0 to 127);
    variable reg_data_rd     : byte_vector_t(0 to 127);
    variable reg_be          : byte_enable_t(0 to 127);
    variable reg_addr        : adb3_ocp_addr_s;
    variable reg_xfer        : natural;
    variable master_idle     : std_logic;
    variable master_timeout  : natural := 64;
    variable master_cnt      : natural := 0;
    variable master_passed   : boolean := true;
    variable clock_running   : std_logic;
    variable ck_passed       : boolean := true;
    variable io_data_ls      : std_logic_vector(31 downto 0);
    variable io_data_ms      : std_logic_vector(31 downto 0);
    variable io_data_x2      : std_logic_vector(63 downto 0);
    variable f_io_passed     : boolean := true;
    variable r_io_passed     : boolean := true;
    variable int_set         : std_logic_vector(31 downto 0);
    variable int_stat        : std_logic_vector(31 downto 0);
    variable int_passed      : boolean := true;
    variable info_passed     : boolean := true;
    variable bram_passed     : boolean := true;
    variable ram_passed      : boolean := true;
    variable ram_init        : std_logic;
    variable ram_done        : std_logic;

  begin
    wait until (mptl_online_long = '1');  -- OCP commands ignored until MPTL online
    ds_comp.simple_complete  <= false;
    ds_comp.master_complete  <= false;
    ds_comp.clock_complete   <= false;
    ds_comp.frontio_complete <= false;
    ds_comp.reario_complete  <= false;
    ds_comp.int_complete     <= false;
    ds_comp.info_complete    <= false;
    ds_comp.bram_complete    <= false;
    ds_comp.ram_complete     <= false;

    ds_pass.simple_passed    <= true;
    ds_pass.master_passed    <= true;
    ds_pass.clock_passed     <= true;
    ds_pass.frontio_passed   <= true;
    ds_pass.reario_passed    <= true;
    ds_pass.int_passed       <= true;
    ds_pass.info_passed      <= true;
    ds_pass.bram_passed      <= true;
    ds_pass.ram_passed       <= true;


    ---------------------------------------------
    -- Simple Test (32-bit register interface) --
    ---------------------------------------------
    reg_be(0 to 3) := conv_byte_enable("1111");

    -- Write a pattern to register
    reg_addr := SIMPLE_ADDR_BASE + SIMPLE_WDATA_REG_ADDR;
    reg_data_wr(0 to 3) := conv_byte_vector(X"cafeface");
    adb3_target_sim_write_reg32(reg_addr,10,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Wrote simple WDATA " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
           " with enable 0b" & conv_string(reg_be(0 to 3)) &
           " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Read back nibble reversed pattern from register
    reg_addr := SIMPLE_ADDR_BASE + SIMPLE_RDATA_REG_ADDR;
    adb3_target_sim_read_reg32(reg_addr,11,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read simple RDATA " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    ds_comp.simple_complete <= true;
    if (reg_data_rd(0 to 3) = conv_byte_vector(X"ecafefac")) then
      assert false report "Test Simple completed: PASSED." severity note;
    else
      assert false report "Test Simple completed: FAILED." severity error;
      ds_pass.simple_passed <= false;
    end if;


    if (MASTER_USE = MASTER_ON) then
      ---------------------------------------------
      -- Master Test (32-bit register interface) --
      ---------------------------------------------
      reg_be(0 to 3) := conv_byte_enable("1111");

      -- Read OOR
      reg_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) + 1;
      adb3_target_sim_read(reg_addr,12,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read OOR Addr 0 " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (reg_data_rd(0 to 3) /= conv_byte_vector(X"DEADC0DE")) then
        master_passed := false;
      end if;

      -- Write direct master write data
      reg_addr := MASTER_ADDR_BASE + MASTER_WDATA_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(X"76543210");  -- 32-bit data
      adb3_target_sim_write_reg32(reg_addr,12,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote master WDATA " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Write direct master configuration (1 = write, 0 = read)
      reg_addr := MASTER_ADDR_BASE + MASTER_SEL_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(X"0000000" & "0001");  -- bit 0 = write
      adb3_target_sim_write_reg32(reg_addr,13,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote master SEL " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Write direct master address and trigger transfer
      reg_addr := MASTER_ADDR_BASE + MASTER_ADDR_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(HOST_MASTER_REG_ADDR(DM_DATA_WIDTH-1 downto 0));
      adb3_target_sim_write_reg32(reg_addr,14,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote master ADDR " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Write direct master configuration (1 = write, 0 = read)
      reg_addr := MASTER_ADDR_BASE + MASTER_SEL_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(X"0000000" & "0000");  -- bit 0 = read
      adb3_target_sim_write_reg32(reg_addr,15,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote master SEL " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Write direct master address and trigger transfer
      reg_addr := MASTER_ADDR_BASE + MASTER_ADDR_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(HOST_MASTER_REG_ADDR(DM_DATA_WIDTH-1 downto 0));
      adb3_target_sim_write_reg32(reg_addr,16,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote master ADDR " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      master_idle := '0';
      while ((master_idle = '0')and(master_cnt < master_timeout)) loop
        -- Read direct master status
        reg_addr := MASTER_ADDR_BASE + MASTER_STAT_REG_ADDR;
        adb3_target_sim_read_reg32(reg_addr,17,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        master_idle         := reg_data_rd(0)(0);
        master_cnt          := master_cnt + 1;
      end loop;

      if (master_cnt < master_timeout) then
        report "Direct Master transfer complete (" & std_logic'image(master_idle) & ")";
      else
        report "Direct Master transfer timeout (" & std_logic'image(master_idle) & ")";
        master_passed := false;
      end if;

      -- Read direct master read data
      reg_addr := MASTER_ADDR_BASE + MASTER_RDATA_REG_ADDR;
      adb3_target_sim_read_reg32(reg_addr,18,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (reg_data_rd(0 to 3) /= conv_byte_vector(X"76543210")) then
        master_passed := false;
      end if;

      ds_comp.master_complete <= true;
      if (master_passed) then
        assert false report "Test Master completed: PASSED." severity note;
      else
        assert false report "Test Master completed: FAILED." severity error;
        ds_pass.master_passed <= false;
      end if;
    end if;


    if (TARGET_USE /= SYN_NGC) then
      -- SIM OCP:  Fast frequency measurement in MHz for simulation
      -- SIM_MPTL: Fast frequency measurement in MHz for simulation
      -- SYN_NGC:  SLOW frequency measurement in Hz for synthesis

      ------------------------------------------------------------------
      -- Clock Frequency Measurement Test (32-bit register interface) --
      ------------------------------------------------------------------
      reg_be(0 to 3) := conv_byte_enable("1111");

      -- Read OOR
      reg_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) + 1;
      adb3_target_sim_read(reg_addr,19,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read OOR Addr 0 " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (reg_data_rd(0 to 3) /= conv_byte_vector(X"DEADC0DE")) then
        ck_passed := false;
      end if;

      -- Clear all flags
      reg_addr := CLK_READ_ADDR_BASE + CLK_READ_CTRL_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(X"80000000");
      adb3_target_sim_write_reg32(reg_addr,19,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote Clear All CTRL " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Select pll_reg_clk
      reg_addr := CLK_READ_ADDR_BASE + CLK_READ_SEL_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(X"000000" & "000" & PLL_REG_CLK_SEL);
      adb3_target_sim_write_reg32(reg_addr,20,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PLL_REG_CLK_SEL SEL " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Wait until clock register has been updated (msb = 1)
      -- Then read clock frequency
      clock_running := '0';
      while (clock_running = '0') loop
        reg_addr := CLK_READ_ADDR_BASE + CLK_READ_STAT_REG_ADDR;
        adb3_target_sim_read_reg32(reg_addr,21,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        clock_running := reg_data_rd(3)(7);
      end loop;

      -- Read frequency data
      reg_addr := CLK_READ_ADDR_BASE + CLK_READ_FREQ_REG_ADDR;
      adb3_target_sim_read_reg32(reg_addr,22,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read PLL_REG_CLK FREQ " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      report "Expected freq = " & natural'image(PLL_REG_CLK_FREQ) & " MHz +/-" & natural'image(CLK_TEST_DIFF) & " MHz";
      report "Actual   freq = " & integer'image(to_integer(unsigned(reg_data_rd(0)))) & " MHz";

      -- Check for truncated  and +1 frequency values
      if ((to_integer(unsigned(reg_data_rd(0))) < (PLL_REG_CLK_FREQ-CLK_TEST_DIFF))or
          (to_integer(unsigned(reg_data_rd(0))) > (PLL_REG_CLK_FREQ+CLK_TEST_DIFF)))then
        ck_passed := false;
      end if;

      -- Select pll_pri_clk
      reg_addr := CLK_READ_ADDR_BASE + CLK_READ_SEL_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(X"000000" & "000" & PLL_PRI_CLK_SEL);
      adb3_target_sim_write_reg32(reg_addr,23,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PLL_PRI_CLK_SEL SEL " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Wait until clock register has been updated (msb = 1)
      -- Then read clock frequency
      clock_running := '0';
      while (clock_running = '0') loop
        reg_addr := CLK_READ_ADDR_BASE + CLK_READ_STAT_REG_ADDR;
        adb3_target_sim_read_reg32(reg_addr,24,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        clock_running       := reg_data_rd(3)(7);
      end loop;

      -- Read frequency data
      reg_addr := CLK_READ_ADDR_BASE + CLK_READ_FREQ_REG_ADDR;
      adb3_target_sim_read_reg32(reg_addr,25,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read PLL_PRI_CLK FREQ " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      report "Expected freq = " & natural'image(PLL_PRI_CLK_FREQ) & " MHz +/-" & natural'image(CLK_TEST_DIFF) & " MHz";
      report "Actual   freq = " & integer'image(to_integer(unsigned(reg_data_rd(0)))) & " MHz";

      if ((to_integer(unsigned(reg_data_rd(0))) < (PLL_PRI_CLK_FREQ-CLK_TEST_DIFF))or
          (to_integer(unsigned(reg_data_rd(0))) > (PLL_PRI_CLK_FREQ+CLK_TEST_DIFF))) then
        ck_passed := false;
      end if;

      -- Select MGT 113 clk0
      reg_addr := CLK_READ_ADDR_BASE + CLK_READ_SEL_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(X"000000" & "000" & MGT113_CLK0_SEL);
      adb3_target_sim_write_reg32(reg_addr,26,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote MGT113_CLK0_SEL SEL " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Wait until clock register has been updated (msb = 1)
      -- Then read clock frequency
      clock_running := '0';
      while (clock_running = '0') loop
        reg_addr := CLK_READ_ADDR_BASE + CLK_READ_STAT_REG_ADDR;
        adb3_target_sim_read_reg32(reg_addr,27,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        clock_running       := reg_data_rd(3)(7);
      end loop;

      -- Read frequency data
      reg_addr := CLK_READ_ADDR_BASE + CLK_READ_FREQ_REG_ADDR;
      adb3_target_sim_read_reg32(reg_addr,28,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read MGT113_CLK0 FREQ " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (BOARD_TYPE = ADM_XRC_6TL) then
        report "Expected freq = " & natural'image(CLK_100_FREQ) & " MHz +/-" & natural'image(CLK_TEST_DIFF) & " MHz";
      end if;
      if (BOARD_TYPE = ADM_XRC_6T1) then
        report "Expected freq = " & natural'image(CLK_250_FREQ) & " MHz +/-" & natural'image(CLK_TEST_DIFF) & " MHz";
      end if;
      report "Actual  freq = " & integer'image(to_integer(unsigned(reg_data_rd(0)))) & " MHz";

      if (BOARD_TYPE = ADM_XRC_6TL) then
        if ((to_integer(unsigned(reg_data_rd(0))) < (CLK_100_FREQ-CLK_TEST_DIFF))or
            (to_integer(unsigned(reg_data_rd(0))) > (CLK_100_FREQ+CLK_TEST_DIFF))) then
          ck_passed := false;
        end if;
      end if;
      if (BOARD_TYPE = ADM_XRC_6T1) then
        if ((to_integer(unsigned(reg_data_rd(0))) < (CLK_250_FREQ-CLK_TEST_DIFF))or
            (to_integer(unsigned(reg_data_rd(0))) > (CLK_250_FREQ+CLK_TEST_DIFF))) then
          ck_passed := false;
        end if;
      end if;

      ds_comp.clock_complete <= true;
      if (ck_passed) then
        assert false report "Test Clock Read completed: PASSED." severity note;
      else
        assert false report "Test Clock Read completed: FAILED." severity error;
        ds_pass.clock_passed <= false;
      end if;
    end if;


    if XRM_GPIO_VALID then
      -----------------------------------------------
      -- XRM GPIO Test (32-bit register interface) --
      -----------------------------------------------
      reg_be(0 to 3) := conv_byte_enable("1111");

      -- Read OOR
      reg_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) + 1;
      adb3_target_sim_read(reg_addr,12,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read OOR Addr 0 " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (reg_data_rd(0 to 3) /= conv_byte_vector(X"DEADC0DE")) then
        f_io_passed := false;
      end if;

      -- Write io data
      io_data_ls := X"76543210";

      reg_addr := GPIO_ADDR_BASE + XRM_GPIO_DD_DATAO_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(io_data_ls);
      adb3_target_sim_write_reg32(reg_addr,29,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote XRM_GPIO_DD DATAO " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Write io tri-state enable
      reg_addr := GPIO_ADDR_BASE + XRM_GPIO_DD_TRI_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(X"00000000");
      adb3_target_sim_write_reg32(reg_addr,30,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote XRM_GPIO_DD TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Wait for data to get to outputs
      adb3_wait_cycles(50,ocp_clk);

      -- Read io data
      reg_addr := GPIO_ADDR_BASE + XRM_GPIO_DD_DATAI_REG_ADDR;
      adb3_target_sim_read_reg32(reg_addr,31,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read XRM_GPIO_DD DATAI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (conv_vector(reg_data_rd(0 to 3)) /= io_data_ls) then
        f_io_passed := false;
      end if;

      -- Write io tri-state disable
      reg_addr := GPIO_ADDR_BASE + XRM_GPIO_DD_TRI_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(X"FFFFFFFF");
      adb3_target_sim_write_reg32(reg_addr,32,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote XRM_GPIO_DD TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      ds_comp.frontio_complete <= true;
      if (f_io_passed) then
        assert false report "Test Front IO completed: PASSED." severity note;
      else
        assert false report "Test Front IO completed: FAILED." severity error;
        ds_pass.frontio_passed <= false;
      end if;
    end if;


    if PN4_GPIO_VALID then
      -----------------------------------------------
      -- PN4 GPIO Test (64-bit register interface) --
      -----------------------------------------------
      reg_be(0 to 3) := conv_byte_enable("1111");

      -- Write io data
      io_data_ms := PN4_DATA;
      io_data_ls := not(PN4_DATA);

      reg_addr := GPIO_ADDR_BASE + PN4_GPIO_P_DATAO_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(io_data_ms);
      adb3_target_sim_write_reg32(reg_addr,33,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN4_GPIO_P DATAO " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      reg_addr := GPIO_ADDR_BASE + PN4_GPIO_N_DATAO_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(io_data_ls);
      adb3_target_sim_write_reg32(reg_addr,34,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN4_GPIO_N DATAO " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Write io enable
      reg_addr := GPIO_ADDR_BASE + PN4_GPIO_P_TRI_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(X"00000000");
      adb3_target_sim_write_reg32(reg_addr,35,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN4_GPIO_P TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      reg_addr := GPIO_ADDR_BASE + PN4_GPIO_N_TRI_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(X"00000000");
      adb3_target_sim_write_reg32(reg_addr,36,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN4_GPIO_N TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Wait for data to get to outputs
      adb3_wait_cycles(50,ocp_clk);

      -- Read io data
      reg_addr := GPIO_ADDR_BASE + PN4_GPIO_P_DATAI_REG_ADDR;
      adb3_target_sim_read_reg32(reg_addr,37,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read PN4_GPIO_P DATAI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (conv_vector(reg_data_rd(0 to 3))(PN4_GPIO_WIDTH-1 downto 0) /= io_data_ms(PN4_GPIO_WIDTH-1 downto 0)) then
        assert false report "Test PN4 P Rear IO completed: FAILED." severity error;
        r_io_passed := false;
      end if;

      reg_addr := GPIO_ADDR_BASE + PN4_GPIO_N_DATAI_REG_ADDR;
      adb3_target_sim_read_reg32(reg_addr,38,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read PN4_GPIO_N DATAI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (conv_vector(reg_data_rd(0 to 3))(PN4_GPIO_WIDTH-1 downto 0) /= io_data_ls(PN4_GPIO_WIDTH-1 downto 0)) then
        assert false report "Test PN4 N Rear IO completed: FAILED." severity error;
        r_io_passed := false;
      end if;


      -- Write io tri-state disable
      reg_addr := GPIO_ADDR_BASE + PN4_GPIO_P_TRI_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(X"FFFFFFFF");
      adb3_target_sim_write_reg32(reg_addr,39,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN4_GPIO_P TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      reg_addr := GPIO_ADDR_BASE + PN4_GPIO_N_TRI_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(X"FFFFFFFF");
      adb3_target_sim_write_reg32(reg_addr,40,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN4_GPIO_N TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));
    end if;


    if PN6_GPIO_VALID then
      -----------------------------------------------
      -- PN6 GPIO Test (32-bit register interface) --
      -----------------------------------------------
      reg_be(0 to 3) := conv_byte_enable("1111");

      -- Write io data
      io_data_ms := PN6_DATA(63 downto 32);
      io_data_ls := PN6_DATA(31 downto 0);

      reg_addr := GPIO_ADDR_BASE + PN6_GPIO_MS_DATAO_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(io_data_ms);
      adb3_target_sim_write_reg32(reg_addr,41,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN6_GPIO_MS DATAO " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      reg_addr := GPIO_ADDR_BASE + PN6_GPIO_LS_DATAO_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(io_data_ls);
      adb3_target_sim_write_reg32(reg_addr,42,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN6_GPIO_LS DATAO " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Write io enable
      reg_addr := GPIO_ADDR_BASE + PN6_GPIO_MS_TRI_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(X"00000000");
      adb3_target_sim_write_reg32(reg_addr,43,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN6_GPIO_MS TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      reg_addr := GPIO_ADDR_BASE + PN6_GPIO_LS_TRI_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(X"00000000");
      adb3_target_sim_write_reg32(reg_addr,44,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN6_GPIO_LS TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Wait for data to get to outputs
      adb3_wait_cycles(50,ocp_clk);

      -- Read io data
      reg_addr := GPIO_ADDR_BASE + PN6_GPIO_MS_DATAI_REG_ADDR;
      adb3_target_sim_read_reg32(reg_addr,45,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read PN6_GPIO_MS DATAI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      io_data_x2(63 downto 32) := conv_vector(reg_data_rd(0 to 3));

      reg_addr := GPIO_ADDR_BASE + PN6_GPIO_LS_DATAI_REG_ADDR;
      adb3_target_sim_read_reg32(reg_addr,46,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read PN6_GPIO_LS DATAI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      io_data_x2(31 downto 0) := conv_vector(reg_data_rd(0 to 3));

      if (io_data_x2(PN6_GPIO_WIDTH-1 downto 0) /= PN6_DATA(PN6_GPIO_WIDTH-1 downto 0)) then
        assert false report "Test PN6 Rear IO completed: FAILED." severity error;
        r_io_passed := false;
      end if;

      -- Write io tri-state disable
      reg_addr := GPIO_ADDR_BASE + PN6_GPIO_MS_TRI_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(X"FFFFFFFF");
      adb3_target_sim_write_reg32(reg_addr,47,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN6_GPIO_MS TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      reg_addr := GPIO_ADDR_BASE + PN6_GPIO_LS_TRI_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(X"FFFFFFFF");
      adb3_target_sim_write_reg32(reg_addr,48,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN6_GPIO_LS TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      ds_comp.reario_complete <= true;
      if (r_io_passed) then
        assert false report "Test Rear IO completed: PASSED." severity note;
      else
        assert false report "Test Rear IO completed: FAILED." severity error;
        ds_pass.reario_passed <= false;
      end if;
    end if;


    --------------------
    -- Interrupt Test --
    --------------------
    reg_be(0 to 3) := conv_byte_enable("1111");

    -- Read OOR
    reg_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) + 1;
    adb3_target_sim_read(reg_addr,12,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read OOR Addr 0 " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (reg_data_rd(0 to 3) /= conv_byte_vector(X"DEADC0DE")) then
      int_passed := false;
    end if;

    -- Write interrupt MASK
    reg_addr := INT_ADDR_BASE + INT_WMASK_REG_ADDR;
    reg_data_wr(0 to 3) := conv_byte_vector(MASK_EN_ALL);
    adb3_target_sim_write_reg32(reg_addr,49,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Wrote Interrupt MASK " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
           " with enable 0b" & conv_string(reg_be(0 to 3)) &
           " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Read interrupt MASK
    reg_addr := INT_ADDR_BASE + INT_RMASK_REG_ADDR;
    adb3_target_sim_read_reg32(reg_addr,50,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Interrupt MASK " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (reg_data_rd(0 to 3) /= conv_byte_vector(MASK_EN_ALL)) then
      int_passed := false;
    end if;

    -- Write interrupt COUNT
    reg_addr := INT_ADDR_BASE + INT_WCOUNT_REG_ADDR;
    reg_data_wr(0 to 3) := conv_byte_vector(X"FFFFFFFF");
    adb3_target_sim_write_reg32(reg_addr,51,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Wrote Interrupt COUNT " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
           " with enable 0b" & conv_string(reg_be(0 to 3)) &
           " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Read interrupt COUNT
    reg_addr := INT_ADDR_BASE + INT_RCOUNT_REG_ADDR;
    adb3_target_sim_read_reg32(reg_addr,52,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Interrupt COUNT " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (reg_data_rd(0 to 3) /= conv_byte_vector(X"FFFFFFFF")) then
      int_passed := false;
    end if;

    -- Cycle through set-detect-clear for all 32 interrupts
    for i in 0 to (DO_INTERRUPT_NUM-1) loop
      int_set := (others => '0');
      int_set(i) := '1';

      -- Write interrupt SET
      reg_addr := INT_ADDR_BASE + INT_SET_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(int_set);
      adb3_target_sim_write_reg32(reg_addr,53,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
--      report "Wrote Interrupt SET " & integer'image(reg_xfer) &
--             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
--             " with enable 0b" & conv_string(reg_be(0 to 3)) &
--             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      loop
        exit when (linti_l_edge = '1');   -- Wait for next edge on linti_l (Shared variable defined in uber_tb_pkg)
        adb3_wait_cycles(1,ocp_clk);
      end loop;

      -- Interrupt response time of 15 cycles
      -- Unrealistically short for purposes of simulation
      adb3_wait_cycles(15,ocp_clk);

      -- Check interrupt status register
      reg_addr := INT_ADDR_BASE + INT_STAT_REG_ADDR;
      adb3_target_sim_read_reg32(reg_addr,54,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
--      report "Read Interrupt STAT " & integer'image(reg_xfer) &
--             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
--             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));
      int_stat := conv_vector(reg_data_rd(0 to 3));

      -- Mask the STAT register value just read with the value of the MASK register
      int_stat := int_stat and not(MASK_EN_ALL);

      -- Clear the unmasked interrupts
      reg_addr := INT_ADDR_BASE + INT_CLEAR_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(int_stat);
      adb3_target_sim_write_reg32(reg_addr,55,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Interrupt Handler: Cleared interrupt(s), masked STAT = 0x" & conv_string_hex(reg_data_wr(0 to 3));

      -- Re-arm the interrupt logic
      linti_l_edge := '0';
      reg_addr := INT_ADDR_BASE + INT_ARM_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(X"deadbeef"); -- actual value does not matter for ARM register
      adb3_target_sim_write_reg32(reg_addr,56,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    end loop;

    -- Read interrupt STAT data
    reg_addr := INT_ADDR_BASE + INT_STAT_REG_ADDR;
    adb3_target_sim_read_reg32(reg_addr,57,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Interrupt STAT " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (reg_data_rd(0 to 3) /= conv_byte_vector(X"00000000")) then
      int_passed := false;
    end if;

    ds_comp.int_complete <= true;
    if (int_passed) then
      assert false report "Test Interrupt completed: PASSED." severity note;
    else
      assert false report "Test Interrupt completed: FAILED." severity error;
      ds_pass.int_passed <= false;
    end if;


    -------------------------------------------
    -- Info Test (32-bit register interface) --
    -------------------------------------------
    reg_be(0 to 3) := conv_byte_enable("1111");

    -- Read OOR
    reg_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) + 1;
    adb3_target_sim_read(reg_addr,12,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read OOR Addr 0 " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (reg_data_rd(0 to 3) /= conv_byte_vector(X"DEADC0DE")) then
      info_passed := false;
    end if;

    -- Read DATE from register
    reg_addr := INFO_ADDR_BASE + INFO_DATE_REG_ADDR;
    adb3_target_sim_read_reg32(reg_addr,58,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Info DATE " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (conv_vector(reg_data_rd(0 to 3)) /= TODAYS_DATE) then
      info_passed := false;
    end if;

    -- Read TIME from register
    reg_addr := INFO_ADDR_BASE + INFO_TIME_REG_ADDR;
    adb3_target_sim_read_reg32(reg_addr,59,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Info TIME " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (conv_vector(reg_data_rd(0 to 3)) /= TODAYS_TIME) then
      info_passed := false;
    end if;

    -- Read BRAM addr base from register
    reg_addr := INFO_ADDR_BASE + INFO_BRAM_BASE_REG_ADDR;
    adb3_target_sim_read_reg32(reg_addr,60,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Info BRAM BASE " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (conv_vector(reg_data_rd(0 to 3))(DS_ADDR_WIDTH-1 downto 0) /= BRAM_ADDR_BASE(DS_ADDR_WIDTH-1 downto 0)) then
      info_passed := false;
    end if;

    -- Read BRAM addr mask from register
    reg_addr := INFO_ADDR_BASE + INFO_BRAM_MASK_REG_ADDR;
    adb3_target_sim_read_reg32(reg_addr,61,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Info BRAM MASK " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (conv_vector(reg_data_rd(0 to 3))(DS_ADDR_WIDTH-1 downto 0) /= BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0)) then
      info_passed := false;
    end if;

    -- Read RAM addr base from register
    reg_addr := INFO_ADDR_BASE + INFO_RAM_BASE_REG_ADDR;
    adb3_target_sim_read_reg32(reg_addr,60,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Info RAM BASE " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (conv_vector(reg_data_rd(0 to 3))(DS_ADDR_WIDTH-1 downto 0) /= RAM_WIN_ADDR_BASE(DS_ADDR_WIDTH-1 downto 0)) then
      info_passed := false;
    end if;

    -- Read BRAM addr mask from register
    reg_addr := INFO_ADDR_BASE + INFO_RAM_MASK_REG_ADDR;
    adb3_target_sim_read_reg32(reg_addr,61,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Info RAM MASK " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (conv_vector(reg_data_rd(0 to 3))(DS_ADDR_WIDTH-1 downto 0) /= RAM_WIN_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0)) then
      info_passed := false;
    end if;

    -- Read RAM info from register
    reg_addr := INFO_ADDR_BASE + INFO_RAM_INFO_REG_ADDR;
    adb3_target_sim_read_reg32(reg_addr,61,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Info RAM INFO " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (conv_vector(reg_data_rd(0 to 3))(3 downto 0) /= std_logic_vector(to_unsigned(MEM_BANKS,4))) then
      info_passed := false;
    end if;

    ds_comp.info_complete <= true;
    if (info_passed) then
      assert false report "Test Info completed: PASSED." severity note;
    else
      assert false report "Test Info completed: FAILED." severity error;
      ds_pass.info_passed <= false;
    end if;


    -----------------------------------
    -- BRAM Test (128-bit interface) --
    -----------------------------------
    reg_be( 0 to 15) := conv_byte_enable("11111111" & "11111111");
    reg_be(16 to 31) := conv_byte_enable("11111111" & "11111111");

    -- Test lower address limit
    -- Write BRAM
    reg_addr := BRAM_ADDR_BASE;
    reg_data_wr(0 to 3) := conv_byte_vector(X"2389EF45");
    adb3_target_sim_write_reg32(reg_addr,66,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Wrote BRAM Addr base " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
           " with enable 0b" & conv_string(reg_be(0 to 3)) &
           " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Read BRAM
    reg_addr := BRAM_ADDR_BASE;
    adb3_target_sim_read_reg32(reg_addr,67,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read BRAM Addr base " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (reg_data_rd(0 to 3) /= conv_byte_vector(X"2389EF45")) then
      bram_passed := false;
    end if;

    -- Test burst length = 1, 1 x 16-byte transfer
    -- Write BRAM
    reg_addr := BRAM_ADDR_BASE;
    reg_data_wr(0 to 15) := conv_byte_vector(X"56789ABC" & X"DEF12345" & X"6789ABCD" & X"EF123456");
    adb3_target_sim_write(reg_addr,62,reg_be(0 to 15),reg_data_wr(0 to 15),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Wrote BRAM Addr base " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_wr(0 to 15)) &
           " with enable 0b" & conv_string(reg_be(0 to 15)) &
           " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Read BRAM
    reg_addr := BRAM_ADDR_BASE;
    adb3_target_sim_read(reg_addr,63,reg_data_rd(0 to 15),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read BRAM Addr base " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 15)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (reg_data_rd(0 to 15) /= conv_byte_vector(X"56789ABC" & X"DEF12345" & X"6789ABCD" & X"EF123456")) then
      bram_passed := false;
    end if;

    -- Test burst length > 1, 2 x 16-byte transfer
    -- Write BRAM
    reg_addr := BRAM_ADDR_BASE;
    reg_data_wr( 0 to 15) := conv_byte_vector(X"12345678" & X"9ABCDEF1" & X"23456789" & X"ABCDEF12");
    reg_data_wr(16 to 31) := conv_byte_vector(X"FEDCBA98" & X"7654321F" & X"EDCBA987" & X"654321FE");
    adb3_target_sim_write(reg_addr,64,reg_be(0 to 31),reg_data_wr(0 to 31),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Wrote BRAM Addr base " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_wr(0 to 31)) &
           " with enable 0b" & conv_string(reg_be(0 to 31)) &
           " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Read BRAM
    reg_addr := BRAM_ADDR_BASE;
    adb3_target_sim_read(reg_addr,65,reg_data_rd(0 to 31),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read BRAM Addr base " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 31)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (reg_data_rd( 0 to 15) /= conv_byte_vector(X"12345678" & X"9ABCDEF1" & X"23456789" & X"ABCDEF12"))or
       (reg_data_rd(16 to 31) /= conv_byte_vector(X"FEDCBA98" & X"7654321F" & X"EDCBA987" & X"654321FE")) then
      bram_passed := false;
    end if;

    -- Test beyond lower address limit (if > 0)
    -- Write BRAM
    if not(OR_reduce(BRAM_ADDR_BASE) = '0') then
      reg_addr := BRAM_ADDR_BASE - 4;
      reg_data_wr(0 to 3) := conv_byte_vector(X"369CF258");
      adb3_target_sim_write_reg32(reg_addr,68,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote OOR Addr base-4 " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read BRAM
      reg_addr := BRAM_ADDR_BASE - 4;
      adb3_target_sim_read_reg32(reg_addr,69,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read OOR Addr base-4 " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (reg_data_rd(0 to 3) /= conv_byte_vector(X"DEADC0DE")) then
        bram_passed := false;
      end if;
    end if;

    -- Test beyond upper address limit
    -- Write OOR
    reg_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) + 1;
    reg_data_wr(0 to 3) := conv_byte_vector(X"258BE147");
    adb3_target_sim_write_reg32(reg_addr,70,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Wrote OOR Addr top+1 " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
           " with enable 0b" & conv_string(reg_be(0 to 3)) &
           " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Read OOR
    reg_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) + 1;
    adb3_target_sim_read_reg32(reg_addr,71,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read OOR Addr top+1 " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (reg_data_rd(0 to 3) /= conv_byte_vector(X"DEADC0DE")) then
      bram_passed := false;
    end if;

    -- Test burst length > 1, 2 x 16-byte transfer
    -- Write OOR
    reg_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) + 1;
    reg_data_wr( 0 to 15) := conv_byte_vector(X"12345678" & X"9ABCDEF1" & X"23456789" & X"ABCDEF12");
    reg_data_wr(16 to 31) := conv_byte_vector(X"FEDCBA98" & X"7654321F" & X"EDCBA987" & X"654321FE");
    adb3_target_sim_write(reg_addr,64,reg_be(0 to 31),reg_data_wr(0 to 31),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Wrote OOR Addr top+1 " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_wr(0 to 31)) &
           " with enable 0b" & conv_string(reg_be(0 to 31)) &
           " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Read OOR
    reg_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) + 1;
    adb3_target_sim_read(reg_addr,65,reg_data_rd(0 to 31),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read OOR Addr top+1 " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 31)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (reg_data_rd( 0 to 15) /= conv_byte_vector(X"DEADC0DE" & X"DEADC0DE" & X"DEADC0DE" & X"DEADC0DE"))or
       (reg_data_rd(16 to 31) /= conv_byte_vector(X"DEADC0DE" & X"DEADC0DE" & X"DEADC0DE" & X"DEADC0DE")) then
      bram_passed := false;
    end if;

    -- Test upper address limit
    -- Write BRAM
    reg_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) - 3;
    reg_data_wr(0 to 3) := conv_byte_vector(X"147AD036");
    adb3_target_sim_write_reg32(reg_addr,66,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Wrote BRAM Addr top " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
           " with enable 0b" & conv_string(reg_be(0 to 3)) &
           " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Read BRAM
    reg_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) - 3;
    adb3_target_sim_read_reg32(reg_addr,67,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read BRAM Addr top " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (reg_data_rd(0 to 3) /= conv_byte_vector(X"147AD036")) then
      bram_passed := false;
    end if;


    ds_comp.bram_complete <= true;
    if (bram_passed) then
      assert false report "Test BRAM completed: PASSED." severity note;
    else
      assert false report "Test BRAM completed: FAILED." severity error;
      ds_pass.bram_passed <= false;
    end if;


    if MEM_VALID then
      -----------------------------------
      -- RAM Test (128-bit interface) --
      -----------------------------------
      -- Start internal memory test
      reg_be(0 to 3) := conv_byte_enable("1111");

      report "Waiting for on-board RAM bank " & integer'image(TEST_RAM_BANK) & " to initialise...";

      -- Wait until internal RAM bank init done
      ram_init := '0';
      while (ram_init = '0') loop
        reg_addr := RAM_REG_ADDR_BASE + RAM_BANK0_STAT_REG_ADDR + TEST_RAM_BANK*32;
        adb3_target_sim_read_reg32(reg_addr,0,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
--        report "Read RAM Bank Stat Reg " & integer'image(reg_xfer) &
--               " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
--               " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));
        ram_init := reg_data_rd(0)(3);
      end loop;

      report "On-board RAM bank " & integer'image(TEST_RAM_BANK) & " initialised";

      -- Read RAM Bank Info register
      reg_addr := RAM_REG_ADDR_BASE + RAM_BANK0_INFO_REG_ADDR + TEST_RAM_BANK*32;
      adb3_target_sim_read_reg32(reg_addr,1,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM Bank Info Reg " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));


      if DO_INT_RAM_TEST then
        -- Write RAM test offset (16-byte words)
        -- Do not use last 16-byte word as this is will conflict with DS PAGE TOP test below
        reg_addr := RAM_REG_ADDR_BASE + RAM_BANK0_OFFSET_REG_ADDR + TEST_RAM_BANK*32;
        reg_data_wr(0 to 3) := conv_byte_vector(RAM_TEST_OFF);
        adb3_target_sim_write_reg32(reg_addr,2,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        report "Wrote RAM Bank Offset Reg " & integer'image(reg_xfer) &
               " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
               " with enable 0b" & conv_string(reg_be(0 to 3)) &
               " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

        -- Read RAM test offset
        reg_addr := RAM_REG_ADDR_BASE + RAM_BANK0_OFFSET_REG_ADDR + TEST_RAM_BANK*32;
        adb3_target_sim_read_reg32(reg_addr,3,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        report "Read RAM Bank Offset Reg " & integer'image(reg_xfer) &
               " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
               " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

        -- Write RAM test length-1 (16-byte words)
        reg_addr := RAM_REG_ADDR_BASE + RAM_BANK0_LENGTH_REG_ADDR + TEST_RAM_BANK*32;
        reg_data_wr(0 to 3) := conv_byte_vector(RAM_TEST_LEN);
        adb3_target_sim_write_reg32(reg_addr,4,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        report "Wrote RAM Bank Length Reg " & integer'image(reg_xfer) &
               " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
               " with enable 0b" & conv_string(reg_be(0 to 3)) &
               " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

        -- Read RAM test length
        reg_addr := RAM_REG_ADDR_BASE + RAM_BANK0_LENGTH_REG_ADDR + TEST_RAM_BANK*32;
        adb3_target_sim_read_reg32(reg_addr,5,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        report "Read RAM Bank Length Reg " & integer'image(reg_xfer) &
               " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
               " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

        -- Start internal test
        -- Write ram ctrl
        reg_addr := RAM_REG_ADDR_BASE + RAM_BANK0_CTRL_REG_ADDR + TEST_RAM_BANK*32;
        reg_data_wr(0 to 3) := conv_byte_vector(X"00000100");
        adb3_target_sim_write_reg32(reg_addr,6,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        report "Wrote RAM Bank Ctrl Reg " & integer'image(reg_xfer) &
               " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
               " with enable 0b" & conv_string(reg_be(0 to 3)) &
               " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));
      end if;

      -- Set up access to RAMs using DS window
      reg_be(0 to 3) := conv_byte_enable("1111");

      -- Write DS bank
      reg_addr := RAM_REG_ADDR_BASE + RAM_DS_BANK_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(DS_WIN_RAM_BANK);
      adb3_target_sim_write_reg32(reg_addr,8,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM DS Bank Reg Addr " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Write DS bottom page
      reg_addr := RAM_REG_ADDR_BASE + RAM_DS_PAGE_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(DS_WIN_RAM_PAGE_BOT);
      adb3_target_sim_write_reg32(reg_addr,9,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM DS Page Reg Addr " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Test lower address limit
      -- Write RAM
      reg_addr := RAM_WIN_ADDR_BASE;
      reg_data_wr(0 to 3) := conv_byte_vector(X"349AF056");
      adb3_target_sim_write_reg32(reg_addr,10,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read RAM
      reg_addr := RAM_WIN_ADDR_BASE;
      adb3_target_sim_read_reg32(reg_addr,11,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (reg_data_rd(0 to 3) /= conv_byte_vector(X"349AF056")) then
        ram_passed := false;
      end if;

      -- Write DS top page
      reg_addr := RAM_REG_ADDR_BASE + RAM_DS_PAGE_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(DS_WIN_RAM_PAGE_TOP);
      adb3_target_sim_write_reg32(reg_addr,9,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM DS Page Reg Addr " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Test upper address limit
      -- Write RAM
      reg_addr := RAM_WIN_ADDR_BASE + RAM_WIN_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) - 3;
      reg_data_wr(0 to 3) := conv_byte_vector(X"47AD0369");
      adb3_target_sim_write_reg32(reg_addr,12,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr top " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read RAM
      reg_addr := RAM_WIN_ADDR_BASE + RAM_WIN_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) - 3;
      adb3_target_sim_read_reg32(reg_addr,13,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM Win Addr top " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (reg_data_rd(0 to 3) /= conv_byte_vector(X"47AD0369")) then
        ram_passed := false;
      end if;

      -- Write DS bottom page
      reg_addr := RAM_REG_ADDR_BASE + RAM_DS_PAGE_REG_ADDR;
      reg_data_wr(0 to 3) := conv_byte_vector(DS_WIN_RAM_PAGE_BOT);
      adb3_target_sim_write_reg32(reg_addr,9,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM DS Page Reg Addr " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      reg_be( 0 to 15) := conv_byte_enable("11111111" & "11111111");
      reg_be(16 to 31) := conv_byte_enable("11111111" & "11111111");
      reg_be(32 to 47) := conv_byte_enable("11111111" & "11111111");

      -- Test even burst length, 32-byte aligned
      -- Write RAM
      reg_addr := RAM_WIN_ADDR_BASE;
      reg_data_wr( 0 to 15) := conv_byte_vector(X"12345678" & X"9ABCDEF1" & X"23456789" & X"ABCDEF12");
      reg_data_wr(16 to 31) := conv_byte_vector(X"FEDCBA98" & X"7654321F" & X"EDCBA987" & X"654321FE");
      adb3_target_sim_write(reg_addr,14,reg_be(0 to 31),reg_data_wr(0 to 31),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 31)) &
             " with enable 0b" & conv_string(reg_be(0 to 31)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read RAM
      reg_addr := RAM_WIN_ADDR_BASE;
      adb3_target_sim_read(reg_addr,15,reg_data_rd(0 to 31),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 31)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (reg_data_rd( 0 to 15) /= conv_byte_vector(X"12345678" & X"9ABCDEF1" & X"23456789" & X"ABCDEF12"))or
         (reg_data_rd(16 to 31) /= conv_byte_vector(X"FEDCBA98" & X"7654321F" & X"EDCBA987" & X"654321FE")) then
        ram_passed := false;
      end if;

      -- Test odd burst length, 32-byte aligned
      -- Write RAM
      reg_addr := RAM_WIN_ADDR_BASE;
      reg_data_wr( 0 to 15) := conv_byte_vector(X"23456789" & X"ABCDEF12" & X"3456789A" & X"BCDEF123");
      reg_data_wr(16 to 31) := conv_byte_vector(X"EDCBA987" & X"654321FE" & X"DCBA9876" & X"54321FED");
      reg_data_wr(32 to 47) := conv_byte_vector(X"3456789A" & X"BCDEF123" & X"456789AB" & X"CDEF1234");
      adb3_target_sim_write(reg_addr,14,reg_be(0 to 47),reg_data_wr(0 to 47),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 47)) &
             " with enable 0b" & conv_string(reg_be(0 to 47)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read RAM
      reg_addr := RAM_WIN_ADDR_BASE;
      adb3_target_sim_read(reg_addr,15,reg_data_rd(0 to 47),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 47)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (reg_data_rd( 0 to 15) /= conv_byte_vector(X"23456789" & X"ABCDEF12" & X"3456789A" & X"BCDEF123"))or
         (reg_data_rd(16 to 31) /= conv_byte_vector(X"EDCBA987" & X"654321FE" & X"DCBA9876" & X"54321FED"))or
         (reg_data_rd(32 to 47) /= conv_byte_vector(X"3456789A" & X"BCDEF123" & X"456789AB" & X"CDEF1234")) then
        ram_passed := false;
      end if;

      -- Test even burst length, non 32-byte aligned
      -- Write RAM
      reg_addr := RAM_WIN_ADDR_BASE+16;
      reg_data_wr( 0 to 15) := conv_byte_vector(X"456789AB" & X"CDEF1234" & X"56789ABC" & X"DEF12345");
      reg_data_wr(16 to 31) := conv_byte_vector(X"DCBA9876" & X"54321FED" & X"CBA98765" & X"4321FEDC");
      adb3_target_sim_write(reg_addr,14,reg_be(0 to 31),reg_data_wr(0 to 31),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 31)) &
             " with enable 0b" & conv_string(reg_be(0 to 31)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read RAM
      reg_addr := RAM_WIN_ADDR_BASE+16;
      adb3_target_sim_read(reg_addr,15,reg_data_rd(0 to 31),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 31)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (reg_data_rd( 0 to 15) /= conv_byte_vector(X"456789AB" & X"CDEF1234" & X"56789ABC" & X"DEF12345"))or
         (reg_data_rd(16 to 31) /= conv_byte_vector(X"DCBA9876" & X"54321FED" & X"CBA98765" & X"4321FEDC")) then
        ram_passed := false;
      end if;

      -- Test odd burst length, non 32-byte aligned
      -- Write RAM
      reg_addr := RAM_WIN_ADDR_BASE+16;
      reg_data_wr( 0 to 15) := conv_byte_vector(X"56789ABC" & X"DEF12345" & X"6789ABCD" & X"EF123456");
      reg_data_wr(16 to 31) := conv_byte_vector(X"DCBA9876" & X"54321FED" & X"CBA98765" & X"4321FEDC");
      reg_data_wr(32 to 47) := conv_byte_vector(X"6789ABCD" & X"EF123456" & X"789ABCDE" & X"F1234567");
      adb3_target_sim_write(reg_addr,14,reg_be(0 to 47),reg_data_wr(0 to 47),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 47)) &
             " with enable 0b" & conv_string(reg_be(0 to 47)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read RAM
      reg_addr := RAM_WIN_ADDR_BASE+16;
      adb3_target_sim_read(reg_addr,15,reg_data_rd(0 to 47),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 47)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (reg_data_rd( 0 to 15) /= conv_byte_vector(X"56789ABC" & X"DEF12345" & X"6789ABCD" & X"EF123456"))or
         (reg_data_rd(16 to 31) /= conv_byte_vector(X"DCBA9876" & X"54321FED" & X"CBA98765" & X"4321FEDC"))or
         (reg_data_rd(32 to 47) /= conv_byte_vector(X"6789ABCD" & X"EF123456" & X"789ABCDE" & X"F1234567")) then
        ram_passed := false;
      end if;

      -- Test byte enables
      -- Write RAM
      reg_addr := RAM_WIN_ADDR_BASE;
      reg_be(0 to 3) := conv_byte_enable("1000");
      reg_data_wr(0 to 3) := conv_byte_vector(X"45000000");
      adb3_target_sim_write(reg_addr,16,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      reg_addr := RAM_WIN_ADDR_BASE;
      reg_be(0 to 3) := conv_byte_enable("0100");
      reg_data_wr(0 to 3) := conv_byte_vector(X"00AB0000");
      adb3_target_sim_write(reg_addr,17,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      reg_addr := RAM_WIN_ADDR_BASE;
      reg_be(0 to 3) := conv_byte_enable("0010");
      reg_data_wr(0 to 3) := conv_byte_vector(X"00000100");
      adb3_target_sim_write(reg_addr,18,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      reg_addr := RAM_WIN_ADDR_BASE;
      reg_be(0 to 3) := conv_byte_enable("0001");
      reg_data_wr(0 to 3) := conv_byte_vector(X"00000067");
      adb3_target_sim_write(reg_addr,19,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(reg_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read RAM
      reg_addr := RAM_WIN_ADDR_BASE;
      adb3_target_sim_read_reg32(reg_addr,20,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (reg_data_rd(0 to 3) /= conv_byte_vector(X"45AB0167")) then
        ram_passed := false;
      end if;

      if DO_INT_RAM_TEST then
        report "Waiting for internal test of on-board RAM bank " & integer'image(TEST_RAM_BANK) & " to complete...";

        -- Wait until internal RAM test done by reading done status
        ram_done := '0';
        while (ram_done = '0') loop
          reg_addr := RAM_REG_ADDR_BASE + RAM_BANK0_STAT_REG_ADDR + TEST_RAM_BANK*32;
          adb3_target_sim_read_reg32(reg_addr,21,reg_data_rd(0 to 3),reg_xfer,ocp_clk,ds_m2s,ds_s2m);
--          report "Read RAM Bank Stat Reg " & integer'image(reg_xfer) &
--                 " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
--                 " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));
          ram_done := reg_data_rd(2)(0);
        end loop;

        report "Read RAM Bank Stat Reg " & integer'image(reg_xfer) &
               " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
               " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

        report "Internal test of on-board RAM bank " & integer'image(TEST_RAM_BANK) & " complete";

        -- Test stat bit 23 = mem_app_err
        if (reg_data_rd(2)(7) = '1') then
          report "Internal test of on-board RAM bank " & integer'image(TEST_RAM_BANK) & " failed";
          ram_passed := false;
        end if;
      end if;

      ds_comp.ram_complete <= true;
      if (ram_passed) then
        assert false report "Test RAM completed: PASSED." severity note;
      else
        assert false report "Test RAM completed: FAILED." severity error;
        ds_pass.ram_passed <= false;
      end if;
    end if;

    wait;
  end process;

  -- Connect direct slave output
  direct_slave_m2s <= ds_m2s;

end test;