--==============================================================================
--
--     File Name: uber_pkg.vhd
--       Version: 0.2
--          Date: 09 Feb 2011
--         Model: Synthesisable
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Package uber_pkg
--
--   Description: Package of components used in uber target
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 25 Mar 10: Initial design
--                0.2, TW, 09 Feb 2011:
--                  - Now references adb3_ocp_comp package instead of
--                    adb3_target_pkg, for certain definitions.
--                  - pll_ref_clk renamed to pll_pri_clk (primary clock).
--                  - pll_ref_clk is now a (guaranteed) 1x copy of the reference
--                    clock used to derived pll_pri_clk and others.
--                  - Deleted ref_clk port from blk_clocks.
--                  - FPGA interrupt signal is now active low, to facilitate
--                    packing into IOB.
--                  - Added "instantiate" generic to blk_chipscope component,
--                    to allow instantiation of chipscope cores to easily be
--                    enabled or disabled.
--                  - Created address range table DS_DDR3_ADDR_RANGE_TABLE that
--                    is used when instantiating a splitter for direct slave
--                    access to DDR3 memory banks.
--
--==============================================================================

--
-- Exported components:
--
--   blk_clocks
--   blk_direct_slave
--   blk_ds_simple_test
--   blk_ds_master_test
--   blk_ds_clk_freq
--   blk_ds_io_test
--   blk_ds_int_test
--   blk_ds_ram_reg
--   blk_ds_info
--   blk_dma_switch
--   blk_bram
--   blk_ddr3_if
--   blk_mem_app
--   blk_chipscope
--   blk_clock_freq
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_comp.all;
use work.adb3_target_pkg.all;
use work.adb3_target_inc_pkg.all;
use work.mem_if_pkg.all;

package uber_pkg is

  ---------------------------------------
  -- Top level signal type definitions --
  ---------------------------------------

  -- Clock inputs
  type clks_in_t is record
    ref_clk      : std_logic_dbl_t;
    lclk         : std_logic;
    xrm_gclk_m2c : std_logic_dbl_t;
  end record;

  -- MGT clock inputs
  type clks_mgt_in_t is record
    mgt112_clk0 : std_logic_dbl_t;
    mgt113_clk0 : std_logic_dbl_t;
    mgt113_clk1 : std_logic_dbl_t;
    mgt114_clk0 : std_logic_dbl_t;
    mgt115_clk0 : std_logic_dbl_t;
    mgt116_clk0 : std_logic_dbl_t;
    mgt117_clk0 : std_logic_dbl_t;
  end record;

  -- Clock outputs
  type clks_out_t is record
    xrm_mgtclk_c2m : std_logic_dbl_t;
  end record;


  -- XRM GPIO interface
  type xrm_gpio_t is record
    xrm_da_p    : std_logic_vector(15 downto 0);
    xrm_da_n    : std_logic_vector(15 downto 0);
    xrm_db_p    : std_logic_vector(15 downto 0);
    xrm_db_n    : std_logic_vector(15 downto 0);
    xrm_dc_p    : std_logic_vector(15 downto 0);
    xrm_dc_n    : std_logic_vector(15 downto 0);
    xrm_dd_p    : std_logic_vector(15 downto 0);
    xrm_dd_n    : std_logic_vector(15 downto 0);
    xrm_da_cc_p : std_logic;
    xrm_da_cc_n : std_logic;
    xrm_db_cc_p : std_logic;
    xrm_db_cc_n : std_logic;
    xrm_dc_cc_p : std_logic;
    xrm_dc_cc_n : std_logic;
    xrm_dd_cc_p : std_logic;
    xrm_dd_cc_n : std_logic;
    xrm_sa      : std_logic_vector(1 downto 0);
    xrm_sb      : std_logic_vector(1 downto 0);
    xrm_sc      : std_logic_vector(1 downto 0);
    xrm_sd      : std_logic_vector(3 downto 0);
  end record;

  -- PN4 GPIO interface
  type pn4_gpio_t is record
    gpio_p : std_logic_vector(PN4_GPIO_WIDTH downto 1);
    gpio_n : std_logic_vector(PN4_GPIO_WIDTH downto 1);
  end record;

  constant PN4_GPIO_Z : pn4_gpio_t :=
    (gpio_p => (others => 'Z'),
     gpio_n => (others => 'Z'));

  -- PN6 GPIO interface
  type pn6_gpio_t is record
    gpio : std_logic_vector(PN6_GPIO_WIDTH downto 1);
  end record;

  constant PN6_GPIO_Z : pn6_gpio_t :=
    (gpio => (others => 'Z'));

  -- GPIO interface
  type gpio_inout_t is record
    xrm_gpio : xrm_gpio_t;
    pn4_gpio : pn4_gpio_t;
    pn6_gpio : pn6_gpio_t;
  end record;


  ---------------------------------------
  -- Direct Slave interface memory map --
  ---------------------------------------

  constant SIMPLE_ADDR_BASE    : adb3_ocp_addr_s := ds_base_conv(X"000000");  -- Used by Simple test application
  constant SIMPLE_ADDR_MASK    : adb3_ocp_addr_s := ds_mask_conv(X"00003F");
  constant CLK_READ_ADDR_BASE  : adb3_ocp_addr_s := ds_base_conv(X"000040");  -- Used by Clock Read application
  constant CLK_READ_ADDR_MASK  : adb3_ocp_addr_s := ds_mask_conv(X"00003F");
  constant MASTER_ADDR_BASE    : adb3_ocp_addr_s := ds_base_conv(X"000080");  -- Unused
  constant MASTER_ADDR_MASK    : adb3_ocp_addr_s := ds_mask_conv(X"00003F");
  constant INT_ADDR_BASE       : adb3_ocp_addr_s := ds_base_conv(X"0000C0");  -- Used by Interrupt test application
  constant INT_ADDR_MASK       : adb3_ocp_addr_s := ds_mask_conv(X"00003F");

  constant INFO_ADDR_BASE      : adb3_ocp_addr_s := ds_base_conv(X"000140");  -- Used by multiple applications
  constant INFO_ADDR_MASK      : adb3_ocp_addr_s := ds_mask_conv(X"00003F");

  constant GPIO_ADDR_BASE      : adb3_ocp_addr_s := ds_base_conv(X"000200");  -- Used by GPIO test applications
  constant GPIO_ADDR_MASK      : adb3_ocp_addr_s := ds_mask_conv(X"00007F");

  constant RAM_REG_ADDR_BASE   : adb3_ocp_addr_s := ds_base_conv(X"000300");  -- Used by RAM test application
  constant RAM_REG_ADDR_MASK   : adb3_ocp_addr_s := ds_mask_conv(X"0000FF");

  constant BRAM_ADDR_BASE      : adb3_ocp_addr_s := ds_base_conv(X"080000");  -- Used by DMA test application
  constant BRAM_ADDR_MASK      : adb3_ocp_addr_s := ds_mask_conv(X"07FFFF");

  constant RAM_WIN_ADDR_BASE   : adb3_ocp_addr_s := ds_base_conv(X"200000");  -- Used by Memory test application
  constant RAM_WIN_ADDR_MASK   : adb3_ocp_addr_s := ds_mask_conv(X"1FFFFF");


  constant SIMPLE_ADDR_RANGE   : addr_range_t := (SIMPLE_ADDR_BASE, SIMPLE_ADDR_MASK);
  constant CLK_READ_ADDR_RANGE : addr_range_t := (CLK_READ_ADDR_BASE, CLK_READ_ADDR_MASK);
  constant GPIO_ADDR_RANGE     : addr_range_t := (GPIO_ADDR_BASE, GPIO_ADDR_MASK);
  constant INT_ADDR_RANGE      : addr_range_t := (INT_ADDR_BASE, INT_ADDR_MASK);
  constant RAM_REG_ADDR_RANGE  : addr_range_t := (RAM_REG_ADDR_BASE, RAM_REG_ADDR_MASK);
  constant MASTER_ADDR_RANGE   : addr_range_t := (MASTER_ADDR_BASE, MASTER_ADDR_MASK);
  constant INFO_ADDR_RANGE     : addr_range_t := (INFO_ADDR_BASE, INFO_ADDR_MASK);
  constant BRAM_ADDR_RANGE     : addr_range_t := (BRAM_ADDR_BASE, BRAM_ADDR_MASK);
  constant RAM_WIN_ADDR_RANGE  : addr_range_t := (RAM_WIN_ADDR_BASE, RAM_WIN_ADDR_MASK);

  constant DS_ADDR_RANGE_TABLE : addr_range_table_t(0 to 7) := (0 => SIMPLE_ADDR_RANGE,
                                                                1 => CLK_READ_ADDR_RANGE,
                                                                2 => GPIO_ADDR_RANGE,
                                                                3 => INT_ADDR_RANGE,
                                                                4 => RAM_REG_ADDR_RANGE,
                                                                5 => INFO_ADDR_RANGE,
                                                                6 => BRAM_ADDR_RANGE,
                                                                7 => RAM_WIN_ADDR_RANGE);



  ----------------------------
  -- Simple test memory map --
  ----------------------------

  -- =============================================================================
  -- Offset Type Name    Function
  -- =============================================================================
  --
  -- 0x00   WR   DATA    Simple Write/Read data
  --
  -- =============================================================================

  -- Register offsets from base address in bytes, eg 0,4,8,... for 32 bit registers
  constant SIMPLE_WDATA_REG_OFF : natural := 16#00#;
  constant SIMPLE_RDATA_REG_OFF : natural := 16#00#;

  constant SIMPLE_WDATA_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(SIMPLE_WDATA_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant SIMPLE_RDATA_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(SIMPLE_RDATA_REG_OFF,ADB3_OCP_ADDR_WIDTH));


  -------------------------------------------------
  -- Clock read frequency measurement memory map --
  -------------------------------------------------

  -- =============================================================================
  -- Offset Type Name    Function
  -- =============================================================================
  --
  -- 0x00   WR   SEL     Selects STAT and FREQ register clock as follows:
  --
  --                       [4:0] select: 00000 => pll_reg_clk (Internal register OCP clock)
  --                                     00001 => pll_pri_clk (Internal primary OCP clock)
  --                                     00010 => pll_ref_clk (Internal ref_clk clock)
  --                                      ..
  --                                     01100 => lclk        (External)
  --                                     01101 => xrm_clkin   (External)
  --                                     10010 => mgt112_clk0 (External)
  --                                     10100 => mgt113_clk0 (External)
  --                                     10101 => mgt113_clk1 (External)
  --                                     10110 => mgt114_clk0 (External)
  --                                     11010 => mgt116_clk0 (External)
  --                                     11100 => mgt117_clk0 (External)
  --
  -- 0x04   W    CTRL    All clock frequency measurement control as follows:
  --
  --                       [31] Clear all measurement updated flags
  --
  -- 0x04   R    STAT    Selected clock frequency measurement status as follows:
  --
  --                       [31] Measurement status: 1 => measurement updated
  --                                                0 => measurement not updated
  --                       [30] Clock valid for board
  --                       [29] Clock running
  --
  -- 0x08   RO   FREQ    Selected clock frequency measurement as follows:
  --
  --                       [31:0] Clock frequency (Hz)
  --
  -- =============================================================================

  -- Register offsets from base address in bytes, eg 0,4,8,... for 32 bit registers
  constant CLK_READ_SEL_REG_OFF    : natural := 16#00#;
  constant CLK_READ_CTRL_REG_OFF   : natural := 16#04#;
  constant CLK_READ_STAT_REG_OFF   : natural := 16#04#;
  constant CLK_READ_FREQ_REG_OFF   : natural := 16#08#;

  constant CLK_READ_SEL_REG_ADDR   : adb3_ocp_addr_s := std_logic_vector(to_unsigned(CLK_READ_SEL_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant CLK_READ_CTRL_REG_ADDR  : adb3_ocp_addr_s := std_logic_vector(to_unsigned(CLK_READ_CTRL_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant CLK_READ_STAT_REG_ADDR  : adb3_ocp_addr_s := std_logic_vector(to_unsigned(CLK_READ_STAT_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant CLK_READ_FREQ_REG_ADDR  : adb3_ocp_addr_s := std_logic_vector(to_unsigned(CLK_READ_FREQ_REG_OFF,ADB3_OCP_ADDR_WIDTH));


  ------------------------
  -- IO test memory map --
  ------------------------

  -- =============================================================================
  -- Offset Type Name    Function
  -- =============================================================================
  --
  -- 0x00   WR   DA_DATAO  XRM GPIO DA data output (31:0)
  -- 0x04   RO   DA_DATAI  XRM GPIO DA data input  (31:0)
  -- 0x08   WR   DA_TRI    XRM GPIO DA data output enables (31:0) (active low)
  -- 0x0C   WR   DB_DATAO  XRM GPIO DB data output (31:0)
  -- 0x10   RO   DB_DATAI  XRM GPIO DB data input  (31:0)
  -- 0x14   WR   DB_TRI    XRM GPIO DB data output enables (31:0) (active low)
  -- 0x18   WR   DC_DATAO  XRM GPIO DC data output (31:0)
  -- 0x1C   RO   DC_DATAI  XRM GPIO DC data input  (31:0)
  -- 0x20   WR   DC_TRI    XRM GPIO DC data output enables (31:0) (active low)
  -- 0x24   WR   DD_DATAO  XRM GPIO DD data output (31:0)
  -- 0x28   RO   DD_DATAI  XRM GPIO DD data input  (31:0)
  -- 0x2C   WR   DD_TRI    XRM GPIO DD data output enables (31:0) (active low)
  -- 0x30   WR   CS_DATAO  XRM GPIO CS data output (31:0)
  -- 0x34   RO   CS_DATAI  XRM GPIO CS data input  (31:0)
  -- 0x38   WR   CS_TRI    XRM GPIO CS data output enables (31:0) (active low)
  --
  -- 0x3C   WR   PN4_P_DATAO  XRM GPIO PN4_P  data output (PN4_GPIO_WIDTH-1:0)
  -- 0x40   RO   PN4_P_DATAI  XRM GPIO PN4_P  data input  (PN4_GPIO_WIDTH-1:0)
  -- 0x44   WR   PN4_P_TRI    XRM GPIO PN4_P  data output enables (PN4_GPIO_WIDTH-1:0) (active low)
  -- 0x48   WR   PN4_N_DATAO  XRM GPIO PN4_N  data output (PN4_GPIO_WIDTH-1:0)
  -- 0x4C   RO   PN4_N_DATAI  XRM GPIO PN4_N  data input  (PN4_GPIO_WIDTH-1:0)
  -- 0x50   WR   PN4_N_TRI    XRM GPIO PN4_N  data output enables (PN4_GPIO_WIDTH-1:0) (active low)
  --
  -- 0x54   WR   PN6_MS_DATAO XRM GPIO PN6_MS data output (PN6_GPIO_WIDTH-1:0)
  -- 0x58   RO   PN6_MS_DATAI XRM GPIO PN6_MS data input  (PN6_GPIO_WIDTH-1:0)
  -- 0x5C   WR   PN6_MS_TRI   XRM GPIO PN6_MS data output enables (PN6_GPIO_WIDTH-1:0) (active low)
  -- 0x60   WR   PN6_LS_DATAO XRM GPIO PN6_LS data output (PN6_GPIO_WIDTH-1:0)
  -- 0x64   RO   PN6_LS_DATAI XRM GPIO PN6_LS data input  (PN6_GPIO_WIDTH-1:0)
  -- 0x68   WR   PN6_LS_TRI   XRM GPIO PN6_LS data output enables (PN6_GPIO_WIDTH-1:0) (active low)
  --
  -- =============================================================================

  -- Register offsets from base address in bytes, eg 0,4,8,... for 32 bit registers
  constant XRM_GPIO_DA_DATAO_REG_OFF : natural := 16#00#;
  constant XRM_GPIO_DA_DATAI_REG_OFF : natural := 16#04#;
  constant XRM_GPIO_DA_TRI_REG_OFF   : natural := 16#08#;

  constant XRM_GPIO_DB_DATAO_REG_OFF : natural := 16#0C#;
  constant XRM_GPIO_DB_DATAI_REG_OFF : natural := 16#10#;
  constant XRM_GPIO_DB_TRI_REG_OFF   : natural := 16#14#;

  constant XRM_GPIO_DC_DATAO_REG_OFF : natural := 16#18#;
  constant XRM_GPIO_DC_DATAI_REG_OFF : natural := 16#1C#;
  constant XRM_GPIO_DC_TRI_REG_OFF   : natural := 16#20#;

  constant XRM_GPIO_DD_DATAO_REG_OFF : natural := 16#24#;
  constant XRM_GPIO_DD_DATAI_REG_OFF : natural := 16#28#;
  constant XRM_GPIO_DD_TRI_REG_OFF   : natural := 16#2C#;

  constant XRM_GPIO_CS_DATAO_REG_OFF : natural := 16#30#;
  constant XRM_GPIO_CS_DATAI_REG_OFF : natural := 16#34#;
  constant XRM_GPIO_CS_TRI_REG_OFF   : natural := 16#38#;

  constant PN4_GPIO_P_DATAO_REG_OFF  : natural := 16#3C#;
  constant PN4_GPIO_P_DATAI_REG_OFF  : natural := 16#40#;
  constant PN4_GPIO_P_TRI_REG_OFF    : natural := 16#44#;

  constant PN4_GPIO_N_DATAO_REG_OFF  : natural := 16#48#;
  constant PN4_GPIO_N_DATAI_REG_OFF  : natural := 16#4C#;
  constant PN4_GPIO_N_TRI_REG_OFF    : natural := 16#50#;

  constant PN6_GPIO_MS_DATAO_REG_OFF : natural := 16#54#;
  constant PN6_GPIO_MS_DATAI_REG_OFF : natural := 16#58#;
  constant PN6_GPIO_MS_TRI_REG_OFF   : natural := 16#5C#;

  constant PN6_GPIO_LS_DATAO_REG_OFF : natural := 16#60#;
  constant PN6_GPIO_LS_DATAI_REG_OFF : natural := 16#64#;
  constant PN6_GPIO_LS_TRI_REG_OFF   : natural := 16#68#;

  constant XRM_GPIO_DA_DATAO_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(XRM_GPIO_DA_DATAO_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant XRM_GPIO_DA_DATAI_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(XRM_GPIO_DA_DATAI_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant XRM_GPIO_DA_TRI_REG_ADDR   : adb3_ocp_addr_s := std_logic_vector(to_unsigned(XRM_GPIO_DA_TRI_REG_OFF,ADB3_OCP_ADDR_WIDTH));

  constant XRM_GPIO_DB_DATAO_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(XRM_GPIO_DB_DATAO_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant XRM_GPIO_DB_DATAI_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(XRM_GPIO_DB_DATAI_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant XRM_GPIO_DB_TRI_REG_ADDR   : adb3_ocp_addr_s := std_logic_vector(to_unsigned(XRM_GPIO_DB_TRI_REG_OFF,ADB3_OCP_ADDR_WIDTH));

  constant XRM_GPIO_DC_DATAO_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(XRM_GPIO_DC_DATAO_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant XRM_GPIO_DC_DATAI_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(XRM_GPIO_DC_DATAI_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant XRM_GPIO_DC_TRI_REG_ADDR   : adb3_ocp_addr_s := std_logic_vector(to_unsigned(XRM_GPIO_DC_TRI_REG_OFF,ADB3_OCP_ADDR_WIDTH));

  constant XRM_GPIO_DD_DATAO_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(XRM_GPIO_DD_DATAO_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant XRM_GPIO_DD_DATAI_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(XRM_GPIO_DD_DATAI_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant XRM_GPIO_DD_TRI_REG_ADDR   : adb3_ocp_addr_s := std_logic_vector(to_unsigned(XRM_GPIO_DD_TRI_REG_OFF,ADB3_OCP_ADDR_WIDTH));

  constant XRM_GPIO_CS_DATAO_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(XRM_GPIO_CS_DATAO_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant XRM_GPIO_CS_DATAI_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(XRM_GPIO_CS_DATAI_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant XRM_GPIO_CS_TRI_REG_ADDR   : adb3_ocp_addr_s := std_logic_vector(to_unsigned(XRM_GPIO_CS_TRI_REG_OFF,ADB3_OCP_ADDR_WIDTH));

  constant PN4_GPIO_P_DATAO_REG_ADDR  : adb3_ocp_addr_s := std_logic_vector(to_unsigned(PN4_GPIO_P_DATAO_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant PN4_GPIO_P_DATAI_REG_ADDR  : adb3_ocp_addr_s := std_logic_vector(to_unsigned(PN4_GPIO_P_DATAI_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant PN4_GPIO_P_TRI_REG_ADDR    : adb3_ocp_addr_s := std_logic_vector(to_unsigned(PN4_GPIO_P_TRI_REG_OFF,ADB3_OCP_ADDR_WIDTH));

  constant PN4_GPIO_N_DATAO_REG_ADDR  : adb3_ocp_addr_s := std_logic_vector(to_unsigned(PN4_GPIO_N_DATAO_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant PN4_GPIO_N_DATAI_REG_ADDR  : adb3_ocp_addr_s := std_logic_vector(to_unsigned(PN4_GPIO_N_DATAI_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant PN4_GPIO_N_TRI_REG_ADDR    : adb3_ocp_addr_s := std_logic_vector(to_unsigned(PN4_GPIO_N_TRI_REG_OFF,ADB3_OCP_ADDR_WIDTH));

  constant PN6_GPIO_MS_DATAO_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(PN6_GPIO_MS_DATAO_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant PN6_GPIO_MS_DATAI_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(PN6_GPIO_MS_DATAI_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant PN6_GPIO_MS_TRI_REG_ADDR   : adb3_ocp_addr_s := std_logic_vector(to_unsigned(PN6_GPIO_MS_TRI_REG_OFF,ADB3_OCP_ADDR_WIDTH));

  constant PN6_GPIO_LS_DATAO_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(PN6_GPIO_LS_DATAO_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant PN6_GPIO_LS_DATAI_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(PN6_GPIO_LS_DATAI_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant PN6_GPIO_LS_TRI_REG_ADDR   : adb3_ocp_addr_s := std_logic_vector(to_unsigned(PN6_GPIO_LS_TRI_REG_OFF,ADB3_OCP_ADDR_WIDTH));


  -------------------------------
  -- Interrupt test memory map --
  -------------------------------

  -- =============================================================================
  -- Offset Type Name    Function
  -- =============================================================================
  --
  -- 0x00   W1S  SET     Set interrupt register bits
  --
  -- 0x04   W1C  CLEAR   Clear interrupt register bits
  --        R    STAT    Read interrupt register
  --
  -- 0x08   WR   MASK    Write/read mask register
  --
  -- 0x0C   WO   ARM     Re-arm
  --
  -- 0x10   WR   COUNT   Elapsed cycle counter (write only on STAT=0)
  --
  -- =============================================================================

  -- Register offsets from base address in bytes, eg 0,4,8,... for 32 bit registers
  constant INT_SET_REG_OFF    : natural := 16#00#;
  constant INT_CLEAR_REG_OFF  : natural := 16#04#;
  constant INT_STAT_REG_OFF   : natural := 16#04#;
  constant INT_WMASK_REG_OFF  : natural := 16#08#;
  constant INT_RMASK_REG_OFF  : natural := 16#08#;
  constant INT_ARM_REG_OFF    : natural := 16#0C#;
  constant INT_WCOUNT_REG_OFF : natural := 16#10#;
  constant INT_RCOUNT_REG_OFF : natural := 16#10#;

  constant INT_SET_REG_ADDR    : adb3_ocp_addr_s := std_logic_vector(to_unsigned(INT_SET_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant INT_CLEAR_REG_ADDR  : adb3_ocp_addr_s := std_logic_vector(to_unsigned(INT_CLEAR_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant INT_STAT_REG_ADDR   : adb3_ocp_addr_s := std_logic_vector(to_unsigned(INT_STAT_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant INT_WMASK_REG_ADDR  : adb3_ocp_addr_s := std_logic_vector(to_unsigned(INT_WMASK_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant INT_RMASK_REG_ADDR  : adb3_ocp_addr_s := std_logic_vector(to_unsigned(INT_RMASK_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant INT_ARM_REG_ADDR    : adb3_ocp_addr_s := std_logic_vector(to_unsigned(INT_ARM_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant INT_WCOUNT_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(INT_WCOUNT_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant INT_RCOUNT_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(INT_RCOUNT_REG_OFF,ADB3_OCP_ADDR_WIDTH));


  ---------------------
  -- Info memory map --
  ---------------------

  -- =============================================================================
  -- Offset Type Name       Function
  -- =============================================================================
  --
  -- 0x00   RO   DATE       Date
  -- 0x04   RO   TIME       Time
  -- 0x08   RO   SPLIT      DS split block Write and Read address OOR counters
  -- 0x0C   RO   BRAM BASE  BRAM DS address base value
  -- 0x10   RO   BRAM MASK  BRAM DS address mask value
  -- 0x14   RO   RAM BASE   RAM window DS address base value
  -- 0x18   RO   RAM MASK   RAM window DS address mask value
  -- 0x1C   RO   RAM INFO   RAM information
  --
  -- =============================================================================

  -- Register offsets from base address in bytes, eg 0,4,8,... for 32 bit registers
  constant INFO_DATE_REG_OFF      : natural := 16#00#;
  constant INFO_TIME_REG_OFF      : natural := 16#04#;
  constant INFO_SPLIT_REG_OFF     : natural := 16#08#;
  constant INFO_BRAM_BASE_REG_OFF : natural := 16#0C#;
  constant INFO_BRAM_MASK_REG_OFF : natural := 16#10#;
  constant INFO_RAM_BASE_REG_OFF  : natural := 16#14#;
  constant INFO_RAM_MASK_REG_OFF  : natural := 16#18#;
  constant INFO_RAM_INFO_REG_OFF  : natural := 16#1C#;

  constant INFO_DATE_REG_ADDR      : adb3_ocp_addr_s := std_logic_vector(to_unsigned(INFO_DATE_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant INFO_TIME_REG_ADDR      : adb3_ocp_addr_s := std_logic_vector(to_unsigned(INFO_TIME_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant INFO_SPLIT_REG_ADDR     : adb3_ocp_addr_s := std_logic_vector(to_unsigned(INFO_SPLIT_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant INFO_BRAM_BASE_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(INFO_BRAM_BASE_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant INFO_BRAM_MASK_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(INFO_BRAM_MASK_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant INFO_RAM_BASE_REG_ADDR  : adb3_ocp_addr_s := std_logic_vector(to_unsigned(INFO_RAM_BASE_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant INFO_RAM_MASK_REG_ADDR  : adb3_ocp_addr_s := std_logic_vector(to_unsigned(INFO_RAM_MASK_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant INFO_RAM_INFO_REG_ADDR  : adb3_ocp_addr_s := std_logic_vector(to_unsigned(INFO_RAM_INFO_REG_OFF,ADB3_OCP_ADDR_WIDTH));


  ----------------------------
  -- Master test memory map --
  ----------------------------

  -- =============================================================================
  -- Offset Type Name    Function
  -- =============================================================================
  --
  -- 0x00   WO   ADDR    Address for direct master transfer
  --                     Writing this register also triggers a transfer
  --
  -- 0x04   W    SEL     Selects type of direct master transfer
  --                     Bit fields as follows:
  --
  --                     [0] select: 1 => perform a write
  --                                 0 => perform a read
  --
  --        R    STAT    Indicates status of direct master transfer
  --                     Bit fields as follows:
  --
  --                     [0] status: 1 => no direct master transfer in
  --                                      progress
  --                                 0 => direct master transfer in
  --                                      progress
  --
  -- 0x08   WR   DATA    Direct Master Write/Read transfer data
  --
  -- =============================================================================

  -- Register offsets from base address in bytes, eg 0,4,8,... for 32 bit registers
  constant MASTER_ADDR_REG_OFF  : natural := 16#00#;
  constant MASTER_SEL_REG_OFF   : natural := 16#04#;
  constant MASTER_STAT_REG_OFF  : natural := 16#04#;
  constant MASTER_WDATA_REG_OFF : natural := 16#08#;
  constant MASTER_RDATA_REG_OFF : natural := 16#08#;

  constant MASTER_ADDR_REG_ADDR  : adb3_ocp_addr_s := std_logic_vector(to_unsigned(MASTER_ADDR_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant MASTER_SEL_REG_ADDR   : adb3_ocp_addr_s := std_logic_vector(to_unsigned(MASTER_SEL_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant MASTER_STAT_REG_ADDR  : adb3_ocp_addr_s := std_logic_vector(to_unsigned(MASTER_STAT_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant MASTER_WDATA_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(MASTER_WDATA_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant MASTER_RDATA_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(MASTER_RDATA_REG_OFF,ADB3_OCP_ADDR_WIDTH));


  ------------------------
  -- RAM reg memory map --
  ------------------------

  -- =============================================================================
  -- Offset Type Name    Function
  -- =============================================================================
  --
  -- 0x00   RW   DS BANK   Direct Slave memory window bank select
  -- 0x04   RW   DS PAGE   Direct Slave memory window page select

  -- 0x20   RW   CTRL      Memory bank 0 control
  -- 0x24   RW   OFFSET    Memory test 0 offset (16-byte address)
  -- 0x28   RW   LENGTH    Memory test 0 length (16-byte address)
  -- 0x2C   RO   INFO      Memory bank 0 info
  -- 0x30   RO   STATUS    Memory bank 0 status
  -- 0x34   RO   ERR_ADDR  Memory test 0 error address (16-byte address)
  -- 0X38   RO   MUX ERR   Memory bank 0 switching mux error info
  -- 0x3C   RO   DDR3 ERR  Memory bank 0 DDR3 interface error info
  --
  -- Bank 0: 0x20-0x3F
  -- Bank 1: 0x40-0x5F
  -- Bank 2: 0x60-0x7F
  -- Bank 3: 0x80-0x9F
  -- =============================================================================

  -- Register offsets from base address in bytes, eg 0,4,8,... for 32 bit registers
  constant RAM_DS_BANK_REG_OFF        : natural := 16#00#;
  constant RAM_DS_PAGE_REG_OFF        : natural := 16#04#;

  constant RAM_BANK0_CTRL_REG_OFF     : natural := 16#20#;
  constant RAM_BANK0_OFFSET_REG_OFF   : natural := 16#24#;
  constant RAM_BANK0_LENGTH_REG_OFF   : natural := 16#28#;
  constant RAM_BANK0_INFO_REG_OFF     : natural := 16#2C#;
  constant RAM_BANK0_STAT_REG_OFF     : natural := 16#30#;
  constant RAM_BANK0_ERR_ADDR_REG_OFF : natural := 16#34#;
  constant RAM_BANK0_MUX_ERR_REG_OFF  : natural := 16#38#;
  constant RAM_BANK0_DDR3_ERR_REG_OFF : natural := 16#3C#;
  -- Repeated for all banks up to 16#FC#

  constant RAM_DS_BANK_REG_ADDR        : adb3_ocp_addr_s := std_logic_vector(to_unsigned(RAM_DS_BANK_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant RAM_DS_PAGE_REG_ADDR        : adb3_ocp_addr_s := std_logic_vector(to_unsigned(RAM_DS_PAGE_REG_OFF,ADB3_OCP_ADDR_WIDTH));

  constant RAM_BANK0_CTRL_REG_ADDR     : adb3_ocp_addr_s := std_logic_vector(to_unsigned(RAM_BANK0_CTRL_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant RAM_BANK0_OFFSET_REG_ADDR   : adb3_ocp_addr_s := std_logic_vector(to_unsigned(RAM_BANK0_OFFSET_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant RAM_BANK0_LENGTH_REG_ADDR   : adb3_ocp_addr_s := std_logic_vector(to_unsigned(RAM_BANK0_LENGTH_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant RAM_BANK0_INFO_REG_ADDR     : adb3_ocp_addr_s := std_logic_vector(to_unsigned(RAM_BANK0_INFO_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant RAM_BANK0_STAT_REG_ADDR     : adb3_ocp_addr_s := std_logic_vector(to_unsigned(RAM_BANK0_STAT_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant RAM_BANK0_ERR_ADDR_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(RAM_BANK0_ERR_ADDR_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant RAM_BANK0_MUX_ERR_REG_ADDR  : adb3_ocp_addr_s := std_logic_vector(to_unsigned(RAM_BANK0_MUX_ERR_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  constant RAM_BANK0_DDR3_ERR_REG_ADDR : adb3_ocp_addr_s := std_logic_vector(to_unsigned(RAM_BANK0_DDR3_ERR_REG_OFF,ADB3_OCP_ADDR_WIDTH));
  -- Repeated for all banks up to 16#FC#


  ------------------------------
  -- DMA interface memory map --
  ------------------------------

  constant DMA_BRAM_ADDR_BASE   : adb3_ocp_addr_s := dma_base_conv(X"0000080000");
  constant DMA_BRAM_ADDR_MASK   : adb3_ocp_addr_s := dma_mask_conv(X"000007FFFF");

  constant DMA_RAM0_ADDR_BASE   : adb3_ocp_addr_s := dma_base_conv(X"1000000000");
  constant DMA_RAM0_ADDR_MASK   : adb3_ocp_addr_s := dma_mask_conv(X"0FFFFFFFFF");
  constant DMA_RAM1_ADDR_BASE   : adb3_ocp_addr_s := dma_base_conv(X"2000000000");
  constant DMA_RAM1_ADDR_MASK   : adb3_ocp_addr_s := dma_mask_conv(X"0FFFFFFFFF");
  constant DMA_RAM2_ADDR_BASE   : adb3_ocp_addr_s := dma_base_conv(X"3000000000");
  constant DMA_RAM2_ADDR_MASK   : adb3_ocp_addr_s := dma_mask_conv(X"0FFFFFFFFF");
  constant DMA_RAM3_ADDR_BASE   : adb3_ocp_addr_s := dma_base_conv(X"4000000000");
  constant DMA_RAM3_ADDR_MASK   : adb3_ocp_addr_s := dma_mask_conv(X"0FFFFFFFFF");

  constant DMA_BRAM_ADDR_RANGE  : addr_range_t := (DMA_BRAM_ADDR_BASE, DMA_BRAM_ADDR_MASK);
  constant DMA_RAM0_ADDR_RANGE  : addr_range_t := (DMA_RAM0_ADDR_BASE, DMA_RAM0_ADDR_MASK);
  constant DMA_RAM1_ADDR_RANGE  : addr_range_t := (DMA_RAM1_ADDR_BASE, DMA_RAM1_ADDR_MASK);
  constant DMA_RAM2_ADDR_RANGE  : addr_range_t := (DMA_RAM2_ADDR_BASE, DMA_RAM2_ADDR_MASK);
  constant DMA_RAM3_ADDR_RANGE  : addr_range_t := (DMA_RAM3_ADDR_BASE, DMA_RAM3_ADDR_MASK);

  constant DMA_ADDR_RANGE_TABLE : addr_range_table_t(0 to 4) := (0 => DMA_BRAM_ADDR_RANGE,
                                                                 1 => DMA_RAM0_ADDR_RANGE,
                                                                 2 => DMA_RAM1_ADDR_RANGE,
                                                                 3 => DMA_RAM2_ADDR_RANGE,
                                                                 4 => DMA_RAM3_ADDR_RANGE);

  -- Address range table used in blk_dma_switch splitter for direct slave access to DDR3 memory banks
  constant DS_DDR3_ADDR_RANGE_TABLE : addr_range_table_t(0 to 3) := DMA_ADDR_RANGE_TABLE(1 to 4);


  ---------------------------------
  -- Clock frequency measurement --
  ---------------------------------

  -- Max of 2x9 = 18 MGT clocks (MGT110 CLK0 -> MGT118 CLK1) in Virtex6
  -- Max of 2+18 = 20 external clocks (LCLK + XRM_CLKIN + MGT)
  subtype clk_vec_sel_t is std_logic_vector(4 downto 0);
  subtype clk_vec_range_t is natural range 0 to 31;

  -- Common Internal user defined clocks
  constant PLL_REG_CLK_SEL : clk_vec_sel_t := "00000"; -- 00
  constant PLL_PRI_CLK_SEL : clk_vec_sel_t := "00001"; -- 01
  constant PLL_REF_CLK_SEL : clk_vec_sel_t := "00010"; -- 02
  constant PLL_MEM_CLK_SEL : clk_vec_sel_t := "00011"; -- 03

  -- External non-MGT clocks
  constant LCLK_SEL        : clk_vec_sel_t := "01100"; -- 12
  constant XRM_CLKIN_SEL   : clk_vec_sel_t := "01101"; -- 13

  -- External MGT clocks
  constant MGT110_CLK0_SEL : clk_vec_sel_t := "01110"; -- 14
  constant MGT110_CLK1_SEL : clk_vec_sel_t := "01111"; -- 15
  constant MGT111_CLK0_SEL : clk_vec_sel_t := "10000"; -- 16
  constant MGT111_CLK1_SEL : clk_vec_sel_t := "10001"; -- 17
  constant MGT112_CLK0_SEL : clk_vec_sel_t := "10010"; -- 18
  constant MGT112_CLK1_SEL : clk_vec_sel_t := "10011"; -- 19
  constant MGT113_CLK0_SEL : clk_vec_sel_t := "10100"; -- 20
  constant MGT113_CLK1_SEL : clk_vec_sel_t := "10101"; -- 21
  constant MGT114_CLK0_SEL : clk_vec_sel_t := "10110"; -- 22
  constant MGT114_CLK1_SEL : clk_vec_sel_t := "10111"; -- 23
  constant MGT115_CLK0_SEL : clk_vec_sel_t := "11000"; -- 24
  constant MGT115_CLK1_SEL : clk_vec_sel_t := "11001"; -- 25
  constant MGT116_CLK0_SEL : clk_vec_sel_t := "11010"; -- 26
  constant MGT116_CLK1_SEL : clk_vec_sel_t := "11011"; -- 27
  constant MGT117_CLK0_SEL : clk_vec_sel_t := "11100"; -- 28
  constant MGT117_CLK1_SEL : clk_vec_sel_t := "11101"; -- 29
  constant MGT118_CLK0_SEL : clk_vec_sel_t := "11110"; -- 30
  constant MGT118_CLK1_SEL : clk_vec_sel_t := "11111"; -- 31

  -- Integer versions of clock selects
  -- Common Internal user defined clocks
  constant PLL_REG_CLK_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(PLL_REG_CLK_SEL));
  constant PLL_PRI_CLK_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(PLL_PRI_CLK_SEL));
  constant PLL_REF_CLK_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(PLL_REF_CLK_SEL));
  constant PLL_MEM_CLK_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(PLL_MEM_CLK_SEL));

  -- External non-MGT clocks
  constant LCLK_NUM        : clk_vec_range_t := TO_INTEGER(UNSIGNED(LCLK_SEL));
  constant XRM_CLKIN_NUM   : clk_vec_range_t := TO_INTEGER(UNSIGNED(XRM_CLKIN_SEL));

  -- External MGT clocks
  constant MGT110_CLK0_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(MGT110_CLK0_SEL));
  constant MGT110_CLK1_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(MGT110_CLK1_SEL));
  constant MGT111_CLK0_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(MGT111_CLK0_SEL));
  constant MGT111_CLK1_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(MGT111_CLK1_SEL));
  constant MGT112_CLK0_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(MGT112_CLK0_SEL));
  constant MGT112_CLK1_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(MGT112_CLK1_SEL));
  constant MGT113_CLK0_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(MGT113_CLK0_SEL));
  constant MGT113_CLK1_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(MGT113_CLK1_SEL));
  constant MGT114_CLK0_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(MGT114_CLK0_SEL));
  constant MGT114_CLK1_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(MGT114_CLK1_SEL));
  constant MGT115_CLK0_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(MGT115_CLK0_SEL));
  constant MGT115_CLK1_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(MGT115_CLK1_SEL));
  constant MGT116_CLK0_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(MGT116_CLK0_SEL));
  constant MGT116_CLK1_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(MGT116_CLK1_SEL));
  constant MGT117_CLK0_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(MGT117_CLK0_SEL));
  constant MGT117_CLK1_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(MGT117_CLK1_SEL));
  constant MGT118_CLK0_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(MGT118_CLK0_SEL));
  constant MGT118_CLK1_NUM : clk_vec_range_t := TO_INTEGER(UNSIGNED(MGT118_CLK1_SEL));

  -- Types for buffer instantiation (blk_clocks) and clock frequency measurement (blk_ds_clk_read)
  type mgt_clk_pin_t    is array(MGT110_CLK0_NUM to MGT118_CLK1_NUM) of std_logic_dbl_t;
  type mgt_clk_buf_t    is array(MGT110_CLK0_NUM to MGT118_CLK1_NUM) of std_logic;

  type clk_vec_t        is array(LCLK_NUM to MGT118_CLK1_NUM) of std_logic;
  type clk_vec_stat_t   is array(LCLK_NUM to MGT118_CLK1_NUM) of std_logic_vector(2 downto 0);
  type clk_vec_freq_t   is array(LCLK_NUM to MGT118_CLK1_NUM) of std_logic_vector(31 downto 0);


  -- Enables for buffer instantiation (blk_clocks) and clock frequency measurement (blk_ds_clk_read)
  constant CLKS_IN_VALID : clk_vec_t := LCLK_VALID & XRM_GCLK_M2C_VALID &
                                        MGT_CLK0_VALID(0) & MGT_CLK1_VALID(0) &
                                        MGT_CLK0_VALID(1) & MGT_CLK1_VALID(1) &
                                        MGT_CLK0_VALID(2) & MGT_CLK1_VALID(2) &
                                        MGT_CLK0_VALID(3) & MGT_CLK1_VALID(3) &
                                        MGT_CLK0_VALID(4) & MGT_CLK1_VALID(4) &
                                        MGT_CLK0_VALID(5) & MGT_CLK1_VALID(5) &
                                        MGT_CLK0_VALID(6) & MGT_CLK1_VALID(6) &
                                        MGT_CLK0_VALID(7) & MGT_CLK1_VALID(7) &
                                        MGT_CLK0_VALID(8) & MGT_CLK1_VALID(8);


  ----------------------
  -- Memory interface --
  ----------------------

  -- Types for memory interface
  type mem_if_stat_array_t   is array(0 to MEM_BANKS-1) of std_logic_vector(3 downto 0);
  type mem_if_err_array_t    is array(0 to MEM_BANKS-1) of std_logic_vector(3 downto 0);
  type mem_if_rdy_array_t    is array(0 to MEM_BANKS-1) of std_logic;
  type mem_if_debug_array_t  is array(0 to MEM_BANKS-1) of std_logic_vector(31 downto 0);


  ------------------------
  -- Memory application --
  ------------------------

  -- Types for memory application interface
  type mem_app_go_array_t       is array(0 to MEM_BANKS-1) of std_logic;
  type mem_app_offset_array_t   is array(0 to MEM_BANKS-1) of std_logic_vector(DDR3_16_BYTE_ADDR_WIDTH-1 downto 0);
  type mem_app_length_array_t   is array(0 to MEM_BANKS-1) of std_logic_vector(DDR3_16_BYTE_ADDR_WIDTH-1 downto 0);
  type mem_app_done_array_t     is array(0 to MEM_BANKS-1) of std_logic;
  type mem_app_err_array_t      is array(0 to MEM_BANKS-1) of std_logic;
  type mem_app_err_ph_array_t   is array(0 to MEM_BANKS-1) of std_logic_vector(2 downto 0);
  type mem_app_err_addr_array_t is array(0 to MEM_BANKS-1) of std_logic_vector(DDR3_16_BYTE_ADDR_WIDTH-1 downto 0);


  ---------------------------
  -- uber top level blocks --
  ---------------------------

  component blk_clocks
  port(
    -- Clock pins
    clks_in        : in    clks_in_t;
    clks_out       : out   clks_out_t;
    clks_mgt_in    : in    clks_mgt_in_t;
    -- Reset output
    pll_rst        : out   std_logic;
    -- Clock outputs
    pll_pri_clk    : out   std_logic;
    pll_reg_clk    : out   std_logic;
    pll_mem_clk    : out   std_logic;
    pll_ref_clk    : out   std_logic;
    clk_vec        : out   clk_vec_t;
    mptl_clk       : out   std_logic_dbl_t);
  end component;

  component blk_direct_slave
  port(
    -- Reset input
    rst                 : in    std_logic;
    -- Clock inputs
    pll_pri_clk         : in    std_logic;
    pll_reg_clk         : in    std_logic;
    pll_ref_clk         : in    std_logic;
    pll_mem_clk         : in    std_logic;
    -- Direct slave primary OCP interface
    direct_slave_m2s    : in    adb3_ocp_m2sT;
    direct_slave_s2m    : out   adb3_ocp_s2mT;
    -- Clock frequency measurement register interface
    clk_vec             : in    clk_vec_t;
    -- GPIO signal test register interface
    gpio_inout          : inout gpio_inout_t;
    -- Interrupt test register interface
    interrupt_l         : out   std_logic;
    -- Memory interface status
    mem_if_stat         : in    mem_if_stat_array_t;
    mem_if_err          : in    mem_if_err_array_t;
    -- Memory application control
    mem_app_go          : out   mem_app_go_array_t;
    mem_app_offset      : out   mem_app_offset_array_t;
    mem_app_length      : out   mem_app_length_array_t;
    -- Memory application status
    mem_app_done        : in    mem_app_done_array_t;
    mem_app_err         : in    mem_app_err_array_t;
    mem_app_err_ph      : in    mem_app_err_ph_array_t;
    mem_app_err_addr    : in    mem_app_err_addr_array_t;
    -- BRAM secondary OCP interface
    ds_bram_m2s         : out   adb3_ocp_m2sT;
    ds_bram_s2m         : in    adb3_ocp_s2mT;
    -- RAM secondary OCP interface
    ds_ram_m2s          : out   adb3_ocp_m2sT;
    ds_ram_s2m          : in    adb3_ocp_s2mT;
    -- Debug
    bram_mux_err_info   : in    std_logic_vector(31 downto 0);
    ram_mux_err_info    : in    mem_if_debug_array_t;
    mem_if_err_info     : in    mem_if_debug_array_t);
  end component;

  component blk_ds_simple_test
  generic(
    addr_width : natural range 1 to ADB3_OCP_ADDR_WIDTH);
  port(
    -- OCP interface
    ocp_rst          : in    std_logic;
    ocp_clk          : in    std_logic;
    direct_slave_m2s : in    adb3_ocp_m2sT;
    direct_slave_s2m : out   adb3_ocp_s2mT);
  end component;

  component blk_ds_clk_read
  generic(
    addr_width : natural range 1 to ADB3_OCP_ADDR_WIDTH);
  port(
    -- OCP interface
    ocp_rst          : in    std_logic;
    ocp_clk          : in    std_logic;
    direct_slave_m2s : in    adb3_ocp_m2sT;
    direct_slave_s2m : out   adb3_ocp_s2mT;
    -- Clock inputs
    pll_pri_clk      : in    std_logic;
    pll_reg_clk      : in    std_logic;
    pll_ref_clk      : in    std_logic;
    pll_mem_clk      : in    std_logic;
    clk_vec          : in    clk_vec_t);
  end component;

  component blk_ds_io_test
  generic(
    addr_width : natural range 1 to ADB3_OCP_ADDR_WIDTH);
  port(
    -- OCP interface
    ocp_rst          : in    std_logic;
    ocp_clk          : in    std_logic;
    direct_slave_m2s : in    adb3_ocp_m2sT;
    direct_slave_s2m : out   adb3_ocp_s2mT;
    -- IO interface
    gpio_inout       : inout gpio_inout_t);
  end component;

  component blk_ds_int_test
  generic(
    addr_width : natural range 1 to ADB3_OCP_ADDR_WIDTH);
  port(
    -- OCP interface
    ocp_rst          : in    std_logic;
    ocp_clk          : in    std_logic;
    direct_slave_m2s : in    adb3_ocp_m2sT;
    direct_slave_s2m : out   adb3_ocp_s2mT;
    -- Interrupt interface
    interrupt_l      : out   std_logic);
  end component;

  component blk_ds_mem_reg
  generic(
    addr_width           : natural range 1 to ADB3_OCP_ADDR_WIDTH;
    bank_addr_width      : natural;
    page_addr_width_ddr3 : natural);
  port(
    -- OCP interface
    ocp_rst          : in    std_logic;
    ocp_clk          : in    std_logic;
    direct_slave_m2s : in    adb3_ocp_m2sT;
    direct_slave_s2m : out   adb3_ocp_s2mT;
    -- DS RAM interface
    ds_bank          : out   std_logic_vector(bank_addr_width-1 downto 0);
    ds_page_ddr3     : out   std_logic_vector(page_addr_width_ddr3-1 downto 0);
    -- Memory interface status
    mem_if_stat      : in    mem_if_stat_array_t;
    mem_if_err       : in    mem_if_err_array_t;
    -- Memory application control
    mem_app_go       : out   mem_app_go_array_t;
    mem_app_offset   : out   mem_app_offset_array_t;
    mem_app_length   : out   mem_app_length_array_t;
    -- Memory application status
    mem_app_done     : in    mem_app_done_array_t;
    mem_app_err      : in    mem_app_err_array_t;
    mem_app_err_ph   : in    mem_app_err_ph_array_t;
    mem_app_err_addr : in    mem_app_err_addr_array_t;
    -- Debug
    ram_mux_err_info : in    mem_if_debug_array_t;
    mem_if_err_info  : in    mem_if_debug_array_t);
  end component;

  component blk_ds_info
  generic(
    addr_width : natural range 1 to ADB3_OCP_ADDR_WIDTH);
  port(
    -- OCP interface
    ocp_rst           : in    std_logic;
    ocp_clk           : in    std_logic;
    direct_slave_m2s  : in    adb3_ocp_m2sT;
    direct_slave_s2m  : out   adb3_ocp_s2mT;
    -- Debug
    ds_split_err_info : in    std_logic_vector(31 downto 0));
  end component;

  component blk_dma_switch
  port(
    ocp_rst           : in    std_logic;
    ocp_clk           : in    std_logic;
    -- Source: BRAM direct slave OCP interface
    ds_bram_m2s       : in    adb3_ocp_m2sT;
    ds_bram_s2m       : out   adb3_ocp_s2mT;
    -- Source: On-board RAM DS window OCP interface
    ds_ram_m2s        : in    adb3_ocp_m2sT;
    ds_ram_s2m        : out   adb3_ocp_s2mT;
    -- Source: DMA OCP interface
    dma_channels_m2s  : in    adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
    dma_channels_s2m  : out   adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
    -- Source: On-board RAM app OCP interface
    mem_app_m2s       : in    adb3_ocp_m2sT_array(0 to MEM_BANKS-1);
    mem_app_s2m       : out   adb3_ocp_s2mT_array(0 to MEM_BANKS-1);
    -- Dest: BRAM OCP interface
    bram_if_m2s       : out   adb3_ocp_m2sT;
    bram_if_s2m       : in    adb3_ocp_s2mT;
    -- Dest: On-board RAM OCP interface
    ram_if_m2s        : out   adb3_ocp_m2sT_array(0 to MEM_BANKS-1);
    ram_if_s2m        : in    adb3_ocp_s2mT_array(0 to MEM_BANKS-1);
      -- Debug
    bram_mux_err_info : out   std_logic_vector(31 downto 0);
    ram_mux_err_info  : out   mem_if_debug_array_t);
  end component;

  component blk_bram
  port(
    ocp_rst     : in    std_logic;
    ocp_clk     : in    std_logic;
    -- BRAM OCP interface
    bram_if_m2s : in    adb3_ocp_m2sT;
    bram_if_s2m : out   adb3_ocp_s2mT);
  end component;

  component blk_mem_if
  port(
    -- OCP interface
    ocp_rst         : in    std_logic;
    ocp_clk         : in    std_logic;
    mem_if_m2s      : in    adb3_ocp_m2sT_array(0 to MEM_BANKS-1);
    mem_if_s2m      : out   adb3_ocp_s2mT_array(0 to MEM_BANKS-1);
    -- DDR3 clocking
    ddr3_rst        : in    std_logic;
    ddr3_ref_clk    : in    std_logic;
    ddr3_clk        : in    std_logic;
    -- Memory status
    mem_if_rdy      : out   mem_if_rdy_array_t;
    mem_if_stat     : out   mem_if_stat_array_t;
    mem_if_err      : out   mem_if_err_array_t;
    -- Memory physical interface
    mem_addr_out    : out   mem_addr_out_t;
    mem_ctrl_out    : out   mem_ctrl_out_t;
    mem_data_inout  : inout mem_data_inout_t;
    mem_clk_out     : out   mem_clk_out_t;
    -- Debug info
    mem_if_err_info : out   mem_if_debug_array_t);
  end component;

  component blk_mem_app
  port(
    -- OCP interface
    ocp_rst       : in    std_logic;
    ocp_clk       : in    std_logic;
    app_m2s       : out   adb3_ocp_m2sT_array(0 to MEM_BANKS-1);
    app_s2m       : in    adb3_ocp_s2mT_array(0 to MEM_BANKS-1);
    -- Control interface
    mem_if_rdy    : in    mem_if_rdy_array_t;
    app_go        : in    mem_app_go_array_t;
    app_offset    : in    mem_app_offset_array_t;
    app_length    : in    mem_app_length_array_t;
    -- Status interface
    app_done      : out   mem_app_done_array_t;
    app_err       : out   mem_app_err_array_t;
    app_err_addr  : out   mem_app_err_addr_array_t;
    app_err_ph    : out   mem_app_err_ph_array_t);
  end component;

  component blk_chipscope
  generic(
    instantiate   : boolean := true);
  port(
    -- OCP channel 0 interface
    ocp_ch0_clk   : in    std_logic;
    ocp_ch0_m2s   : in    adb3_ocp_m2sT;
    ocp_ch0_s2m   : in    adb3_ocp_s2mT;
    ocp_ch0_trig  : in    std_logic_vector(7 downto 0);
    -- OCP channel 1 interface
    ocp_ch1_clk   : in    std_logic;
    ocp_ch1_m2s   : in    adb3_ocp_m2sT;
    ocp_ch1_s2m   : in    adb3_ocp_s2mT;
    ocp_ch1_trig  : in    std_logic_vector(7 downto 0);
    -- OCP channel 2 interface
    ocp_ch2_clk   : in    std_logic;
    ocp_ch2_m2s   : in    adb3_ocp_m2sT;
    ocp_ch2_s2m   : in    adb3_ocp_s2mT;
    ocp_ch2_trig  : in    std_logic_vector(7 downto 0));
  end component;

  component blk_clock_freq
  generic(
    ref_clk_tcval      : natural range 2 to 2147483647;   -- Measurement period, in 'ref_clk' cycles
    smp_clk_div_stages : natural range 0 to 30 := 0);     -- Number of ripple-divide stages for 'smp_clk'
  port(
    rst             : in  std_logic;                      -- Asynchronous reset
    ref_clk         : in  std_logic;                      -- Reference clock
    smp_clk         : in  std_logic;                      -- Clock to be measured
    -- 'read_clk' domain
    read_clk        : in  std_logic;                      -- Output clock
    do              : in  std_logic;                      -- Pulse for at least 1 cycle to start a measurement
    count           : out std_logic_vector(31 downto 0);  -- Number of 'smp_clk' cycles counted, qualified by 'valid'
    running         : out std_logic;                      -- '1' => clock is running (i.e. count /= 0), qualified by 'valid'
    valid           : out std_logic;                      -- '1' => 'count' & 'running' are valid
    done            : out std_logic;                      -- '1' (pulsed) => measurement completed
    idle            : out std_logic);                     -- '1' => measurement not in progress
  end component;

end;


package body uber_pkg is

end;