--==============================================================================
--
--     File Name: uber_tb_pkg.vhd
--       Version: 0.1
--          Date: 25 Mar 2010
--         Model: Behavioral (NOT SYNTHESISABLE)
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Package uber_tb_pkg
--
--   Description: Testbench package for uber target FPGA
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 25 Mar 10: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_comp.all;
use work.adb3_target_pkg.all;
use work.adb3_target_inc_pkg.all;
use work.uber_pkg.all;
use work.ddr3_sdram_pkg.all;

package uber_tb_pkg is

  ----------------------
  -- Clock Generation --
  ----------------------

  -- Clock periods
  constant CLK_250_PERIOD : time :=  4.00 ns;    -- 250 MHz
  constant CLK_200_PERIOD : time :=  5.00 ns;    -- 200 MHz
  constant CLK_150_PERIOD : time :=  6.67 ns;    -- 150 MHz
  constant CLK_166_PERIOD : time :=  6.00 ns;    -- 166 MHz
  constant CLK_125_PERIOD : time :=  8.00 ns;    -- 125 MHz
  constant CLK_100_PERIOD : time := 10.00 ns;    -- 100 MHz
  constant CLK_80_PERIOD  : time := 12.50 ns;    --  80 MHz

  -- Clock frequencies in MHz (truncated division)
  constant CLK_250_FREQ   : natural := 1 us / CLK_250_PERIOD;
  constant CLK_200_FREQ   : natural := 1 us / CLK_200_PERIOD;
  constant CLK_150_FREQ   : natural := 1 us / CLK_150_PERIOD;
  constant CLK_166_FREQ   : natural := 1 us / CLK_166_PERIOD;
  constant CLK_125_FREQ   : natural := 1 us / CLK_125_PERIOD;
  constant CLK_100_FREQ   : natural := 1 us / CLK_100_PERIOD;
  constant CLK_80_FREQ    : natural := 1 us / CLK_80_PERIOD;


  --------------
  -- DMA Test --
  --------------

  -- Define DMA OCP channel function
  constant DMA_WRITE_CHANNEL  : natural range 0 to DMA_CHANNELS-1 := 0;
  constant DMA_READ_CHANNEL   : natural range 0 to DMA_CHANNELS-1 := 1;
  constant DMA_SINGLE_CHANNEL : natural range 0 to DMA_CHANNELS-1 := 0;

  -- BRAM is the only part of the DMA address space that is connected to multiple DMA channels
  -- Start DMA at BRAM address 0x100 to avoid clash with DS testing
--  constant DMA_ADDR_WR : adb3_ocp_addr_s := DMA_BRAM_ADDR_BASE + dma_base_conv(X"0000000100");
--  constant DMA_ADDR_RD : adb3_ocp_addr_s := DMA_BRAM_ADDR_BASE + dma_base_conv(X"0000000100");

  -- On-board RAM is connected to DMA channel 0 only
  -- Start DMA at RAM1 address 0x7F00 to avoid clash with DS testing
  constant DMA_ADDR_WR : adb3_ocp_addr_s := DMA_RAM1_ADDR_BASE + dma_base_conv(X"0000007F00");
  constant DMA_ADDR_RD : adb3_ocp_addr_s := DMA_RAM1_ADDR_BASE + dma_base_conv(X"0000007F00");

  -- Constants for dma channel 0/1 test
  constant DMA_SIZE     : natural := 4032;  -- Must be multiple of 16
  constant DMA_BL_WRITE : natural := 8;
  constant DMA_BL_READ  : natural := 8;


  ---------------------------------
  -- On-board RAM Part Selection --
  ---------------------------------

  -- DDR3 component model
  constant TB_DDR3_1G_PART : part_t := MT41J64M16_187E;
  constant TB_DDR3_2G_PART : part_t := MT47J128M16_187E;

  -- Select DDR3 part for board simulation
  constant TB_DDR3_PART : part_t := TB_DDR3_1G_PART;

  constant TB_DDR3_ROW  : natural := TB_DDR3_PART.part_size.arow_width;
  constant TB_DDR3_COL  : natural := TB_DDR3_PART.part_size.acol_width;
  constant TB_DDR3_BANK : natural := TB_DDR3_PART.part_size.bank_width;

  -- MIG interface address width = Bank(3)+Row(13/14)+Col(10) and convert to byte addressing (+2)
  constant TB_DDR3_BYTE_ADDR_WIDTH    : natural := TB_DDR3_BANK + TB_DDR3_ROW + TB_DDR3_COL + 2;
  constant TB_DDR3_16_BYTE_ADDR_WIDTH : natural := TB_DDR3_BYTE_ADDR_WIDTH - 4;


  ------------------------
  -- Direct Master Test --
  ------------------------

  -- Select Direct Master inclusion
  type master_use_t is (MASTER_ON,MASTER_OFF);

  constant MASTER_USE : master_use_t := MASTER_OFF;
  constant DM_CHANNEL : natural range 0 to DM_CHANNELS-1 := 0;
  constant DM_DATA_WIDTH : natural range 8 to ADB3_OCP_DATA_WIDTH:= 32;

  -- Constants for testbench direct master register interface to simulate host memory
  -- Register offsets from base address in bytes, eg 0,4,8,... for 32 bit registers
  constant HOST_MASTER_REG_NUM  : natural := 5;
  constant HOST_MASTER_REG_OFF  : natural := HOST_MASTER_REG_NUM * (DM_DATA_WIDTH/8);
  constant HOST_MASTER_REG_ADDR : std_logic_vector(DM_ADDR_WIDTH-1 downto 0) := std_logic_vector(TO_UNSIGNED((HOST_MASTER_REG_OFF),DM_ADDR_WIDTH));


  --------------------
  -- Interrupt Test --
  --------------------

  constant MASK_EN_ALL : std_logic_vector(31 downto 0) := X"00000000";

  -- Used in test_uber and test_uber_ds to indicate falling edge on linti_l
  shared variable linti_l_edge : std_logic := '0';

  -- Test status
  type top_comp_t is record
    mptl_gpio_complete : boolean;
    dma_abort_complete : boolean;
    clk_out_complete   : boolean;
  end record;

  type ds_comp_t is record
    simple_complete    : boolean;
    master_complete    : boolean;
    clock_complete     : boolean;
    frontio_complete   : boolean;
    reario_complete    : boolean;
    int_complete       : boolean;
    info_complete      : boolean;
    bram_complete      : boolean;
    ram_complete       : boolean;
  end record;

  type dma_comp_t is record
    dma_write_cmd_complete : boolean;
    dma_read_cmd_complete  : boolean;
    dma_read_resp_complete : boolean;
  end record;

  type top_pass_t is record
    mptl_gpio_passed   : boolean;
    dma_abort_passed   : boolean;
    clk_out_passed     : boolean;
  end record;

  type ds_pass_t is record
    simple_passed      : boolean;
    master_passed      : boolean;
    clock_passed       : boolean;
    frontio_passed     : boolean;
    reario_passed      : boolean;
    int_passed         : boolean;
    info_passed        : boolean;
    bram_passed        : boolean;
    ram_passed         : boolean;
  end record;

  type dma_pass_t is record
    dma_write_cmd_passed : boolean;
    dma_read_cmd_passed  : boolean;
    dma_read_resp_passed : boolean;
  end record;

end;


package body uber_tb_pkg is

end;